	 
  (**************************************************************************)
  (******** Functions required by the GUI for reading data structures *******)
  (**************************************************************************)

let stack_to_string stack = 
  match stack with
      Region.EveryStack -> "[_]"
    | Region.CallStack stack_list -> 
	begin
	  let stack_string = 
	    List.fold_left 
	      (fun a -> fun b -> (a^(C_SD.get_location_fname b)^";\n"))
	      "" stack_list
	  in
	    "[ "^stack_string^" ]"
	end


let locations_of_stack stack =
  match stack with
      Region.EveryStack -> []
    | Region.CallStack stack_list -> stack_list

	
let data_to_string_list data =
  let pred_is_atomic p = 
    match p with
      Predicate.Atom _ -> true
    | Predicate.Not (Predicate.Atom _) -> true
    | Predicate.True | Predicate.False -> true 
    | _ -> false
  in
    match data with
	Region.Concrete p ->
	  begin
	    match p.Region.pred with
		Predicate.And l -> 
		  let _ = 
		    if (List.exists (fun x -> not (pred_is_atomic x)) l) 
		    then ( 
		      Message.msg_string 
			Message.Debug 
			("Warning: region has non-atomic conjuncts "^(Predicate.toString p.Region.pred)))
		  in
		    List.map Predicate.toString l
	      | Predicate.True -> []
	      |	Predicate.Atom a -> [ Predicate.toString p.Region.pred ]
	      |	Predicate.Not (Predicate.Atom _) -> [ Predicate.toString p.Region.pred ]
	      | _ -> [(Predicate.toString p.Region.pred)]
	end
    | Region.Abstract adr ->
        let s = PredTable.bdd_to_string_pretty adr.Region.bdd in
        [s]


let reg_to_strings reg = 
  match reg with
      Region.Union _ -> failwith "Bad trace -- union region"
    | Region.Atomic r -> 
	begin
	  let loc = r.Region.location in
	  let data = r.Region.data in
	  let stack = r.Region.stack in
          let s2 = 
            if (Options.getValueOfBool "fmc" 
                || Options.getValueOfBool "fmcc" 
                || Options.getValueOfBool "cf") 
            then 
              let (x,y) = C_SD.location_coords loc in
              (string_of_int x)^"#"^(string_of_int y)
            else
            stack_to_string stack
          in
	   (data_to_string_list data,s2)
	end
    | Region.ParallelAtomic _
    | Region.ParallelUnion _ ->failwith "reg_to_strings: bad region"
    | Region.Empty -> (["empty"],"")

let atomic_reg_function_name r =
  begin
    let loc = r.Region.location in
    C_SD.get_location_fname loc
  end
  
let reg_function_name reg = 
  match reg with
      Region.Union _ -> failwith "Bad trace -- union region"
    | Region.Atomic r -> atomic_reg_function_name r
    | Region.ParallelAtomic _
    | Region.ParallelUnion _ 
    | Region.Empty ->
	failwith "reg_function_name: bad region"

let reg_to_string reg = 
  match reg with
      Region.Union _ -> failwith "Bad trace -- union region"
    | Region.Atomic r -> 
	begin
	  let loc = r.Region.location in
	  C_SD.location_to_string loc
	end
    | Region.ParallelAtomic _
    | Region.ParallelUnion _ 
    | Region.Empty ->
	failwith "reg_function_name: bad region"

let loc_to_string loc =
  match C_SD.get_source_position loc with
      Some(file,line,col) -> (string_of_int line^":")
    | None -> "-1:"

let loc_to_int loc =
  match C_SD.get_source_position loc with
      Some(file,line,col) -> line (* (string_of_int line^":")*)
    | None -> -1

let op_to_string op =
  let print_fncall fcallexp = (* when printing the function calls, also print the formal argument names *)
    let fncall_to_string (fname, args) =
      let argstring = 
        match fname with Expression.Lval (Expression.Symbol s) -> 
          begin
            try let formals = C_SD.lookup_formals s in
            match formals with 
              C_SD.Fixed l ->  List.fold_right2 
               (fun thisformal thisactual sofar -> (thisformal ^ " = " ^ (Expression.toString thisactual) ^"," ^ sofar))
               l args ""
            | C_SD.Variable l -> List.fold_right (fun thisone sofar -> (Expression.toString thisone) ^" , " ^ sofar) args ""
            with _ -> List.fold_right (fun thisone sofar -> (Expression.toString thisone) ^" , " ^ sofar) args ""
          end
        | _ -> List.fold_right (fun thisone sofar -> (Expression.toString thisone) ^" , " ^ sofar) args ""
      in
      (Expression.toString fname) ^ "("  ^ argstring ^ ")"
    in
    match fcallexp with
      Expression.FunctionCall(fname, args) ->
       fncall_to_string (fname, args)
    | Expression.Assignment (_, t, Expression.FunctionCall (fname, args)) ->
       (Expression.lvalToString t) ^ " = " ^ (fncall_to_string (fname, args))
    | _ -> failwith "Internal error in op_to_string" 
  in
  let s_loc = C_SD.get_source op in
  let x = C_SD.get_command op in
  let s1 = C_SD.get_target op in
  let cmd_string = 
    match x.C_SD.Command.code with 
      C_SD.Command.FunctionCall f -> "FunctionCall("^ (print_fncall f ) ^ ")"
    | _ -> C_SD.Command.to_string x 
  in
  (loc_to_int s_loc, loc_to_int s1, (Printf.sprintf "%s \t %s" (loc_to_string s_loc) (cmd_string)))
  
(**********************End GUI functions *********************************)
