
(*
 * Copyright (c) 1990-2002 The Regents of the University of California. All rights reserved. 
 *
 * Permission is hereby granted, without written agreement and without 
 * license or royalty fees, to use, copy, modify, and distribute this 
 * software and its documentation for any purpose, provided that the 
 * above copyright notice and the following two paragraphs appear in 
 * all copies of this software. 
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN 
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE. 
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS 
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION 
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *)

(* $Id: absutil.ml,v 1.3 2005/05/18 03:02:00 rupak Exp $

 *
 * This file is part of the BLAST Project.
 *)

open BlastArch

module Symbol = Ast.Symbol
module Constant = Ast.Constant
module Expression = Ast.Expression
module Predicate = Ast.Predicate


module Stats = Bstats

module Make_Absutil = 
  functor(PredTable : PREDTABLE with module C_SD.Command = BlastCSystemDescr.C_Command) ->
    struct
		 
module C_SD = PredTable.C_SD 
(* BlastCSystemDescr.C_System_Descr *)
(***********************************************************************************)
  (** Code to generate new temporary variable names **)
  let newTmpVarIndex = ref (-1)
  let getNextTmpVarIndex () =
    newTmpVarIndex := !newTmpVarIndex +1; !newTmpVarIndex
	
  (* Check if a symbol is a temporary *)
  (* a string is temporary if it starts with a number (obtained from getNewTmpVar)
     or contains the character '%' (obtained from brand_new_suffix)
     *)
  let isTemporary sym =
    if (Misc.is_digit (String.get sym 0) || String.contains sym '%') then
      true
    else
      false

    
  (* This is used by the strongest postcondition to give names
     to the existentiallly quantified variables.
     This is also used by the wp when substitutions are explicit.
  *)
	 (* jhala -- making this go inside recursively -- ask rupak if this is ok *)
 let getNewTmpVar (t : Expression.lval) : Expression.expression =
    let _makeStr s = (string_of_int (getNextTmpVarIndex ()))^"_"^s 
    in
      Expression.Chlval (t, _makeStr "")

 let clock = ref 0

 let clock_tick () = clock := !clock + 1 ; !clock

 let getNewTmpVar_clock (t : Expression.lval) : Expression.expression =
   Expression.Chlval (t, string_of_int !clock)


  let get_new_unknown fname = 
    getNewTmpVar (Expression.Symbol ( "BLAST_UNKNOWN@" ^ fname ) )



  (****************************************************************************************)
  (* Code to handle the aliasing between variables in the wp and sp computations. *)

  (* Code to manage hashtable mapping each predicate to its list of variables of the form
     x, x.f, *x, &x, and ( *x ).f *)
  let predTbl = Hashtbl.create 11
  let predTbl_deep = Hashtbl.create 11



  (* get the set of variable expressions in a predicate.  we ask this both for a predicate and its
     negation, so we have to deal with both cases 
     The returned list contains either lvalues or Unary(AddressOf, lvalue).
     *)
  let getVarExps pred  = 
    try Hashtbl.find predTbl pred
    with Not_found -> 
      (try Hashtbl.find predTbl (Predicate.negate pred)
       with Not_found -> 
         let f = Predicate.allVarExps pred
         in
           Hashtbl.add predTbl pred f;
           f)

(* RJ: must generalize this for pointers *)
  let getVarExps_deep pred  = 
      try Hashtbl.find predTbl_deep pred
      with Not_found -> 
	(try Hashtbl.find predTbl_deep (Predicate.negate pred)
	with Not_found -> 
	  let f = Predicate.allVarExps_deep pred in
	  let f' = List.filter 
		     (fun x -> 
			match x with
			    (Expression.Symbol _) -> true 
			  | _ -> false
		     ) f 
	  in
	  let f''= List.map (fun x ->
			       match x with
				  (Expression.Symbol y) -> y 
				 | _ -> failwith "should only receive symbols here !"
			    ) f'
	  in
	  Hashtbl.add predTbl_deep pred f'';
	  f'')	   


 (* like getVarExps, but it filters out the expressions of the form &x.  Such expressions are
     unnecessary for alias queries, since they are never aliased to anything.  *)
  let getVarsAndDerefs (exp : Predicate.predicate) : Expression.expression list =
    let filter_addrofs = 
      List.filter 
	(function e -> 
          match e with
            Expression.Unary(Expression.Address, _) -> false
          |        _ -> true)
	(getVarExps exp)
    in
    List.map 
      (fun e -> 
	match e with
	  Expression.Lval _ -> e
	| Expression.Chlval (_,_) -> e
	| _ -> failwith ("getVarsAndDerefs: Non lval expression! "^(Expression.toString e)))
      filter_addrofs
      
(* a cheap hack: I want the constituent varsandderefs of an expression -- but dont want to rewrite the function above so... *)
  let getVarsAndDerefs_exp (e : Expression.expression) : Expression.expression list =
    let _pred = Predicate.Atom (Expression.Binary (Expression.Eq, e, Expression.Constant (Constant.Int 1))) 
    in
      getVarsAndDerefs _pred 

  (* this function returns a list of expressions (of the form x, *x, x.f, or ( *x ).f) in 
     pred that may be aliased to target. 
     we assume that the given C program 
     is in a normal form such that we only need to do deal with one level of 
     pointers.  for example, a statement of the form **p = 3 will be 
     represented as (tmp = *p; *tmp = 3). *)

	(* this table caches results to getPossibleAliases *)
  let queryAlias_table = Hashtbl.create 1009
  
  let stats_nb_queryAlias_cache_hits = ref 0	
  let stats_nb_queryAlias_cache_called = ref 0
  let queryAlias_cache x y =
    stats_nb_queryAlias_cache_called := !stats_nb_queryAlias_cache_called + 1;
    Message.msg_string Message.Debug
      (Printf.sprintf "queryAlias_cache %s %s" (Expression.toString x) (Expression.toString y));
    let returnval = 
      (* if (Options.getValueOfBool "nocache") then
	AliasAnalyzer.queryAlias x y
      else *)
	try 
          let rv = Hashtbl.find queryAlias_table (x,y) in
          stats_nb_queryAlias_cache_hits := 1 + !stats_nb_queryAlias_cache_hits;
          rv
	with Not_found ->
	  begin
            let get_query_result () = AliasAnalyzer.queryAlias x y in
	    let rv = Stats.time "queryAlias" get_query_result () in
                
	      Hashtbl.add queryAlias_table (x,y) rv;
	      rv
	  end
    in
    Message.msg_string Message.Debug (Printf.sprintf "alias result: %b" returnval);
    returnval

  module LvalueSet =
      Set.Make (struct
                  type t = Expression.lval
                  let compare x y =  compare x  y
                end) 
         
  let lv_rd_aliases_table = Hashtbl.create 1001  
  let stats_nb_get_lvrd_aliases_compute = ref 0
  let stats_nb_get_lvrd_aliases_cached = ref 0
  let stats_total_alias_set_size = ref 0 
  let stats_max_alias_set_size = ref 0
  
  let get_scoped_aliases lv =
      let aliasfn = 
        if Options.getValueOfBool "cf" then AliasAnalyzer.get_lval_aliases_scope
        else (fun fn -> AliasAnalyzer.get_lval_aliases)
      in
      let scopes = Misc.sort_and_compact (C_SD.scope_of_lval lv) in
      let raw () = 
              (List.map (fun fn -> aliasfn fn lv)
              scopes)
      in       
      let ac_aliases = Stats.time "inside AliasIter_scope" raw () in
      let rv = Misc.sort_and_compact (List.flatten ac_aliases) in
      let size = List.length rv in
        stats_total_alias_set_size := !stats_total_alias_set_size + size;
        if (!stats_max_alias_set_size < size) then
                stats_max_alias_set_size := size;
        rv
        
   (* a version of the above that uses "alias_iter" instead. *)
  let queryAlias_iter_cache e_wr e_rd =
    match (e_wr,e_rd) with
          (Expression.Lval lv_wr,Expression.Lval lv_rd) ->
        (* 1. get the set of lv_rd's aliases *)
          let lv_rd_alias_set () = 
                try 
                        let rv = Hashtbl.find lv_rd_aliases_table lv_rd in
                        stats_nb_get_lvrd_aliases_cached := 1 +
                        !stats_nb_get_lvrd_aliases_cached;
                        rv
                with Not_found -> 
                   begin
                        let _ = stats_nb_get_lvrd_aliases_compute := 1 +
                           !stats_nb_get_lvrd_aliases_compute 
                        in 
                        let lv_rd_aliases : Expression.lval list = 
                             Stats.time "compute lv_rd_aliases"
                                get_scoped_aliases lv_rd 
                        in
                        let lv_aset : LvalueSet.t = 
                            List.fold_left 
                              (fun _set _elt -> LvalueSet.add _elt _set)
                                 LvalueSet.empty lv_rd_aliases
                        in
                            Hashtbl.replace lv_rd_aliases_table lv_rd lv_aset;
                            lv_aset
                   end
          in    
          let lvaset = Stats.time "get lv_rd_alias_set" lv_rd_alias_set () in
          LvalueSet.mem lv_wr lvaset
          | _ -> false
  let getPossibleAliases target pred =
    if (Options.getValueOfString "alias" = "") then []
    else 
      begin
	let predVars = Stats.time "getVarsAndDerefs" getVarsAndDerefs pred in
	  (* we only need to check alias pairs where at least one is a deref or a field access of a deref *)
	let candidates : Expression.expression list =
	  if Expression.isDeref (Expression.Lval target) then predVars
	  else List.filter Expression.isDeref predVars in
	  (* remove any candidates that are definitely not aliased to target.  
	     such candidates are determined by consulting the results of a pointer 
	     analysis. *)
	let _ = Message.msg_string Message.Debug "Alias candidates" in

        (* let _ = 
                List.iter (Message.msg_printer Message.Debug Expression.print)
                candidates in *)
        let do_filter () =  
                let alias_query_fn = 
                        if (Options.getValueOfBool "aiter") then
                                queryAlias_iter_cache
                        else    queryAlias_cache
                in
                (List.filter (alias_query_fn (Expression.Lval target))) candidates
	    (* JHALA: this should not return ANY chlvals.... *) 
        in
                Stats.time "filter alias candidates" do_filter ()
      end 

                       
      
(***************************************************************************)
(* Jhala: Misc functions required for the pre of function calls *)
  (* retTarget always points to the last target -- it is set when doing a 
     post from the stack info of the region 
     *)

  let retTarget = ref C_SD.junkRet
  let retCallSubs = ref [] (* the asgns done to pass actuals into formals -- a sin that follows from retTarget *)
  let retCallerName = ref C_SD.__NotImplementedFunctionName
			
  let wpCapRegion = ref Predicate.True

  (* this is the only way to write to retTarget -- it should not be used in any other way! *)		      
  let set_ret_pointers targ subs caller_name =
    retTarget := targ;
    retCallSubs := subs;
    retCallerName := caller_name
    
  let get_ret_pointers () = (!retTarget,!retCallSubs,retCallerName)
  let reset_ret_pointers () = set_ret_pointers C_SD.junkRet [] C_SD.__NotImplementedFunctionName
			      
   let sat_filter cubes_disjl cap_reg =
     let print_info_flag = (cap_reg <> Predicate.True) in
     let _ = if print_info_flag then Message.msg_string Message.Debug ("Cap Region: "^(Predicate.toString cap_reg)) in
     let _candidate_terms = Predicate.to_dnf (cubes_disjl) in
     let is_term_satisfiable term = 
       if (print_info_flag) 
       then
	 begin
	   Printf.printf (".");
	   not ((Options.getValueOfBool "wpsat") && (PredTable.askTheoremProver (Predicate.implies (Predicate.conjoinL [term;!wpCapRegion]) Predicate.False)))
	 end
       else
	 true
     in
     let _surviving_terms = List.filter is_term_satisfiable _candidate_terms in
     if (true) then
       begin
	 Message.msg_string Message.Debug ("Original cubes: "^(string_of_int (List.length _candidate_terms)));
	 Message.msg_string Message.Debug ("Remaining cubes: "^(string_of_int (List.length _surviving_terms)))
       end;
     Predicate.disjoinL _surviving_terms



(* Misc functions to play with parameters etc. *)

let _get_param exp = match exp with
  Expression.Lval (Expression.Symbol(x)) -> x
| _ -> failwith ("Invalid function Parameter " ^ (Expression.toString exp))
(* All function call params should be variables ... or you need
   unification ... ! *)

let listToMap l = function x -> (try List.assoc x l with _ -> x)

let real_fun_name fcexp = 
   match fcexp with 
     Expression.FunctionCall(Expression.Lval (Expression.Symbol(fname)), elist) -> fname
   | Expression.Assignment (Expression.Assign,_ , 
			   Expression.FunctionCall(Expression.Lval (Expression.Symbol(fname)), _)) ->
			     fname
   | _ -> failwith "bad call to real_fun_name"


  (**
   * This module defines the operations of this abstraction.
   *)


      
	
module Make_Operation =
  functor(C_SD : SYSTEM_DESCRIPTION with module Command = BlastCSystemDescr.C_Command) ->
struct
  type t = C_SD.edge
  let print = C_SD.print_edge
  let to_string = Misc.to_string_from_printer print

    (* presumably, these are written here solely for brevity in the remainder of the module ? *)
type exp = Ast.Expression.expression
type lval = Ast.Expression.lval
	      
type info =
  | Normal
  | Call of Predicate.predicate (* the guard, for guarded calls *) * lval option * string * exp list
  | Ret of exp
	      
  open BlastCSystemDescr.C_Command

  let get_info edge =
      match (C_SD.edge_to_command edge).code with
	Block [Return e] -> Ret e
      |	FunctionCall e ->
	  begin
	    match e with
		Expression.FunctionCall (Expression.Lval (Expression.Symbol n), el) ->
		  if C_SD.is_defined n then 
		      Call (Predicate.True, None, n, el) 
		    else Normal
	      | Expression.Assignment (Expression.Assign, targ, Expression.FunctionCall
					 (Expression.Lval (Expression.Symbol n), el)) ->
		  if C_SD.is_defined n then 
		    Call (Predicate.True, Some targ, n, el) (* in this case n=n' *)
		  else Normal

	      | e -> Normal (* raise (Failure ("Function call edge has non-call expression"^(Expression.toString e))) *)
	  end
      | GuardedFunctionCall (p,e) ->
	  begin
	    match e with
		Expression.FunctionCall (Expression.Lval (Expression.Symbol n), el) ->
		  if C_SD.is_defined n then 
		      Call (p, None, n, el) 
		    else Normal
	      | Expression.Assignment (Expression.Assign, targ, Expression.FunctionCall
					 (Expression.Lval (Expression.Symbol n), el)) ->
		  if C_SD.is_defined n then 
		    Call (p, Some targ, n, el) (* in this case n=n' *)
		  else Normal

	      | e -> Normal (* raise (Failure ("Function call edge has non-call expression"^(Expression.toString e))) *)
	  end

      |	_ -> Normal
	  
    let paren_fun op = 
      match get_info op with
	  Call _ -> "("
	| Ret _  -> ")"
	| _ -> ""
 end

    
      

let funCallSubstitute l = Predicate.substitute 
      (List.map (function (x,y) -> (Expression.Lval (Expression.Symbol(x)), Expression.Lval (Expression.Symbol(y)))) l)

(*  removeFunPreds fname predList = (preds that remain ,delPreds) 
    the ones that remain are the ones that do not contain any
    variables  local to that function *)

  let removeFunPreds fname predList = 
    let formalList = C_SD.lookup_formals fname in
    let formals = 
      match formalList with
	C_SD.Variable l -> l
      |	C_SD.Fixed l -> l
    in
    let _keepPred i = 
      let pred_i = PredTable.getPred i in
      let vars_pred_i = getVarExps_deep pred_i in
      if ( List.exists (function elt -> List.mem elt formals) vars_pred_i) then false else true 
    in
    let remain_preds = List.filter _keepPred predList in
    let delPreds = List.filter (function x -> (not (_keepPred x))) predList in
    (remain_preds,delPreds) 

				
let asgn_of_funcall (targ_sym_flag,exp_sym_flag) glvm_filter_o fcallExp = 
  let glv_filter = match glvm_filter_o with Some f -> f | None -> (fun lv -> true) in 
  let (fname,_subs,target) = (C_SD.deconstructFunCall fcallExp) in
  let exists_fun = 
    try 
      let _ = C_SD.lookup_entry_location fname 
      in true 
    with C_SD.NoSuchFunctionException _ -> false
  in
    Message.msg_string Message.Debug "Funcall assignment";
    let _ = 
      List.iter 
	( fun (x,e) ->
	    Message.msg_string 
	    Message.Debug 
	    (Printf.sprintf "%s := %s" x (Expression.toString e))
	)
	_subs
    in
    let asgn_list = 
      if (exists_fun || (target = C_SD.junkRet)) then
	List.map (fun (x,e) ->
		    C_SD.Command.Expr( Expression.Assignment (Expression.Assign,Expression.Symbol(x),e)))
	  _subs 
      else
	[C_SD.Command.Expr( Expression.Assignment (Expression.Assign, target, get_new_unknown fname ))] (*TBD:SECURITY --  should also smash all the pointers passed into function *)
    in
    let mk_sym flag symv = if flag then symv else (Expression.peel_symbolic symv) 
    in
    let lhs symv = mk_sym (true) (*targ_sym_flag*) symv in
    let rhs symv = Expression.Lval (mk_sym false (* exp_sym_flag *) symv) in
    let sym_const_hooks =
      if (Options.getValueOfBool "cf" && Options.getValueOfBool "incmod") then
	begin
	  (* Get the relevant symvars *)
	  let rel_sym_vars = 
            if targ_sym_flag then (PredTable.get_relevant_symvar fname)
            else 
              let glv_mod = List.filter glv_filter (C_SD.global_lvals_modified fname) in
              List.map (Expression.make_symvar fname) glv_mod        
          in
          let _ = Message.msg_string Message.Debug 
          ("Symvars: "^(Misc.strList (List.map Expression.lvalToString rel_sym_vars)))
          in
          let make_asgn symv =
	    C_SD.Command.Expr ( Expression.Assignment
				  (Expression.Assign,(lhs symv),(rhs symv)))
	  in
	    List.map make_asgn rel_sym_vars
	end
      else []
    in
      (* PREV: List.fold_right (wp explicit_flag) asgn_list pred *)
      (* WITH ANNOT: NOT SURE THIS IS CORRECT: check with Jhala *)
      asgn_list @ sym_const_hooks (* NOTE: the sym_const_hooks come AFTER the asgn_list *)
  
  
(* should only be called with x_lv which can possible be modified by the call,
    i.e. x_lv is either a global or a formal lvalue *)
	(* says if a DEREF of x_lv is modified -- not if x_lv itself is!
	   -- useful only if a local parameter has been passed to it *)
(* only used below *)
	
let is_smashed_fname callee_name =
  let smashed =
    let formals_smashed = (C_SD.formal_lvals_modified callee_name) in
    let globals_smashed = (C_SD.global_lvals_modified callee_name) in
      formals_smashed @ globals_smashed
  in
    fun lv_l -> Misc.nonempty_intersect compare lv_l (smashed)
            
let return_tweak callee_name is_smashed sym_flag x_lv =
  let modopt = (Options.getValueOfBool "modopt" ) in
    if (not modopt) || (is_smashed [x_lv] && sym_flag) then
      Expression.make_symvar callee_name x_lv
    else x_lv
      

let modified_by_call_table = Hashtbl.create 101

let modified_by_call callee_name sym_flag x_lv =
  Message.msg_string Message.Debug ("In modified_by_call: "^(Expression.lvalToString x_lv));
  try
    Hashtbl.find modified_by_call_table (sym_flag,x_lv)
  with Not_found ->
    let is_smashed = is_smashed_fname callee_name in
    let modopt = (Options.getValueOfBool "modopt" ) in
    let is_modified =
      if not (C_SD.is_ptr x_lv) then false
      else
	if modopt then
	  let deref_x_lv = (Expression.Dereference (Expression.Lval x_lv)) in
            is_smashed (C_SD.lvalue_closure deref_x_lv) 
	else true
    in
    let tweaked = return_tweak callee_name is_smashed sym_flag x_lv in
      Hashtbl.replace modified_by_call_table (sym_flag,x_lv) (is_modified,tweaked);
      (is_modified,tweaked)
    

(* the above functions are atrociously slow! we need a new modified_by_call -- or judicious use of caching -- as done above.*)
(* x_lv is unique to each callee_name, so we can hash on it *)
let modified_by_call2 callee_name sym_flag x_lv = 
  (* 1. do we need to copy back.
     2. if so, is x_lv itself modified ? -- *)
  try 
    Hashtbl.find modified_by_call_table (sym_flag,x_lv)
  with Not_found ->
    begin
      let all_smashed = (C_SD.formal_lvals_modified callee_name) in
      let x_smashed = List.mem x_lv all_smashed in
      let x_cl_smashed = 
	let x_deref = (Expression.Dereference (Expression.Lval x_lv)) in
	  List.exists (fun lv' -> 
			 Expression.occurs_check 
			 (Expression.Lval x_deref) (Expression.Lval lv')) all_smashed
      in
      let tweaked = 
	if x_smashed && sym_flag then Expression.make_symvar callee_name x_lv else x_lv
      in
      	Hashtbl.replace modified_by_call_table (sym_flag,x_lv) (x_cl_smashed,tweaked);
	(x_cl_smashed,tweaked)
    end 

(* only when sym_flag is on do we bother to plug in the symbolic names *)
let copy_back_asgns sym_flag param_subs callee_name  =
  let _ = Message.msg_string Message.Debug "In copy_back_asgns" in
  let tweaker = modified_by_call2  callee_name sym_flag in
  let copy_back_asgn (x,e) =
    let x_lv = (Expression.Symbol x)  in
    let (is_modified,tweaked_x_lv) = Stats.time "tweaker" tweaker x_lv in
      if is_modified
      then 
        Some 
	  (Expression.push_deref(Expression.Dereference e), 
	   Expression.Lval
	     (Expression.Dereference (Expression.Lval (tweaked_x_lv))))
      else None
  in
   if Options.getValueOfBool "cf" then Misc.map_partial copy_back_asgn param_subs else []
  
    
let asgn_of_return callee_name  =
  Message.msg_string Message.Debug "In asgn_of_return";
  let (ret_target,param_subs,_) = get_ret_pointers () in
  let cb_asgns = copy_back_asgns true (* MAKE FLEXIBLE ? *) param_subs callee_name in
  let ret_pair_option = if ret_target = C_SD.junkRet then None else Some (ret_target) in
    (ret_pair_option,cb_asgns)

(******************************************************************************************************)
(******************************************************************************************************)
(* Useful utilities for concurrent Blast *) 
(*****************************************************************************************************)
(*****************************************************************************************************)
  let is_local v =
    Message.msg_string Message.Debug ("[Abstraction.is_local " ^(Expression.lvalToString v) ^ "]");
    let rec _is_local lv =
      match lv with
	Expression.Symbol s ->
	  String.contains s '@'
      | Expression.This -> failwith "is this local?"
      | Expression.Access (_, e, _) -> _is_local_e e
      | Expression.Dereference e -> _is_local_e e
      | Expression.Indexed (e1, e2) -> _is_local_e e1 || _is_local_e e2
    and
	_is_local_e e =
      match e with
	Expression.Lval l -> _is_local l
      |	Expression.Chlval (l, _) -> _is_local l
      |	Expression.Binary (_, e1, e2) -> _is_local_e e1 || _is_local_e e2
      |	Expression.Unary (_, e) -> _is_local_e e
      |	Expression.Constant _ -> false
      |	_ -> failwith ("is_local : Strange expression "^(Expression.toString e))
    in
    _is_local v
      
  (* function to substitute tid -> notTid and locals with fresh copies.
     Uses the global phi_com_counter to get new names for locals 
     this is used to give new names to all the local variables inside the phi *)
 
  let new_name suffix (lv : Expression.lval) = 
    let rec _new_name lv =
      match lv with
	Expression.Symbol v ->
	  Expression.Symbol 
	    (if (Symbol.compare v "tid") = 0 
	    then "notTid"
	    else
	      if (Symbol.compare v "notTid") = 0 then "tid" else
	      if (Symbol.compare v "self") = 0 then "self" else
	      if String.contains v '@' 
	      then (v^suffix)
	      else v)
      | Expression.This -> failwith "this : in new_name" 
      | Expression.Access (aop, e, fld) -> 
	  Expression.Access(aop, _new_name_e e, fld)
      |	Expression.Dereference e -> Expression.Dereference (_new_name_e e)
      |	Expression.Indexed (e1, e2) -> Expression.Indexed (_new_name_e e1, _new_name_e e2)
    and
	_new_name_e e =
      match e with
	Expression.Lval l -> Expression.Lval (_new_name l)
      |	Expression.Chlval (l, suff) -> 
	  begin
	    let l' = 
	      match l with
		Expression.Symbol _ -> l
              | Expression.This -> l
	      | Expression.Dereference e -> Expression.Dereference (_new_name_e e)
	      | Expression.Access (op, e1, fld) -> Expression.Access (op, _new_name_e e1, fld)
	      | Expression.Indexed (e1, e2) -> failwith "_new_name : array not handled"
	    in
	    Expression.Chlval (l', suff)
	  end
      |	Expression.Binary (op, e1, e2) -> Expression.Binary (op, _new_name_e e1, _new_name_e e2)
      |	Expression.Unary (op, e) -> Expression.Unary (op, _new_name_e e)
      |	_ -> e
    in
    _new_name lv
      


  (******************************************************************************)
	    
      (* every so often we need fresh names for variables. The
	 suffix_counter variable, and the brand_new_suffix
	 function provides the utility: essentially these
	 provide a "brand new suffix" that is tagged on to the current
	 id / lvalue.

	 Sometimes, we need to know the current suffix. get_current_suffix returns
	 the current suffix.
	 *)

  let suffix_counter = ref 0 
  let brand_new_suffix () = 
    suffix_counter := !suffix_counter + 1;
    "%_bsuf_"^(string_of_int !suffix_counter)
		 
  let get_current_suffix () =
    "%_bsuf_"^(string_of_int !suffix_counter)

  let expression_mentions_locals e =
    try
      let reject_local (name : Expression.lval) = if is_local name then raise Exit; name in
      let _ = Ast.Expression.alpha_convert reject_local e in false
    with Exit -> true

  let predicate_mentions_locals p = 
    let atoms = Predicate.getAtoms p in
    let xtract_xps p' = 
      match p' with 
	  Predicate.Atom x -> x
	| _ -> failwith "bad call to xtract_xps"
    in
    let exps = List.map xtract_xps atoms in (* this should not fail due to the above ! *)
      List.exists expression_mentions_locals exps

(******************************************************************************************)
(******************************************************************************************)
(******************************************************************************************)
(* Utilities related to focus *)
(******************************************************************************************)
(******************************************************************************************)

 (******** extracting new predicates from **********)
(* from a cube *)

let getNewPreds orlist =
    let newp = 
      try 
	List.fold_left 
	  (function prlst -> function p -> prlst @ (Stats.time "getUsefulPreds" TheoremProver.getUsefulPredicates p)) [] orlist 
      with ((Failure "simplify") as e) -> (Message.msg_string Message.Normal "Simplify exception!"; raise e)
    in
      match newp with 
	  [] ->
            begin
              Message.msg_string Message.Error  "\n do_focus: Vampyre bug: cannot prove region to be empty!\n" ;
              Message.msg_string Message.Error "Taking all predicates!\n" ;
              Message.msg_string Message.Error ("Vampyre failed on query:\n "^(Predicate.toString (Predicate.Or (orlist))));
	      List.fold_left (function plist -> function p -> plist @ (Predicate.getAtoms p)) [] orlist
            end
	| _ -> newp             


(* tweaks a given pred to drop primes and so on ... *)

let _newTemporariesAdded = ref []

let _substOriginalVars pred =
  (* substitute original variables for temporaries *)
  let _massage symbol = (* get back originl varname from temporary name *)
    let drop_initial_number =
      if (Misc.is_digit (String.get symbol 0)) then
        let _ind = String.index symbol '_' in
        let len = String.length symbol in
          (String.sub symbol (_ind+1) (len - _ind - 1))
      else symbol
    in
      if (String.contains drop_initial_number '%') then
	  begin
	    let _ind = String.index symbol '%' in
	      String.sub drop_initial_number 0 _ind
	  end
      else
	drop_initial_number
  in
  let rec massage_lval l =
    match l with
      | Expression.Symbol s -> Expression.Symbol (_massage s)
      | Expression.This -> l
      | Expression.Access(op, e1, s) -> Expression.Access(op, massage e1, s)
      |	Expression.Dereference e -> Expression.push_deref (Expression.Dereference (massage e))
      |	Expression.Indexed (e1, e2) -> Expression.Indexed (massage e1, massage e2)
  and massage e =
    match e with
	Expression.Lval l -> Expression.Lval (massage_lval l)
      |	Expression.Chlval (l,_) -> Expression.Lval (massage_lval l)
      |	Expression.Binary(op, e1, e2) -> Expression.Binary(op, massage e1, massage e2)
      | Expression.Unary(Expression.Address, e) -> Expression.addressOf (massage e)
      | Expression.Unary(op, e) -> Expression.Unary(op, massage e)
      | Expression.Constant _   -> e
      | Expression.FunctionCall (Expression.Lval fnname, arglist) -> 
          begin
            (* worry about sel4 used to model pointer dereferences.
               We should rename sel4 to select, no? sel4 is a remnant of
               Necula's sel4 -- but there we were keeping track of bytes for memory safety
            *)
            let _getFnName fnname =
              match fnname with
                  Expression.Symbol s -> s
                | _ -> failwith ("focus: unexpected function call "^(Expression.lvalToString fnname))
            in
            let _fnNm = _getFnName fnname in
              if(_fnNm = "sel4" (* case: Arrow *)) then
                begin
                  if (List.length arglist <> 2) then
                    failwith "focus: unexpected use of sel4! More than two arguments."
                  else
                    begin
                      let _var = massage (List.hd arglist) and _deref = List.nth arglist 1 in
                        match _deref with
                            Expression.Constant i -> Expression.Lval (Expression.Dereference _var)
                          | Expression.Lval (Expression.Symbol s) -> 
				Expression.Lval (Expression.Access (Expression.Arrow, _var, _massage s))
                          | _ -> failwith "focus: unexpected use of sel4! Strange arguments."
                    end 
                end
              else
                if(_fnNm = "select" (* case Dot *)) then
                  begin
                    if (List.length arglist <> 2) then
                      failwith ("focus: unexpected use of select! More than two arguments. "^(Expression.toString e))
                    else
                      begin
                        (*
                          let _var = _massage (_getFnName (List.hd arglist)) and _deref = List.nth arglist 1 in
                        *)
                        let _var = (massage (List.hd arglist)) and _deref = List.nth arglist 1 in
                          match _deref with
                            | Expression.Lval (Expression.Symbol s) -> 
				Expression.Lval (Expression.Access (Expression.Dot, _var, _massage s))
			    | Expression.Constant (Constant.Int _) ->
				Expression.Lval (Expression.Dereference _var)
                            | _ -> failwith ("focus: unexpected use of select! Strange arguments. (deref)"^(Expression.toString e))
                      end
                  end
                else
                  if(_fnNm = "addrOf") then
                    begin
                      if (List.length arglist <> 1) then
                        failwith "focus: unexpected use of addrOf! More than one argument."
                      else
                        begin
                          let _var = massage (List.hd arglist) in
                            Expression.Unary(Expression.Address, _var)
                        end 
                    end
                  else
                    (* Other function calls can only be due to unary/binary operators 
                       that were converted to uninterpreted functions in the theoremprover
                    *)
                    begin
                      match _fnNm with
                          "BitAnd" -> Expression.Binary(Expression.BitAnd, 
							massage (List.hd arglist), 
							massage (List.nth arglist 1))
                        | "BitOr" -> Expression.Binary(Expression.BitOr, 
						       massage (List.hd arglist), 
						       massage (List.nth arglist 1))
                        | "Xor" -> Expression.Binary(Expression.Xor, 
                                                     massage (List.hd arglist), 
                                                     massage (List.nth arglist 1))
                        | "Div" -> Expression.Binary(Expression.Div, 
                                                     massage (List.hd arglist), 
                                                     massage (List.nth arglist 1))
                        | "LShift" -> Expression.Binary(Expression.LShift, 
							massage (List.hd arglist), 
							massage (List.nth arglist 1))
                        | "RShift" -> Expression.Binary(Expression.RShift, 
							massage (List.hd arglist), 
							massage (List.nth arglist 1))
                        | "Rem" -> Expression.Binary(Expression.Rem, 
                                                     massage (List.hd arglist), 
                                                     massage (List.nth arglist 1))
                        | "BitNot" -> Expression.Unary(Expression.BitNot, 
						       massage (List.hd arglist))
			| "chlval" -> massage (List.hd arglist)
			| "foffset" -> Expression.Binary(Expression.FieldOffset,
							 massage (List.hd arglist),
							 massage (List.nth arglist 1))

                        | _ -> failwith ("focus: unexpected function name "^(_getFnName fnname))
                    end
          end
      | _ -> failwith "focus: Expression type not handled.."
  in
    match pred with
        Predicate.Atom (Expression.Binary(Expression.Eq, Expression.Lval (Expression.Symbol s), e)) ->
          begin
            (*Message.msg_string Message.Error ("IN focus pred = "^(Predicate.toString pred));*)
            if (isTemporary s) then
              begin
                let orig_s = _massage s in
                let rec massage_check_lval l orig_s =
		  match l with
		    | Expression.Access(op, e1, s) -> Expression.Access(op, massage_check e1 orig_s, s)
                    | Expression.Symbol str -> 
                        begin
                          let ss = (_massage str)
                          in
                            if (ss = orig_s) then (* need annotations *)
                              begin
                                try
                                  let tempVarName = "$tmp_" in
                                    _newTemporariesAdded := 
                                    if List.mem (tempVarName^ss, ss) (!_newTemporariesAdded) then
                                      (!_newTemporariesAdded)
                                    else ("$tmp_"^ss, ss) :: (!_newTemporariesAdded);
                                    Expression.Symbol ss
                                with Not_found -> 
                                  Expression.Symbol ss
                                  
                              end
                            else (Expression.Symbol ss)
                        end
                    | Expression.This -> l
		    | Expression.Dereference e -> Expression.push_deref (Expression.Dereference (massage_check e orig_s))
		    | Expression.Indexed (e1, e2) -> Expression.Indexed (massage_check e1 orig_s, 
									 massage_check e2 orig_s)
		and massage_check e orig_s =
                  match e with
		    | Expression.Lval l -> Expression.Lval (massage_check_lval l orig_s)
		    | Expression.Chlval _ -> failwith "massage_check: Chlval not handled"
                    |   Expression.Binary(op, e1, e2) -> Expression.Binary(op, massage_check e1 orig_s, massage_check e2 orig_s)
                    | Expression.Unary(Expression.Address, e) -> Expression.addressOf (massage_check e orig_s)
                    | Expression.Unary(op, e) -> Expression.Unary(op, massage_check e orig_s)
                    | Expression.Constant _   -> e
                    | Expression.FunctionCall (Expression.Lval fnname, arglist) -> 
			begin
                          (* worry about sel4 used to model pointer dereferences.
			     We should rename sel4 to select, no? sel4 is a remnant of
			     Necula's sel4 -- but there we were keeping track of bytes for memory safety
			  *)
			  let _getFnName fnname =
			    match fnname with
				Expression.Symbol s -> s
			      | _ -> failwith ("focus: unexpected function call here: "^(Expression.lvalToString fnname))
			  in
			  let _fnNm = _getFnName fnname in
			    if(_fnNm = "sel4") then
			      begin
				if (List.length arglist <> 2) then
				  failwith "focus: unexpected use of sel4! More than two arguments."
				else
				  begin
				    let _var = (massage_check (List.hd arglist) orig_s) and _deref = List.nth arglist 1 in
				      match _deref with
					  Expression.Constant i -> Expression.Lval (Expression.Dereference  _var)
					| Expression.Lval Expression.Symbol s -> 
					    Expression.Lval (Expression.Access (Expression.Arrow, _var, _massage s))
					| _ -> failwith "focus: unexpected use of sel4! Strange arguments."
				  end 
			      end
			    else
			      if(_fnNm = "select" (* case Dot *)) then
				begin
				  if (List.length arglist <> 2) then
				    failwith ("focus: unexpected use of select! More than two arguments."^(Expression.toString e))
				  else
				    begin
				      let _var = (massage_check (List.hd arglist) orig_s) and _deref = List.nth arglist 1 in
					match _deref with
					  | Expression.Lval (Expression.Symbol s) -> Expression.Lval (Expression.Access (Expression.Dot, _var, _massage s))
					  | Expression.Constant (Constant.Int _) -> Expression.Lval (Expression.Dereference _var)
					  | _ -> failwith ("focus: unexpected use of select! Strange arguments."^(Expression.toString e))
				    end
				end
			      else
				
				if(_fnNm = "addrOf") then
				  begin
				    if (List.length arglist <> 1) then
				      failwith "focus: unexpected use of addrOf! More than one argument."
				    else
				      begin
					let _var =  (massage_check (List.hd arglist) orig_s) in
					  Expression.addressOf _var
				      end 
				  end
				else  
				  (* Other function calls can only be due to unary/binary operators 
				     that were converted to uninterpreted functions in the theoremprover
				  *)
				  begin
				    match _fnNm with
					"BitAnd" -> Expression.Binary(Expression.BitAnd, 
								      massage (List.hd arglist), 
								      massage (List.nth arglist 1))
				      | "BitOr" -> Expression.Binary(Expression.BitOr, 
								     massage (List.hd arglist), 
								     massage (List.nth arglist 1))
				      | "Xor" -> Expression.Binary(Expression.Xor, 
								   massage (List.hd arglist), 
								   massage (List.nth arglist 1))
				      | "Div" -> Expression.Binary(Expression.Div, 
								   massage (List.hd arglist), 
								   massage (List.nth arglist 1))
				      | "LShift" -> Expression.Binary(Expression.LShift, 
								      massage (List.hd arglist), 
								      massage (List.nth arglist 1))
				      | "RShift" -> Expression.Binary(Expression.RShift, 
								      massage (List.hd arglist), 
								      massage (List.nth arglist 1))
				      | "Rem" -> Expression.Binary(Expression.Rem, 
								   massage (List.hd arglist), 
								   massage (List.nth arglist 1))
				      | "BitNot" -> Expression.Unary(Expression.BitNot, 
								     massage (List.hd arglist))
				      | "chlval" -> massage (List.hd arglist)
				      | "foffset" -> Expression.Binary(Expression.FieldOffset, 
								   massage (List.hd arglist), 
								   massage (List.nth arglist 1))
				      | _ -> failwith ("focus: unexpected function name "^(_getFnName fnname))
				  end
			end
			
                    | _ -> failwith "do_focus: expression type not handled: TBD"
                in
                  Predicate.Atom (Expression.Binary(Expression.Eq, Expression.Lval (Expression.Symbol orig_s), massage_check e orig_s))
              end
            else
              Predicate.Atom (Expression.Binary (Expression.Eq, Expression.Lval (Expression.Symbol s), massage e))
          end
      | Predicate.Atom e -> Predicate.Atom (massage e)
      | _  -> failwith "do_focus: obtained non-atomic predicate: unexpected"
	  
let callsToFocus = ref 0

let clean_new_predicates newpreds = 
  let _ = _newTemporariesAdded := [] in
  let _substnewpreds = 
    try List.map _substOriginalVars 
	newpreds 
    with e -> 
      begin
	Message.msg_string Message.Debug ("_substOrigVars raises :"^(Printexc.to_string e));
	Message.msg_string Message.Error ("_substOrigVars raises :"^(Printexc.to_string e));
	[]
      end
  in
  let rv = 
    List.fold_left
    (fun curlist -> fun p -> 
      match p with
	Predicate.Atom e ->
	  begin
	    match e with
	      Expression.Binary(Expression.Eq, lhs, (Expression.Binary(op, _, _) as rhs)) when Expression.isRelOp op ->
		(Predicate.Atom (Expression.Binary(Expression.Eq, lhs, Expression.Constant (Constant.Int 0))) ::
		(Predicate.Atom (Expression.Binary(Expression.Eq, lhs, Expression.Constant (Constant.Int 1))) ::
		 (Predicate.Atom rhs :: curlist)
		 ))
	    | _ -> p :: curlist
	  end
      | _ -> failwith "clean_new_predicates: Nonatomic predicate!"
	    )
    []
    _substnewpreds 
  in
    if (Options.getValueOfBool "cf") then
      (* deal with symbolic constants *)
      let get_syms pred = 
	let extract_sym_pred lv =
	  let lv' = Expression.peel_symbolic lv in
	    if lv' <> lv then (* lv is a symbolic constant *)
	      let p_eq = (Predicate.expr_equate (Expression.Lval lv) (Expression.Lval lv')) in
	      let p_sub = Predicate.substitute [(Expression.Lval lv),(Expression.Lval lv')] pred in
		Some ([p_eq;p_sub]) (* all this overkill -- ! *)
	      
	    else None
	in
	let sym_preds = List.flatten (Misc.map_partial extract_sym_pred (Predicate.allVarExps_deep pred))
	in
	  (pred::sym_preds)
      in
	List.flatten (List.map get_syms rv)
    else rv
      
    
let extractUsefulPredicates emptyPred = 
  let _ = incr callsToFocus in
    (* let newpreds = Stats.time "getUsefulPreds" TheoremProver.getUsefulPredicates emptyPred *)
    (* while the above should have been suficient, Vampyre seems to have trouble with Or's.
       So we convert to DNF, and check if each conjunct is satisfiable, collecting useful predicates. *)
  let emptyPredDNF_form = Predicate.convertDNF emptyPred in
    (* Printf.printf "Converting to DNF and printing...\n%s\n" (Predicate.toString emptyPredDNF_form);*)
  let orlist = 
     match emptyPredDNF_form with
	 Predicate.Or plist -> plist
       | a -> [a]
  in
  let newpreds = getNewPreds orlist in
  let l = clean_new_predicates newpreds in
  l

    
(* Hacks to identify predicates from for loops.
   No longer used. Remove?
*)
let predsFromForLoops b =
  (* Hack for 'for' loops 
     Here is what we do: for each temporary variable added, we find lower and upper bounds for
               the real variable "b" associated with the temporary. Then we add all predicates real = i
     for i in (lower, upper). This is a hack, so we do not try to be very clever about this.
     The hope is that simple cases will be caught.
  *)
  Message.msg_string Message.Error "";
  Message.msg_string Message.Error "";
  Message.msg_string Message.Error ("Warning: Adding predicates for a possible for loop involving index variable "^b);
  Message.msg_string Message.Error ("Warning: The current handling of such for loops in BLAST is extremely elementary");
  Message.msg_string Message.Error ("Warning: and is a major cause for state explosion.");
  Message.msg_string Message.Error ("Warning: We recommend that you manually remove the for loop from your program");
  Message.msg_string Message.Error ("Warning: by unrolling, or by replacing the concrete upper bound by an uninitialized variable.");
  Message.msg_string Message.Error ("Blast shall continue in the meantime.");
  Message.msg_string Message.Error "The Control key is at the bottom left of your keyboard, and C is in the third line of the alphabet, between X and V.";
  Message.msg_string Message.Error "";
  Message.msg_string Message.Error "";
  let find_bounds str = (* lower <= upper *)
    let preds_with_str = PredTable.getPredsContainingVar str in
    let rec upper_lower str predList (u, l) =
      match predList with 
          [] -> (u,l)
        | p :: rest ->
            begin
              match p with
                  Predicate.Atom (Expression.Binary (Expression.Eq, Expression.Lval (Expression.Symbol s), Expression.Constant (Constant.Int a))) ->
                    if (s = str) then 
                      let newu = if (a > u) then a else u and newl = if (a < l) then a else l in
			upper_lower str rest (newu, newl)
                    else upper_lower str rest (u,l)
		| Predicate.Atom (Expression.Binary (Expression.Lt, Expression.Lval (Expression.Symbol s), Expression.Constant (Constant.Int a)))
		| Predicate.Atom (Expression.Binary (Expression.Gt, Expression.Lval (Expression.Symbol s), Expression.Constant (Constant.Int a)))
		| Predicate.Atom (Expression.Binary (Expression.Le, Expression.Lval (Expression.Symbol s), Expression.Constant (Constant.Int a)))
		| Predicate.Atom (Expression.Binary (Expression.Ge, Expression.Lval (Expression.Symbol s), Expression.Constant (Constant.Int a))) 
		| Predicate.Atom (Expression.Binary (Expression.Lt, Expression.Constant (Constant.Int a), Expression.Lval (Expression.Symbol s)))
		| Predicate.Atom (Expression.Binary (Expression.Gt, Expression.Constant (Constant.Int a), Expression.Lval (Expression.Symbol s)))
		| Predicate.Atom (Expression.Binary (Expression.Le, Expression.Constant (Constant.Int a), Expression.Lval (Expression.Symbol s)))
		| Predicate.Atom (Expression.Binary (Expression.Ge, Expression.Constant (Constant.Int a), Expression.Lval (Expression.Symbol s))) ->
                    if (s = str) then 
                      let newu = if (a > u) then a else u and newl = if (a < l) then a else l in
			upper_lower str rest (newu, newl)
                    else upper_lower str rest (u,l)
		| _ -> upper_lower str rest (u,l)
            end
    in
      upper_lower str preds_with_str (min_int, max_int)
  in
  let (u, l) = find_bounds b in
    if (l = min_int || u = max_int) then [] (* Uh oh -- trouble! *)
    else
      (let localPredList = ref [] in
       for i = l to u do
         localPredList := 
         (Predicate.Atom (Expression.Binary (Expression.Eq, 
                                             Expression.Lval (Expression.Symbol b), 
                                             Expression.Constant (Constant.Int i)))) :: !localPredList     
       done;
	 !localPredList)
      
      
      
(*************************************************************************************)
(*************************************************************************************)
(*************************************************************************************)
(* Utilities for interpolants anbd interpolant based predicate discovery *)

(* note that the function below changes the PredTable.loc_pred_table *)

let _extract_foci_preds post_process arg = 
  let raw_preds = TheoremProver.extract_foci_preds arg in
  let _ = Message.msg_string Message.Debug "raw foci preds" in
  let _ = List.iter (fun x -> Message.msg_string Message.Debug (Predicate.toString x)) raw_preds 
  in
  let less_raw_preds = post_process raw_preds in
  let _ = Message.msg_string Message.Debug "post-proc foci preds" in
  let _ = List.iter (fun x -> Message.msg_string Message.Debug (Predicate.toString x))
	    less_raw_preds 
  in
  let cleaned_preds = (clean_new_predicates less_raw_preds) in
  let _ = Message.msg_string Message.Debug "cleaned foci preds" in
  let _ = List.iter (fun x -> Message.msg_string Message.Debug (Predicate.toString x)) cleaned_preds 
  in
  let _ = 
    (* if (Options.getValueOfInt "craig" >= 2 && Options.getValueOfString "checkRace" <> "") 
    then *)
    let globvars = List.filter C_SD.is_global 
		     (List.map (fun s -> Expression.Symbol s)
			(List.flatten (List.map (fun p -> getVarExps_deep p) cleaned_preds)))
    in
      List.iter PredTable.update_global_lvals globvars 
  in	
  let pred_ids = List.map PredTable.addPred cleaned_preds in
    List.filter (fun x -> x <> -1) pred_ids
 
let extract_foci_preds = _extract_foci_preds (fun x -> x)

let extract_foci_preds_cf (callee_name, pre_call_cons) =
  let pre_call_atoms = Predicate.getAtoms pre_call_cons in
  let is_chlv e =
    match e with
	Expression.Chlval (_) -> true
      | _ -> false
  in
  let is_symv e =
    (is_chlv e) &&
    List.exists
      (fun x ->
	 match x with
	     Predicate.Atom e_atom -> Expression.occurs_check e e_atom
	   | _ -> false)
      pre_call_atoms
  in
  let transformer e =
    if is_symv e then
      begin
	let lv = Expression.lv_of_expr (Expression.replaceChlvals e) in
	let symlv = (Expression.make_symvar callee_name lv) in
	  PredTable.add_relevant_symvar callee_name symlv;
	  Expression.Lval symlv
      end
    else e
  in
  let pproc preds =
    let proc_atom atm =
      match atm with
	  Predicate.Atom e ->
	    Predicate.Atom (Expression.deep_transform_top transformer e)
	| _ -> atm
    in
      List.map proc_atom
	(List.flatten (List.map Predicate.getAtoms preds))
  in
    _extract_foci_preds pproc


(* THIS DOESNOT WORK. DELETE       
let extract_foci_preds_incref_cf (callee_name, _ ) =
  let is_chlv e =
    match e with
	Expression.Chlval (_) -> true
      | _ -> false
  in
  let pproc preds =
    (* TBD:ARRAYS *)
    (* let us first find the symbolic variables:
         An lv is a symbolic variable if in the preds,
         there are two distinctly chlvalled occurrences of it. *)
    let plist_chlv_table = Hashtbl.create 31 in (* map lv -> chlvs found for it *)
    let gather_chlv e =
      if (is_chlv e) then
	begin
	  let lv_e = Expression.lv_of_expr (Expression.replaceChlvals e) in
	    Misc.hashtbl_check_update plist_chlv_table lv_e e;
	    ()
	end;
      e
    in
    let all_atoms = (List.flatten (List.map Predicate.getAtoms preds)) in
    let extract_atom_lv atm =
      match atm with
	  Predicate.Atom e -> (Expression.deep_transform gather_chlv e);()
	| _ -> ()
    in
    let _ = List.iter extract_atom_lv all_atoms in
      (* we now know all the symvars *)
    let is_symv e =
      (is_chlv e) &&
      try
	let lv_e = Expression.lv_of_expr (Expression.replaceChlvals e) in
	let chlv = Hashtbl.find plist_chlv_table lv_e in
	  (List.length chlv) > 1
      with _ -> false
    in
    let transformer e =
      if is_symv e then
	let lv = Expression.lv_of_expr (Expression.replaceChlvals e) in
	let symlv = (Expression.make_symvar callee_name lv) in
	  add_relevant_symvar callee_name symlv;
	  Expression.Lval symlv
      else e
    in
    let proc_atom atm =
      match atm with
	  Predicate.Atom e ->
	    Predicate.Atom (Expression.deep_transform_top transformer e)
	| _ -> atm
    in
      List.map proc_atom
	(List.flatten (List.map Predicate.getAtoms preds))
  in
    _extract_foci_preds pproc
*)
      
let extract_foci_preds_parity arg = 
  let blist_to_string bl = Misc.strList (List.map string_of_bool bl) in
  let raw_preds = TheoremProver.extract_foci_preds_parity arg in
  let _ = Message.msg_string Message.Debug "raw foci preds (parity)" in
  let _ = List.iter 
	    (fun (x,p) -> Message.msg_string Message.Debug ((Predicate.toString x)^(string_of_bool p))) 
	    raw_preds 
  in 
  let (predlist,parity_list) = Misc.unzip raw_preds in
  let  cleaned_preds =   (List.map (fun p -> clean_new_predicates [p]) predlist) in
  let  cleaned_parity_list = List.combine cleaned_preds parity_list in
  let _ =  Message.msg_string Message.Debug "cleaned foci preds" in
  let _ =  List.iter 
	     (fun x -> Message.msg_string Message.Debug 
	       (Misc.strList (List.map Predicate.toString x))) 
  	    cleaned_preds 
   in
   let _ = 
   if (Options.getValueOfInt "craig" >= 2 && Options.getValueOfString "checkRace" <> "") 
   then
   let globvars = List.filter C_SD.is_global 
   (List.map (fun s -> Expression.Symbol s)
   (List.flatten (List.map (fun p -> getVarExps_deep p) (List.flatten cleaned_preds))))
   in
   List.iter PredTable.update_global_lvals globvars 
   in	
   let ppl = List.combine cleaned_preds parity_list in
   let pred_ids = 
     List.map ( fun (pl,par) -> ((List.map PredTable.addPred pl), par))  cleaned_parity_list in
   let proc (pl,par) = 
     List.iter 
       (fun p -> 
	  let pres_par = 
	    try Hashtbl.find PredTable.pred_parity_table p 
			      with Not_found -> [] 
	  in
	    if (List.mem par pres_par) then ()
	    else
	      begin
		Message.msg_string Message.Normal "Adding new parity...";
		PredTable.block_a_t_flag := true;
		Hashtbl.replace PredTable.pred_parity_table p (par::pres_par)
	      end) 
       pl 
   in
   let _ = List.iter proc pred_ids in 
     List.filter (fun x -> x <> -1) (List.flatten (List.map fst pred_ids)) 
     (* the value that will be returned*)
  
      
(* KILL VARIABLES *)
let killPredTable : (Predicate.predicate, bool) Hashtbl.t = Hashtbl.create 37
let kill_filter_fun x  = Hashtbl.mem killPredTable (PredTable.getPred x) 
			   
let hg_exists kill_ff bdd = 
  let soc_bdd = ref PredTable.bddZero in
  let make_cube_bdd cube = 
    let tmp_cube_bdd = ref PredTable.bddOne in
      for i = 0 to (Array.length cube - 1) do
	if (cube.(i)==2 || (kill_ff i) ) 
        then 
	  begin
	    if (not (cube.(i)==2)) then Message.msg_string Message.Debug ("killing pred: "^(Predicate.toString (PredTable.getPred i))^"\n")
          end
        else 
	  begin
   	    if (cube.(i)==1) then 
	      tmp_cube_bdd := CaddieBdd.bddAnd (!tmp_cube_bdd) (PredTable.getPredIndexBdd i)
	    else
	      tmp_cube_bdd := CaddieBdd.bddAnd (!tmp_cube_bdd) (CaddieBdd.bddNot (PredTable.getPredIndexBdd i))
	  end
      done;
      soc_bdd := CaddieBdd.bddOr (!soc_bdd) (!tmp_cube_bdd)
  in
    (Message.msg_string Message.Debug ("Calling hg_exists = "^(Predicate.toString (PredTable.convertBddToPred bdd))));
    CaddieBdd.bddForeachCube bdd make_cube_bdd;
    (Message.msg_string Message.Debug ("Bye bye hg_exists = "^(Predicate.toString (PredTable.convertBddToPred !soc_bdd))));
    !soc_bdd

end






















    
