(*
 * Copyright  1990-2002 The Regents of the University of California. All rights reserved. 
 *
 * Permission is hereby granted, without written agreement and without 
 * license or royalty fees, to use, copy, modify, and distribute this 
 * software and its documentation for any purpose, provided that the 
 * above copyright notice and the following two paragraphs appear in 
 * all copies of this software. 
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN 
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE. 
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS 
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION 
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *)

(* $Id: cfAbstraction.ml.m4,v 1.7 2005/05/21 05:36:59 jhala Exp $   
 *
 * This file is part of the BLAST Project.
 *)


(**
 * This module implements the abstraction interface connecting the model
 * checker to the concrete transition system.  It defines a Region module
 * based on Predicate Abstraction and BDDs, the symbolic pre and post,
 * and routines for refinment used by the model checker.
 *) 


open BlastArch

module Symbol = Ast.Symbol
module Constant = Ast.Constant
module Expression = Ast.Expression
module Predicate = Ast.Predicate
module Counter = Ast.Counter
module Stats = Bstats


let valOf = function
    None -> failwith "valOf"
  | Some v -> v

module Make_C_Abstraction =
  functor(C_SD : SYSTEM_DESCRIPTION with module Command = BlastCSystemDescr.C_Command) ->
struct

  module PredBdd =  PredBdd.Make_PredBdd(C_SD) 
  module PredTable = PredTable.Make_PredTable(PredBdd)
  module Absutil = Absutil.Make_Absutil(PredTable)

  open PredTable
  open Absutil
	
  let fprintf = Format.fprintf Format.std_formatter

  let check_races = ref None

  let assumed_invariants = ref []
  let maxStackLen = ref 0

  let getCurrentPredIndex () = PredTable.getCurrentIndex ()


				 
  let dump_abs () = PredTable.dump_abs ()
  let is_fun_refined fn = PredTable.is_fun_refined fn
		      

  let type_of_edge e = failwith "unimp cf"



  (**************************************************************************)
  (*********************************************************************************************)
  (*********************************************************************************************)
  (* The operation, DataLattice and Region modules  *)
				  include(abs-region.ml)
  (*********************************************************************************************)
  (*********************************************************************************************)


  (*******************************************************************************************)
				  (* for m4 *)
				  (* GUI functions *)
				  include(abs-gui.ml)
				  (* PRE/POST CONDITIONS + lvalue_maps *)
				  include(abs-wpsp.ml)
  (*******************************************************************************************)			 

  (* Pre and Post *)

  (******************************************************************************)

  (* Abstract pre operator. Keep track of predicates mentioned 
     in reg.preds
  *)
  (* Adamc: Changing abstract_data_pre to drop predicates regarding out-of-scope vars *)
  (* Type: Region.abstract_data_region -> C_SD.Command.t -> Region.abstract_data_region *)
  let abstract_data_pre abst_data_reg cmd edge phi_table =
    Message.msg_string Message.Debug "In abstract_data_pre -- arguments are:" ;
    Message.msg_printer Message.Debug Region.print_abstract_data_region abst_data_reg ;
    Message.msg_printer Message.Debug C_SD.Command.print cmd ;

    (* Implementation of the H function from DasDillPark-CAV99 *)
    let rec _H (bdd, p) predlist =
      (* optimization: check if (abst_data_reg, p) is satisfiable? *)
      if (isSatisfiable (bdd, p)) then
        begin
          match predlist with
              [] -> if (isSatisfiable (bdd, p)) then bddOne else bddZero
            | a:: rest ->
                let pr = getPred a in
                let cr = concrete_data_post { Region.pred = pr ; } cmd edge phi_table in
                let wppos = cr.Region.pred in
                let wpneg = Predicate.negate wppos in
		  CaddieBdd.bddOr 
                    (CaddieBdd.bddAnd (getPredIndexBdd a) (_H (bdd, wppos :: p) rest))
                    (CaddieBdd.bddAnd (CaddieBdd.bddNot (getPredIndexBdd a)) (_H (bdd, wpneg :: p) rest))
        end
      else
        bddZero
    in
      {
	Region.lelement = (failwith "foo" ; DataLattice.pre abst_data_reg.Region.lelement cmd ); 
	Region.bdd = _H (abst_data_reg.Region.bdd, []) abst_data_reg.Region.preds ;
	Region.preds = abst_data_reg.Region.preds ; }
	

	
  (* This is a bogus and buggy function ... DO NOT USE ! *)
  let abstract_data_post_das_dill abst_data_reg cmd edge post_location phi_table =
    Message.msg_string Message.Debug "In abstract_data_post -- arguments are:" ;
    Message.msg_printer Message.Debug Region.print_abstract_data_region abst_data_reg ;
    Message.msg_printer Message.Debug C_SD.Command.print cmd ;
    failwith "abstract_data_post_das_dill: This is a bogus and buggy function ... DO NOT USE !" 
      (* *********************************************************************************************
      (* Implementation of the H function from DasDillPark-CAV99 *)
	 let rec _H_cmd _cmd (bdd,p) predlist =
	 Message.msg_string Message.Debug ("In _H of post, bdd = "^(Predicate.toString (convertBddToPred bdd)));
	 Message.msg_string Message.Debug ("pred = "^(match p with [] -> "" | _ -> Predicate.toString (List.hd p)));
      (* optimization: check if (abst_data_reg, p) is satisfiable? *)
	 if (isSatisfiable (bdd,p)) then
         begin
         match predlist with
         [] -> if (isSatisfiable (bdd,p)) then bddOne else bddZero
         | a:: rest ->
         let pr = getPred a in
      (*
         let cr = concretePre {Region.conclocation = reg.Region.abslocation; Region.pred = pr} op in
      *)
         let cr = concrete_data_pre false false { Region.pred = pr ; } _cmd edge phi_table in
         let wppos = cr.Region.pred in
         let ncr = concrete_data_pre false false { Region.pred = Predicate.negate pr ; } _cmd edge phi_table in
         let wpneg = ncr.Region.pred
         in
         CaddieBdd.bddOr 
         (CaddieBdd.bddAnd (getPredIndexBdd a) (_H_cmd _cmd (bdd, wppos :: p) rest))
         (CaddieBdd.bddAnd (CaddieBdd.bddNot (getPredIndexBdd a)) (_H_cmd _cmd (bdd, wpneg :: p) rest))
         end
	 else
	 bddZero
	 in
	 let _H = _H_cmd cmd in

	 match cmd.C_SD.Command.code with 
   	 C_SD.Command.FunctionCall fcallExp ->
	 let (fname,_subs,_) = (deconstructFunCall fcallExp) in 
         let asgn_block =
         { cmd with C_SD.Command.code = C_SD.Command.Block (List.map (fun (x,e) -> C_SD.Command.Expr( Expression.Assignment (Expression.Assign,Expression.Symbol(x), e))) _subs) } in
         let region = { 
      	 Region.preds = abst_data_reg.Region.preds;
      	 Region.bdd =   _H_cmd asgn_block (abst_data_reg.Region.bdd, []) abst_data_reg.Region.preds;
	 } in
      (*print_string "Restricting region across function call....\n";*)
	 Region.restrictAbstractData (Region.locals_and_globals fname) region

      (* What was just done for the function call ... *) 
      (* Push stuff onto the stack: ... in the rest of post 
	 1. edge, on return, the location will be
	 the other end of the edge ...
	 2. if edge was a y = f(x) then post on return reflects this *)
      (* Go through list of predicates and compute "copies" *)
      (* Compute BDD of reg ^_i (pred_i = copy_pred_i) *)
      (* Start location = start location of function *)
	 |	C_SD.Command.Block l  -> 
	 if ( not (List.exists (function x -> match x with C_SD.Command.Return e -> true | _ -> false) l))
	 then 
	 let _ = Absutil.reset_ret_pointers () in
	 { 
      	 Region.preds = abst_data_reg.Region.preds;
      	 Region.bdd =   _H  (abst_data_reg.Region.bdd, []) abst_data_reg.Region.preds;
	 }
	 else 
	 let (targ, fname, retloc ,subs) = lastTarget () in
      (* Jhala: the t.o.s. is passed using last_loc_pointer  reg.Region.stack *) 
	 let _ = (Absutil.set_ret_pointers targ subs (C_SD.get_location_fname retloc)) in
	 let (newPreds,delPreds) = removeFunPreds fname abst_data_reg.Region.preds in
         let _ = Message.msg_string Message.Debug "Return: Calling _H" in
	 let newBdd = (_H  (abst_data_reg.Region.bdd, []) abst_data_reg.Region.preds) in
         let _ = Message.msg_string Message.Debug "Return: Back from _H" in
	 let _predToVar i =  -1 in (* JHALA: To be done ... ??? *)
	 let ex_newBdd = newBdd in
      (*try CaddieBdd.exists (CaddieBdd.fromList bddManager (List.map
	 _predToVar
	 delPreds)) newBdd 
	 with _ -> newBdd (* Jhala: TBD... how to go from pred to
	 bdd variable (int) ? wrapper needs
	 work *)*)
	 
	 let _ = Absutil.reset_ret_pointers ()  in
	 { 
      	 Region.preds = newPreds@delPreds;
      	 Region.bdd =  ex_newBdd;
	 } 
	 | _ -> { 
         Region.bdd = _H (abst_data_reg.Region.bdd, []) abst_data_reg.Region.preds ;
         Region.preds = abst_data_reg.Region.preds ;
         }
      (* End of abstract_data_post_das_dill *)
	 **************************************************************************************** *)


  (* -----  JHALA: The new abstract_data_post : SLAM_POST *)
  let assume_pred_ref = ref None 
			  (* We use this to check if a given predicate is unchanged in this post ie if its wp = p 
			     ... which also means <?> wpnp = np this is cleared every time postBdd is called *)

  let unchangedPredTable = Hashtbl.create 37
			     
  let dontcareArray = Array.create maxPossiblePreds false
  let outOfScopeArray = Array.create maxPossiblePreds false

  let extract_scope var = 
    let names = Misc.chop var "@" in
      try List.nth names 1 with _ -> "__BLAST_GLOBAL" 
	

  let postBdd bdd _pred_list _post_split_preds =
    statPostBddCalls := !statPostBddCalls + 1;
    Message.msg_string Message.Debug "In postBdd..\n";
    Message.msg_string Message.Debug ("post of assume"^(match !assume_pred_ref with Some(e) -> ("YES:"^(Predicate.toString e)) | _ -> "NO")^"\n");
    Message.msg_string Message.Debug ("Bdd = "^(Predicate.toString (convertBddToPred bdd)));
    Message.msg_string Message.Debug "\nList of predicates = ";
    let pred_list = List.map (fun (a,b,c,d) -> (a,b,c)) _pred_list in
    let post_split_preds = List.map (fun (a,b,c,d) -> (a,b,c)) _post_split_preds in
    let craig_flag = (Options.getValueOfInt "craig" = 2) in
    let craig_flag1 = (Options.getValueOfInt "craig" = 1) in
    let dc_flag = (Options.getValueOfBool "dc") in
    let _ = 
      if (Options.getValueOfBool "scope") then 
	begin
	  Misc.array_reset outOfScopeArray false;
	  List.iter (fun (a,_,_,d) -> Array.set outOfScopeArray a d) _pred_list 
	end
    in
    let rec loop lst = 
      match lst with 
	  [] ->  () 
	| (a,b,c,d)::rest -> 
	    begin
	      ignore (Message.msg_string Message.Debug (Predicate.toString (getPred a)));
	      ignore (Message.msg_string Message.Debug (Predicate.toString b));
	      ignore (Message.msg_string Message.Debug (Predicate.toString c)); 
	      ignore (Message.msg_string Message.Debug (string_of_bool d));
	      loop rest
	    end
    in
    let _ = loop _pred_list in
    let _ =  Message.msg_string Message.Debug ("Will split on: \n");
      loop  _post_split_preds
    in

    (* assuming that if p = wp, then np = wpnp ...*)
    let wp_different_filter (p,wp,_) = 
      if craig_flag && (!assume_pred_ref <> None ) then true 
      else
	match Options.getValueOfString "post" with
	    "slam" -> true
	  | "slam-dumb" -> not (getPred p = wp)
	  | "slam1" -> not ((getPred p = wp) && (!assume_pred_ref = None))
	  | "slam2" -> let p_pred = getPred p in 
	      (not (p_pred = wp) || 
	       (match !assume_pred_ref with 
		    None -> false
		  |	Some(asm) -> not (Misc.intersection (getVarExps_deep p_pred) (getVarExps_deep asm) = [])))  
		(* JHALA: Instead of this intersection check I would like to say 
		   if the two preds are textually the same or imply 
		   each other or some such ...*)
	  |  a -> failwith ("Unknown post algorithm "^a)
    in
      (* I dont have a good idea of when an "if" affects/doesnt affect a predicate, so for now, leave them all in ... there should be some cone-check here ...*)
    let (_changed_pred_list, unchanged_pred_list) = Stats.time "wp_diff_filter" (Misc.filter_cut wp_different_filter) pred_list in
    let cpl_addon = if craig_flag1 then post_split_preds else [] in
    let changed_pred_list = _changed_pred_list @ cpl_addon in
      
    let _ = Hashtbl.clear unchangedPredTable in
    let _ = List.iter (function x -> Hashtbl.add unchangedPredTable (Misc.fst3 x) true) unchanged_pred_list in
      (* all the unchanged preds are dumped into this table *)
    let sumofcubesBdd = ref bddZero in
      
    let conjunctL = ref [] in
      (* conjunct that the present cube produces post-command c *)
    let _ = Misc.array_reset dontcareArray false in
      
    (* START TIME HOLE *)
    let _check_Pred_list_simplify assumptions pred_triple_list conjunct_pointer =
      let asm_conj = Predicate.conjoinL assumptions in
      let _ = Stats.time "_assume" (TheoremProver._assume ) asm_conj in 
      let _check_pred (*_asm_conj*) (p,wpp,wpnp) = 
	if ( dontcareArray.(p) && dc_flag) then false 
	  (* this tuple will never be called if I dont care about it if craig is true *)
	else 
          let eToCheck1 = Predicate.implies asm_conj wpp in
          let eToCheck2 = Predicate.implies asm_conj wpnp in
          let ans1 = Stats.time "simp-a1:" (askTheoremProverContext eToCheck1) wpp in
          let ans2 = Stats.time "simp-a2:" (askTheoremProverContext eToCheck2) wpnp in
            (*let ans1 = (* Stats.time "simp-a1:"*) askTheoremProver eToCheck1 in
              let ans2 = (* Stats.time "simp-a2:"*) askTheoremProver eToCheck2 in*)
            match ans1 with
		true ->  
		  if  (ans2 = true) then 
		    begin
		      Message.msg_string Message.Debug ("Cube implies both wp(p) and wp(neg p): "^(Predicate.toString (getPred p))); 
		      conjunct_pointer := (p,true)::(p,false)::(!conjunct_pointer);
		      true
		    end
		  else
		    begin 
	              conjunct_pointer := (p,true)::(!conjunct_pointer);
	              true               
		    end
              | false ->  
		  if (ans2 = true) then  
		    begin
	              conjunct_pointer := (p,false)::(! conjunct_pointer);
	              true
		    end
		  else 
		    false
      in
      let _ = Stats.time "_check_pred" (List.map (_check_pred (* asm_conj *))) pred_triple_list in
	statWorstCaseQueriesForPost :=  2*(List.length pred_list) + !statWorstCaseQueriesForPost;
	statActualQueriesForPost := 2*(List.length pred_triple_list) + (!statActualQueriesForPost);
	statNumPost := !statNumPost + 1;
	(match (!assume_pred_ref) with
	     Some _ -> 
	       (statAssumeQueriesForPost := 2*(List.length pred_triple_list) + (!statAssumeQueriesForPost); statNumAssumePost := !statNumAssumePost + 1)
	   | None -> ())
	;
	TheoremProver._pop ();
	true
    in
      (* END TIME HOLE *)

    let _check_Pred_list_vampyre assumptions pred_triple_list conjunct_pointer = 
      let is_cube_consistent  = TheoremProver.assume assumptions in
      let asm_conj = Predicate.conjoinL assumptions in
      let _check_pred assumptions (p,wpp,wpnp) = 
	if ((dontcareArray.(p)) && dc_flag) then false
	else 
	  let eToCheck1 = Predicate.implies asm_conj wpp in
	  let eToCheck2 = Predicate.implies asm_conj wpnp in
	  let ans1 = (*Stats.time "ask1:"*) (askTheoremProverContext eToCheck1) wpp in
	  let ans2 = (*Stats.time "ask2:"*) (askTheoremProverContext eToCheck2) wpnp in
	    match ans1 with
		true ->  
		  if  (ans2 = true) then 
		    begin
		      Message.msg_string Message.Debug ("Cube implies both wp(p) and wp(neg p): "^(Predicate.toString (getPred p))); 
		      conjunct_pointer := (p,true)::(p,false)::(!conjunct_pointer);
		      true
		    end
		      
		  else
		    begin 
		      conjunct_pointer := (p,true)::(!conjunct_pointer);
		      true               
		    end
	      | false -> 
		  if (ans2 = true) then  
		    begin
		      conjunct_pointer := (p,false)::(!conjunct_pointer);
		      true
		    end
		  else 
		    false
      in
	
	if (is_cube_consistent) then 
	  begin 
            let _ = List.map (_check_pred asm_conj) pred_triple_list in
              true
	  end 
	else false
    in
      
    let _check_Pred_list = 
      match (Options.getValueOfString "s" ) with
	  "vampyre" -> _check_Pred_list_vampyre
	| _ -> _check_Pred_list_simplify
    in
      (* the preds below are the integer identifiers *)
    let rec make_conjunct_bdd conjl = 
      match conjl with
	  [] -> bddOne 
	| (p,true)::tl ->  CaddieBdd.bddAnd (getPredIndexBdd p) (make_conjunct_bdd tl) 
	| (p,false)::tl -> CaddieBdd.bddAnd (CaddieBdd.bddNot (getPredIndexBdd p))  (make_conjunct_bdd tl) 
    in
      
    let rec refine_cube pre_reg cube_list split_preds = 
      (* Jhala: this next line is very slow and should be optimized *)
      let has_no_value (x,wp,wpn) = List.for_all (fun y -> x <> fst y) cube_list in
      let preds_to_split_on = List.filter (has_no_value) split_preds in 
	match preds_to_split_on with
	    [] -> make_conjunct_bdd cube_list
	  | (h,wp,wpneg)::t -> 
	      begin
		let cube_plus = (h,true)::cube_list in
		let pre_reg_plus = wp::pre_reg in
		let conj_plus_ptr = ref cube_plus in
		let _ = _check_Pred_list pre_reg_plus preds_to_split_on conj_plus_ptr in
		let plus_bdd = refine_cube pre_reg_plus !conj_plus_ptr preds_to_split_on in
		  
		let cube_minus = (h,false)::cube_list in
		let pre_reg_minus = wpneg::pre_reg in
		let conj_minus_ptr = ref cube_minus in
		let _ = _check_Pred_list pre_reg_minus preds_to_split_on conj_minus_ptr in
		let minus_bdd = refine_cube pre_reg_minus !conj_minus_ptr preds_to_split_on in
		  
		  CaddieBdd.bddOr plus_bdd minus_bdd
              end
    in
      (* let custom_post_cube assumptions  =
	 Message.msg_string Message.Debug "In custom_post_cube";
	 let asm_conj = Predicate.conjoinL assumptions in
	 let _ = Stats.time "_assume" (TheoremProver._assume) asm_conj in
	 let is_sat (pi,wp,wpnp) = 
      (* check if this pred is satifiable in the post - returns a boolean *)
	 let query = Predicate.implies (Predicate.conjoinL (wp::assumptions)) Predicate.False in
	 let ac_query = Predicate.implies wp Predicate.False in 
         not (askTheoremProverContext query ac_query) 
      (* if the predicate is NOT valid iff fmla is satisfiable *) 
	 in
	 let _process_var_preds l = 
	 List.fold_left (fun x y -> CaddieBdd.bddOr x (getPredIndexBdd (Misc.fst3 y))) bddZero (List.filter is_sat l)
	 in
	 let rv =  List.fold_left (fun x y -> CaddieBdd.bddAnd (_process_var_preds y) x) bddOne (!custom_preds_wp)
	 in
	 TheoremProver._pop (); 
	 Message.msg_string Message.Debug "Leaving custom_post_cube";
	 rv
      (* Jhala: ideally i should only fire queries for those variables that are affected by the cmd *)
	 in
      *)

    let _check_Cube _pred_list (cube : int array)  =
      (Message.msg_string Message.Debug (" In _check_Cube"));
      statReachedCubes := !statReachedCubes + 1;
      conjunctL := [];   
      let _ = 
	if (dc_flag) then Misc.array_reset dontcareArray false;
	if (craig_flag) then 
	  begin
	    Misc.array_reset dontcareArray true; (* RJ: whoa! *)
	    List.iter (fun (i,_,_) -> Array.set dontcareArray i false) pred_list
	      (* make sure we care about the predicates that we care about *)
	  end
      in
	(* if (_syntaxCheck cube modeledPredIndexList) then 
	   () else --syntactic heuristic *)
      let cubePostIsEmpty = 
	match (!assume_pred_ref) with 
            Some(e) -> (askTheoremProver (Predicate.implies (Predicate.conjoinL [(convertCubeToPred cube);e]) Predicate.False))
          | None -> false
      in 
	if (cubePostIsEmpty) then ()
	else 
          begin   
            let pref = match (!assume_pred_ref) with None -> ref [] | Some(e) -> ref [e] in
            let _plist = ref _pred_list in
	      begin
		for i = 0 to (Array.length cube - 1) do
		  if (cube.(i)==2) 
		  then 
		    begin
		      try
			if ((dc_flag) && (outOfScopeArray.(i)))  then
			  begin
			    Message.msg_string Message.Debug ("DC about: "^(Predicate.toString (getPred i))^" \n"); 
			    Array.set dontcareArray i true
			  end
		      with _ -> ()
		    end
		  else 
		    begin
		      let i_pred = getPred i in
			(* Assume the invariant that i_pred : Predicate.Atom *)
                      let i_unchanged = (Hashtbl.mem unchangedPredTable i)  
		      in
			if (cube.(i)==1) then
			  begin
			    pref := i_pred::(!pref);
			    if (i_unchanged && not(dontcareArray.(i))) 
			    then conjunctL := (i,true)::(!conjunctL);
			  end
			else
			  begin
			    pref := (Predicate.negate i_pred)::(!pref);
			    if (i_unchanged && not(dontcareArray.(i))) 
			    then conjunctL := (i,false)::(!conjunctL);
			  end
		    end
		done;
		
		(* Now ask the thmprover if pref -> each fellow in predlist *)
		if (Stats.time "_check_Pred_list" (_check_Pred_list !pref)  changed_pred_list conjunctL) 
		then 
		  begin
		    (* Jhala: I'm being paranoid about the soundness and aliasing. something to assuage me *)
		    let _ =
		      match (Options.getValueOfBool "scheck") with
			  false -> ()
			| true ->
			    begin
			      let cube_l = List.map 
					     (fun (i,b) -> if b then (getPred i) else Predicate.Not(getPred i))
					     !conjunctL 
			      in
				if (TheoremProver.queryExp (Predicate.Implies ((Predicate.And cube_l),(Predicate.False))) && (!assume_pred_ref = None)) 
				then 
				  begin
				    Message.msg_string Message.Normal "SCHECK: Post of block is FALSE !!";
				    Message.msg_string Message.Error  "SCHECK: Post of block is FALSE !!";
				    ()
				  end
				else 
				  ()
			    end
		    in
		      (*  this is where we plug in the post_split_cube *)
		    let splitup_bdd = 
		      if (false (*Options.getValueOfBool "cref"*)) then (refine_cube !pref !conjunctL post_split_preds)
		      else (make_conjunct_bdd !conjunctL) 
		    in
		      sumofcubesBdd := CaddieBdd.bddOr (splitup_bdd) (!sumofcubesBdd)                
			(*
			  let os_succ_cube = (make_conjunct_bdd !conjunctL) in
			  if (Options.getValueOfString "custom" = "") 
			  then 
			  sumofcubesBdd := CaddieBdd.bddOr (make_conjunct_bdd !conjunctL) (!sumofcubesBdd)                
			  else
			  begin
			  let custom_succ_bdd = (custom_post_cube !pref) in
			  sumofcubesBdd := CaddieBdd.bddOr 
			  (CaddieBdd.bddAnd custom_succ_bdd (make_conjunct_bdd !conjunctL)) 
			  (!sumofcubesBdd); 
			  end
			*)
  		  end; () 
		    (* !conjunctL *)
              end;
	  end
    in
      begin
	
	if (CaddieBdd.bddEqual bdd bddZero) then
	  sumofcubesBdd := bddZero
	else 
	  begin
	    if (CaddieBdd.bddEqual bdd bddOne) 
	    then
              let _tmpcube = (* build a temporary cube with nothing in it *)
		Array.make (*size *)(getCurrentIndex ()+2) (* content *)2 
		  
	      (* Note: The size is certainly a vast	overestimate, but this
		 is the easiest way. The "+2" removes additional complications relating
		 to index being -1 initially. Note that this is a fake cube -- we should
		 never try to print it or convert it to a predicate!Ugly hack, but 
		 works. So far *)
              in
		_check_Cube changed_pred_list _tmpcube
	    else
              ignore (Stats.time "bddForEachCube-post" (CaddieBdd.bddForeachCube bdd) (_check_Cube changed_pred_list));
	  end;
	(Message.msg_string Message.Debug ("Returning from Post-Bdd = "^(Predicate.toString (convertBddToPred !sumofcubesBdd))));
	!sumofcubesBdd 
      end

  let wp_table = Hashtbl.create 1009 
  let stackInfo = ref None
  let my_loc_print l = 
    let (i,j) = C_SD.location_coords l in
      Message.msg_string Message.Debug (Printf.sprintf "location (%d,%d)" i j)



  let query_fn bdd (qpred:Predicate.predicate)  =
    let trueval = not (PredTable.isSatisfiable (bdd, [ Predicate.Not qpred ]))
    in let falseval = not (PredTable.isSatisfiable (bdd, [ qpred]))
    in
      if (trueval = falseval) then Ast.Dontknow
      else if trueval then  Ast.True
      else Ast.False

  let rec sane_pred_list pred_list =
    match pred_list with
      | [] -> []
      | head::tail ->
	  if (head = Ast.Predicate.True) then sane_pred_list tail
	  else [head] @ (sane_pred_list tail)

	      
  let pred_to_assume_cmd _pred =
    let _ = print_string ("pushing assume::"^(Predicate.toString _pred)^"\n") in
      {C_SD.Command.code = C_SD.Command.Pred _pred;
       C_SD.Command.read = [];
       C_SD.Command.written = []}

  let rec s_get_bdd_for_pred_list lpredlist __H_cmd tmpbdd regpreds =
    match lpredlist with
      | [] -> tmpbdd
      | hd::tail ->
	  let assume_cmd = pred_to_assume_cmd hd in
	  let str = C_SD.Command.to_string assume_cmd(*.C_SD.Command.code*)
	  in
	    assume_pred_ref := Some (hd) ;
	    let newtmpbdd = __H_cmd (pred_to_assume_cmd hd) tmpbdd regpreds
	    in
	      assume_pred_ref := None ;
	      s_get_bdd_for_pred_list tail __H_cmd newtmpbdd regpreds
		


  (* Self-recursion. See test/cftest/rec4.c for an example.
   * If there is a call to f() inside f... then we rename:
     * 1. the "Formals" of f, 
     * 2. the symlvals of f,
     * 3. the predicates inside f
     So that they don't clash with the caller region.
     Note that the funcall assignments are: fml = actual and sym_fml = fml.
     For the former, we rename only the fml, in the latter we rename both
     sym_fml and fml as they both correspond to the callee region.
   *)
  let rename_locals lv =
    match lv with
    Expression.Symbol s ->
      if (String.contains s '@') then (Expression.Symbol (s^"BLAST"))
      else lv
      | _ -> lv

  let self_rename_pred p = Predicate.alpha_convert rename_locals p
  let self_rename_lval lv = Expression.alpha_convert_lval rename_locals lv
  let self_rename_expr e = Expression.alpha_convert rename_locals e
  let self_rename_cmd cmd =
    let self_rename_stm stmt =
      match stmt with
      C_SD.Command.Expr (Expression.Assignment (Expression.Assign,lv,e)) ->
        let (lv',e') = 
          if Expression.is_symbolic lv then (self_rename_lval lv, self_rename_expr e)
          else (self_rename_lval lv, e)
        in C_SD.Command.Expr (Expression.Assignment (Expression.Assign, lv',e'))
      | _ -> stmt
    in
    let new_code = 
      match cmd.C_SD.Command.code with
      C_SD.Command.Block el -> C_SD.Command.Block (List.map self_rename_stm el)
      | _ -> cmd.C_SD.Command.code
    in
    {cmd with C_SD.Command.code = new_code}
    
  let rec abstract_data_post_graf_saidi abst_data_reg cmd edge post_location phi_table =
    Message.msg_string Message.Debug "In abstract_data_post GRAF_SAIDI_POST -- arguments are:" ;
    Message.msg_printer Message.Debug Region.print_abstract_data_region abst_data_reg ;
    Message.msg_printer Message.Debug C_SD.Command.print cmd ;
    let src,dst = (C_SD.get_source edge,C_SD.get_target edge) in
    let self_recursion_flag =
      match Operation.get_info edge with
      Operation.Call (_,_,callee_name,_) -> callee_name = C_SD.get_location_fname src
      | _ -> false
    in
    let _ = my_loc_print src in
    let target_loc  = C_SD.location_coords post_location (* (C_SD.get_target edge) *) in
    let _ = Message.msg_string Message.Debug
	      (Printf.sprintf "absdatapost postloc: (%d,%d)" (fst target_loc)  (snd target_loc)) 
    in
    let craig_flag = Options.getValueOfInt "craig" = 2 in
    let craig_flag1 = Options.getValueOfInt "craig" = 1 in
    let craig_preds () = 
      try
	Hashtbl.find loc_pred_table target_loc (* doesnt work if this isnt the real successor! *)
      with
	  Not_found -> []
    in
    let predlist =
      (!globally_useful_preds 
      @ (try Hashtbl.find fun_pred_table (fst target_loc) with Not_found -> []))
    in
    let _H_cmd _cmd bdd actual_predlist =
      Message.msg_string Message.Debug ("In _H of abstract_data_post, bdd = "^(Predicate.toString (convertBddToPred bdd)));
      let (tos,tos2,call) = match (!stackInfo) with
	  Some (t,t2,c) -> (t,t2,c) 
	| None -> failwith "scope info NOT supplied !"
      in
      let _ = Message.msg_string Message.Debug
		(Printf.sprintf "tos=%s , tos2=%s , call=%s" tos tos2 (string_of_bool call)) in
      let out_of_scope pred = 
	if (Options.getValueOfBool "scope") then
	  let _stay_ff x = 
	    let scope_x = (extract_scope x) in
	      ((tos = scope_x) || (call && tos2 = scope_x) || (scope_x = "__BLAST_GLOBAL"))
	  in
	  let predVars = getVarExps_deep pred in
	    List.exists (fun x -> not (_stay_ff x)) predVars
	else 
	  true
      in
      let _cmd = if (self_recursion_flag) then self_rename_cmd _cmd else _cmd in
      let make_pred_triple () (a : int) = 
	let pr = getPred a in
        let pr = if (self_recursion_flag) then self_rename_pred pr else pr in
          match (!assume_pred_ref) with 
	      None ->
		let cr = concrete_data_pre false false { Region.pred = pr ; } _cmd edge phi_table in
		let wppos = cr.Region.pred in
		let ncr = concrete_data_pre false false { Region.pred = Predicate.negate pr ; } _cmd edge phi_table in
		let wpneg = ncr.Region.pred in
		  
                  (*if (wppos,wpneg) = (_check_wppos,_check_wpneg) then*)
                  (a,wppos,wpneg, (out_of_scope pr))
                    (* else failwith ("bad wp_table")*)
            | Some(e) -> (a,pr,Predicate.negate pr,(out_of_scope pr) )
      in
      let post_split_preds = 
	if craig_flag1 then craig_preds ()
	else []
	  (*if (Options.getValueOfBool "cref") 
	    then (try Hashtbl.find post_split_table post_location with Not_found -> []) 
	    else [] *)
      in
	Stats.time "postBdd" (postBdd bdd  (List.map ( Stats.time "mk_pred_triple" make_pred_triple ()) actual_predlist)) (List.map (make_pred_triple ()) post_split_preds)
    in
    let _H = _H_cmd cmd in

      if (predlist = [] && (CaddieBdd.bddEqual abst_data_reg.Region.bdd PredTable.bddOne))
      then
	let (_lpredlist, rlelement) = DataLattice.post abst_data_reg.Region.lelement edge (query_fn PredTable.bddOne)
	in let lpredlist = sane_pred_list _lpredlist
	in if (List.length lpredlist = 0)
	  then
	    { 
	      Region.lelement = rlelement;
      	      Region.preds = [];
      	      Region.bdd =  PredTable.bddOne 
	    }
	  else
	    let tmppreds = []
	    in let tmpbdd = PredTable.bddOne
	    in let tmpbdd2 = s_get_bdd_for_pred_list lpredlist _H_cmd  tmpbdd  predlist
	    in
	      {
		Region.lelement = rlelement;
		Region.preds = tmppreds;
		Region.bdd = tmpbdd2
	      }
      else
	match cmd.C_SD.Command.code with 
	    C_SD.Command.FunctionCall fcallExp 
	  | C_SD.Command.GuardedFunctionCall (_, fcallExp) -> (* TBD: RUPAK: treat as an assume followed by a functioncall *)
	      let _ = Message.msg_string Message.Debug "NEW CDP" in
	      let caller = abst_data_reg in
	      let (fname, _ , _ ) = (C_SD.deconstructFunCall fcallExp) in
	      let asgn_block =
		{cmd with
		   C_SD.Command.code = C_SD.Command.Block (Absutil.asgn_of_funcall
							     (true,false) None fcallExp)}
	      in
	      let (_lpredlist, rlelement) = DataLattice.post caller.Region.lelement edge (query_fn caller.Region.bdd)
	      in let lpredlist = sane_pred_list _lpredlist
	      in let region =
		  (if (List.length lpredlist = 0)
		   then 
		     {
		       Region.lelement = rlelement;
      		       Region.preds = caller.Region.preds;
      		       Region.bdd =  _H_cmd asgn_block (caller.Region.bdd) predlist;
		     }
		   else
		     let tmppreds = caller.Region.preds
		     in let tmpbdd = _H_cmd asgn_block (caller.Region.bdd) predlist
		     in let tmpbdd2 = s_get_bdd_for_pred_list lpredlist _H_cmd tmpbdd predlist
		     in
		       {
			 Region.lelement = rlelement;
			 Region.preds = tmppreds;
			 Region.bdd = tmpbdd2
		       }
		  )
	      in
		if not (C_SD.is_defined fname) then
		  region
		else
		  begin
		    Message.msg_string Message.Debug 
		      ("Restricting region across call to: "^fname);
		    Message.msg_printer Message.Debug Region.print_abstract_data_region region;
		    let region = Region.restrictAbstractData (Region.locals_and_globals fname) region in
		      Message.msg_string Message.Debug "-->";
		      Message.msg_printer Message.Debug Region.print_abstract_data_region  region;
		      region
		  end
		    (* What was just done for the function call ... *) 
		    (* Push stuff onto the stack: ... in the rest of post 
		       1. edge, on return, the location will be
		       the other end of the edge ...
		       2. if edge was a y = f(x) then post on return reflects this *)
		    (* Go through list of predicates and compute "copies" *)
		    (* Compute BDD of reg ^_i (pred_i = copy_pred_i) *)
		    (* Start location = start location of function *)


	  | C_SD.Command.Skip ->  abst_data_reg
	  | C_SD.Command.SymbolicHook _ -> abst_data_reg 
	  | C_SD.Command.Block [] ->  abst_data_reg
	  | C_SD.Command.Block l  -> 
	      if ( not (List.exists (function x -> match x with C_SD.Command.Return e -> true | _ -> false) l))
	      then 
		let _ = (Absutil.reset_ret_pointers ()) in
		let (_lpredlist, rlelement) = DataLattice.post abst_data_reg.Region.lelement edge (query_fn abst_data_reg.Region.bdd)
		in let lpredlist = sane_pred_list _lpredlist
		in if (List.length lpredlist = 0)
		  then 
		    { 
	              Region.lelement = rlelement;
      		      Region.preds = abst_data_reg.Region.preds;
      		      Region.bdd =   _H  (abst_data_reg.Region.bdd) predlist (*abst_data_reg.Region.preds*);
		    }
		  else
		    let tmppreds = abst_data_reg.Region.preds
		    in let tmpbdd =  _H  (abst_data_reg.Region.bdd) predlist
		    in let tmpbdd2 = s_get_bdd_for_pred_list lpredlist _H_cmd  tmpbdd  predlist
		    in 
		    {
		      Region.lelement = rlelement;
		      Region.preds = tmppreds;
		      Region.bdd = tmpbdd2
		    }
	      else 
		let (targ, fname, retloc ,subs) = lastTarget () in
		  (* Jhala: the t.o.s. is passed using last_loc_pointer  reg.Region.stack *) 
		let _ = Absutil.set_ret_pointers targ subs (C_SD.get_location_fname retloc) in
		let (newPreds,delPreds) = removeFunPreds fname abst_data_reg.Region.preds in
		let _ = Message.msg_string Message.Debug "Return: Calling _H" in
		let newBdd = (_H  (abst_data_reg.Region.bdd) predlist (*abst_data_reg.Region.preds*)) in
		let _ = Message.msg_string Message.Debug "Return: Back from _H" in
		let _predToVar i =  -1 in (* JHALA: To be done ... ??? *)
		let ex_newBdd = newBdd in
		  (*try CaddieBdd.exists (Bdd.fromList bddManager (List.map
		    _predToVar
		    delPreds)) newBdd 
		    with _ -> newBdd (* Jhala: TBD... how to go from pred to
		    bdd variable (int) ? wrapper needs
		    work *)*)
		  
		let _ = Absutil.reset_ret_pointers () in
		let (_lpredlist, rlelement) = DataLattice.post abst_data_reg.Region.lelement edge (query_fn ex_newBdd)
		in let lpredlist = sane_pred_list _lpredlist
		in if (List.length lpredlist = 0) 
		  then
		    { 
		      Region.lelement = rlelement;
      		      Region.preds = newPreds@delPreds;
      		      Region.bdd =  ex_newBdd;
		    }
		  else
		    let tmppreds = newPreds@delPreds
		    in let tmpbdd = ex_newBdd
		    in let tmpbdd2 = s_get_bdd_for_pred_list lpredlist _H_cmd tmpbdd (newPreds@delPreds)
		    in 
		    {
		      Region.lelement = rlelement;
		      Region.preds = tmppreds;
		      Region.bdd = tmpbdd2
		    }
	  | C_SD.Command.Pred e -> 
	      
	      if (e = Predicate.True) then ( Message.msg_string Message.Debug "thinks its true!";abst_data_reg)
	      else 
		begin
		  assume_pred_ref := Some(e);
		  Message.msg_string Message.Debug "calling _H";
		  let (_lpredlist, rlelement) = DataLattice.post abst_data_reg.Region.lelement edge (query_fn abst_data_reg.Region.bdd)
		  in let lpredlist = sane_pred_list _lpredlist
		  in 
		  let retReg =
		    (if (List.length lpredlist = 0)
		     then
		       { 
			 Region.lelement = rlelement;
			 Region.bdd = _H (abst_data_reg.Region.bdd) predlist (*abst_data_reg.Region.preds*) ;
			 Region.preds = abst_data_reg.Region.preds ;
		       }
		     else
		       let tmppreds = abst_data_reg.Region.preds
		       in let tmpbdd = _H (abst_data_reg.Region.bdd) predlist (*abst_data_reg.Region.preds*)
		       in let tmpbdd2 = s_get_bdd_for_pred_list lpredlist _H_cmd tmpbdd predlist
		       in 
		       {
			 Region.lelement = rlelement;
			 Region.preds = tmppreds;
			 Region.bdd = tmpbdd2
		       }
		    )
		  in  
		    assume_pred_ref := None;          
		    retReg
		end
	  | C_SD.Command.Phi (sym, flag) -> 
	      let e = phi_command_to_pred sym flag phi_table in
		(* Now treat just like C_SD.Command.Pred e *)
		if (e = Predicate.True) then abst_data_reg
		else 
		  begin
		    assume_pred_ref := Some(e);
		    let retReg = 
		      {
			Region.lelement = failwith "Do not handle Phi" ;
			Region.bdd = _H (abst_data_reg.Region.bdd) predlist (* abst_data_reg.Region.preds*) ;
			Region.preds = abst_data_reg.Region.preds ;
		      }
		    in  
		      assume_pred_ref := None;          
		      retReg
		  end
	  | C_SD.Command.Havoc sym -> 
	      let new_var = 
		match getNewTmpVar (Expression.Symbol "__foo") with 
		    Expression.Chlval (Expression.Symbol x, s) -> Expression.Lval (Expression.Symbol (s^x))
		  | _ -> failwith "Abstraction.concrete_data_pre: Havoc: TBD"
	      in
	      let assign = Expression.Assignment(Expression.Assign, sym, new_var) in
	      let stmt = C_SD.Command.Expr assign in
	      let cmd' = {cmd with C_SD.Command.code = C_SD.Command.Block [stmt]} in
		abstract_data_post_graf_saidi abst_data_reg cmd' edge post_location phi_table
	  | C_SD.Command.HavocAll -> failwith "Abstraction.abstract_data_post_graf_Saidi: HavocAll: TBD"       
	      

  (* Starting the definition of PRE *)


  let setWPCapRegion reg =
    let conc_reg = Region.concretize reg in
      wpCapRegion :=
      match conc_reg with 
          Region.Empty -> Predicate.False
        | Region.Atomic atomic_reg -> 
            begin
	      match atomic_reg.Region.data with
                  Region.Concrete foo -> foo.Region.pred
                | _ -> failwith ("bad call to setWPCapRegion")
	    end
	| Region.ParallelAtomic p_atomic_reg ->
	    begin
	      match p_atomic_reg.Region.p_data with
		  Region.Concrete foo -> foo.Region.pred
		| _ -> failwith ("bad call to setWPCapRegion")
	    end
	| _ -> failwith ("bad call to setWPCapRegion: Strange region type")

  let resetWPCapRegion () = 
    wpCapRegion := Predicate.True
      
  (* the symbolic pre function: (approximate) weakest pre-condition *)
  let pre_template subs_flag sat_flag reg op phi_table =
    Message.msg_string Message.Debug "In pre -- arguments are:" ;
    Message.msg_printer Message.Debug Region.print reg ;
    Message.msg_printer Message.Debug Operation.print op ;
    Message.msg_string Message.Debug (if subs_flag then "Substitute \n" else "Explicit (no substitute) \n");
    Message.msg_string Message.Debug (if sat_flag then "sat-filter-on \n" else "sat-filter-off \n");
    match reg with
        Region.Empty ->
          Region.bot
      | Region.Atomic atomic_reg ->
          if ((C_SD.location_coords atomic_reg.Region.location <> C_SD.location_coords (C_SD.get_target op)) 
		&& (not (isReturn op)) && (not (isFunCall op))) then
            (* Jhala: This is only if this is not a return op *)
            begin
              Message.msg_string Message.Error "WARNING: pre: blast internal error suspected" ;
              Message.msg_string Message.Error "              operation do not match this region's location" ;
	      failwith "location mismatch";
              Region.bot
            end
          else
            begin
              let new_stack =
                if (isFunCall op) then
                  begin 
                    match (C_SD.get_command op).C_SD.Command.code with
			
			C_SD.Command.FunctionCall fcexp -> 
			  let function_is_defined fcexp = 
			    match fcexp with
				Expression.FunctionCall(Expression.Lval (Expression.Symbol(fname)), _) -> 
				  C_SD.is_defined fname
			      | Expression.Assignment (_, _, Expression.FunctionCall (Expression.Lval (Expression.Symbol fname), _)) -> 
				  C_SD.is_defined fname
			      | _ -> 
				  (* Rupak: this case happens if there is a function pointer call.
				     Previously we failed here with:
				     failwith ("Bad function call :"^(Expression.toString fcexp))
				     However we can punt this by assuming that the function pointer call
				     does not call a known function, and continue.
				     
				     Do this only if nofp is set.
				  *)
				  if (Options.getValueOfBool "nofp") then
				    false
				  else
				    failwith ("pre_template: Bad function call (Possible function pointer?):"^(Expression.toString fcexp))
			  in
			    if (not (function_is_defined fcexp)) then
			      atomic_reg.Region.stack
			    else 
			      begin
				let (_fname,_,target) = C_SD.deconstructFunCall fcexp in
				  if (_fname = C_SD.__NotImplementedFunctionName) 
				  then 
				    atomic_reg.Region.stack
				  else   
				    if (_fname <> C_SD.__BLAST_DummyFunctionName) then 
				      try (Region.CallStack (List.tl (Region.list_of_stack atomic_reg.Region.stack))) 
				      with e -> (print_string (Printexc.to_string e);failwith "Incorrect callstack type !!")
				    else 
				      atomic_reg.Region.stack (* was a skip *)
					(*begin
					  let vars_in_support = List.flatten (List.map getVarExps_deep (preds_of_region reg)) in   
					  if (not (List.mem target vars_in_support)) then 
					  atomic_reg.Region.stack 
					  else failwith ("Undeclared important function"^_fname)
					  end*)
			      end
                      |    _ -> failwith "This can never happen: isFuncall checked that this was a function call!"
                  end
                else
                  if (isReturn op)
                  then
                    begin
                      Message.msg_string Message.Debug "Pre: Hit isReturn\n";
                      (* recall that we put the call points in the stack *)
                      (* we assume that a function call edge is the only edge
			 going to its target location *)
                      Message.msg_string Message.Debug "Ingoing edges : ";
                      List.iter (fun a -> Message.msg_string Message.Debug (Operation.to_string a)) (C_SD.get_ingoing_edges atomic_reg.Region.location);
                      Message.msg_string Message.Debug "\nNow matching...\n";
                      let rec get_fn_call_edge lst =
			match lst with
                            a::rest -> 
                              begin
				if (isFunCall a) then
				  (* check there are no other function calls *)
				  if (List.exists isFunCall rest) then
                                    failwith "pre: More than one function call in incoming edges!"
				  else a
				else
				  get_fn_call_edge rest
                              end
			  | [] -> failwith ("pre: No corresponding function call found")
                      in
                      let callLoc = (C_SD.get_source (get_fn_call_edge (C_SD.get_ingoing_edges atomic_reg.Region.location))) in
			last_loc_pointer := Some(callLoc);
			let (rt,_,retloc,subs) = lastTarget () in
			  Absutil.set_ret_pointers rt subs (C_SD.get_location_fname retloc);
			  last_loc_pointer := None; (* RJ: why the hell do we need this silly last_loc_pointer! *)
                          ( Region.CallStack
			      (callLoc::(Region.list_of_stack atomic_reg.Region.stack)))
			    
                    end
                  else
                    atomic_reg.Region.stack
              in
              let new_location = (C_SD.get_source op)
              and new_data =
                match atomic_reg.Region.data with
                    Region.Abstract abst_data_reg ->
                      Region.Abstract (abstract_data_pre abst_data_reg (C_SD.get_command op) op phi_table)
                  | Region.Concrete conc_data_reg ->
                      Region.Concrete (concrete_data_pre subs_flag sat_flag conc_data_reg (C_SD.get_command op) (op) phi_table)
              in
              Absutil.reset_ret_pointers ();
              Region.Atomic { Region.location = new_location ;
                              Region.data = new_data ;
                              Region.stack = new_stack ; }
		  
            end
      |	Region.ParallelAtomic p_atomic_reg ->
	  Region.bot
	    (* i am here *)
      | Region.Union _ ->
          failwith "pre: argument is a union: unimplemented"
      | Region.ParallelUnion _ ->
          failwith "pre: argument is a parallel union: unimplemented"


  (* Jhala: am setting sat flag to false for now ... *)
  let pre (* reg op phi_table *) = pre_template true false (* reg op phi_table *)
  let spre (* reg op phi_table *) = pre_template false false (* reg op phi_table *)
				      
				      
  let abstract_data_post abst_data_reg  cmd edge post_location phi_table = 
    match (Options.getValueOfString "post") with
        "H" -> (* Stats.time "Post H: " *) (abstract_data_post_das_dill abst_data_reg cmd) edge post_location phi_table
      | _  -> (* Stats.time "Post Graf-Saidi"*) (abstract_data_post_graf_saidi abst_data_reg  cmd) edge post_location phi_table 


  let abstract_predicate pred pred_list = 
    Message.msg_string Message.Debug "abstract_predicate called" ;
    assume_pred_ref := Some(pred);
    let abstract_bdd = postBdd bddOne (List.map (function i -> (i,getPred i,Predicate.negate (getPred i),true)) pred_list) [] 
    in
      assume_pred_ref := None;
      Region.Abstract { Region.lelement = DataLattice.top (* RUPAK: NEED TO DO:
				DataLattice.post DataLattice.top ... *) ; 
			Region.bdd = abstract_bdd; Region.preds = pred_list}


  let scope_kill data_region fname = 
    Message.msg_string Message.Debug "Inside scope_kill:\n";
    match data_region with
	Region.Abstract(abst_data_reg) ->
	  begin
	    Message.msg_printer Message.Debug Region.print_abstract_data_region abst_data_reg;
	    let _kill_ff x =  
	      let predVars = getVarExps_deep (getPred x) in
		List.exists (fun x -> Misc.is_substring x ("@"^fname)) predVars
	    in
	    let new_bdd = hg_exists _kill_ff abst_data_reg.Region.bdd in
	    let ret_abst_data_reg  = { Region.lelement = abst_data_reg.Region.lelement ;
				       Region.bdd = new_bdd; Region.preds = abst_data_reg.Region.preds} in
	      Message.msg_string Message.Debug "Returning from kill_data_region with :\n";
	      Message.msg_printer Message.Debug Region.print_abstract_data_region ret_abst_data_reg;
	      Region.Abstract(ret_abst_data_reg)
          end
      | _ -> 
	  begin
	    Message.msg_string Message.Debug "bad call to kill"; 
	    data_region
	  end

  (* the symbolic post function: (approximate) strongest post-condition *)
  (* TODO: if checking races, add havocs *)
  let rec post reg op phi_table =
    Message.msg_string Message.Debug "In post -- arguments are:" ;
    Message.msg_printer Message.Debug Region.print reg ;
    Message.msg_printer Message.Debug Operation.print op ;
    let _ = Hashtbl.replace statVisitedFunctionTable (C_SD.get_location_fname (C_SD.get_target op)) true 
    in
    let preds_of_region reg = 
      match reg with
	  Region.Atomic atr -> 
            begin     
              match atr.Region.data with 
		  Region.Abstract abst_data_reg -> List.map getPred abst_data_reg.Region.preds
		| _ -> failwith "Bad call to preds_of_region"
            end
	| Region.ParallelAtomic p_atr ->
            begin     
              match p_atr.Region.p_data with 
		  Region.Abstract abst_data_reg -> List.map getPred abst_data_reg.Region.preds
		| _ -> failwith "Bad call to preds_of_region"
            end
	| _ -> failwith "Bad call to preds_of_region"
    in
      match reg with
	  Region.Empty -> Region.bot
	| Region.Atomic atomic_reg ->
	    begin
              try 
		if (not (Region.check_locations_eq atomic_reg.Region.location  (C_SD.get_source op))) then
		  begin
		    Message.msg_string Message.Error "WARNING: post: blast internal error suspected" ;
		    Message.msg_string Message.Error "               operation do not match this region's location" ;
		    
		    Region.bot
		  end
		else
		  begin
                    try 

		      let callStack = Region.list_of_stack atomic_reg.Region.stack in
	              let tos_fname = 
			if (not (callStack = [])) 
			then (last_loc_pointer := Some (List.hd callStack);C_SD.get_location_fname (C_SD.get_target op)) 
			else ( last_loc_pointer := None; C_SD.__BLAST_DummyFunctionName) 
		      in
		      let tos_2_fname = 
			if (List.length callStack >= 1) then (C_SD.get_location_fname (List.hd callStack)) else C_SD.__BLAST_DummyFunctionName 
		      in
		      let _ = Message.msg_string Message.Debug (Printf.sprintf "tos_fname: %s, tos_2_fname: %s" tos_fname tos_2_fname) in

 		      (* So last_loc_pointer points to the top of the stack at all times or to errorloc Rif the stack is empty *)
                      let (new_location, new_stack) = 
			if (isFunCall op) then
                          begin
                            match (C_SD.get_command op).C_SD.Command.code with
				C_SD.Command.FunctionCall fcexp  
                              | C_SD.Command.GuardedFunctionCall (_, fcexp) ->
				  begin
				    try
				      let (_fname,_,target) = C_SD.deconstructFunCall fcexp in
				      let _ = stackInfo := Some (_fname,tos_fname,true) in
					if (_fname = C_SD.__NotImplementedFunctionName) 
					then 
					  ((C_SD.get_target op), atomic_reg.Region.stack)
					else   
					  begin
					    if ( C_SD.is_defined _fname) then 
					      (let newStack =
						 (C_SD.get_source op) :: callStack
					       in
						 ((C_SD.lookup_entry_location _fname),
						  Region.CallStack newStack))
					    else				 
					      ((C_SD.get_target op), atomic_reg.Region.stack) (* is a skip *)
					  end
				    with C_SD.NoSuchFunctionException _ -> failwith "AAHA"
				  end (*try*)
                              | _ -> failwith "what the (post) !"
				  
                          end
			else 
			  try 
			    if (isReturn op) then 
			      begin
				let _ = stackInfo := Some (tos_fname,tos_2_fname,true) in
				let (ret, _, nextLoc,subs) = lastTarget () in
				  Absutil.set_ret_pointers ret subs (C_SD.get_location_fname nextLoc);
				  let newStack =
				    List.tl callStack
				  in
				    (nextLoc,(Region.CallStack newStack) ) 
			      end
			    else 
			      begin
				let _ = stackInfo := Some (tos_fname,tos_2_fname,false) in
				  (C_SD.get_target op,atomic_reg.Region.stack) 
			      end
			  with Failure ("tl") -> 
			    begin
			      let _ = stackInfo := Some (tos_fname,tos_2_fname,false) in
				(C_SD.get_target op, atomic_reg.Region.stack ) 
			    end
		      in 
		      let get_new_data () () =
			let _new_data =
			  try 
			    match atomic_reg.Region.data with
				Region.Abstract abst_data_reg ->
				  Region.Abstract (abstract_data_post abst_data_reg (C_SD.get_command op) op new_location phi_table)
			      | Region.Concrete conc_data_reg ->
				  Region.Concrete 
				    (concrete_data_post conc_data_reg (C_SD.get_command op) op phi_table)
			  with C_SD.NoSuchFunctionException f -> failwith "Foo"
			in
			  (* Adding scoping hack *)
			  (* Adding kill *)
			  match ((Options.getValueOfBool "scope") && (isReturn op) (*&& false*) ) with
			      true  -> 
				begin
				  Message.msg_printer Message.Debug C_SD.print_location (C_SD.get_source op);
				  Message.msg_printer Message.Debug C_SD.print_location new_location;
				  let (_,arg,_,_) = lastTarget () in
				    Stats.time "scope_kill" (scope_kill _new_data) arg 
				end
			    | false -> _new_data
		      in
		      let new_data = Stats.time "get_new_data" (get_new_data ()) () in
		      let _ = Absutil.reset_ret_pointers () in 
			Message.msg_string Message.Debug "Writing back post region" ;
			begin
			  Region.Atomic { Region.location = new_location ;
					  Region.data = new_data ;
					  Region.stack = new_stack ; }
			end
		    with
			Failure("Empty Call Stack") -> Region.bot
		  end
	      with 
		  C_SD.NoSuchFunctionException f -> (* treat it as a skip *)
		    Message.msg_string Message.Minor ("Unexpected NoSuchFunctionException "^f);
		    (* THIS SHOULD BE COMMENTED OUT IF UNDEFINED FUNCTIONS ARE TO BE AVOIDED *)
		    (* failwith ("Unexpected NoSuchFunctionException "^f); *)
		    Region.Atomic
                      { Region.location = C_SD.get_target op;
			Region.data = (failwith "Here" ;atomic_reg.Region.data) ;
			Region.stack = atomic_reg.Region.stack;
		      } 
		      (* COMMENT OUT UPTO HERE *)

            end
	| Region.ParallelAtomic p_atomic_reg ->
	    failwith "post: argument is a parallel atomic: unimplemented"
	| Region.Union _ ->
            failwith "post: argument is a union: unimplemented"
	| Region.ParallelUnion _ ->
            failwith "post: argument is a parallel union: unimplemented"




  let extract_pred reg = 
    match reg with
	Region.Atomic r -> 
	  begin
	    match r.Region.data with 
		Region.Concrete x -> x.Region.pred 
	      | Region.Abstract a -> convertBddToPred a.Region.bdd
		  (*
		    (Message.msg_string Message.Error "Bad call to extract_pred"; failwith ("Bad call to extract pred"))*)
	  end
      | _ -> (Message.msg_string Message.Error "Bad call to extract_pred"; failwith ("Bad call to extract pred"))

  let join_abstract_data d1 d2 =
    { Region.lelement = DataLattice.meet d1.Region.lelement d2.Region.lelement ;
      Region.bdd = CaddieBdd.bddAnd d1.Region.bdd d2.Region.bdd ;
      Region.preds = Misc.union d1.Region.preds d2.Region.preds }

      
  let modvar_table = Hashtbl.create 31 (* filled with the globals that the callee may change *)




  (*******************************************************************************************)

  let statsMaxNumberOfPredsActive = ref 0
				      (* Keeps track of the number of predicates active
                                         at any point, discounting scope
                                      *)

  (* do_focus : get new predicates to add to the current set *)
  exception AnnotationException  
  exception NoNewPredicatesException
  exception RecursionException
    

  (* this is a highly stateful function! *)
  let last_no_of_preds = ref 0
  let predicate_fixpoint_reached () = 
    if (!last_no_of_preds = List.length !tableOfPreds) then 
      begin
	(* printTableOfPreds (); *)
	true
      end
    else
      begin
	last_no_of_preds := List.length !tableOfPreds;
	false
      end

	
  let do_focus abst_data_reg empty_conc_data_reg =
    Message.msg_string Message.Debug "In do_focus -- arguments are:" ;
    Message.msg_printer Message.Debug Region.print_abstract_data_region abst_data_reg ;
    Message.msg_printer Message.Debug Region.print_concrete_data_region empty_conc_data_reg ;
    let emptyPred = empty_conc_data_reg.Region.pred in
    let allPreds = extractUsefulPredicates emptyPred in
      (*let _= 
	begin
	Message.msg_string Message.Error "Candidate preds";
	List.iter (fun p -> Message.msg_string Message.Error (Predicate.toString p) ) allPreds
	end
	in*)
      (* if this round did not add any new predicates, then we should restart *)
      (*
	let _ = Misc.write_to_file ".newpred" " " in     
      *)
    let oldnum = getCurrentIndex () in
    let added_preds = List.map addPred allPreds in
    let newnum = getCurrentIndex () in
      if (oldnum = newnum) then
	begin
          Message.msg_string Message.Debug "Restarting..." ;
	  Options.setValueOfBool "localrestart" true ;
        end
      else
        begin
          Message.msg_string Message.Debug "At least one predicate added ..." ;
	  Options.setValueOfBool "localrestart" false ;
        end ;
      
      let newpredlist = Misc.compact (Misc.difference (added_preds) (-1 :: abst_data_reg.Region.preds)) 
                          (* This is not correct ... as there may be a descendant along the error path 
                             that does not have a certain pred that is in this list which the 
                             ancestor ie the current data region has... if you are doing PATH that is ...*)
      in
        if (newpredlist = []) then (Message.msg_string Message.Error "No new predicates ! Unexpected ...";
                                    printTableOfPreds () ;
				    if (Options.getValueOfInt "predH" < 6) then
        			      raise NoNewPredicatesException;
				   );
        begin  
	  (* let _ =  Message.msg_string Message.Debug ("NEW PREDICATES FOUND \n") in *)
          let _listOfAllActivePreds = abst_data_reg.Region.preds @ newpredlist in
            ignore (if (List.length _listOfAllActivePreds) > !statsMaxNumberOfPredsActive then 
                      begin
			statsMaxNumberOfPredsActive := List.length _listOfAllActivePreds;
			Message.msg_string Message.Debug ("Size of pred set: "^(string_of_int (List.length !tableOfPreds)) );
			Message.msg_string Message.Debug "Active Predicates"; 
			List.iter (function i -> Message.msg_string Message.Debug  (Predicate.toString (getPred i)))  
                          _listOfAllActivePreds
                      end);
            { Region.lelement = abst_data_reg.Region.lelement ;
	      Region.bdd = abst_data_reg.Region.bdd ;
              Region.preds = _listOfAllActivePreds ; }
        end


  let replace_data_region reg new_data_reg =
    match reg with 
	Region.Atomic _reg ->
          Region.Atomic 
	  { Region.location = _reg.Region.location;
            Region.data = new_data_reg;
            Region.stack = _reg.Region.stack;
          }
      | _ -> failwith ("bad call to replace_data_region")


  (* ******************* ERROR TRACE FILTER ****************************** *)

  (* the following procedure takes a trace and spits out that subtrace of it which is used to 
     get a contradiction -- it is to help in "manual" counterexample analysis *)

  let  trace_filter region_list op_list = 
    (* the above is the unsatisfiable list of regions and operations -- arranged in "past --> present" order*)
    let reg_list_r = List.rev region_list in
    let op_list_r = List.rev op_list in
    let process_op_reg (phi , map) (op, reg) = () in
      ()
	
  (* ******************* find_all_preds ********************************** *)


  (* this function will look at ALL subtraces of a given trace and find all predicates coming from it *)



  (* again oplist is in increasing order past --> present *)
  (* r1 -> o1 -> r2 -> o2 -> ... -> on -> rn *)

  let extract_abs_data_reg reg = 
    match reg with
	Region.Atomic atomic_reg ->
	  begin
	    match atomic_reg.Region.data with
		Region.Abstract ar -> ar
	      | _ -> failwith "conc_dr given to extract_abs_data_reg"
	  end
      | _ -> failwith "union region given to extract_abs_data_reg"

  let quantify_and_extract_abs_data_reg = extract_abs_data_reg (* RJ: In short, I'm being lazy i.e. indolent *)


  let insert_data_reg reg dr = 
    match reg with
	Region.Atomic atomic_reg ->
	  begin
	    Region.Atomic 
	      { Region.data = dr;
		Region.location = atomic_reg.Region.location;
		Region.stack = atomic_reg.Region.stack;
	      }
	  end
      | _ -> failwith "bad call to insert_data_reg"


  (* this is called only in "block" mode ... and with noforget *)
  let focus_stamp reg =
    let abst_data_reg = extract_abs_data_reg reg in
    let cur = getCurrentIndex () in
    let present_prec = 
      try List.hd abst_data_reg.Region.preds 
      with _ -> -1 
    in
      if (present_prec = cur && not (refine_check ())) then
	begin
	  dump_abs ();
	  raise NoNewPredicatesException;
	end
      else
	begin
	  insert_data_reg reg 
	    (Region.Abstract 
	       { Region.lelement = abst_data_reg.Region.lelement; 
		 Region.bdd = abst_data_reg.Region.bdd;
		 Region.preds = cur :: abst_data_reg.Region.preds (*[cur]*);
	       })
	end




  (* Cheap hack: go through all the predicates along oplist and add them to the list of predicates. *)

  let find_all_predicates_cheap oplist =
    let rec _more lst accum =
      match lst with
	  [] -> accum
	| a :: rest ->
            begin
              match (C_SD.get_command a).C_SD.Command.code with
		  C_SD.Command.Pred e ->
                    if e = Predicate.True then _more rest accum
                    else 
                      begin
			let ee = match e with Predicate.Not e' -> e' | Predicate.Atom _ -> e | _ -> failwith "nonatomic predicate on path!" in
			  _more rest (ee :: accum)
                      end
		| _ -> _more rest accum
            end
    in
    let all_preds_in_path = _more oplist [] in
    let newpredlist = Misc.compact 
			(Misc.difference 
                           (List.map addPred all_preds_in_path) 
                           (-1 :: ( PredTable.get_all_pred_indexes ()))) 
    in 
    let number = List.length newpredlist in
      Message.msg_string Message.Debug ("find_all_predicates_cheap found "^
					(string_of_int number)^" predicates.") ;
      number


  let find_all_predicates reglist oplist = 
    let list_of_predicates = ref [] in
    let oplist_r = List.rev oplist in
    let reglist_r = List.rev reglist in

    let rec _proc_sub seed ol = 
      (* take wp of seed wrt (ol.head,rl.head) -- query for unsat if unsat take preds... quit and break out ..
	 else if sat, then take region and recur on it with ol.tail, rl.tail *)
      match ol with
	  op::t -> 
	    begin 
	      let emptytable = Hashtbl.create 2 in
	      let pre_reg = pre seed op emptytable in
		if (Region.is_empty pre_reg) then 
		  begin
		    let allpreds = extractUsefulPredicates (extract_pred pre_reg) in
		      Message.msg_string Message.Debug ("Learning new predicates: ");
		      List.iter (fun x -> list_of_predicates := x::(!list_of_predicates); 
				   Message.msg_string Message.Debug ((Predicate.toString x)^" , ")) allpreds;
		      Message.msg_string Message.Debug ("From region: "^(Predicate.toString (extract_pred pre_reg)))
			(* do we or do we not recur on pre_reg ? as of now, NO *)
		  end
		else 
		  _proc_sub pre_reg t
	    end
	| [] -> ()
    in

    let rec process_subtrace ol_r rl_r = 
      match rl_r with
	  h::h1::t ->
	    let seed_reg = (replace_data_region h (Region.Concrete {Region.pred = Predicate.True})) in
	      _proc_sub seed_reg ol_r; 
	      process_subtrace (List.tl ol_r) (h1::t)
	| _ -> ()
    in
      (*
	let current_preds = !tableOfPreds in
	let current_pred_vars = 
	List.map (fun x -> Expression.Symbol x)
	(Misc.compact (List.flatten (List.map (fun x -> getVarExps_deep (Misc.snd3 x)) current_preds)))
	in
	let make_predicate v c = Predicate.Atom (Expression.Binary (Expression.Eq,v,c)) in
	let process_var v = 
	try 
	List.map (make_predicate v) (Hashtbl.find lval_constant_table v)
	with
	Not_found -> failwith ("lval not in lval_constant_table: "^(Expression.toString v))
	in
	let new_const_preds = List.flatten (List.map process_var current_pred_vars) in
	let _ = List.map addPred new_const_preds in
	let _ = 
	if (List.length !tableOfPreds = List.length current_preds) then
	failwith (" No new constant preds learnt! ")
	else 
	begin
	Message.msg_string Message.Debug "new_const_predicates found: \n";
	List.iter (fun x -> Message.msg_string Message.Normal (Predicate.toString x)) new_const_preds;
	Message.msg_string Message.Normal ("# new preds = "^(string_of_int (List.length new_const_preds )));
	end
      *)  
      (* this is the point where the main function is called ! *)
      (*  The stuff in this needs to be worked on ... its *very* slow ... 
	  let _ = process_subtrace oplist_r reglist_r in
	  let newpredlist = Misc.compact (Misc.difference (List.map addPred !list_of_predicates) (-1 :: ( List.map Misc.fst3 !tableOfPreds))) in
	  Message.msg_string Message.Normal "find_all_predicates found: \n";
	  List.iter (fun x -> Message.msg_string Message.Normal (Predicate.toString x)) !list_of_predicates;
	  Message.msg_string Message.Normal ("# new preds = "^(string_of_int (List.length newpredlist)));
	  if (newpredlist = [] ) then failwith ("find_all_predicates also fails to get new predicates !!") *)

      if find_all_predicates_cheap oplist == 0 then
	begin
	  let _ = process_subtrace oplist_r reglist_r in
	  let newpredlist = Misc.compact (Misc.difference (List.map addPred !list_of_predicates) (-1 :: ( PredTable.get_all_pred_indexes ()))) in
	    Message.msg_string Message.Debug "find_all_predicates found: \n";
	    List.iter (fun x -> Message.msg_string Message.Debug (Predicate.toString x)) !list_of_predicates;
	    Message.msg_string Message.Debug ("# new preds = "^(string_of_int (List.length newpredlist)));
	    if (newpredlist = [] ) then failwith ("find_all_predicates also fails to get new predicates !!")
	end


	  (**************************************************************************************)

	  (* this is for m4 *)
	  include(abs-boolean.ml)
	  include(abs-include.ml)
	  include(abs-focimc.ml)
	  (**************************************************************************************)

  (* the refinment operator required by the ABSTRACTION signature *)
  let focus reg_to_refine empty_reg  =
    Message.msg_string Message.Debug "In focus -- arguments are:" ;
    Message.msg_printer Message.Debug Region.print reg_to_refine ;
    Message.msg_printer Message.Debug Region.print empty_reg ;

    match (reg_to_refine, empty_reg) with
        (Region.Atomic abst_atomic_reg, Region.Atomic empty_conc_atomic_reg) ->
          begin
            try 
              match (abst_atomic_reg.Region.data, empty_conc_atomic_reg.Region.data) with
                  (Region.Abstract abst_data_reg, Region.Concrete empty_conc_data_reg) ->
                    Region.Atomic { Region.location = abst_atomic_reg.Region.location ;
                                    Region.data = Region. Abstract (Stats.time "do_focus" (do_focus abst_data_reg) empty_conc_data_reg) ;
                                    Region.stack = abst_atomic_reg.Region.stack ; }
		| _ -> failwith "focus: input regions should be abstract and concrete in this order"
            with NoNewPredicatesException ->
              ( Message.msg_string Message.Debug "In focus -- arguments are:" ;
                Message.msg_printer Message.Debug Region.print reg_to_refine ;
                Message.msg_printer Message.Debug Region.print empty_reg ;
                Message.msg_string Message.Error "No new predicates found. Aborting." ;
                raise NoNewPredicatesException
              )
          end
      | _ -> failwith "focus: arguments should be atomic regions"


  (* the following function takes a concrete data region and a list of predicates and returns an 
     abstract region corresponding to it. Cousot would call it alpha *)


  let abstract atomic_conc_reg pred_list =
    let assumption  =
      match atomic_conc_reg.Region.data with
	  Region.Concrete cr -> cr.Region.pred
	| _ -> failwith ("bad call to atomic_conc_reg")
    in
      assume_pred_ref := Some(assumption);
      let abstract_bdd = postBdd bddOne (List.map (function i -> (i,getPred i,Predicate.negate (getPred i),true)) pred_list) [] 
      in
	assume_pred_ref := None;
	Region.Abstract { Region.lelement = DataLattice.top ; (* I assume I know nothing about the lattice state
								 Each lattice should implement an
								 alpha that takes a concrete predicate and
								 returns a lattice element
							      *) 
			  Region.bdd = abstract_bdd; Region.preds = pred_list}
	  


  let intersect_with_abstraction reg1 reg2 =
    (* reg1 : abstract, reg2 : concrete or abstract, returns reg1 cap reg2 : abstract *)
    let reg2' = precise reg2 in
      match (reg1,reg2') with
	  (Region.Atomic atr1, Region.Atomic atr2) ->
            begin
	      Message.msg_string Message.Debug ("intersect_with_abstraction:\n" ^ 
						"    reg1 = " ^ Region.to_string reg1 ^
						"    reg2 = " ^ Region.to_string reg2);
              match (atr1.Region.data,atr2.Region.data) with
		  (Region.Abstract abst_1,Region.Concrete conc_2) ->
		    let preds = abst_1.Region.preds in
		    let abst_2 = abstract atr2 preds in
                    let newreg2 = Region.Atomic { Region.location = atr2.Region.location;
                                                  Region.data = abst_2;
                                                  Region.stack = atr2.Region.stack; }
                    in
                      Region.cap reg1 newreg2
		| _ -> failwith "bad call to intersect_with_abstraction"
            end
	| _ -> failwith "bad call to intersect_with_abstraction 2"
	    
  let print_abs_stats () = 
    Message.msg_string Message.Normal ("TP Queries:" ^ (string_of_int !statActualQueriesForPost));
    Message.msg_string Message.Normal ("Reached Cubes:" ^ (string_of_int !statReachedCubes));
    ()
      

      
  (** this function is the heart of cfAbs. Here's the deal:
    Given:
    1. p_in (the region for the caller -- at the callsite) -- encoded here inside "caller"
    2. p_out (the region at the callee's return site -- encoded here inside "exit"
    we want the post. w.r.t the call of p_in, where the effect of the call is encoded inside p_out.
    -- let p_in' = p_in \cap_{relevant_sym_consts}(c = ac.c)
    where ac.c = (peel c) -- if its a global then, thats that, else, replace all occurrences of formal x, with actual e
    let p_out' = SP.p_in'.(asgn_of_return)
    output: pred_abs.(p_out \cap p_out')
  *)
   
  let statTotalNumberOfSummaryPosts = ref 0
  let summary_post caller exit ret_op =
    let _ = statTotalNumberOfSummaryPosts := 1 + !statTotalNumberOfSummaryPosts
    in
    (* used to rename the callee_exit in case of clash *)
    let (closure_fun,closure_fun_lv) = 
      if (Options.getValueOfBool "fldunroll") then
	(C_SD.expression_closure_stamp false (* (PredTable.is_relevant_field_image) *),
	 C_SD.lvalue_closure_stamp false (*  (PredTable.is_relevant_field_image) *))
      else (C_SD.expression_closure,C_SD.lvalue_closure)
    in
    let _ = Message.msg_string Message.Debug "Summary Post" in
      match (Region.advance_location caller, exit) with
	  (Region.Atomic ({Region.data = Region.Abstract d1} as at1),
	   Region.Atomic ({Region.data = Region.Abstract d2} as at2)) ->
	    begin (* TBD: RUPAK: I NEED TO LOOK AT THE PRED HERE! *)
	      Message.msg_string Message.Debug "Printing abstract data regions: ";
	      Message.msg_printer Message.Debug Region.print (Region.Atomic at1);
	      Message.msg_printer Message.Debug Region.print (Region.Atomic at2);
	      let pred_i_list = 
		!globally_useful_preds @
		(try 
		   Hashtbl.find PredTable.fun_pred_table 
		     (fst (C_SD.location_coords (at1.Region.location)))
		 with Not_found -> [])
	      in
	      let get_fcexp edge = 
		match (C_SD.get_command edge).C_SD.Command.code with
		    C_SD.Command.FunctionCall fcexp -> (Predicate.True, fcexp )
		  | C_SD.Command.GuardedFunctionCall (p, fcexp) -> (p, fcexp)  
		  | _ -> failwith "Bad op/edge arg: summary_post"
	      in
		(* I am folding in the edge, guard, and function call information into one place.
		   This allows me to handle calls arising from function pointers locally *)
	      let (edge (* edge of call *), guard (* the fn ptr assumption *), (callee_name, subs, target) (* result of deconstructFunCall on the call *) ) =  
		(* guard IS NOT USED YET, BUT WILL TELL THE GUARD OF THE FUNCTION CALL *)
		match enabled_ops caller with
	            [e] -> let (grd, fce) = get_fcexp e in (e, grd, C_SD.deconstructFunCall fce) 
		  | ((e :: _) as l) ->   (* This can only happen for function pointer calls. In this case, we return any edge  and get the right function
					    later using a devilish hack that I have documented below.
					 *)
                      begin
			(* The location of the exit region will be in the called function *)
			let (exitloc, _) = Region.getPcStack exit in
			let fname = (C_SD.get_location_fname exitloc) in
			  try 
			    let rec _find_call l =
			      match l with [] -> raise Not_found
				| e :: rest -> 
				    begin
				      let (g, fce) = get_fcexp e in
				      let (c, s, t) = C_SD.deconstructFunCall fce in
					if (fname = c) then (e, g, (c, s, t) )
					else _find_call rest
				    end 
			    in
			      _find_call l
			  with Not_found -> failwith "summary post: incorrect function pointer?"
                      end
	      in
	      let _ = Message.msg_string Message.Debug ("Target: "^(Expression.lvalToString target)) in
	      let retres = getRetexp ret_op in
		if (CaddieBdd.bddEqual d1.Region.bdd PredTable.bddOne &&
		    CaddieBdd.bddEqual d2.Region.bdd PredTable.bddOne
		    && 
		    (not (PredTable.is_relevant_lval target)
		     || (pred_i_list = []) 
		     || (target = C_SD.junkRet) 
		     || (Ast.Expression.is_lval retres)
		    )
		   )
		then
		  begin
		    Message.msg_string Message.Debug "Trivial post";
		    Region.Atomic {at1 with Region.data = at2.Region.data}
		  end
		else
		  let creg_caller = extract_pred caller in (* p_in *)
		  let ac_creg_callee_exit = extract_pred exit in (* p_out *)
		  let caller_name = reg_function_name caller in
		  let name_clash = (caller_name = callee_name) in
		  		  let (retres,creg_callee_exit) = 
	    	    if not name_clash then (retres, ac_creg_callee_exit) 
		    else 
                      ((Expression.alpha_convert rename_locals retres),(Predicate.alpha_convert rename_locals ac_creg_callee_exit))
		  in
		  let _ = Absutil.set_ret_pointers target subs caller_name in 
		    (* let callee_name = reg_function_name exit in *)
		  let _ =
		    Message.msg_string Message.Debug
		      ("Caller (input) Region:"^(Predicate.toString creg_caller));
		    Message.msg_string Message.Debug
		      ("Callee (output) Region:"^(Predicate.toString creg_callee_exit))
		  in
		  let rel_globs_mod =
		    Stats.time "relevent_globals_mod"
		      ((fun () -> PredTable.relevant_globals_mod) ()) callee_name
		  in
		  let _ = Message.msg_string Message.Debug 
			    ("rel_globs_mod: "^(Misc.strList (List.map Expression.lvalToString rel_globs_mod)))
			    
		  in
		  let symvar_fun = 
		    if not name_clash then Expression.make_symvar callee_name
		    else
		      (fun lv -> 
			 match Expression.make_symvar callee_name lv with
			     Expression.Symbol s -> Expression.Symbol (s^"BLAST")
			   | _ -> failwith "bad return from make_symvar")
		  in	
		  let get_sym_preds () = 
		    let make_eq replace_fn lv =
		      let lv_clos = Stats.time "lv-closure" closure_fun_lv lv in 
		      let _ = Message.msg_string Message.Debug 
				(Printf.sprintf "lv (%s)--- clos (%s)" (Expression.lvalToString lv)
				   (Misc.strList (List.map Expression.lvalToString lv_clos)))
		      in
			Predicate.equate
			  (List.map (fun lv -> Expression.Lval (symvar_fun lv)) lv_clos)
			  (List.map (fun lv -> replace_fn (Expression.Lval lv)) lv_clos)
		    in
		    let g_pred =
		      Predicate.conjoinL
			(List.map (make_eq (fun x -> x)) rel_globs_mod)  
		    in
		    let l_pred = Predicate.conjoinL
				   (List.map
				      (fun (x,e) ->
					 make_eq
					 (Expression.substituteExpression
					    [(Expression.Lval (Expression.Symbol x), e)]) 
					 (Expression.Symbol x))
				      subs)
		    in
		      (g_pred,l_pred)
		  in
		  let creg_callee_exit' = (* p_out' *)
		    (* if (CaddieBdd.bddEqual d1.Region.bdd PredTable.bddOne) 
		       then Predicate.True
		       else *)
		    begin
		      let (g_sym_pred,l_sym_pred) = Stats.time "get sym preds" get_sym_preds () in
		      let creg_caller' = Predicate.conjoinL [creg_caller;g_sym_pred;l_sym_pred] in (* p_in' *)
		      let get_asgn_list () =
			let params_list = (* could do an optimization similar to the global situation here *)
			  let _ = Message.msg_string Message.Debug "Getting copy_back asgns" in
			  let (_,copy_back_asgns) = Stats.time "asgn of ret" Absutil.asgn_of_return callee_name in
			  let _cba = 
			    if target = C_SD.junkRet then copy_back_asgns
			    else
			      (target,retres)::copy_back_asgns
			  in 
			    List.filter 
                              (fun (l,_) -> 
				 let rv = PredTable.is_relevant_lval l in                                         Message.msg_string Message.Debug 
														    (Printf.sprintf "Check relevant cb: %s : %b" 
														       (Expression.lvalToString l) rv);
				   rv)
                              _cba
			in
			let globals_list =
			  let glvs = rel_globs_mod in 
			    (* earlier:   C_SD.global_lvals_modified callee_name   in *) 
			    List.map (fun glv -> (glv,Absutil.get_new_unknown callee_name)) glvs
			in
			  globals_list@params_list
		      in
		      let asgn_list = Stats.time "get asgn list" get_asgn_list () in
		      let dump_asgn (lv,e) = Message.msg_string Message.Debug
					       (Printf.sprintf "%s := %s" (Expression.lvalToString lv) (Expression.toString e))
		      in
		      let _ = List.iter dump_asgn asgn_list in
			(* now do the "return" using the following trick. the copy backs, i.e the WP step *)
		      let pred_list = List.map PredTable.getPred pred_i_list in
		      let relevant_future_lvals : (Expression.lval list) =
			let cands = List.flatten (List.map getVarsAndDerefs (creg_caller'::pred_list)) in
			let cand_lvs =
			  Misc.map_partial
			    (fun e -> try Some (Expression.lv_of_expr e) with _ -> None)
			    cands
			in
			  Misc.sort_and_compact cand_lvs
		      in
			
		      let _ = Message.msg_string Message.Debug "Now building creg_callee_exit'" in
			
			Stats.time "sp_via_block" (sp_via_block closure_fun relevant_future_lvals creg_caller') asgn_list
		    end
		  in 
		  let _ = Message.msg_string Message.Debug ("callee_exit' = "^(Predicate.toString creg_callee_exit'))
		  in
		  let lift_conj () = (abstract_predicate
					(Predicate.conjoinL [creg_callee_exit';creg_callee_exit])
					pred_i_list) 
		  in
		  let _ = Message.msg_string Message.Debug ("Using preds:"^(Misc.strList (List.map (fun i -> Predicate.toString (PredTable.getPred i)) pred_i_list)))
		  in
		  let new_data = Stats.time "lift conj" lift_conj () in
		  let reg = Region.Atomic {at1 with Region.data = new_data} (* abs.Gamma.(p_out' \cap p_out) *)
		  in 
		    Message.msg_string Message.Debug "Summary Post: Region is ";
		    Message.msg_printer Message.Debug  Region.print  reg;
		    Message.msg_string Message.Debug (Printf.sprintf "@.After restrict to %s: " caller_name);
		    (* this restrict business should not be needed if craig is being used *)
		    let reg' = Stats.time "Region.restrict"
				 (Region.restrict (Region.locals_and_globals caller_name)) reg
		    in
		      Message.msg_printer Message.Debug Region.print reg';
		      reg'
            end
	| _ -> failwith "One or both regions passed to summary_post are not Abstract-Atomic"
	    
	    
	    
  let initialize_abstraction () =
    PredTable.load_abs ();
    Message.msg_string Message.Normal "Initialized Abstraction";
    if (Options.getValueOfBool "cf") then DataLattice.initialize ()
      
  (* Saving final abs from tree *)
  let save_region_abs reg = 
    let l_id = C_SD.location_coords (location_coords_of_region reg) in
    let preds = PredTable.bdd_support (extract_abs_data_reg reg).Region.bdd in
      PredTable.add_local_preds l_id preds

	
  (* this function will allow us to create the initial and error regions
     to give to the model checking function *)

  let ac_create_region abs_conc_flag loclist stack_bool seedPreds initialStatePredicate = failwith "unimplemented: cfAbs ac_create_region"
  let add_alias_override _ = failwith "CF: alias override unimplemented"

  let create_region loclist stack_bool seedPreds initialStatePredicate =
    
    (* JHALA: The init and error regions have different stacks -- 
       I dont know how to navigate the module hierarchy so as to be able to export 
       this so I will say that true means EveryStack and false means emptyStack *)
    let _ =  
      if (Options.getValueOfBool "simpquery") then
	TheoremProver.set_simplify_query_flag ()
    in
    let atom_list = 
      (Predicate.getAtoms seedPreds) @ (Predicate.getAtoms initialStatePredicate)
    in
    let ind_list = List.map (fun x -> addPred x) atom_list in 
    let pred_indexes = List.filter (fun a -> if (a> -1) then true else false) ind_list in
    let gpred_indexes = 
      if (Options.getValueOfBool "scope" || Options.getValueOfBool "cf") then
    	Misc.map_partial (PredTable.add_pred_to_scope) pred_indexes
      else pred_indexes
    in
    let _ = Message.msg_string Message.Normal "Glob useful preds" in
    let _ = Message.msg_string Message.Normal 
	      (Misc.strList (List.map (fun i ->
					 Predicate.toString
					 (PredTable.getPred i)) gpred_indexes)) in
    let _ =	globally_useful_preds := gpred_indexes @ !globally_useful_preds in
      (*let our_bdd = convertPredToBdd p in *)
    let our_bdd =
      try
	convertPredToBdd initialStatePredicate
      with Not_found -> failwith "create_region: Some predicate in the initial state not found in the predicate table!"
    in
      (*
	if (Options.getValueOfBool "par") then
	Region.ParallelAtomic { Region.p_control = 
	List.map 
	(fun x -> 
	{ 
	Region.p_location = x; 
       	Region.p_stack = if stack_bool then Region.EveryStack else Region.CallStack [] 
	}  
	)
	loclist ;
	Region.p_data = Region.Abstract { Region.lelement = DataLattice.top ;
	Region.bdd = our_bdd ;
	Region.preds = pred_indexes ; } ;
	} 
	else
      *)
      Region.Atomic {
	Region.location = List.hd loclist ;
	Region.data = Region.Abstract { Region.lelement = DataLattice.top ; Region.bdd = our_bdd ; Region.preds = pred_indexes ; } ;
	Region.stack = if stack_bool then Region.EveryStack else Region.CallStack [] ;
      }


	
  let print_stats fmt () =
    Message.msg_string Message.Normal ("CF Abstraction Statistics");
    Message.msg_string Message.Normal ("Total number of summary-posts = "^(string_of_int (!statTotalNumberOfSummaryPosts)));
    Message.msg_string Message.Normal ("Total number of queries = "^(string_of_int (!statTotalNumberOfQueries))^"\n");
    Message.msg_string Message.Normal ("Total number of non-post queries = "^(string_of_int (!statCoveredQueries))^"\n");
    Message.msg_string Message.Normal ("PostBdd Calls = "^(string_of_int
    !statPostBddCalls));
    Message.msg_string Message.Normal ("Worst-case number of post queries = "^(string_of_int (!statWorstCaseQueriesForPost))^"\n");
    Message.msg_string Message.Normal ("Actual number of post queries = "^(string_of_int (!statActualQueriesForPost))^"\n");
    Message.msg_string Message.Normal ("Assume post queries = "^(string_of_int (!statAssumeQueriesForPost))^"\n");
    Message.msg_string Message.Normal ("Total posts = "^(string_of_int (!statNumPost))^"\n");
    Message.msg_string Message.Normal ("Total assume posts = "^(string_of_int (!statNumAssumePost))^"\n");

    Message.msg_string Message.Normal ("Total cached = "^(string_of_int (!statTotalCached))^"\n\n");
    Message.msg_string Message.Normal ("Total foci queries = "^(string_of_int (!TheoremProver.stats_nb_foci_queries))^" ");

    Message.msg_string Message.Normal ("Total number of alias queries = "^(string_of_int !AliasAnalyzer.stats_nb_alias_query));
    Message.msg_string Message.Normal ("Total number of cached exp_closures ="^(string_of_int !C_SD.stats_nb_cached_exp_closure));
    Message.msg_string Message.Normal ("Total number of queryAlias_cache calls = "^(string_of_int !Absutil.stats_nb_queryAlias_cache_called ));
    Message.msg_string Message.Normal ("Total number of queryAlias_cache hits = "^(string_of_int !Absutil.stats_nb_queryAlias_cache_hits ));
    Message.msg_string Message.Normal ("Total number of tproj alias queries = "^(string_of_int !stats_nb_tproj_alias_queries));
    Message.msg_string Message.Normal ("Total number of lv_rd alias queries (computed) = "^(string_of_int !stats_nb_get_lvrd_aliases_compute));
    Message.msg_string Message.Normal ("Total number of lv_rd alias queries (cached) = "^(string_of_int !stats_nb_get_lvrd_aliases_cached));
    if !stats_nb_get_lvrd_aliases_compute > 0 then
      Message.msg_string Message.Normal ("Average Alias Set size number of lv_rd alias queries (computed) = "^(string_of_int (!stats_total_alias_set_size / !stats_nb_get_lvrd_aliases_compute)));
    Message.msg_string Message.Normal ("Max Alias Set size = "^(string_of_int !stats_max_alias_set_size));
    Message.msg_string Message.Normal ("Total number of recomp exp_closures ="^(string_of_int !C_SD.stats_nb_recomp_exp_closure));
    Message.msg_string Message.Normal ("The List of predicates:\n");
    printTableOfPreds ();
    Message.msg_string Message.Normal 
      ("Maximum number of predicates active together (discounting scope) = "^(string_of_int (!statsMaxNumberOfPredsActive))^"\n");
    Message.msg_string Message.Normal ("Functions visited:");
    let fvcnt = ref 0 in
      Hashtbl.iter (fun x y -> fvcnt := !fvcnt +1;
		      print_string (x^" , ")) statVisitedFunctionTable;
      
    Message.msg_string Message.Normal (Printf.sprintf "\nNumber of visited functions:  %d" !fvcnt);
    dump_abs (); ()




  (* **************** CONCURRENCY **************** *)

  (* this function will reset all the appropriate global data structures ... *)

  let reset () = 
    Hashtbl.clear theoremProverCache;
    Hashtbl.clear predTbl;
    Hashtbl.clear predTbl_deep; 
    Hashtbl.clear unchangedPredTable; 
    (* [RUPAK] Do not reset the predicates
       tableOfPreds := [];
       predIndex := -1;
    *)
    Absutil.reset_ret_pointers ();
    last_loc_pointer := (None);
    assume_pred_ref := None;
    ()

  let exit_abstraction ()  = PredTable.kill_predbdd () 

  (*  let iter_global_variables = C_SD.iter_global_variables
  *)
  let iter_global_variables = C_SD.iterate_all_lvals


  let expression_mentions_locals e =
    try
      let reject_local (name : Expression.lval) = if is_local name then raise Exit; name in
      let _ = Ast.Expression.alpha_convert reject_local e in false
    with Exit -> true


  type access_type = Anone | Aread | Awrite | Aboth
  let accessing op v =
    let reads, writes = C_SD.reads_and_writes_at_edge op in
      match (List.mem v reads, List.mem v writes) with
	| false, false -> Anone
	| false, true -> Awrite
	| true, false -> Aread
	| true, true -> Aboth

  let get_all_writes op =
    snd (C_SD.reads_and_writes_at_edge op)

  let get_all_reads op =
    fst (C_SD.reads_and_writes_at_edge op)

  let reads_and_writes op = C_SD.reads_and_writes_at_edge op 


  let globally_important_lvals () = !PredTable.globally_important_lvals

  let aliasesOf lval = if can_escape lval then ( lval::(AliasAnalyzer.getAllAliases lval)) else []

  let eq_addr_constraint_region lval1 lval2 = 
    let p = 
      Predicate.Atom 
	(Expression.Binary 
	   (Expression.Eq, 
	    Expression.addressOf (Expression.Lval lval1),
	    Expression.addressOf (Expression.Lval lval2)))
    in
      predicate_to_region p

  (* this should return the list of possible aliases in the whole program, of some lval 
     maybe we should keep a Hashtbl or something and query that every time 
  *)

  let make_multiply_instantiable_region r1 r2 lval1 lval2 = 
    Message.msg_string Message.Debug "In make_multiply instantiable\n\n" ;
    Message.msg_string Message.Debug ("r1 is"^(Region.to_string r1)) ;
    Message.msg_string Message.Debug ("r2 is"^(Region.to_string r2)) ;
    let rename_fun = new_name (brand_new_suffix ()) in
    let eq_addr_const_reg = eq_addr_constraint_region lval1 (Expression.alpha_convert_lval rename_fun lval2) in 
      Region.cap 
	(Region.cap !assumed_invariants_region eq_addr_const_reg)
	(Region.cap r1 (Region.rename_vars r2 rename_fun))
	
	
  (* the below function should ONLY be called on two regions known to be such that:
     1. lval_1, lval_2 are such that they are accessed in region r1 r2 respectively
     2. neither of the lvals is a lock.
  *)
  let multiply_instantiable r1 r2 lval1 lval2 =
    let r = make_multiply_instantiable_region r1 r2 lval1 lval2 
    in
      if Region.is_empty r then None else Some r

	
	
	
        (* **************** END concurrency **************** *)
	(************* FOR BLASTPCC ************************)
	(* for m4 -- include blast-pcc.ml *)
	include(blast-pcc.ml)


  (************************************ MAP functions **************************************) 





  let get_region_bdd reg =
    match reg with
	Region.Empty -> bddZero
      | Region.Atomic {Region.data = Region.Abstract adr} -> adr.Region.bdd
      | _ -> failwith "Bad region to get_region_bdd"

	  
  type spec = string * (CaddieBdd.bdd * CaddieBdd.bdd) list * CaddieBdd.bdd

  let replace = ref None

  let init_spec () =
    (* First, designate the current predicates as relative to "preconditions" and create a
     * corresponding "postcondition" version of each predicate. *)
    let preds = Hashtbl.fold (fun p n l -> (p, n)::l) hashTableOfPreds [] in
    let prime_lval lv =
      match lv with
	  Expression.Symbol v ->
	    (try
	       let pos = String.index v '@' in
		 Expression.Symbol
		   (String.sub v 0 pos ^ "'" ^ String.sub v pos (String.length v - pos))
	     with Not_found -> lv)
	| _ -> lv in
    let prime_pred = Predicate.alpha_convert prime_lval in
      
    let add_post replace (p, n) =
      let p' = prime_pred p in
      let n' = actual_addPred p' in
	(n, n') :: replace in
      replace := Some ((* BDDFIX: CaddieBdd.toMap bddManager*) make_varMap (List.fold_left add_post [] preds))

  module StringSet = Set.Make(struct
				type t = string
				let compare = compare
			      end)

  let build_spec fname summaries =
    let formals = match C_SD.lookup_formals fname with
	C_SD.Fixed args -> args
      | C_SD.Variable args -> args in
    let formals = List.map (function s -> String.sub s 0 (String.index s '@')) formals in
    let allowed = List.fold_right StringSet.add formals
		    (StringSet.add "__retres" StringSet.empty) in
      print_string "ALLOWED: ";
      StringSet.iter (fun s -> print_string s; print_string "; ") allowed;
      print_char '\n';
      let isInternal sym =
	try
	  let idx = String.index sym '@' in
	    not (StringSet.mem (String.sub sym 0 idx) allowed)
	with Not_found -> false in
      let restrictInternal = Region.restrict isInternal in
      let isNonoutput sym =
	try
	  let idx = String.index sym '@' in
	    String.sub sym 0 idx <> "__retres"
	with Not_found -> false in
      let restrictNonoutput = Region.restrict isNonoutput in

      let replace = valOf (!replace) in
	(* This function transforms a "precondition" bdd into the corresponding "postcondition." *) 
      let make_post bdd = CaddieBdd.replace bdd replace in

      (* Now we build a formula describing function behavior per observed entry point *)
      let map_entry (entry, exits) =
	let pre_bdd = get_region_bdd (restrictInternal entry) in
	let post_bdds = List.map (fun r -> make_post
				    (get_region_bdd (restrictNonoutput r))) exits in
	let post_bdd = List.fold_left CaddieBdd.bddOr bddZero post_bdds in
	  ((pre_bdd, post_bdd), CaddieBdd.bddImp pre_bdd post_bdd) in
      let (casePairs, cases) = List.split (List.map map_entry summaries) in
	fprintf "@.Cases:@.";
	List.iter (fun bdd ->
		     (*CaddieBdd.bddPrint bdd;*)
		     Predicate.print Format.std_formatter (convertBddToPred bdd);
		     fprintf "@.") cases;

	(* Finally, a conjunction of the entry point formulae describes all behavior. *)
	(fname, casePairs, List.fold_left CaddieBdd.bddAnd bddOne cases)


  let genTests reg = []

  type spec_result =
    | Yes
    | No
    | Maybe of Ast.Predicate.predicate

  let check_spec (f1, c1, s1) (f2, c2, s2) =
    let replace_lval lv =
      match lv with
	  Expression.Symbol x ->
	    (try
	       let idx = String.index x '@' in
		 if String.sub x (idx+1) (String.length x - idx-1) = f1 then
		   Expression.Symbol (String.sub x 0 (idx+1) ^ f2)
		 else
		   lv
	     with Not_found -> lv)
	|	_ -> lv in
    let preds = Hashtbl.fold (fun p n l -> (p, n)::l) hashTableOfPreds [] in
    let change_pred = Predicate.alpha_convert replace_lval in
      
    let add_changed replace (p, n) =
      let p' = change_pred p in
      let n' = actual_addPred p' in
	(n, n') :: replace in
    let replace = (* BDDFIX CaddieBdd.toMap bddManager*) make_varMap (List.fold_left add_changed [] preds) in

    let s1 = CaddieBdd.replace s1 replace in

    let caseBdd (pre, post) =
      fprintf "@.  Pre: ";
      Predicate.print Format.std_formatter (convertBddToPred pre);
      fprintf "@. Post: ";
      Predicate.print Format.std_formatter (convertBddToPred post);
      fprintf "@.Known: ";
      Predicate.print Format.std_formatter (convertBddToPred s1);
      let bdd = CaddieBdd.bddImp (CaddieBdd.bddAnd pre s1) post in
	fprintf "@. Case: ";
	Predicate.print Format.std_formatter (convertBddToPred bdd);
	fprintf "@.";
	bdd in
    let caseBdds = List.map caseBdd c2 in
    let casesBdd = List.fold_left CaddieBdd.bddAnd bddOne caseBdds in

    let pred = convertBddToPred casesBdd in
      fprintf "All cases: ";
      Predicate.print Format.std_formatter pred;
      fprintf "@.";

      block_reset ();
      block_assert pred;
      if TheoremProver._is_contra () then
	(block_reset ();
	 No)
      else
	(block_reset ();
	 block_assert (Predicate.negate pred);
	 if TheoremProver._is_contra () then
	   (block_reset ();
	    Yes)
	 else
	   (block_reset ();
	    Maybe pred))

  (* Extended TAR Stuff -- Not implemented *)
  (**********************************************************************************************)
  (**********************************************************************************************)
  (* The environment automaton model in the TAR algorithm *)
  (**********************************************************************************************)
  (**********************************************************************************************)
  type opKind = ProgramOperation of Operation.t | AutomatonOperation of (int * int) 

  module EnvAutomaton =
  struct
    type onode_label =
	{
	  oid : int;
	  oadr : Region.abstract_data_region;
	  oregion : Region.t
	}
	  (* Invariants: Node 0 is always the stuck node *)      
	  
    type anode_label =
	{ id   : int ; (* the id of the abstract automaton node *)
	  root : int ; (* the id of the root node in the cluster *)
	  mutable adr  : Region.abstract_data_region ; (* the abstract data region (is this on globals?) *)
          mutable atomic : bool ;
	  mutable task : bool ;
	  mutable event : bool ;
	  mutable modifies : Expression.lval list ; (* the list of globals that are modified by this transition *)

	}	
	  
    type aedge_label = 
	Thread of Expression.lval list              (* Solid edge *) 
      | Env                 (* Dotted edge *)
	  
    type oautomaton_node = (onode_label, opKind) BiDirectionalLabeledGraph.node
  	
    type automaton_node = ((* node label *) anode_label, (* edge label *) aedge_label) BiDirectionalLabeledGraph.node 
	
    type  ('a, 'b) automaton = {
      
      autom_root : automaton_node ;
      number_of_states : int ;
      
      id_to_node_ht : (int, automaton_node) Hashtbl.t ;
      
      o_auto_root : oautomaton_node;
      cover_set : int Region.coverers;
      (* other maps *)
      o_map : (int, oautomaton_node) Hashtbl.t;  (* auto_map *)
      o_to_a_map : (int, int) Hashtbl.t;         (* auto_shrink_map *)
      a_to_o_map : (int, int list) Hashtbl.t;    (* shrink_elts_map *)
      (*  
	  id_to_bdd : (CaddieBdd.bdd) array ; (* map from state id to bdd encoding *)
	  mutable stVarMap : (int * int) list ;     (* maps between the unprimed (s) and primed (t) *)
	  mutable tsVarMap : (int * int) list ;     (* state variables *)

	  mutable env_transrel    : CaddieBdd.bdd ;
      (* Not required at the moment 
	  mutable thread_transrel : CaddieBdd.bdd ;
      *)
      *)  
      
    } 

    let getNodeFromId autom i = (* raises Not_found *)
      Hashtbl.find autom.id_to_node_ht i

    let getEnvEdges autom i =
      let thisnode = getNodeFromId autom i in
      let outgoing = BiDirectionalLabeledGraph.get_out_edges thisnode in
	List.filter (fun e -> match (BiDirectionalLabeledGraph.get_edge_label e) with Env -> true | Thread _ -> false) outgoing

    let getThreadEdges autom i =
      let thisnode = getNodeFromId autom i in
      let outgoing = BiDirectionalLabeledGraph.get_out_edges thisnode in
	List.filter (fun e -> match (BiDirectionalLabeledGraph.get_edge_label e) with Env -> false | Thread _ -> true) outgoing
	
  end 

  (**********************************************************************************************)
  (**********************************************************************************************)
  (* End of The environment automaton model in the TAR algorithm *)
  (**********************************************************************************************)
  (**********************************************************************************************)

  let tar_enabled_ops _ = failwith "tar_enabled_ops unimplemented"      
  let tarpost reg op automaton = failwith "tarpost not implemented"
  let exists_stuck reg = failwith "exists stuck not implemented"
  let get_modifies autom reg = failwith "get_modifies not implemented"
  let create_parallel_atomic_region autom reg ctr = failwith "Not implemented"
      

  let parallel_block_analyze_trace autom a b c = ()

  (* random things put in so that the typechecker doesnt complain *)

  (* focimodelchecker *)
  let foci_post (reg,orig_lvmap) op = failwith "unimplemented foci_post"
  let pred_to_data_region pred = Region.Concrete {Region.pred = pred}
  let fociModelChecker_refine_trace reg_array op_array block_array lvmap_array = 
    failwith "unimplemented: fmc refine_trace"
  let ti_init_region _ = failwith "unimp:cf:ti_init"
end
