(*
 * Copyright © 1990-2002 The Regents of the University of California. All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *)

(* $Id: unionlattice.ml,v 1.2 2005/05/06 00:13:50 fischer Exp $

 *
 * This file is part of the BLAST Project.
 *)

(** Module for lattice implementations *)

open BlastArch

open Lattice

module Make_UnionLattice =
  functor(C_SD : SYSTEM_DESCRIPTION with module Command = BlastCSystemDescr.C_Command) ->
    functor (Op : OPERATION ) ->
struct

  module Command = C_SD.Command
  (* Whenever you write a new lattice, add an entry here *)
  module TrivialLattice = (Lattice.Make_TrivialLattice (C_SD)) (Op)
  module SymbolicStoreLattice = (Lattice.Make_SymbolicStoreLattice (C_SD)) (Op)
  module ListLattice = (ListLattice.Make_ListLattice (C_SD)) (Op)

  (* This code assumes that all lattices have the same edge type *)
  type edge = C_SD.edge
      
  (* Whenever you write a new lattice, add an entry here *)
  type elements = 
      Trivial of TrivialLattice.lattice
    | Symb of SymbolicStoreLattice.lattice
    | LList of ListLattice.lattice

  type lattice = elements list

  let checklattice l =
    match l with
        Trivial t -> "trivial"
      | Symb s -> "symbolic"
      | LList ll -> "list"
      
  let top = [LList ListLattice.top; Symb SymbolicStoreLattice.top; Trivial TrivialLattice.top]

  let bottom = [LList ListLattice.bottom; Symb SymbolicStoreLattice.bottom; Trivial TrivialLattice.bottom]
		 
  (* We need to maintain flags to indicate whether each lattice type is enabled. This is used
     by post exclude disabled lattices. Originally, this was done in the definitions for
     top and bottom. However, this doesn't work because command line options are processed
     after this file is initialized. JF 3/13/05 *)
  let list_enabled = ref false
  let symb_enabled = ref false
  let const_enabled = ref false
  let trivial_enabled = ref true

  let sane a =
    let is_bottom l = 
      match l with
          Trivial t -> if t = TrivialLattice.bottom then true else false
	| Symb s -> if s = SymbolicStoreLattice.bottom then true else false
	| LList ll -> if ll = ListLattice.bottom then true else false
    in
      if List.exists is_bottom a then bottom else a 

  let predlist_bottom = 
      (if (*not (List.mem "list" excludes)*)true then [ ([], LList ListLattice.bottom) ] else [] ) @
      (if (*not (List.mem "symb" excludes)*)true then [ ([], Symb SymbolicStoreLattice.bottom) ] else [] ) @
      [ ([], Trivial TrivialLattice.bottom) ] 

	
  let aux_sane (a:(Ast.Predicate.predicate list * elements) list) =
    let is_bottom l = 
      match l with
          (_,Trivial t) -> if t = TrivialLattice.bottom then true else false
	| (_,Symb s) -> if s = SymbolicStoreLattice.bottom then true else false
	| (_,LList ll) -> if ll = ListLattice.bottom then true else false
    in
      if List.exists is_bottom a then predlist_bottom else a 
	
  let rec conjoin_results results cumval=
    match results with
      | [] ->
	  if (cumval = 0) then Ast.False
	  else if(cumval = 1) then Ast.True
	  else Ast.Dontknow
      | Ast.True::tail ->
	  if (cumval = 0) then failwith "fatal contradiction - true : conjoin_results@unionlattice.ml"
	  else conjoin_results tail 1
      | Ast.False::tail ->
	  if (cumval = 1) then failwith "fatal contradiction - false : conjoin_results@unionlattice.ml"
	  else conjoin_results tail 0
      | Ast.Dontknow::tail ->
	  conjoin_results tail cumval

  let aux_query_fn intval l pred  =
    let f a  =
      match a with
	  Trivial t1 ->
	    if intval = 1 then Ast.Dontknow
	    else TrivialLattice.query_fn t1 pred
	| Symb s1 ->
	    if intval = 2 then Ast.Dontknow
	    else SymbolicStoreLattice.query_fn s1 pred
	| LList l1 ->
	    if intval = 3 then Ast.Dontknow
	    else ListLattice.query_fn l1 pred
    in
      List.map f l
	
  let query_fn (id:int) l (mother_query_fn:Ast.Predicate.predicate->Ast.predicateVal) (pred:Ast.Predicate.predicate) =
    let qresults:Ast.predicateVal list = aux_query_fn id l pred 
    in let aug_results = (qresults @ [mother_query_fn pred])
    in let cumval = -1
    in conjoin_results aug_results cumval
	
  let join (l1:lattice) (l2:lattice) =  
    let f a b =
      match (a,b) with
          (Trivial t1, Trivial t2) -> Trivial (TrivialLattice.join t1 t2)
	| (Symb s1, Symb s2) -> Symb (SymbolicStoreLattice.join s1 s2)
	| (LList ll1, LList ll2) -> LList (ListLattice.join ll1 ll2)
	| _ -> failwith "UnionLattice join: Mismatched or unknown lattice elements"
    in
      sane (List.map2 f l1 l2)
	
  let meet (l1:lattice) (l2:lattice) =
    let f a b =
      match (a,b) with
          (Trivial t1, Trivial t2) -> Trivial (TrivialLattice.meet t1 t2)
	| (Symb s1, Symb s2) -> Symb (SymbolicStoreLattice.meet s1 s2)
	| (LList ll1, LList ll2) -> LList (ListLattice.meet ll1 ll2)
	| _ ->
	    let str1 = checklattice a
	    in let str2 = checklattice b
	    in print_string ("failed with:"^str1^"&"^str2);
	      failwith "UnionLattice meet: Mismatched or unknown lattice elements"
    in
      sane (List.map2 f l1 l2)

  let leq  (l1: lattice) (l2: lattice) =
    let f a b =
      match (a,b) with
          (Trivial t1, Trivial t2) -> (TrivialLattice.leq t1 t2)
	| (Symb s1, Symb s2) -> (SymbolicStoreLattice.leq s1 s2)
	| (LList ll1, LList ll2) -> (ListLattice.leq ll1 ll2)
	| _ -> failwith "UnionLattice leq: Mismatched or unknown lattice elements"
    in
      List.for_all2 f l1 l2 
	
  let eq  (l1: lattice) (l2: lattice) =
    let f a b =
      match (a,b) with
          (Trivial t1, Trivial t2) -> (TrivialLattice.eq t1 t2)
	| (Symb s1, Symb s2) -> (SymbolicStoreLattice.eq s1 s2)
	| (LList ll1, LList ll2) -> (ListLattice.eq ll1 ll2)
	| _ -> failwith "UnionLattice eq: Mismatched or unknown lattice elements"
    in
      List.for_all2 f l1 l2 

  let focus () = () (* Unimplemented *) 

  (** This needs to be reimplemented. 
    Right now it checks if the bdd is consistent with each
    individual lattice. How can we be sure that the lattice
    states themselves are consistent?
  *)
  let is_consistent  (l: lattice)  (bdd: CaddieBdd.bdd) =
    let f a =
      match a with
          Trivial t1 -> (TrivialLattice.is_consistent t1 bdd)
	| Symb s1 -> (SymbolicStoreLattice.is_consistent s1 bdd)
	| LList l1 -> (ListLattice.is_consistent l1 bdd)
    in
      List.for_all f l 

  let rec aux_clean_post_output in_list out_list =
    if ((List.length in_list) = 0) then out_list
    else
      begin
	let head = List.hd in_list
	in let (predhead, lathead) = head
	in let (predlist, latlist) = out_list
	in let npredlist = predhead @ predlist
	in let nlatlist =  latlist @ [lathead]
	in let nout_list = (npredlist, nlatlist)
	in let tail = List.tl in_list
	in
	  if ((List.length tail) = 0) then nout_list
	  else aux_clean_post_output tail nout_list
      end
    
	
  let clean_post_output in_list =
    aux_clean_post_output in_list ([],[])

  (* JF 3/13/05
     TODO: need to improve performance - either use a static structure for union lattice or define a
     global top and bottom which can be expanded to the set of lattices enabled. *)      
  (** Compute the lattice which is result after applying
    computation in command to initial lattice *)
  let post  (l: lattice) (e: edge) (fn:Ast.Predicate.predicate->Ast.predicateVal) = 
    Message.log_string_norm Message.UnionLat "Running post on command:";
    Message.log_print_norm Message.UnionLat C_SD.Command.print (C_SD.get_command e);
    let f a = 
      match a with
        | Trivial t1 ->
	    if !trivial_enabled then
	      let (predlist, lelement) = (TrivialLattice.post t1 e (query_fn 1 l fn))
	      in (predlist, Trivial lelement)
	    else ([], Trivial TrivialLattice.top)
	| Symb s1 ->
	    if !symb_enabled then
	      let (predlist, lelement) = (SymbolicStoreLattice.post s1 e (query_fn 2 l fn))
	      in (predlist, Symb lelement)
	    else ([], Symb SymbolicStoreLattice.top)
	| LList l1 ->
	    if !list_enabled then
	      let (predlist, lelement) = (ListLattice.post l1 e (query_fn 3 l fn))
	      in (predlist, LList lelement)
	    else ([], LList ListLattice.top)
    in
    let (pred_list, out_lattice) = clean_post_output (aux_sane (List.map f l))
    in
      Message.log_string_norm Message.UnionLat"Post returning predicates:";
      Message.log_print_norm Message.UnionLat Ast.Predicate.print (Ast.Predicate.And pred_list);
      (pred_list, out_lattice)
	
  let pre  (l: lattice) e = (** Compute the lattice which is result after applying
                                      computation in command to initial lattice *)
    let f a = 
      match a with
          Trivial t1 -> Trivial (TrivialLattice.pre t1 e)
	| Symb s1 -> Symb (SymbolicStoreLattice.pre s1 e)
	| LList l1 -> LList (ListLattice.pre l1 e)
    in
      sane (List.map f l )

  let print (f: Format.formatter)  (l:lattice) = (** Pretty printer for this lattice *)
    let f a = 
      match a with
          Trivial t1 ->
	    if !trivial_enabled then (TrivialLattice.print f t1; Format.pp_force_newline f ())
	| Symb s1 ->
	    if !symb_enabled then (SymbolicStoreLattice.print f s1; Format.pp_force_newline f ())
	| LList l1 ->
	    if !list_enabled then (ListLattice.print f l1; Format.pp_force_newline f ())
    in
      List.iter f l 

  (* Call the initialize function of each enabled lattice and set its enabled flag *)
  let initialize () =
    let excludes = Options.getValueOfStringList "exclude-lattice" in
      begin
	if not (Options.getValueOfBool "nolattice") then
	  begin
	    if not (List.mem "symb" excludes) then
	      (SymbolicStoreLattice.initialize ();
	       symb_enabled := true;
	       Message.log_string_always Message.UnionLat "Symbolic Store Lattice enabled");
	    if not (List.mem "list" excludes) then
	      (ListLattice.initialize ();
	       list_enabled := true;
	       Message.log_string_always Message.UnionLat "List Lattice enabled");
	  end;
	if !symb_enabled || !list_enabled then
	  trivial_enabled := false
	else
	  Message.log_string_always Message.UnionLat "Only trivial lattice enabled"
      end

end

      
