(*
 * Copyright © 1990-2005 The Regents of the University of California. All rights reserved. 
 *
 * Permission is hereby granted, without written agreement and without 
 * license or royalty fees, to use, copy, modify, and distribute this 
 * software and its documentation for any purpose, provided that the 
 * above copyright notice and the following two paragraphs appear in 
 * all copies of this software. 
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN 
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE. 
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS 
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION 
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *)

(* $Id: cfbLazyModelChecker.ml,v 1.3 2005/05/21 00:35:09 jhala Exp $
 *
 * This file is part of the BLAST Project.
 *)


(* CF Reach with arbitrary search order.

Data Structures:
-- BiDirLabGraph.ml
-- Labels on node:
	{id : int;
	 region: Region.t;
	 fname : string; 
	 entries : int set;
	 exits : int set; // only needed for entries
	 callsites : int set; // only needed for entries
         status : Unprocessed aka to be processed | Processed
	}
	
-- parents,children from the BDLG data structure
-- use the tnc object from focimodelchecker so that we have a clean API
   to find a node given id/delete node etc.
   
-- Assorted predicates:
	- is_exit n : if it has one out-edge marked with a return op
	- is_entry n : if it has no parents
	

-- Global "Worklist" Object: stores just integers
	- add node(id) to worklist
	- is_empty
	- pick (and remove) from worklist
	
   INVARIANTS:
   1. For all  n in WKL:
	a. n's children WILL be added
   	b. n.region in the REACH set
	c. n is NOT covered by anyone else
   
   2. For all n: n.entries contains all the entries of n.fname
      that have a path to n in the graph of n.fname

   3. For all n', for all n \in n'.entries:
        is_exit n' ==> n' \in n.exits 
 
   4. Exit(n): n has an outgoing "Return" op.
      For all n. Exit(n) ==> (n \in wkl \/ n.kind = Exit _)

   Connectivity Invariants: (Predecessors always exist Graph)
   5. Forall n. 
      a. n.entries <> []
      b. Forall n' \in n.entries: exists(n')
   
   6. Forall n.
      is_entry n ==> (is_root n || n.callsites <> [])
*)

open Ast
open BlastArch
module Stats = Bstats
module C_System_Descr = BlastCSystemDescr.C_System_Descr
module C_Command = BlastCSystemDescr.C_Command
    
module Make_Lazy_Model_Checker =
  functor(Abstraction : ABSTRACTION) ->
struct
  module Region = Abstraction.Region
  module Operation = Abstraction.Operation
  module Graph = BiDirectionalLabeledGraph

  
(***********************************************************)
(********************* Data Types **************************)
(***********************************************************)

module IntSet = Set.Make(struct
  type t = int
  let compare = compare
end)

type node_kind = Entry | Internal | Exit of (Operation.t)

type node_label =
  {
    id : int;
    region : Region.t;
    mutable kind : node_kind;
    mutable entries : IntSet.t;
    mutable callsites : (Operation.t * int) list ;
    mutable exits : int list;
    mutable callees : int list; (* only needed for refinement *)
    mutable returns : int list; (* only needed for refinement *)
  }
	
type edge_label = Intra of Operation.t | Summary of (Operation.t * int * int * Operation.t)

type graph_node = (node_label,edge_label) Graph.node

type error_path = (Region.t * Operation.t list * Region.t)

let get_node_id n = (Graph.get_node_label n).id

let get_node_region n = (Graph.get_node_label n).region

let get_node_fname n = let reg = get_node_region n in
  Abstraction.reg_function_name reg

let get_node_callsites n = (Graph.get_node_label n).callsites

let get_node_entries n = (Graph.get_node_label n).entries

let get_node_exits n = (Graph.get_node_label n).exits

let get_node_children n = Graph.get_children n

let get_node_calldepth n = 
  try Region.call_depth (get_node_region n) with e ->
    (let es = Printexc.to_string e in
    Message.msg_string Message.Error es; -1) 

let get_node_callees n = (Graph.get_node_label n).callees

let get_node_returns n = (Graph.get_node_label n).returns

let get_node_kind n = (Graph.get_node_label n).kind

let add_node_callsite n c = 
  let cs = (Graph.get_node_label n).callsites in
  (Graph.get_node_label n).callsites <- (c::cs)
  
let add_node_callee n c = 
  let cs = (Graph.get_node_label n).callees in
  (Graph.get_node_label n).callees <- (c::cs)

let add_node_returns n r = 
  let rs = (Graph.get_node_label n).returns in 
  (Graph.get_node_label n).returns <- (r::rs)

let add_node_exits n e = 
  let es = (Graph.get_node_label n).exits in
  if not (List.mem e es) then ((Graph.get_node_label n).exits <- (e::es))

let del_node_callsite c_id n = 
  let cs = (Graph.get_node_label n).callsites in
  (Graph.get_node_label n).callsites <- (List.filter (fun (_,id) -> id <> c_id) cs)
  
let set_node_entries n es = (Graph.get_node_label n).entries <- es

let set_node_kind n k = (Graph.get_node_label n).kind <- k

let is_exit n = 
  match (Graph.get_node_label n).kind with
    Exit _ -> true
  | _ -> false

let is_entry n = 
  match (Graph.get_node_label n).kind with
   Entry -> true
  | _ -> false



  
  
let node_label_to_string nl =
  let (regs_l,loc_s) = Abstraction.reg_to_strings (nl.region) in
  let id = nl.id in
  Printf.sprintf "%d $$ (%s,%s)" id loc_s (Misc.strList regs_l)

let edge_label_to_string el =
  match el with
  Intra(op) -> let (_,_,op_s) = Abstraction.op_to_string op in op_s
  | Summary(call_op,in_id,out_id,ret_op) -> 
      let (_,_,op_s) = Abstraction.op_to_string call_op in
      Printf.sprintf "Summary(%s,%d,%d)" op_s in_id out_id

      
let print_graph_edge fmt e = 
  Format.fprintf fmt ";@ in_op=" ;
  Format.fprintf fmt ";@ %s" (edge_label_to_string (Graph.get_edge_label e)) ;
  Format.fprintf fmt ";@ parent_id=%d" (get_node_id (Graph.get_source e))

  
let print_graph_node fmt n =
  Format.fprintf fmt "@[GNode(@[label=" ;
  Format.fprintf fmt ";@ %s" (node_label_to_string (Graph.get_node_label n)) ;
  List.iter (print_graph_edge fmt) (Graph.get_in_edges n);
  let children_ids = List.map get_node_id (Graph.get_children n) in
  Format.fprintf fmt ";@ children_ids=" ;
      Misc.list_printer_from_printer Format.pp_print_int fmt children_ids ;
  Format.fprintf fmt "@])@]"
    

(************************************************************)
(********************* Statistics Variables *****************)
(************************************************************)
let inc r = r := !r + 1
let stats_nb_iterations = ref 0
let stats_nb_created_nodes = ref 0
let stats_nb_refinment_processes = ref 0
let stats_nb_refined_nodes = ref 0
let stats_nb_proof_tree_nodes = ref 0
let stats_nb_proof_tree_covered_nodes = ref 0
let stats_nb_deleted_nodes = ref 0
let local_restart_counter = ref 1

let reset_stats =
  stats_nb_iterations := 0 ;
  stats_nb_refinment_processes := 0 ;
  stats_nb_created_nodes := 0 ;
  stats_nb_refined_nodes := 0 ;
  stats_nb_proof_tree_nodes := 0;
  stats_nb_proof_tree_covered_nodes := 0
       
let print_stats fmt () =
  Format.fprintf fmt "@[<v>@[Nb iterations of reachability:@ %d@]" !stats_nb_iterations ;
  Format.fprintf fmt "@ " ;
  Format.fprintf fmt "@[Nb created nodes:@ %d@]" !stats_nb_created_nodes ;
  Format.fprintf fmt "@ " ;
  Format.fprintf fmt "@[Nb refinment processes:@ %d@]" !stats_nb_refinment_processes ;
  Format.fprintf fmt "@ " ;
  Format.fprintf fmt "@[Nb refined nodes:@ %d@]" !stats_nb_refined_nodes ;
  Format.fprintf fmt "@ " ;
  Format.fprintf fmt "@[Nb proof tree nodes:@ %d@]" !stats_nb_proof_tree_nodes;
  Format.fprintf fmt "@ " ;
  Format.fprintf fmt "@[Nb proof tree covered nodes:@ %d@]@]" !stats_nb_proof_tree_covered_nodes;
  Format.fprintf fmt "@[Nb deleted nodes:@ %d@]@]" !stats_nb_deleted_nodes;
  ()

(************************************************************)
(************************ Global Variables ******************)
(************************************************************)

(* reach set *)
let cov = Region.emptyCoverers ()

(* needed for post *)
let gvt = Hashtbl.create 2  

(* worklist *)
let wkl =  
  if (Options.getValueOfBool "bfs") then new Search.bfsiterator
  else new Search.dfsiterator

(* root node *)
let root_id_ref = ref (-1)

(* to prevent recomputation of summary-post *)
let sumpost_table = Hashtbl.create 1001

let is_sumpost_computed (caller_id,exit_id) = 
  let rv =  Hashtbl.mem sumpost_table (caller_id,exit_id) in
  Message.msg_string Message.Debug 
  (Printf.sprintf "Check sumpost_comp: (%d,%d) %b" caller_id exit_id rv);
  rv
  
let add_sumpost_computed (caller_id,exit_id) =
   Message.msg_string Message.Debug 
   (Printf.sprintf "Add sumpost_comp: (%d,%d)" caller_id exit_id);
   Hashtbl.replace sumpost_table (caller_id,exit_id) true 
 
let del_sumpost_computed (caller_id,exit_id) = 
  Hashtbl.remove sumpost_table (caller_id,exit_id)

class graph_node_creator = 
  object (self)
    val mutable next_id = 0
    
    val graph_node_table = Hashtbl.create 1009 
    (* hash int -> node ONLY persistent pointer to a node, single point of
     * deletion *)
    
    val entry_node_table : (string,int list) Hashtbl.t = Hashtbl.create 101
    val visited_fun_table : (string, bool) Hashtbl.t = Hashtbl.create 101
    val visited_fun_counter = ref 0

    
    method create_node is_entry reg = 
      let this_id = next_id in
      next_id <- next_id + 1;
      let es = 
        if is_entry then
          begin
          let fname = Abstraction.reg_function_name reg in
          ignore(Misc.hashtbl_check_update entry_node_table fname this_id);
          if not (Hashtbl.mem visited_fun_table fname) then
            (Hashtbl.replace visited_fun_table fname true;
            visited_fun_counter := !visited_fun_counter + 1;
            Message.msg_string Message.Normal (Printf.sprintf "[CFB] Visited Fun %d : %s" !visited_fun_counter fname));
          IntSet.singleton this_id
          end
      else IntSet.empty
      in
      let label = 
        {id = this_id;region = reg;
        kind = if is_entry then Entry else Internal;
        entries = es;callsites = [];callees=[];exits = [];returns = []}
      in
      let rv = ((Graph.create_root label) : graph_node) in
      Hashtbl.add graph_node_table this_id rv;
      rv
    
    method node_from_id id = 
      try Some(Hashtbl.find graph_node_table id) with Not_found -> None
      
    method delete_node (id : int) = 
      let proc_in_e e = 
        match Graph.get_edge_label e with Intra _ -> () 
        | Summary(_,_,ex_id,_) -> 
          let src_id = get_node_id (Graph.get_source e) in
          del_sumpost_computed (src_id,ex_id)
      in
      if Hashtbl.mem graph_node_table id then
        begin
        let n = Hashtbl.find graph_node_table id in
        let callees = Misc.map_partial (self#node_from_id) (get_node_callees n) in
        let reg = get_node_region n in
        let in_edges = Graph.get_in_edges n in
        List.iter proc_in_e in_edges;
        List.iter (del_node_callsite id) callees;        
        Region.deleteCoverer cov reg id;
        Graph.delete_node n;
        Hashtbl.remove graph_node_table id
        end;

    
    method node_exists id = 
      try ignore(Hashtbl.find graph_node_table id);true with Not_found -> false
     

    method get_all_entries () = 
      let elist = Misc.hashtbl_to_list entry_node_table in
      let proc (fname,n_id_l) = 
        (Misc.map_partial (self#node_from_id) n_id_l, fname)
      in List.map proc elist
     
    method get_entries fn = 
      let e_id_l = try Hashtbl.find entry_node_table fn with Not_found -> [] in
      Misc.map_partial (self#node_from_id) e_id_l
      
  end

let gnc = new graph_node_creator


(************************************************************)
(******************* Printing Functions *********************)
(************************************************************)

  
let print_reach_graph () =  
  Message.msg_string Message.Normal "Printing Reach Graph";
  let file = "reachgraph" in
  let tups = gnc#get_all_entries () in
  Graph.output_multi_graph_dot 
        node_label_to_string edge_label_to_string (file^".dot") tups;
  Sys.command (Printf.sprintf "dot -Tps -o %s.ps %s.dot" file file);
  ()
        

(************************************************************)
(****************** Propagation Functions *******************)
(************************************************************)

(* mods: WKL *) 
let rec push_entries ens n =
  let n_en = get_node_entries n in
  let n_en' = IntSet.union n_en ens in
  if (IntSet.cardinal n_en') > (IntSet.cardinal n_en) then
    begin 
      set_node_entries n n_en';
      if (is_exit n) then wkl#add_element (get_node_id n)
      else List.iter (push_entries ens) (get_node_children n)
  end
    
(* mods: cov, wkl this is the ONLY place that:
    a. cov is added to/read, 
    b. new nodes are created using GNC *)
    
let find_node is_entry reg = 
  (* get node, and a bool telling you if its new *)
  let ac_find_node reg =
    let n_opt = 
       match Region.findExactCoverer cov reg with
         None -> None
       | Some id -> gnc#node_from_id id
    in
    match n_opt with
      Some n -> (n,false)
    | None -> 
      let n = gnc#create_node is_entry reg in 
      Region.addCoverer cov reg (get_node_id n); 
      (n, true)
  in
  let (n,is_new) = ac_find_node reg in
  let wkl_add = 
    if is_new then [(get_node_id n)] else 
    if is_entry then (get_node_exits n) else [] 
  in
  List.iter wkl#add_element wkl_add;
  n
  

let post_intra n e_label = 
  (* assert (e_label <> Intra(call/ret edge));*)
  let reg = get_node_region n in
  let (n'_opt,n_ex_opt) = 
    let (reg'_opt,n_ex_o) = 
    match e_label with
      Summary (callexp,_,id_ex,ret_op) ->
        let id = get_node_id n in
        if ( Options.getValueOfBool "sumcache" && is_sumpost_computed (id,id_ex)) then (None,None)
        else
          begin
            add_sumpost_computed (id,id_ex);
            match (gnc#node_from_id id_ex) with None -> failwith "assert_failure!"
            | Some (n_ex) ->
               let n_ex_reg = get_node_region n_ex in
               (Some(Abstraction.summary_post reg n_ex_reg ret_op), Some n_ex)
         end;
    | Intra (op) -> (Some(Abstraction.post reg op gvt),None)
    in
    let n'_o = match reg'_opt with None -> None | Some(reg') -> Some(find_node false reg') in
    (n'_o,n_ex_o) 
  in 
  match n'_opt with None -> () | Some n' ->
    begin
      let _ = 
        match n_ex_opt with None -> () 
        | Some n_ex -> add_node_returns n_ex (get_node_id n') 
      in
      push_entries (get_node_entries n) n';
      Graph.hookup_parent_child n n' e_label
    end
    
let post_exit n ret_op =
  set_node_kind n (Exit ret_op);
  let n_id = get_node_id n in
  let proc_callsite n_en_id (call_exp,call_id) = 
     match gnc#node_from_id call_id with
       None -> ()
     | Some n_call -> post_intra n_call (Summary (call_exp,n_en_id,n_id,ret_op))
  in
  let proc_entry n_en_id =
    match gnc#node_from_id n_en_id with
      None -> ()
    | Some n_en ->
        begin
          add_node_exits n_en n_id;
          List.iter (proc_callsite n_en_id) (get_node_callsites n_en)
        end
  in
  IntSet.iter proc_entry (get_node_entries n)
  
let post_call n op = 
  let reg = get_node_region n in
  let reg_en = Abstraction.post reg op gvt in
  let n_en = find_node true reg_en in 
  let cs = get_node_callsites n_en in
  add_node_callsite n_en (op, get_node_id n);
  add_node_callee n (get_node_id n_en)
    
let process_op n op = 
    match Operation.get_info op with
      Operation.Normal -> post_intra n (Intra op)
    | Operation.Call (_) -> post_call n op
    | Operation.Ret _  -> post_exit n (op)

(***************************************************************)
(********************* Refinement Functions ********************)  
(***************************************************************)

(* A. find_path : node ->  (node * op) list *)


let efilter stk n e_label = 
  match e_label with
  Intra _ -> true
  | Summary(_,_,ex_id,_) -> not (List.mem (get_node_id n,ex_id) stk)
  
let rec expand_se stk (n,e_label) = 
  match e_label with
      Intra(op) -> [(n,e_label)]
    | Summary(call_op,en_id,ex_id,ret_op) ->
        begin
          match Misc.map_partial gnc#node_from_id [en_id;ex_id] with
          [n_en;n_ex] -> 
            let stk' = (get_node_id n,ex_id)::stk in
              let p = sp_intra stk' n_en n_ex in 
                [n,Intra(call_op)] @ p @ [n_ex,Intra(ret_op)]
          | _ -> failwith "expand_se: bad number of nodes!"
        end
and sp_intra stk n_en n_ex =
  let ip = Graph.shortest_path (efilter stk) n_en n_ex in 
  List.flatten (List.map (expand_se stk) ip)
 

  
let get_best_entry n =
  let proc en =
    let en_id = get_node_id en in
    if (en_id = !root_id_ref) then (en,-2,None) else
    match get_node_callsites en with
      [] -> failwith "get_best_entry: non-root with no callsites!"
    | cs -> 
        let ((cexp,id),d) = Misc.list_argmin snd cs in
    (en,d,Some(cexp,id))
  in
  match Misc.map_partial gnc#node_from_id (IntSet.elements (get_node_entries n)) with
   [] -> failwith "get_best_entry: no entries!"
  | entries -> 
      (let ((en,_,call_opt),_) = (Misc.list_argmin Misc.snd3 (List.map proc entries)) in
      assert (call_opt <> None || (get_node_id en = !root_id_ref));
      Message.msg_string Message.Normal ("[GBE] Out Node Depth:"^(string_of_int (get_node_calldepth en)));
      (en,call_opt))

let rec find_path n =
  let (n_en,caller_opt) = get_best_entry n in
  let ip = sp_intra [] n_en n in
  let prefix = 
    match caller_opt with None -> []
    | Some (cexp,id) -> 
        begin
          match gnc#node_from_id id with None -> failwith "caller doesnt exist!"
          | Some (n_call) -> 

              let calledge = (id,get_node_id n_en) in
              (find_path n_call) @ [n_call,Intra(cexp)]
        end
  in
  prefix @ ip
    
(* B: delete nodes that were refined *)

let adjacent n =
  let intra_pred = Graph.get_parents n in
  let intra_succ = Graph.get_children n in
  let inter_pred = 
    let cs = get_node_callsites n in
    Misc.map_partial (fun (_,id) -> gnc#node_from_id id) cs
  in
  let call_succ = 
    let cs = get_node_callees n in
    let proc id = 
      let n_opt = gnc#node_from_id id in
      let (c1,c2) = match n_opt with None -> (true,true)
          | Some n -> ((List.length (get_node_callsites n) = 1),
                        Abstraction.is_fun_refined (get_node_fname n)) in
      let take = match Options.getValueOfString "comp" with 
        "lazy" -> c1 | "demand" -> c1 || c2 | _ -> true
      in if take then n_opt else None
    in 
    Misc.map_partial proc cs
  in 
  let ret_succ = Misc.map_partial (gnc#node_from_id) (get_node_returns n) in
  (intra_pred@inter_pred, intra_succ@call_succ@ret_succ)
  
let rec remove_node n =
  let n_id = get_node_id n in
  if (n_id = !root_id_ref) then 
    let s = "Warning: cannot remove root node!" in
    (Message.msg_string Message.Normal s;Message.msg_string Message.Error s)
  else
    begin
    let (preds,succs) = adjacent n in
    gnc#delete_node (get_node_id n);
    List.iter (fun n -> wkl#add_element (get_node_id n)) preds;
    List.iter remove_node succs
    end

let rec refine_path n_id_l = 
  match Misc.get_first_cont_suffix (gnc#node_exists) n_id_l with [] -> ()
  | h::t ->
      begin
      match gnc#node_from_id h with None -> failwith "assert_failure!"
      | Some(n) -> (remove_node n;refine_path t) 
      end
 
let refine_function fn =
  let entries = gnc#get_entries fn in
  let del_roots = 
    if (List.mem fn (Options.getValueOfStringList "main")) then
    List.flatten (List.map Graph.get_children entries)
    else entries
  in  
  List.iter remove_node del_roots
      
exception FeasiblePath


let feasible_path op_l = 
  let print_op (x : Operation.t) = 
    let (l1,l2,rv) = Abstraction.op_to_string x in
    Message.msg_string Message.Normal (Printf.sprintf "%d :: %s :: %d \n\n" l1 rv l2) 
  in
  Message.msg_string Message.Error "Real Error Found!  :-(";
  List.iter print_op op_l
  


let del_print_debug id = 
  match gnc#node_from_id id with 
    None -> failwith "deleting dead node!" 
  | Some (n) -> Message.msg_printer Message.Debug print_graph_node n 


  
let check_refine n = 
  Message.msg_string Message.Debug "Check Refine:";
  let (n_l',e_l) = Misc.unzip (find_path n) in
  if (Options.getValueOfBool "norefine") then 
    (let s = "No Refine: ctrx size"^(string_of_int (List.length n_l')) in
    Message.msg_string Message.Normal s;
    Message.msg_string Message.Error s;
    raise FeasiblePath);
  let n_l = n_l'@[n] in
  let op_l = 
    List.map (fun e -> match e with Intra o -> o | _ -> failwith "non flat path!") e_l 
  in
  try
    let (rsp,_,_) = Abstraction.block_analyze_trace 
        (List.map get_node_region n_l) op_l  (List.map get_node_id n_l) 
    in
    Message.msg_string Message.Normal ("[CFB] Refining from: "^(string_of_int rsp));
    if (rsp <= 0) then (feasible_path op_l;raise FeasiblePath)    
    else 
      let n_id_l = List.map get_node_id (snd (Misc.cut_list_at_k n_l (rsp - 1))) in
      List.iter del_print_debug n_id_l; (* DEBUG *)
    refine_path n_id_l 
  with Abstraction.RefineFunctionException fn -> refine_function fn 
 
(***************************************************************)
(******************** Top-level MC routine *********************)
(***************************************************************)

let reachloc_table = Hashtbl.create 1001

let add_reachloc reg = 
  let loc_id = Region.get_location_id reg in
  Hashtbl.replace reachloc_table loc_id true


let is_reachable loc_id = Hashtbl.mem reachloc_table loc_id

exception Error of graph_node
(* returns () --> safe | raises Err(n) *)
let mc_loop err =
   while (wkl#has_next ()) do
     Ast.check_time_out ();
     inc stats_nb_iterations;
      if (!stats_nb_iterations mod 100 = 0) then
        Message.msg_string Message.Normal ("While loop tick:"^(string_of_int !stats_nb_iterations));
      let dp_reset = Options.getValueOfInt "simpres" in
      let _ = 
        if dp_reset = 0 then () else
          if (!stats_nb_iterations mod dp_reset = 0) then 
            Abstraction.reset_decision_procedures ()
      in
      let n_id = wkl#next () in
     match gnc#node_from_id n_id with None -> ()
     | Some (n) ->
        let n_region = get_node_region n in
        if (Region.is_empty n_region) then ()
        else
          let err_n = Region.cap n_region err in
          begin
          if not (Region.is_empty err_n) then raise (Error n)
          else
              ( add_reachloc n_region;
                let ops = Abstraction.enabled_ops n_region in
                List.iter (process_op n) ops;)
          end
   done

(* returns () --> safe, or raises FeasiblePath *)
let model_check init err = 
  let rec mc_iter () = 
     try
       Message.msg_string Message.Normal "[CFB] Continue mc_loop";
       mc_loop err; 
       Message.msg_string Message.Normal "Program Safe. Error not reachable! :-)";
       Message.msg_string Message.Error "Program safe. Error not reachable! :-)";
     with Error(n) -> 
       check_refine n;  
       mc_iter ()
  in
  root_id_ref := get_node_id (find_node true init);
  try
    begin
    Message.msg_string Message.Normal "Starting CFB model check";
    Ast.set_time_out_signal ();
    mc_iter ();
    if (Options.getValueOfBool "showtree") then print_reach_graph ();
    1
    end
  with FeasiblePath -> (Message.msg_string Message.Error "Real Error Found! :-(";0)

    

end
