(*
 * Copyright  1990-2002 The Regents of the University of California. All rights reserved. 
 *
 * Permission is hereby granted, without written agreement and without 
 * license or royalty fees, to use, copy, modify, and distribute this 
 * software and its documentation for any purpose, provided that the 
 * above copyright notice and the following two paragraphs appear in 
 * all copies of this software. 
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN 
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE. 
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS 
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION 
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *)

(* $Id: fociModelChecker.ml,v 1.3 2005/05/21 00:35:09 jhala Exp $
 *
 * This file is part of the BLAST Project.
 *)


(**
 * This module implements the foci model checking algorithm.
 *)

open Ast
open BlastArch

module Stats = Bstats
module C_System_Descr = BlastCSystemDescr.C_System_Descr
module C_Command = BlastCSystemDescr.C_Command
    
 
module Make_Foci_Model_Checker =
  functor(Abstraction : ABSTRACTION) ->
struct

  module Region = Abstraction.Region

  module Operation = Abstraction.Operation

  module Tree = BiDirectionalLabeledTree

  type error_path = (Region.t * (Operation.t list) * Region.t)


type marking_data = {
    mutable time_stamp : int ;  (* RJ: Ok, I think the time stamp is useless! *)
    mutable region     : Region.t ;
}
      
let print_marking_data fmt md =
  Format.fprintf fmt "@[time_stamp=%d;@ region=" md.time_stamp ;
  Region.print fmt md.region ;
  Format.fprintf fmt "@]"
    
	  
let print_marking fmt = function
  md ->
      Format.fprintf fmt "@[(" ;
        print_marking_data fmt md ;
      Format.fprintf fmt ")@]" 

type node_data = {
    id           : int ;
    mutable mark : marking_data ; 
  }
		   

let print_node_data fmt nd =
  Format.fprintf fmt "@[Data(@[id=%d;@ mark=" nd.id ;
  print_marking fmt nd.mark ;
  Format.fprintf fmt "@])@]"


type tree_node = (node_data, Operation.t) Tree.node 
    
  (* helper functions to access a tree node's data *)
let get_node_data (n : tree_node) = Tree.get_node_label n
    
let get_op n =
  match (Tree.get_parent_edge n) with
    Some(e) ->
      (Tree.get_edge_label e)
  | None ->
          invalid_arg "get_op: input node has no parent edge"
	
let get_parent n =
  match (Tree.get_parent_edge n) with
    Some(e) ->
      (Tree.get_source e)
  | None ->
      invalid_arg "get_parent: input node has no parent edge"
	
let get_children n = Tree.get_children n
    
let get_id n = (get_node_data n).id
    
let get_marking n = (get_node_data n).mark
    
let get_marking_data n = get_marking n
	  
let get_time_stamp n =
  try (get_marking_data n).time_stamp
  with Invalid_argument "get_marking: input node is unprocessed" ->
    invalid_arg "get_time_stamp: input node is unprocessed"
      
let print_tree_node fmt n =
  Format.fprintf fmt "@[TreeNode(@[label=" ;
  print_node_data fmt (get_node_data n) ;
  begin
    match (Tree.get_parent_edge n) with
      None ->
        Format.fprintf fmt ";@ no parent"
    | Some(e) ->
            Format.fprintf fmt ";@ incoming_op=" ;
        Operation.print fmt ((Tree.get_edge_label e)) ;
        Format.fprintf fmt ";@ parent_id=%d"
          (get_id (Tree.get_source e))
  end ;
  let children_ids = List.map get_id (Tree.get_children n)
  in
  Format.fprintf fmt ";@ children_ids=" ;
      Misc.list_printer_from_printer Format.pp_print_int fmt children_ids ;
  Format.fprintf fmt "@])@]"
    
    
let get_region n =
  try (get_marking_data n).region
  with Invalid_argument "get_marking: input node is unprocessed" ->
    invalid_arg "get_region: input node is unprocessed"
      

  (* helper class to create nodes with appropriate ids *)
class tree_node_creator =
  object
  val mutable next_id = 0
			  
  val tree_node_table = Hashtbl.create 1009
			  (* hash int -> tree node for quick access to a tree node *)
			  (* q: why not an array instead ? -- as nodes get deleted ... This can be changed later *)
			  
  method create_root m =
    let this_id = next_id in
    next_id <- next_id + 1 ;
    let rv = ((Tree.create_root { id  = this_id ;
				  mark = m ;
				}) : tree_node) 
    in
    Hashtbl.add tree_node_table this_id (rv);
    rv
      (* FOCI *)
  method create_child m op par =
    let this_id = next_id in
    next_id <- next_id + 1 ;
    let rv = 
      ((Tree.create_child { id  = this_id ;
			    mark = m ;
			  } op par) : tree_node)
    in
      Hashtbl.add tree_node_table this_id (rv); 
       (* does this store only a pointer ... ? I think so ... but should check *)
      rv
	
  (* Jhala: It would be nice to have a destroy as well -- 
     or are we relying on some sort of Garbage collection to handle that ? *)
  method delete_children par = 
    let children = Tree.get_children (par : tree_node) in
    let child_ids = List.map get_id children in
    let _ = List.iter (Hashtbl.remove tree_node_table) child_ids in
      Tree.delete_children (par : tree_node)

  method node_from_id id = 
    try
      (Hashtbl.find tree_node_table id)
    with Not_found -> failwith "get_node_from_id: unknown id!"
     
  method delete_node preproc n =
    let id = get_id n in
    let pp : unit = preproc n in
    Message.msg_string Message.Debug ("Deleting node: "^(string_of_int id));
    Hashtbl.remove tree_node_table id

  method node_exists n = 
    let i = get_id n in
    Hashtbl.mem tree_node_table i

end

exception RestartException
    
  (* statistic variables *)
let stats_nb_iterations = ref 0
let stats_nb_iterations_of_outer_loop = ref 0 (* added for races *)
let stats_nb_created_nodes = ref 0
let stats_nb_refinment_processes = ref 0
let stats_nb_refined_nodes = ref 0
let stats_nb_proof_tree_nodes = ref 0
let stats_nb_proof_tree_covered_nodes = ref 0
let stats_nb_deleted_nodes = ref 0
let local_restart_counter = ref 1

let reset_stats =
  stats_nb_iterations := 0 ;
  stats_nb_refinment_processes := 0 ;
  stats_nb_created_nodes := 0 ;
  stats_nb_refined_nodes := 0 ;
  stats_nb_proof_tree_nodes := 0;
  stats_nb_proof_tree_covered_nodes := 0
       
let print_stats fmt () =
  Format.fprintf fmt "@[<v>@[Nb iterations of outer while loop:@ %d@]" !stats_nb_iterations_of_outer_loop ;
  Format.fprintf fmt "@ " ;
  Format.fprintf fmt "@[<v>@[Nb iterations of reachability:@ %d@]" !stats_nb_iterations ;
  Format.fprintf fmt "@ " ;
  Format.fprintf fmt "@[Nb created nodes:@ %d@]" !stats_nb_created_nodes ;
  Format.fprintf fmt "@ " ;
  Format.fprintf fmt "@[Nb refinment processes:@ %d@]" !stats_nb_refinment_processes ;
  Format.fprintf fmt "@ " ;
  Format.fprintf fmt "@[Nb refined nodes:@ %d@]" !stats_nb_refined_nodes ;
  Format.fprintf fmt "@ " ;
  Format.fprintf fmt "@[Nb proof tree nodes:@ %d@]" !stats_nb_proof_tree_nodes;
  Format.fprintf fmt "@ " ;
  Format.fprintf fmt "@[Nb proof tree covered nodes:@ %d@]@]" !stats_nb_proof_tree_covered_nodes;
  Format.fprintf fmt "@[Nb deleted nodes:@ %d@]@]" !stats_nb_deleted_nodes;
  ()

  (************************************************************************************************)
  (** Model checker data types **)
type model_check_outcome =
  | No_error of tree_node
  | Error_reached of (error_path * tree_node)


(* To keep the abstraction interface happy, needed in post *)
let global_variable_table = Hashtbl.create 2  

    (* the set of pending nodes to be processed *)
let unprocessed_nodes =  
  if (Options.getValueOfBool "bfs") then new Search.bfsiterator
  else new Search.dfsiterator
  
let cutoff_nodes = new Search.dfsiterator (* aka a stack *)


    (* helper object to create nodes *)
let tnc = new tree_node_creator

    (* the global time used for time stamps *)
let time = ref 0 


let sig_caught_bit = ref false     



  (**************************************************************************************************)

type error_check =
    EmptyErrorAtNode of Region.t * tree_node * (tree_node list)
  | NonEmptyErrorAtRoot of Region.t * tree_node * (tree_node list)

 (*****************************************************************************************************************)
  (************************************  FOCI MODEL CHECKER loop ***************************************************) 
  (*****************************************************************************************************************)

 

let rec reclaim n = 
  if (Tree.has_child n || (not (Tree.has_parent n))) then ()
  else 
    begin
      let par = get_parent n in
	Tree.delete_child n; 
	reclaim par
    end
    

(* covering data structures *)

(* the "reach set" *)
let cov = Region.emptyCoverers ()

let cov_delete n =
  let id = get_id n in
  let reg = get_region n in
  Region.deleteCoverer cov reg id 
  


(*  define maps to track who is covered by whom -- 
    am postponing actually using these data structures *)

(* int -> int list map -- list of nodes that i covers *)
let covers_map = Hashtbl.create 1009

(* int -> int map -- the node that covers i *)
let coveredby_map = Hashtbl.create 1009

let cover_maps_add coverer coveree = 
  let [coverer_id;coveree_id] = List.map get_id [coverer;coveree] in
    Hashtbl.replace coveredby_map coveree_id coverer_id;
    Misc.hashtbl_check_update covers_map coverer_id coveree_id
         
let cover_maps_delete coverer coveree = 
  let [coverer_id;coveree_id] = List.map get_id [coverer;coveree] in
    Hashtbl.remove coveredby_map coveree_id;
    Misc.hashtbl_delete covers_map coverer_id coveree_id
     
(* This adds the successors of a node to the tree *)
let fmc_post reg op = 
    if Options.getValueOfBool "bddpost" then Abstraction.post reg op global_variable_table 
    else Abstraction.foci_post_abstract reg op

let add_children n =
  let n_region = get_region n in
  let add_child op = 
    let new_region = fmc_post n_region op in 
    let n_op_md = 
      { time_stamp = (time := !time + 1 ; !time) ;
        region = new_region; 
      }
    in
      (tnc#create_child (n_op_md) (op)  n)
  in
    List.map add_child (Abstraction.enabled_ops n_region)

    
let delete_subtree n =
  let n_desc = Tree.descendants n in
  let _ = List.iter (tnc#delete_node cov_delete) n_desc in
  Tree.delete_child n 
    
(* this is the core refine function -- used both when:
   1. A counterexample is found,
   2. For covered checking.
*)

(* error_flag = true --> use full trace based update, if required (upon hitting error)
   error_flag = false --> use trel based update only, fail if it fails (for cover checking) *)
                         
     
exception FeasiblePath
(* this raises the exception FeasiblePath *)
let update_path_nodes error_flag n final_reg = 
  let _ = Message.msg_string Message.Debug "In update_path_nodes" in
  let path = Tree.path_to_root n in
  let _ = if error_flag then Message.msg_string Message.Normal ("trace depth: "^(string_of_int (List.length path))) in
  let reg_l = List.map get_region path in
        (* now we have the list of regions *) 
  let op_l  = List.map (get_op) ( List.tl path) in
  let (ref_start_pos, new_reg_l) = 
    let len_reg = List.length reg_l in
    let reg_array = Array.of_list reg_l in
    let _ = if not (Options.getValueOfBool "dasdill") then Array.set reg_array (len_reg - 1) final_reg in
    let op_array = Array.of_list op_l in
    (* Abstraction.fociModelChecker_refine_trace_abstract reg_array op_array *)
    let (rsp_1, nrl_1) = 
      if (Options.getValueOfBool "bddpost") then 
          Abstraction.fociModelChecker_refine_trace_abstract_abstract reg_array op_array 
        else (-1,[])
    in
    if (rsp_1 = -1 && (error_flag || not (Options.getValueOfBool "bddpost")))
    then 
        Abstraction.fociModelChecker_refine_trace_abstract reg_array op_array 
    else (rsp_1,nrl_1)
  in
  let _ = Message.msg_string Message.Debug "Back from fmc_refine_trace" in
    (* this new_reg_l has length one less than path as the thing for the 
     * root node is just true *)
  let update_node node reg = (* we will do the "intersection" here *)
    let _ = Message.msg_string Message.Debug "Updating node:" in
    let old_reg = get_region node in
    let new_reg = Region.cap old_reg reg in
    let _ = Message.msg_string Message.Debug 
	      (Printf.sprintf "Old : %s" (Region.to_string old_reg)) in
    let _ = Message.msg_string Message.Debug
	      (Printf.sprintf "New : %s" (Region.to_string new_reg)) in
      (get_node_data node).mark <- {time_stamp = 0; 
				    region = new_reg} 
  in
    if (ref_start_pos = -1) then raise FeasiblePath
    else
    (try
      if (Options.getValueOfBool "bddpost" && error_flag && Options.getValueOfString "comp" = "cut") 
      then
        begin
         let ref_start_pos = if (Options.getValueOfBool "restart") then 1 else ref_start_pos in 
        let n_pivot = List.nth path ref_start_pos in 
        let n_piv_par = get_parent n_pivot in 
        delete_subtree n_pivot;
        cov_delete n_piv_par;
        unprocessed_nodes#add_element n_piv_par
        end
      else
        List.iter2 update_node (List.tl path) (List.tl new_reg_l)
     with e -> 
       Message.msg_string Message.Error "Failure i update_node !"; 
       raise e)
    

 
let ac_is_covered cover_check_fun n = 
  let _ = Message.msg_string Message.Debug "In ac_is_covered" in
  let n_reg = get_region n in
    if (Region.is_empty n_reg) then true
    else
      let coverable = 
        if (Options.getValueOfBool "cov") then 
          Abstraction.coverable (get_region n) 
        else true
      in
      if (not coverable) then false else
      begin
	let candidates_indices = Region.findCoverCandidates cov (get_region n) in
	let candidates = Misc.map_partial (fun i -> try Some (tnc#node_from_id i) with _ -> None) candidates_indices in
	let rec cover_process l = 
	  match l with
	      [] -> false
	    | h::t -> if cover_check_fun n h then true else cover_process t
	in
        let cand_reg_l = (List.map get_region candidates) in
        if (Options.getValueOfString "covercheck" = "single") then
           cover_process cand_reg_l
        else cover_process [(List.fold_right Region.cup cand_reg_l Region.bot)]
      end

(* check whether the node n is covered -- finds the possible candidates, using cov, and then 
   checks for each candidate using the cover_check_and_update *)
    
let is_covered n =        
  let _ = Message.msg_string Message.Debug "In is_covered" in
  let cover_check_and_update node cand_reg = 
    (* return true if cover success ... in which case all updating is done, false o.w. *)
    let _ = Message.msg_string Message.Debug "In cover_check_and_update" in
    let node_reg = get_region node in
    let (_,cand_loc_s) = Abstraction.reg_to_strings node_reg in
    if (Region.leq node_reg cand_reg) then true
    else if (Abstraction.is_false_adr cand_reg) then false
    else 
    try
      let target_reg = Abstraction.negate_region cand_reg in
      let _ = Message.msg_string Message.Debug ("Candidate region: "^(Region.to_string target_reg)) 
      in
	update_path_nodes false node target_reg;
        let n_data_mark = (get_node_data n).mark in 
        (get_node_data n).mark  <- {n_data_mark with region = cand_reg};
	true 
         (* the path is not feasible, 
           the intermediate nodes get updated,
           but leave the last node region = cover candidate,
           node is covered ... return true ! *)
      with FeasiblePath -> 
        begin
          Message.msg_string Message.Debug ("Cover cand fails at: "^cand_loc_s);
        false
      end       
  in
    ac_is_covered cover_check_and_update n


let is_covered_noupdate n = 
  let _ = Message.msg_string Message.Debug "In is_covered_noupdate" in
  let cover_check node cand_reg = 
    (* returns true if reg.node \leq cand_reg -- doesn't change the tree at all *)
    let _ = Message.msg_string Message.Debug "In cover_check" in
      Region.leq (get_region node) cand_reg
  in
    ac_is_covered cover_check n

let fmc_is_covered  () n = 
    if Options.getValueOfBool "bddpost" then is_covered_noupdate n
    else is_covered n

let output_path n = 
  let path = Tree.path_to_root n in
  let op_l = (List.map (get_op) ( List.tl path)) in
  let print_op (x : Operation.t) = 
    let (l1,l2,rv) = Abstraction.op_to_string x in
    Message.msg_string Message.Normal (Printf.sprintf "%d :: %s :: %d \n\n" l1 rv l2) 
 in
  List.iter print_op op_l


  
let print_tree_from_root root = 
  let nl_to_string nl = 
    let (regs_l,loc_s) = Abstraction.reg_to_strings (nl.mark.region) in
    let id = nl.id in
    Printf.sprintf "%d $$ (%s,%s)" id (loc_s) (Misc.strList regs_l)
  in
  let el_to_string el =
    let (_,_,s) = Abstraction.op_to_string el in
        s
  in
  Message.msg_string Message.Normal "**************************************************";
  Message.msg_string Message.Normal "**** Printing the Reachability Tree **************";
  Message.msg_string Message.Normal "**************************************************";
  let tree_size_ref = ref 0 in
  let node_iter n = 
    tree_size_ref := !tree_size_ref + 1;
    Message.msg_printer Message.Normal print_tree_node n
  in
    Tree.iter_descendants_nodes node_iter root;
    let name = "reachtree" in
    Tree.output_tree_dot nl_to_string el_to_string (name^".dot") [root];
    Sys.command (Printf.sprintf "dot -Tps -o %s.ps %s.dot" name name);
    Message.msg_string Message.Normal ("Tree size: "^(string_of_int !tree_size_ref))

exception PostCheckException of (tree_node * tree_node)

exception CoveredCheckException of tree_node

(* WARNING:DNF *)
let check_tree root = 
  let _ = Message.msg_string Message.Normal "Checking Tree Validity" in
  let reach_reg_ref = ref Region.bot in
  (* first, just check that each edge, is good -- then we have to check the leaves *)
  let check_post (n1, op , n2) = 
    let _ = Message.msg_string Message.Debug ("Checking Node: "^(string_of_int (get_id n2))) in
    let reg = get_region n1 in
    let reg' = get_region n2 in
    if not (Abstraction.check_abstract_transition reg op reg') then
      raise (PostCheckException (n1,n2))
  in
  let gather_reach n = 
    if BiDirectionalLabeledTree.has_child n then 
      reach_reg_ref := Region.cup !reach_reg_ref (get_region n)
  in
  let check_cov n = 
    if BiDirectionalLabeledTree.has_child n then () 
    else 
      if not (Region.leq (get_region n) !reach_reg_ref) then
      raise (CoveredCheckException n)
  in
    try
      Message.msg_string Message.Error "WARNING: not checking edges TBD:FOCIDAG";
      (* BiDirectionalLabeledTree.iter_descendants_edges check_post root;*)
      BiDirectionalLabeledTree.iter_descendants_nodes gather_reach root;
      BiDirectionalLabeledTree.iter_descendants_nodes check_cov root; ()  
    with
	CoveredCheckException n -> 
	  begin
	    Message.msg_string Message.Error "Covered Check Fails!!";
	    Message.msg_printer Message.Error print_tree_node n
	  end
      | PostCheckException (n1,n2) ->
	  begin
	    Message.msg_string Message.Error "Post Check Fails!!";
	    Message.msg_string Message.Error "Source Node:";
	    Message.msg_printer Message.Error print_tree_node n1;
	    Message.msg_string Message.Error "Target Node:";
	    Message.msg_printer Message.Error print_tree_node n2
	  end

let model_check init err = 
  let tree_root = 
    tnc#create_root { time_stamp = !time; region = init }
  in
    unprocessed_nodes#add_element (tree_root) ;
    let rec mc_iter () = 
      while (unprocessed_nodes#has_next ()) do
      (* begin debug info *)
      stats_nb_iterations := !stats_nb_iterations + 1;
      if (!stats_nb_iterations mod 100 == 0) then
	begin
          Message.msg_string Message.Normal ("While loop tick:"^(string_of_int !stats_nb_iterations));
	  flush stdout; flush stderr;
	end ;
      Message.msg_string Message.Debug 
	"\n\n****************************************************************************\n" ;
      Message.msg_string Message.Debug "Next iteration of foci-model-check's big while-loop" ;
      Message.msg_printer Message.Debug Format.pp_print_int (!stats_nb_iterations);
      (* end debug info *)
      (* let's process the next pending node *)
      let (n : tree_node) = unprocessed_nodes#next () in
      if not (tnc#node_exists n) then ()
      else
        begin
      let _ =
	Message.msg_string Message.Debug (*Debug Normal*) "Now processing tree node:" ;
	Message.msg_printer Message.Debug (*Debug Normal*) print_tree_node n 
      in
      let n_region = get_region (n : tree_node) in
      let err_n = Stats.time "cap" (Region.cap n_region) err in
	if (Region.is_empty err_n)
	then
          (* no error found at this node *)
          let cov_check = Stats.time "fmc_is_cov" (fmc_is_covered ()) n in
	  if cov_check then 
	    begin
              let (_,n_loc_s) = Abstraction.reg_to_strings (get_region n) in
	      Message.msg_string Message.Debug 
              (Printf.sprintf "Covered node: %d : %s" (get_id n) n_loc_s);
	      cutoff_nodes#add_element n
	    end
	  else 
	    begin
	      Message.msg_string Message.Debug "Not covered -- add children etc.";
	      Region.addCoverer cov (get_region n) (get_id n);
	      unprocessed_nodes#add_list (add_children n) 
	    end
	else
	  try
            Message.msg_string Message.Normal "Hit error: update";
	    update_path_nodes true n (Abstraction.erase_data (get_region n))
	  with FeasiblePath -> 
                begin
	      Message.msg_string Message.Error "Real Error Found! :-(";
	      output_path n; (* this is here because n is in scope here *)
              raise FeasiblePath
            end
        end  
     done;
    (* Now check if the cutoff nodes are really cut off -- if not, add them to the set of unprocessed nodes *)
    (* This check should be the side-effect free -- i.e. is there someone in the tree that covers them -- without
       inducing further refines along the way *)

    let filter_fun cov_check n = 
      let rv = 
	if cov_check n then true else 
	  (* if the node isn't "directly" covered, then we put it in the 
	     worklist, and when we process it subsequently, we will check if 
	     it is "indirectly" covered *)
	  begin
            Message.msg_string Message.Normal (Printf.sprintf "filter fails: adding leaf node %d" (get_id n));
	    unprocessed_nodes#add_element n;
	    false
	  end
      in
      let _ = Message.msg_string Message.Debug (Printf.sprintf "Really covered check: %d %b " (get_id n) rv) in
	rv
    in
    (* if bddpost is false, then unproc_nodes = 0 at this point, hence just do
        the old thing -- i.e. make a pass using is_covered_noupdate ..
        if bddpost is true, then first make a pass using covered_update. 
                Now, if the unproc is not empty, then repeat mc ..
                     if it is empty, then, make a pass using noupddate, to be
                     double sure... 
        why this will terminate:
          claim: once you have computed the reachset in the tree, the only way a
          node gets added is if i fails the first check. this cannot happen once
          the trel is precise enough. So, rtp, eventually the trel is
          precise enough... suppose you hit error only finitely often. then you
          are done as you will find a fixpoint. how to hit error infinitely
          often...yak yak. TBProved properly. *)
    let cutoff_check () =
      cutoff_nodes#filter (fun n -> tnc#node_exists n); (* first, remove the "danglers" *)
      if (Options.getValueOfBool "bddpost") then cutoff_nodes#filter (filter_fun is_covered);
      if not (unprocessed_nodes#has_next ()) then cutoff_nodes#filter (filter_fun is_covered_noupdate);
    in  
    cutoff_check ();  
    if unprocessed_nodes#has_next () then 
        begin Message.msg_string Message.Normal "uncovered leaves -- repeat mc"; mc_iter () end
      else ()
  in
    try
      mc_iter ();
      if (Options.getValueOfBool "showtree") then print_tree_from_root tree_root;
      if (Options.getValueOfBool "checktree") then Stats.time "check tree" check_tree tree_root;
      if (Options.getValueOfBool "savetree") then 
        begin
          let nodes = Tree.descendants tree_root in
          List.iter (fun n -> Abstraction.save_region_abs (get_region n)) nodes;
          Abstraction.dump_abs ()
        end;
      let size = Tree.subtree_size tree_root in
      let depth = Tree.subtree_depth tree_root in
      let reach_cubes =
        let c = ref 0 in
        let _do n = 
          let adr = Abstraction.extract_abs_data_reg (get_region n) in
          c := !c + (Region.adr_cubes adr)
        in 
        Tree.iter_descendants_nodes _do tree_root;!c
      in
      Message.msg_string Message.Normal ("Tree size: "^(string_of_int size));
      Message.msg_string Message.Normal ("Tree depth: "^(string_of_int depth));
      Message.msg_string Message.Normal ("Tree cubes: "^(string_of_int reach_cubes));
      Message.msg_string Message.Normal "Error not reachable! Program Safe :-)";
      Message.msg_string Message.Error "Program safe. Error not reachable! :-)";
    with FeasiblePath -> raise FeasiblePath
   	(* RJ: Should return "Error_Reached etc. etc." *)
end
