(*
 * Copyright  1990-2002 The Regents of the University of California. All rights reserved. 
 *
 * Permission is hereby granted, without written agreement and without 
 * license or royalty fees, to use, copy, modify, and distribute this 
 * software and its documentation for any purpose, provided that the 
 * above copyright notice and the following two paragraphs appear in 
 * all copies of this software. 
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN 
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE. 
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS 
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION 
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *)

(**
   aliasAnalyzer.ml (the engine that produces the alias database
   and answers alias queries).

   Depends on Bddptsto
   *)
  
open Ast
module Stats = Bstats


(******************************************************************************************)
(********** Code to "manually override" the alias analysis ********************************)
(******************************************************************************************)
(* SLOW *)
type aliasedge_t = bool
type aliasnode_t = ((int * Expression.expression),aliasedge_t) BiDirectionalLabeledGraph.node
let exp_node_table = Hashtbl.create 31 

let exp_ctr_ref = ref 0 

let make_exp_node e =
  try Hashtbl.find exp_node_table e 
  with Not_found ->
    exp_ctr_ref := !exp_ctr_ref + 1;
    let n = BiDirectionalLabeledGraph.create_root (!exp_ctr_ref,e) in
    Hashtbl.add exp_node_table e n;
    n
  
let add_aliases e_list yesno_flag =
  let link flg (n1,n2) = 
    BiDirectionalLabeledGraph.hookup_parent_child n1 n2 flg;
    BiDirectionalLabeledGraph.hookup_parent_child n2 n1 flg
  in
  let n_list = List.map make_exp_node e_list in
  let n_pairs_list = Misc.make_all_pairs n_list in
  List.iter (link yesno_flag) n_pairs_list


(* 1 = yes, -1 = no, 0 = don't know, for now, either YES or don't know *)
let override_alias e1 e2 = 
  try
    begin
    let [n1;n2] = List.map (Hashtbl.find exp_node_table) [e1;e2] in
    match (BiDirectionalLabeledGraph.find_edges n1 n2) with
    [] -> 0
    | e_l -> if (List.exists (fun e -> BiDirectionalLabeledGraph.get_edge_label e)) e_l then 1 else (-1)
    end
  with _ -> 0

let add_alias_override aliasPreds =
  let proc_atom p = 
    match p with
    Predicate.Atom(Expression.Binary (binop,e1,e2)) -> add_aliases [e1;e2] (binop = Expression.Eq)
    | _ -> failwith ("Unimpl. alias predicate :"^(Predicate.toString p))
  in
  List.iter proc_atom (Predicate.getAtoms aliasPreds)

(* ONLY USE: add_alias_override, is_override_alias *)

(******************************************************************************************)

let listOfAliases = ref []
let pointsToInfo : Bddptsto.pointsToInfo ref = ref (Bddptsto.FlowInsensitivePointsTo Bddptsto.bddZero)

let constructAliasDatabase () =
  let aliasFile = Options.getValueOfString "alias" in
  if (aliasFile = "") then
    (* Nothing to do *)
    ()
  else
    if (aliasFile = "bdd") then
    begin
      (* do the bdd based alias analysis *)
	Message.msg_string Message.Normal "Starting the alias analysis" ;
      pointsToInfo := Bddptsto.do_bdd_alias (Options.getValueOfString "pta" = "flow") 
    end
    else 
      begin (* the bdd has been dumped previously *)
	Options.setValueOfString "alias" "bdd" ;
        pointsToInfo := Bddptsto.load_FI_points_to_bdd aliasFile 
      end
(***************************
	Commenting out hokey alias file mode -- Rupak
      else
      begin 
    (* add each pair to the alias database *)
       (* we simulate the effect of an alias analysis by letting the user input a file 
	  with alias pairs.
        *)
	let readAliasPairs filename =
	  let doParse inchan =
	    let _ = VampyreErrormsg.startFile filename in
	    let lexbuf = Lexing.from_channel inchan in
	    let _ = VampyreErrormsg.theLexbuf := lexbuf in
	    Inputparse.main Inputflex.token lexbuf 
	  in
	  ignore (Message.msg_string Message.Normal "Reading in alias pairs...");
	  try
	    let inchan = open_in filename
	    in
	    let a = doParse inchan
	    in
	    ignore (Message.msg_string Message.Normal "Alias pairs read.\n") ;
	    List.iter (fun a -> Message.msg_string Message.Normal (Predicate.toString a)) a ;
	    a
	  with
	    Sys_error _ -> (Message.msg_string Message.Error ("Cannot find alias information in "^filename^".\n"); [])
	  | VampyreErrormsg.Error -> (Message.msg_string Message.Error ("Error raised in reading alias pairs.\n"); [])
	in
	
	let rec loopOverAliasPairs lst =
	  match lst with
	    [] -> ()
	  | p :: rest ->
	      begin
		match p with
		  Predicate.Atom a ->
		  begin
		    match a with
		      Expression.Binary (Expression.Eq, Expression.Lval l1, Expression.Lval l2) ->
			listOfAliases := (l1, l2) :: !listOfAliases ; loopOverAliasPairs rest
		    | _ -> failwith ("Strange expression read by readAliasPairs "^(Expression.toString a))
		  end
		|	_ -> failwith ("Strange predicate read by readAliasPairs "^(Predicate.toString p))
	      end
	in
	loopOverAliasPairs (readAliasPairs aliasFile)(* ;
							transitivelyClose !listOfAliases *)
      end
********************************************)

       	
let askAliasDatabase l1 l2 =
(*
  Message.msg_string Message.Debug "In askAliasDatabase" ;
  Message.msg_string Message.Debug ("Lvalues are "^(Expression.lvalToString l1)^" and "^(Expression.lvalToString l2)) ;
*)
  if Options.getValueOfString "alias" = "bdd" then
    begin
      try 
	let t1 = BlastCSystemDescr.C_System_Descr.get_type_of_lval l1 
						      and 
	  t2 = BlastCSystemDescr.C_System_Descr.get_type_of_lval l2 
	in
	  if t1 = t2 then (* only check aliasing between values of the same type *) 
	    match !pointsToInfo with
		Bddptsto.FlowSensitivePointsTo locbddhashtbl ->
		  (* Not implemented yet *)
		  false
	      |	Bddptsto.FlowInsensitivePointsTo ptstobdd ->
		  (* Look up aliasing relationship from pointsToBdd *)
		  Bddptsto.checkAliasFI (ptstobdd) l1 l2
	  else
	    begin
	      Message.msg_string Message.Debug "The types are different. Hence not aliased." ;
	      false
            end
      with e -> 
	begin
	  (*Message.msg_string Message.Normal ("get_type fails!"^ (Printexc.to_string e));  unsound! -- assuming that I dont know what type one of them is... *)
	  false 
	end

    end
  else
	failwith "No alias information present!" 

  (*
    let rec loop l1 l2 lst =
    match lst with 
    [] -> false
    | (l, l') :: rest ->
    begin
    if ((l = l1) && (l' = l2)) || ((l=l2) && (l'=l1)) then true
    else
    loop l1 l2 rest
    end
    in
    loop l1 l2 !listOfAliases
  *)

let list_of_functions_whose_address_is_taken = ref None

(* given an lvalue, find all functions this lvalue can be pointing to *)
let getFunctionsPointedTo (fcexp : Ast.Expression.expression) =
  Message.msg_string Message.Error ("getFnPtTo called with "  ^ (Expression.toString fcexp)) ;
  let get_pts_to l flist = 
      Message.msg_string Message.Debug ("l is : " ^ (Expression.lvalToString l)) ;

      match !pointsToInfo with
	Bddptsto.FlowSensitivePointsTo locbddhashtbl ->
	  []
      |	Bddptsto.FlowInsensitivePointsTo ptstobdd ->
          List.filter (fun f -> (* type based disambiguation *)
            Bddptsto.checkPointsToFI ptstobdd l f 
            && 
	    (
            Message.msg_string Message.Debug ("f is : " ^ (Expression.lvalToString f)) ;
            let t1 = BlastCSystemDescr.C_System_Descr.get_type_of_lval l 
            and t2 = BlastCSystemDescr.C_System_Descr.get_type_of_lval f in
            Message.msg_string Message.Debug  
            ("Type t1 is "^(Pretty.sprint ~width:256 (Pretty.dprintf "%a\n"
            Cil.d_type  t1))) ;
            Message.msg_string Message.Debug
            ("Type t2 is "^(Pretty.sprint ~width:256 (Pretty.dprintf "%a\n"
            Cil.d_type  t2))) ;
            (* t1 = t2 : This isn't good enough --- because of physical subtyping *)
            match (t1, t2) with
              (Cil.TFun (rtyp1, vlist1, _, _), Cil.TFun (rtyp2, vlist2, _, _) ) -> 
                (rtyp1 = rtyp2) && ( match (vlist1, vlist2) with (Some l1, Some l2) -> (List.length l1 = List.length l2) | (None, None) -> true | _ -> false)
            | _ -> false
            ) ) 
          flist 
  in
  let l = (match fcexp with 
             Expression.Lval lv -> lv 
           | _ -> failwith ("getFunctionsPointedTo: Strange fn ptr call: " ^ (Expression.toString fcexp))) in
  match !list_of_functions_whose_address_is_taken with
    None ->
     (* filter list of functions *)
      let fl = (List.map (fun s -> Expression.Symbol s) (BlastCSystemDescr.C_System_Descr.list_of_functions ())) in
      list_of_functions_whose_address_is_taken := Some fl ;
      let fptrs = get_pts_to l fl in
(*
      List.iter (fun f -> Message.msg_string Message.Error ("Fn ptr: " ^ (Expression.lvalToString f) )) fptrs ;
*)
      (match fptrs with [] -> failwith ("Function pointer call with no aliased function! " ^ (Expression.toString fcexp))
      | _ -> fptrs)
  | Some fl ->
      let fptrs = get_pts_to l fl in
(*
      List.iter (fun f -> Message.msg_string Message.Error ("Fn ptr: " ^ (Expression.lvalToString f)) ) fptrs ;
*)
      (match fptrs with [] -> failwith ("Function pointer call with no aliased function! " ^ (Expression.toString fcexp))
      | _ -> fptrs)


let getAllAliases lv =
  if Options.getValueOfString "alias" = "bdd" then
    begin
      failwith "getAllAliases : bdds : Not implemented yet" ;
      match !pointsToInfo with
	Bddptsto.FlowSensitivePointsTo locbddhashtbl ->
	  []
      |	Bddptsto.FlowInsensitivePointsTo ptstobdd ->
          (* Look up aliasing relationship from pointsToBdd *)
	  []
    end
  else
    let lst =
      Misc.union 
	(List.map snd (List.filter (fun (l,l') -> l = lv) !listOfAliases))
	(List.map fst (List.filter (fun (l,l') -> l' = lv) !listOfAliases))
    in
    Message.msg_string Message.Debug ("In getAllAliases with "^(Expression.lvalToString lv)) ;
    Message.msg_string Message.Debug ("Aliases are ") ;
    List.iter (fun l -> Message.msg_string Message.Debug (Expression.lvalToString l)) lst ;
    lst
      
let can_escape lv globals =
  Message.msg_string Message.Debug ("In can_escape with "^Expression.lvalToString lv) ;
  let rec can_escape_worker lv globals seen =
    let rec exp_can_escape e globals =
      match e with
      |	Expression.Lval l -> can_escape_worker l globals seen
      |	Expression.Chlval (l,_) -> 
	  begin
	    can_escape_worker l globals seen
	  end
      | Expression.Binary (_, e1, e2) -> (exp_can_escape e1 globals) || (exp_can_escape e2 globals)
      |	Expression.Cast (t, e1) -> exp_can_escape e1 globals
      | Expression.Constant(_) -> false
      |	Expression.Constructor (_, _) -> failwith "Constructor not handled in can_escape"
      | Expression.FunctionCall (func, args) ->
	  begin
	    let rec loop lst =
	      match lst with 
		[] -> false
	      | a:: rest -> if exp_can_escape a globals then true else loop rest
	    in
	    exp_can_escape func globals || (loop args) 
	  end
      | Expression.Unary(_, e) -> 
	  exp_can_escape e globals
      |	_ -> failwith "can_escape : Strange expression"
    in
    if List.mem lv seen then
      false else
      begin
	let escapes_directly = 
	  match lv with
	    Expression.Symbol s -> Hashtbl.mem globals lv 
	  | Expression.Dereference e -> exp_can_escape e globals 
	  | Expression.Access (_, e, _) -> exp_can_escape e globals
	  | Expression.Indexed (e1, e2) -> failwith "Indexed not handled in can_escape"
	  | Expression.This -> failwith "This not handled in can_escape"
	in
	escapes_directly || 
	(List.exists (fun l -> 
	  can_escape_worker l globals (lv::seen)) (getAllAliases lv))
      end
  in
  let r = can_escape_worker lv globals []
  in
  Message.msg_string Message.Debug ("can escape returns "^(if r then "true" else "false"));
  r
    
let rec generateAliasConstraints (var1 : Expression.lval) (var2 : Expression.lval) : Predicate.predicate =
  match (var1, var2) with
  | Expression.Symbol v, Expression.Symbol v' -> if askAliasDatabase var1 var2 then Predicate.True else Predicate.False
  | Expression.Access (Expression.Dot, Expression.Lval l, fld1), Expression.Access (Expression.Dot, Expression.Lval l', fld2)
  | Expression.Access (Expression.Arrow, Expression.Lval l, fld1), Expression.Access (Expression.Arrow, Expression.Lval l', fld2)->
      if fld1 = fld2 then
	generateAliasConstraints l l'
      else
	Predicate.False
  | Expression.Indexed (e1, e2), Expression.Indexed (e1', e2') ->
      Predicate.conjoinL 
	[ Predicate.Atom (Expression.Binary (Expression.Eq, e1, e1')); 
	  Predicate.Atom (Expression.Binary (Expression.Eq, e2, e2'))]
  | _ -> 
      if askAliasDatabase var1 var2 then Predicate.True
      else Predicate.False (* not aliased *)



      
  
let  queryAlias (_v1 : Expression.expression) (_v2 : Expression.expression) = 
  let _ = Message.msg_string Message.Debug ("Alias query for "^(Expression.toString _v1)^" and "^(Expression.toString _v2)) in
  let override = 
    if (Options.getValueOfString "aliasfile" <> "") then override_alias _v1 _v2 else 0
  in
  if override <> 0 then 
    begin
    Message.msg_string Message.Debug (Printf.sprintf "Overriding: %s %s (%d)" (Expression.toString _v1) (Expression.toString _v2) override );
    override > 0
    end
  else
  (* now do the real thing *)
  (* symbolic check *)
  let is_sym1 =
    try Expression.is_symbolic (Expression.lv_of_expr _v1)
    with Not_found -> false
  in
  let is_sym2 =
    try Expression.is_symbolic (Expression.lv_of_expr _v2)
    with Not_found -> false
  in
    if (is_sym1 || is_sym2) then false (* Nothing can be aliased to a symbolic constant -- note, we are talking about
					  a symbolic constant, NOT derefs etc. of symbolic constants *)
    else
      let [v1;v2] =
	if (Options.getValueOfBool "cf")
	then
	  List.map
	    (Expression.deep_alpha_convert
	       (fun lv -> Expression.Lval (Expression.peel_symbolic lv)))
	    [_v1;_v2]
	else [_v1;_v2]
      in
	Message.msg_string Message.Debug ("After peeling: "^(Expression.toString v1)^" and "^(Expression.toString v2)) ;
	     let lvalOfExp e =
	       match e with 
		   Expression.Lval l -> l
               | _ -> failwith ("queryAlias: bug in replaceChlvals?" ^ (Expression.toString e))
	
	     in
	       try
		 let var1 = lvalOfExp (Expression.replaceChlvals v1) 
		 and var2 = lvalOfExp (Expression.replaceChlvals v2) 
		 in
		   match (var1,var2) with
		       (Expression.Symbol _ , Expression.Symbol _) -> false
			 (*RJ: two symbols x,y can only be aliased if they are
                         * synt. identical, or identical after "de-symconsting".
                         * In either case, writing to x, either directly affects "y" 
                         * (if they are identical), i.e. is something the caller knows,
			 * or, doesnot affect y -- i.e. they are NOT aliased, and so 
                         * w.log, knowing who is going to call this, we can safely return false.
			 * But really, one should make this function say yes, for the 
                         * same reasons as it should say yes for syntactic identity 
                         * and the case should be filtered out later. *)
		     | _ ->
			 
			 let a' = 
			   if var1 = var2 then
			     (* RJ: silly hack to get around this identity semantics! 
                              * this should say TRUE if =, the function calling it 
                              * should decide what to do if they are the same!!! *)
			     (if  (_v1 <> _v2) then true else false) (* RJ: Is this function even called with Chlvals ? *)
			   else (* RJ: why is this false ? 
				   Answer: If var1 is syntactically equal to var2, then
				   we substitute var2 anyway; and moreover var1 not = var2
				   is false. So in this case we can do without aliasing.
				   RJ: maybe -- but the semantics are wrong. we need a function that says YES
				   if the two lvals are aliased -- without these special cases.
				*)
			     begin
                               let askdb () = askAliasDatabase var1 var2 in
                               let rv = VampyreStats.time "ask aliasdatabase" askdb ()
                               in
                               rv
			     end
			 in
			   if a' = true then 
			     Message.msg_string Message.Debug ("Got back alias answer true for "^(Expression.toString v1)^" and "^(Expression.toString v2)) ;
			   let a''_3v = (* 1 = TRUE, 0 = Dont know, -1 = FALSE *)
			     match (var1,var2) with
				 (Expression.Dereference (Expression.Binary(Expression.Plus,a1,o1))
				    ,Expression.Dereference (Expression.Binary(Expression.Plus,a2,o2))) ->
                                      if (a1 = a2) then 1 else 0
			       | (Expression.Access(_,Expression.Lval(Expression.Dereference(Expression.Binary(Expression.Plus,a1,o1))),
                               f1),
				  Expression.Access(_,Expression.Lval(Expression.Dereference(Expression.Binary(Expression.Plus,a2,o2))), f2))
				 -> if (a1=a2 && f1 = f2) then 1 else 0
                                 (* JHALA: pointer arith + field *)
                               | (Expression.Access(_,Expression.Lval(Expression.Indexed(e1,e1_idx)),f1),
				  Expression.Access(_,Expression.Lval(Expression.Indexed(e2,e2_idx)),f2))
                               -> if (e1=e2 && f1 = f2) then 1 else 0
                                 (* JHALA: index array + field *)

                               
                               | (Expression.Access(_,Expression.Lval lv1,f1),Expression.Access(_,Expression.Lval lv2,f2)) ->
                                   if (f1<>f2) then (-1) (* RJ: change! *)
                                   else (if (askAliasDatabase lv1 lv2) then 1 else (-1)) 
                                   
                               | (Expression.Indexed(e1,e2),Expression.Indexed(e1',e2')) -> if (e1 = e1') then 1 else 0
		               | _ -> 0
			   in 
			   let a = 
                             let rv = 
                               match a''_3v with
                                 0 -> a'
                               | 1 -> true
                               | (-1) -> false
                               | _ -> a'
                             in 
                             rv && (not (var1 = var2))
                           in
			     Message.msg_string Message.Debug (Printf.sprintf "QueryAlias %s %s (%s)\n\n" (Expression.lvalToString var1) (Expression.lvalToString var2) (if a then "true" else "false")); 
			     a
		   with _ -> 
		     begin
		       Message.msg_string Message.Debug "queryAlias raises exception for the query" ;
		       Message.msg_string Message.Debug ((Expression.toString v1) ^" and " ^ (Expression.toString v2)) ;
		       Message.msg_string Message.Debug "returning false" ;
		       false  
		     end
		     
	   let modified_info_bdd = ref Bddptsto.bddZero
				     
	   let constructModifiesInfo () =
		failwith "Don't do this any more" ;
	     (* construct the modified information for each function.
		Basically build a BDD f(l1, l2), where the first
		component is a function name, and the second component
		is an lvalue whose transitive dereference may be modified by the function.
		The second lvalue is assumed to be a formal parameter of the first,
		or a global.

		This requires alias information.
     *)
	     (* we do a simple data flow analysis.
		Initially, we the bdds are all false, and we populate the BDDs by going through
		the cfa until we get a fixpoint.
	     *)
	     (match !pointsToInfo with
		  Bddptsto.FlowSensitivePointsTo _ ->  ()
		| Bddptsto.FlowInsensitivePointsTo ptstobdd ->
		    modified_info_bdd := VampyreStats.time "mod analysis" Bddptsto.do_mod_analysis ptstobdd);
	     (* Test *)
	     CaddieBdd.bddPrint !modified_info_bdd ;
	     Bddptsto.test_mod_analysis !modified_info_bdd
	       
	       
	   let is_modified fname lval =
	     Bddptsto.is_modified !modified_info_bdd fname lval
	       

(********** RJ: this below function is cut out of abstraction.ml because it really should
  be here
*)


 (* this table maps an lval to the set of lvals it may alias... *)
	 (* this table never decreases! Its never cleared! *)
 let lval_alias_table = Hashtbl.create 1009

 (* I tried to be nice about this but circular dependency means I cant be. 
    type iter_key_t = AllLvals | ScopeLvals | TraceLvals
    These are now defined as integers in Misc.ml ...
 *)


 (* this function is ONLY called during REFINE. During image computation,
    the set of relevant preds is culled from the predicates at hand *)
 let stats_nb_alias_query = ref 0
			      
 (* assume that the lval_iterator is CLOSED under closure *)
 let get_lval_aliases_iter iter_key lval_iterator lval = (* TBD:SPEED *)
   let _foo () = 
   if (Options.getValueOfString "alias" = "" (*|| Options.getValueOfBool
   "incref"*))  then [lval]
   else 
     begin
       try
	 Message.msg_string Message.Debug
	   ("get_lval_aliases_iter: "^(Expression.lvalToString lval));
	 Hashtbl.find lval_alias_table (iter_key,lval) 
       with
	   Not_found ->
	     (* first a rather slow way to do things *)
	     let aliases = ref [] in
	     let _ =
	       Stats.time "Lval iterator" 
		 lval_iterator 
		 (fun x ->
		    
		    let _ = (stats_nb_alias_query := !stats_nb_alias_query + 1) in
		      if queryAlias
			(Expression.Lval x)
			(Expression.Lval lval) 
		      then aliases := x::!aliases;)
	     in
	     let rv = Misc.sort_and_compact (lval::!aliases)
	     in (* NOTE that this always adds lval into the set returned *)
	     let rv_l = List.map Expression.lvalToString rv in
	       Message.msg_string Message.Debug ("alias set: "^(Misc.strList rv_l));
	       Hashtbl.replace lval_alias_table (iter_key,lval) rv;
	       rv
     end
    in Stats.time "get_lval_aliases_iter" _foo () 

 (* TBD: alias-union-find *)    
 (* THIS IS SLOW! please try not to use it! *)    
 let get_lval_aliases =
   Stats.time "get_lval_aliases_iter" (get_lval_aliases_iter Misc.allLvals_t_c)
     BlastCSystemDescr.C_System_Descr.iterate_all_lvals
     
 let get_lval_aliases_scope fname =
   Stats.time "get_lval_aliases_iter[scope]" (get_lval_aliases_iter Misc.scopeLvals_t_c)
     (BlastCSystemDescr.C_System_Descr.iterate_all_scope_lvals fname)
    
(***********************************************************************************)
(* Must aliases for temporary mem variables added by Cil                           *)
(* *)
(***********************************************************************************)

let rec is_tempmem lv =
  match lv with
  | Expression.This -> false 
  | Expression.Symbol s -> 
      Misc.is_prefix "mem_temp" s 
  | Expression.Dereference e 
  | Expression.Access (_, e, _) 
  | Expression.Indexed (e, _) ->
      is_tempmem_exp e
and is_tempmem_exp e =
  match e with 
   Expression.Lval l -> is_tempmem l
  | Expression.Chlval _ 
  | Expression.Constant _ -> false
  | _ -> false (* failwith ("is_tempmem: not handled "^(Expression.toString e)) *)

let mustAliasesHashtbl = Hashtbl.create 31 

module C_System_Descr = BlastCSystemDescr.C_System_Descr

(* Must aliases are computed only for temporary variables introduced by Cil *)
let constructMustAliasDatabase () = 
  let work_edge e =
    (* add the RHS's of any mem_temp lval assignments to the table. *)
    let c = (BlastCSystemDescr.C_System_Descr.get_command e) in
    let code = c.C_System_Descr.Command.code in
    match code with
      BlastCSystemDescr.C_System_Descr.Command.Block l ->
	begin
	  let do_assignment a =
	    match a with
	      BlastCSystemDescr.C_System_Descr.Command.Expr (Expression.Assignment(_, target, rhs)) ->
		begin
		  if is_tempmem target then (* bingo *)
                    begin
                      match rhs with
                        Expression.Lval l -> 
			  if Hashtbl.mem mustAliasesHashtbl target then (* sanity check -- can be discarded *)
                            failwith ("mem_temp recycled? "^(Expression.lvalToString target))
                          else
		            Hashtbl.add mustAliasesHashtbl target [ l ]
		      | _ -> ()
                    end
		  else () 
		end
            | _ -> ()
          in
	  List.iter do_assignment l
	end
    | _ -> () (* No other command can cause a must alias *)
  in
  BlastCSystemDescr.C_System_Descr.map_edges work_edge


let get_all_must_aliases lv =
  let myalias =
    if (Hashtbl.mem mustAliasesHashtbl lv) then
    begin
      Message.msg_string Message.Debug ("get_all_must_aliases hit table with lvalue "^(Expression.lvalToString lv)) ;
      let aliases = Hashtbl.find mustAliasesHashtbl lv in
      Message.msg_string Message.Debug "Aliases are :" ;
      List.iter (fun l -> Message.msg_string Message.Debug (Expression.lvalToString l)) aliases ;
      aliases
    end
  else []
  in
  (* additional aliases may be found by replacing the must aliases of the variable.
     Example: mem_temp1 = a->field;
              mem_temp1->field2 = foo;
     then the must_aliases of mem_temp1->field2 are foo and (a->field)->field2.
   *)
  (* full simplify *)
   let basealias = begin
     try
       let slv = Expression.fullsimp_lval (Expression.push_deref lv) in
       if Hashtbl.mem mustAliasesHashtbl slv then
        begin
          let slv' = Hashtbl.find mustAliasesHashtbl slv in
          let r = List.map (fun s -> match (Expression.deep_alpha_convert (fun l -> if l = slv then Expression.Lval s else Expression.Lval l) (Expression.Lval lv )) with
		 Expression.Lval _l -> _l
	        | _ -> failwith ("Error in get_all_must_aliases : "^(Expression.lvalToString lv))
		) slv' in
          r
        end
       else []
       with _ -> []
   end
   in
   Misc.union myalias basealias

(***********************************************************************************)
(* End code for must aliases for temporary mem variables added by Cil              *)
(***********************************************************************************)

