(*
 * Copyright  1990-2002 The Regents of the University of California. All rights reserved. 
 *
 * Permission is hereby granted, without written agreement and without 
 * license or royalty fees, to use, copy, modify, and distribute this 
 * software and its documentation for any purpose, provided that the 
 * above copyright notice and the following two paragraphs appear in 
 * all copies of this software. 
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN 
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE. 
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS 
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION 
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *)

(* $Id: theoremProver.ml,v 1.6 2005/05/21 00:35:09 jhala Exp $
 *
 * This file is part of the BLAST Project.
 *)


(**
 * Interface to Vampyre and ICS.  This module contains the interface to the theorem provers.
 *)


open Ast
      module Stats = Bstats


exception AtomInExpressionException

let tp = ref "" (* ref on which theorem prover to use *)
  (* valid theorem provers: "cvc", "simplify", "simplify-bits" *)

let rec intToHex n x =
  let rec repeatZero m =
    if m = 0 then ""
    else "0"^repeatZero (m-1) 
  in
    if (x = 0) then (repeatZero n) 
    else  (intToHex (n-1) (x / 16)) ^ string_of_int (x mod 16)
 
(***********************************************************)
(************ Hashing for string constants *****************)
(***********************************************************)

let string_const_ctr_ref = ref 0
let string_const_table = Hashtbl.create 101
let const_string_table = Hashtbl.create 101

let get_string_constant s =
  try Hashtbl.find string_const_table s 
  with Not_found ->
     string_const_ctr_ref := 1 + !string_const_ctr_ref;
     let c = string_of_int !string_const_ctr_ref in
     Hashtbl.replace string_const_table s c;
     Hashtbl.replace const_string_table c s;
     c
let get_const_string c = 
  try Hashtbl.find const_string_table c 
  with Not_found -> failwith "Unknown string constant c"

(* only use get_string_const, get_const_string *)
(******************************************************************)


(* This code maintains a mapping from field names of structs to unique indices.  This allows us
   to translate an expression of the form x.f into (select x i), where i is the unique index of f. 
   This is a hack (that works). A more proper translation should look at
   the types, etc.
*)

(* convert an Expression.expression to Vampyre Input *)
let listOfAddrVars = ref []    
(*  let prfnos = string_of_int (!prfno) in*) (* get rid of prfnos once reinit is fixed *)

let rec convertExp e =
  match e with
    Expression.Lval lv -> convertLval lv
  | Expression.Chlval (l, s) -> Input.Fun ("chlval", [ convertLval l; Input.Var ("_"^s) ])
  | Expression.Binary (op, e1, e2) -> 
      begin
	match op with 
	  Expression.Minus -> Input.Minus(convertExp e1, convertExp e2)
	|	Expression.Plus  -> Input.Plus(convertExp e1, convertExp e2)
	|	Expression.Mul   -> Input.Times(convertExp e1, convertExp e2)
	|	Expression.BitAnd 
	|	Expression.BitOr 
	|	Expression.Div 
	|       Expression.RShift 
	|	Expression.Rem 
	|	Expression.Xor 
	|	Expression.Comma ->
	    Input.Fun(Expression.binaryOpToVanillaString op, [convertExp e1; convertExp e2])
	| Expression.LShift ->
	    begin
	      match (e2) with
		Expression.Constant (Constant.Int k) -> (* multiply by 2^k *)
		  Input.Times (convertExp e1, Input.Int (Misc.power 2 k))
	      |	_ -> Input.Fun (Expression.binaryOpToVanillaString op, [convertExp e1; convertExp e2])
	    end
	| Expression.FieldOffset ->
	    begin
	      let ve1 = convertExp e1 and ve2 = convertExp e2 in
              Input.Fun ("foffset", [ve1; ve2])
	    end
	|	_ -> failwith ("convertExp (Vampyre): This operator should not be handled here "^(Expression.toString e))
      end
  | Expression.Constant (c) -> 
      begin
	match c with
	  Constant.Int i -> Input.Int i
	| Constant.String s -> 
            let c = get_string_constant s in Input.Var ("_STRINGCONSTANT"^c)
	| Constant.Float f -> 
	    if (f -. (ceil f) = 0.0) then Input.Int (int_of_float f)
	    else Input.Var "_FLOATINGPTCONSTANT"
	| _ -> failwith ("Cannot handle constant "^(Constant.toString c))
      end

  | Expression.Unary (uop, e) -> 
      begin
	match uop with 
	  Expression.UnaryMinus -> Input.Minus(Input.Int 0, convertExp e)
	| Expression.UnaryPlus -> convertExp e
	| Expression.Address -> 
	    let v = convertExp e in
	    if (List.mem v (!listOfAddrVars)) then () 
	    else
	      listOfAddrVars := v::(!listOfAddrVars);
	    Input.Fun("addrOf", [v])
	|	_ -> Input.Fun(Expression.unaryOpToVanillaString uop, [convertExp e])
      end
  | Expression.FunctionCall (Expression.Lval (Expression.Symbol f), args) -> 
      begin
	Input.Fun (f, List.map convertExp args)
      end
  | _ -> failwith ("convertExp: " ^ (Expression.toString e) ^ "not handled")
	
  and convertLval l =
    match l with
    | Expression.This -> failwith "This not handled in convertLval"
    | Expression.Symbol(str) -> Input.Var ("_"^str) (* prepend all symbols with "_" for later print out in LF *)
    | Expression.Access(Expression.Dot, e, s) -> (* NEEDS DEBUGGING *) 
	let v = convertExp e in
	Input.Fun("select", [v; Input.Var ("_"^(s))])
	  
    | Expression.Access(Expression.Arrow, e, s) ->
	failwith "Arrrow"
    | Expression.Dereference e -> Input.Fun("select", [convertExp e; Input.Int 0])
    | Expression.Indexed (e1, e2) -> Input.Fun("sel4", [convertExp e1; convertExp e2])


  and convertAtom e =
    match e with
      Expression.Binary (op, e1, e2) -> 
	begin
	  match op with 
	    Expression.Eq  -> Input.Eq(convertExp e1, convertExp e2)
	  |	Expression.Ne  -> Input.Neq(convertExp e1, convertExp e2)
	  |	Expression.Gt  -> Input.Gt(convertExp e1, convertExp e2)
	  |	Expression.Lt  -> Input.Lt(convertExp e1, convertExp e2)
	  |	Expression.Ge  -> Input.Geq(convertExp e1, convertExp e2)
	  |	Expression.Le  -> Input.Leq(convertExp e1, convertExp e2)
	  |	_   -> failwith ("convertAtom: Cannot handle this expression: "^(Expression.toString e))
	end
    | Expression.FunctionCall(Expression.Lval (Expression.Symbol f), args) ->
	begin (* a boolean function *)
	  Input.Eq(convertExp e, Input.Int 0)
	end
    | _ -> failwith ("convertAtom: "^(Expression.toString e)^ " Not a binary relation operator")
  
  and convertPred e =
    match e with
      Predicate.True -> Input.True
    | Predicate.False -> Input.False
    | Predicate.And plist -> 
	List.fold_left (function a -> function b -> Input.And(convertPred b, a)) 
	  Input.True (plist)
    | Predicate.Or plist -> 
	List.fold_left (function a -> function b -> Input.Or(convertPred b, a)) 
	  Input.False (plist)
    | Predicate.Implies (e1, e2) -> Input.Imp(convertPred e1, convertPred e2)
    | Predicate.Iff (e1,e2) -> 
        let e1' = convertPred e1 in
        let e2' = convertPred e2 in
        Input.And (Input.Imp(e1',e2'), Input.Imp(e2',e1'))
    | Predicate.Not e1 -> Input.Not(convertPred e1)
    | Predicate.Atom e ->  Input.Atom (convertAtom e)
    | Predicate.All (s,p) -> Input.All (("_"^s), convertPred p)
    | _ -> failwith ("Cannot handle this predicate "^(Predicate.toString e))
 
and convertToVampyreSyntax e =
(*  let prfnos = string_of_int (!prfno) in*) (* get rid of prfnos once reinit is fixed *)(*  prfno := !prfno + 1; *)
  listOfAddrVars := [];
  let v_exp = convertPred (Predicate.constantfold e (Hashtbl.create 1))
  in
  let genAddrOfEq b =
    Input.Atom(Input.Eq(b, Input.Fun("sel4", [Input.Fun("addrOf", [b]); Input.Int 0])))
  in
  begin	
    (* For each address variable, assert that &b != 0, and sel(&b, 0) = b *)
    let vv = List.fold_left (function a -> function b -> Input.Imp (Input.Atom (Input.Neq(b, Input.Int 0)) , Input.Imp(genAddrOfEq b, a)))
	v_exp !listOfAddrVars
    in
    vv
  end
      
      
      
(* Call to prover *)
(* An Engine.Failure is assumed to indicate falsehood *)
      
(* Initialize Vampyre *)
module S = Satproc
module CC : S.SATPROC = Equality
module SX : S.SATPROC = Simplex
module SI : S.SATPROC = Sup_inf
module GT : S.SATPROC = Tactical
module TYP: S.SATPROC = Typing
module SP: S.SATPROC = Special

let (vampyreProver,vampyre_assume_list, vampyre_prove_context, vampyre_prove_context_proof) =
  let allSatProcs = 
    [CC.initialize "CC";
      SX.initialize "SX";
      GT.initialize "GT";
     (* SI.initialize "SI"; 
      TYP.initialize "TYP";
      SP.initialize "SP" *)
    ]
  in
  (* circular dependency hack *)
  ignore(Proof.simp_and_subtract_is_zero := Sup_inf.simp_and_subtract_is_zero;
	 Proof.simp_and_subtract_is_at_least_zero := Sup_inf.simp_and_subtract_is_at_least_zero);
  (* end circular dependency hack *)
  Engine.prove allSatProcs

(* Interface functions*)
let proof = ref None 

let queryExpUsingVampyre exp =
  try
    let vampyreInput = convertToVampyreSyntax exp
    in
    let vampyreInput_pushneg = (Input.pushNeg  vampyreInput) in
    try
(*
	let _ = VampyrePretty.printf "Vampyre Saw queryExp: %a \n" Input.dp vampyreInput_pushneg
    	in
*)
      let _ = proof := Some (vampyreProver "foo" (vampyreInput_pushneg)) in
      ignore !proof ; true
    with Engine.Failure -> (Message.msg_string Message.Debug "Proof failed.\n") ; false
  with
    VampyreErrormsg.Error -> (failwith ("Bug in Vampyre ..."))

(* for now only vampyre supports contexts -- assume etc *)
let assume assumptions = 
  let atomize_asm asm_pred = 
    match asm_pred with
      Predicate.Atom e -> (convertAtom e)
    | Predicate.Not Predicate.Atom e -> Input.negate  (convertAtom e)
    | _ -> 
	Message.msg_string Message.Error "Bad assumption in atomize_asm";
	Message.msg_printer Message.Error Predicate.print asm_pred;
	failwith ("bad asm !")
  in

  (* try *) 
    let vam_assumptions = List.map atomize_asm assumptions 
    in
    vampyre_assume_list vam_assumptions
  (* with
    _ -> failwith "Vampyre Bug due to Jhala !" *)

let printproof pred = 
  let _eToCheck = Predicate.implies pred (Predicate.False) in
  let vampyreInput = convertToVampyreSyntax _eToCheck in
  let vampyreInput_pushneg = (Input.pushNeg  vampyreInput) in
    try
      let proof = vampyreProver "foo" (vampyreInput_pushneg)
      in
	VampyrePretty.printf "Vampyre proof: %a \n" Proof.dp proof; ()
    with Engine.Failure -> (Message.msg_string Message.Debug "Proof failed.\n") ; ()
     
	
(* This converts a vampyre predicate to one that blast can absorb *)
let convertToPredicate w =
  let rec convertExp e =
    match e with
      Input.Var var -> 
	      (* drop initial "_" *)
	if (String.get var 0 = '_') then
	  Expression.Lval (Expression.Symbol (String.sub var 1 ((String.length var) - 1)))
	else (* Huh? ConvertExp did not prepend a "_"! *)
	  Expression.Lval (Expression.Symbol var)
	    
    | Input.Int i   -> Expression.Constant (Constant.Int i)
    | Input.Plus (e1, e2) -> Expression.Binary (Expression.Plus, convertExp e1, convertExp e2)
    | Input.Minus (e1, e2) -> Expression.Binary (Expression.Minus, convertExp e1, convertExp e2)
    | Input.Times (e1, e2) -> Expression.Binary (Expression.Mul, convertExp e1, convertExp e2)
    | Input.Div (e1, e2) -> Expression.Binary (Expression.Div, convertExp e2, Expression.Constant (Constant.Int e1))
    | Input.Mod (e1, e2) -> Expression.Binary (Expression.Rem, convertExp e2, Expression.Constant (Constant.Int e1))
    | Input.Fun (s, el) -> Expression.FunctionCall (Expression.Lval (Expression.Symbol s), List.map convertExp el)
    | _ -> failwith "getUsefulPredicates: Do not handle Floor and Ceiling"
  in
  let rec convertAtom a =
    match a with
      Input.Beq (a1, a2) -> Expression.Binary(Expression.Eq, convertAtom a1, convertAtom a2)
    | Input.Eq (e1, e2) -> Expression.Binary (Expression.Eq, convertExp e1, convertExp e2)
    | Input.Neq (e1, e2) -> Expression.Binary (Expression.Ne, convertExp e1, convertExp e2)
    | Input.Geq (e1, e2) -> Expression.Binary (Expression.Ge, convertExp e1, convertExp e2)
    | Input.Leq (e1, e2) -> Expression.Binary (Expression.Le, convertExp e1, convertExp e2)
    | Input.Gt (e1, e2) -> Expression.Binary (Expression.Gt, convertExp e1, convertExp e2)
    | Input.Lt (e1, e2) -> Expression.Binary (Expression.Lt, convertExp e1, convertExp e2)
    | _ -> failwith "getUsefulPredicates: BFun and NotBFun not handled."
	in
  let _w = convertAtom w in
  Predicate.Atom _w

  

(* Interface functions*)

(* check validity of exp using Ics *)
                                                                                                  
let queryExpUsingIcs exp =
  failwith ("We dont want ICS, thank you very much")
(*
  let icsInput = convertToIcsSyntax exp
  and icsState = Ics.init () 
  in
  Printf.printf "ICS OUTPUT = ";
  let tms = Ics.terms_empty () in
  let tm2 = Ics.terms_add icsInput tms in
  Ics.terms_pp tm2;
  flush stderr;
  Ics.is_valid icsState icsInput
*)    

(******************************************************************)
(* Interface to CVC                                               *)
(******************************************************************)
	     
(* keep around info: do we have a cvc server (= other process) 
	     * running? *)

exception ChannelException

let filed = open_out "blast-trace.cvc" 

let _ = at_exit (fun () -> close_out filed)


let reset_alarm () =
  (* Sys.set_signal Sys.sigalrm (Sys.Signal_ignore) *) ()

let set_alarm () =
  ()
  (*
  ignore (Sys.signal
            Sys.sigalrm
            (Sys.Signal_handle
               (fun i ->
                 Message.msg_string Message.Error "Caught exception sigalrm in stream!" ;
		  reset_alarm ();
                  raise  ChannelException;
               )) ) ;
  ignore (Unix.alarm (5))
  *)

(*
let simplify_line_tap = ref [] (* string list *)
*)

let secure_output_string oc s = (* CONTRACT: may raise channel exception ... *)
  try
    set_alarm ();
(*
    simplify_line_tap := s::!simplify_line_tap;    
*)
    output_string oc s;
    flush oc;
    reset_alarm ()
  with ChannelException -> 
    
    reset_alarm ();
(*
    dump_line_tap ();
*)
    raise ChannelException

let secure_input_line ic = (* CONTRACT: may raise channel exception *)
  
  set_alarm ();
  try
(*
    simplify_line_tap := ""::!simplify_line_tap;
*)
    let rv = input_line ic in
      reset_alarm ();
      rv
  with ChannelException ->
    reset_alarm () ;
(*
    dump_line_tap ();
*)
    raise ChannelException


let cvcServerWithProof = ref None   
let cvcPushes = ref 0
let simp_ctr = ref 0
let clear_varCvcHashtbl_flag = ref true (* the table will get cleared only if this flag is true *)
		    (* moreover -- when simplify_assume is called the flag gets set to false *)
		    (* when pop is called, the flag is again set to true *)

let varCvcHashtbl = Hashtbl.create 53 
let next_cvc_var = ref 0 
let chlvalCvcHashtbl = Hashtbl.create 53 
let unknown_Cvc_var_counter = ref 0 

let getCvcServerWithProof () = 
  match !cvcServerWithProof with
    None -> begin
      Message.msg_string Message.Normal "Forking CVC process (with proofs)..." ;
      flush stdout ;
      let ic,oc = Unix.open_process "cvcl -sat simple -assump -tcc" (* +proof *)
      in
      cvcServerWithProof := Some(ic,oc) ;
      Message.msg_string Message.Normal "done!\n";
      flush stdout ;
      at_exit (fun () -> 
	ignore (Unix.close_process (ic,oc) ) ) ;
      (ic,oc)
	end
  | Some(a) -> a

(* RuGang's functions 03/13/2005:
   getCvcServer
   askCvcIfValid *)

let cvcServer = ref None   

let cvclAsserts = "addrOf:[[REAL, REAL] -> REAL];
 addrof:[REAL->REAL];
 foffset:[[REAL, REAL] -> REAL]; 
 select:[[REAL, REAL] -> REAL]; 
 RShift:[[REAL, REAL] -> REAL]; 
 LShift:[[REAL, REAL] -> REAL]; 
And: [[REAL, REAL] -> REAL]; 
Or:  [[REAL, REAL] -> REAL]; 
 ASSERT (FORALL (x, y: REAL): ((select (addrOf (x,y),0)) = x)); 
 ASSERT (FORALL (x, y: REAL): (addrOf (x,y) /= 0)); 
 ASSERT (FORALL (x, y, z: REAL): (foffset(x, z) = foffset(y,z)) => (x = y));"


let cvc_ground_term_list = ref [] 
    

let add_cvc_ground str = 
  if List.mem str (!cvc_ground_term_list) 
  then () 
  else 
    begin
      if (Misc.is_prefix "ASSERT" str) then
        cvc_ground_term_list := (!cvc_ground_term_list) @ [ str ]
      else 
        cvc_ground_term_list := str::(!cvc_ground_term_list)
    end 
      
let reset_cvc_ground () = 
  cvc_ground_term_list := 
   [
   ]


let getCvcServer () = 
  match !cvcServer with
    None -> 
      begin	
	Message.msg_string Message.Debug "Forking CVC process (no proofs)..." ;
	let ic,oc = Unix.open_process  
	  "cvcl +int  -tcc 2>&1" 
	in
	  cvcServer := Some(ic,oc) ;
	  secure_output_string oc cvclAsserts; 
	  flush stdout ;
	  at_exit (fun () ->
		     ignore (Unix.close_process (ic,oc) ) ) ;
	  (ic,oc)
      end
    | Some(a) -> a

    
(* usually the answer is at the end like
   CVC> CVC> - CVC> - CVC> - CVC> - CVC> - CVC> InValid.
   so using substring 
   other cases might break it? -RuGang, 03-13-05 *)
let rec askCvcIfValid ic =
  Message.msg_string Message.Debug ("asking cvcl ... \n");
  let line = secure_input_line ic in 
  Message.msg_string Message.Debug ("got cvcl ");
  let result = if Misc.is_substring line "InValid." then
    begin 
      Message.msg_string Message.Debug ("CVC: Invalid"^line);
      false
    end

  else if Misc.is_substring line "Valid." then 
    begin
      Message.msg_string Message.Debug ("CVC: Valid"^line);
      true
    end
  else if Misc.is_substring line "Error" then 
   (print_string ("Error: "^line^"\n");
   print_string ((input_line ic)^"\n");
   print_string ((input_line ic)^"\n");
   flush stdout;
   failwith "Error in CVC output!")
  else (print_string line; askCvcIfValid ic)
  in
  result

    




       
let convertToCvcSyntax exp =
  let get_next_var () = next_cvc_var := !next_cvc_var + 1; !next_cvc_var in
  let get_current_var () = !next_cvc_var in

  if (!clear_varCvcHashtbl_flag) then 
   begin
     Hashtbl.clear chlvalCvcHashtbl;
     next_cvc_var :=0
   end;
  (* I have no idea what the convert recursive crap means? *)
  (* let chlval_convert_recursive (lv,s) =  Printf.sprintf "(chlval %s %s)" (convertLval lv) ("FLD"^ s) in    *)
  let chlval_convert_table e = 
   try (Hashtbl.find chlvalCvcHashtbl e) 
      with 
	  Not_found ->
	    begin
	      let s_e = get_next_var () (* Expression.toString e*) in
              let new_id = "chlval"^(string_of_int s_e) in
              if (not (Hashtbl.mem varCvcHashtbl new_id)) then
                  begin
                     add_cvc_ground (new_id^": REAL; ");
                      Hashtbl.add varCvcHashtbl new_id s_e
                  end;
      	      Hashtbl.add chlvalCvcHashtbl e new_id;
              new_id
	    end 
  in
  let rec convertExp e =
    match e with
      | Expression.Lval lv -> convertLval lv
      | Expression.Chlval (lv,s)  -> 
	  if (Options.getValueOfBool "clock") then  Printf.sprintf "(chlval %s %s)" (convertLval lv) ("FLD"^ s)  
                                          else chlval_convert_table e   
    | Expression.Binary (op, e1, e2) -> 
	  begin
	    match op with 
	      Expression.Minus -> 
		begin
		    match e2 with
	              Expression.Constant (Constant.Int 0) -> (convertExp e1)
	            | _ -> Printf.sprintf "(%s - %s)" (convertExp e1) (convertExp e2)
		end
	      |	Expression.Plus  -> 
		  begin
		    match e2 with
                      Expression.Constant (Constant.Int 0) -> (convertExp e1)
	            | _ -> Printf.sprintf "(%s + %s)" (convertExp e1) (convertExp e2)
		  end
	      |	Expression.Mul ->   
	          Printf.sprintf "(%s * %s)" (convertExp e1) (convertExp e2)
	      |	Expression.Div ->
	          Printf.sprintf "(%s / %s)" (convertExp e1) (convertExp e2) 
	      |	Expression.BitAnd ->
	          Printf.sprintf "And(%s,%s)" (convertExp e1) (convertExp e2)
	      |	Expression.BitOr -> 
	          Printf.sprintf "Or(%s,%s)" (convertExp e1) (convertExp e2)
	      |	Expression.Rem 
	      |	Expression.Xor 
	      |	Expression.Comma ->
	          Printf.sprintf "guf(%s,%s)" (convertExp e1) (convertExp e2)
	      |	Expression.LShift ->
		  begin
		    match (e2) with
		      Expression.Constant (Constant.Int k) -> (* multiply by 2^k *)
			Printf.sprintf "(%s * %d)" (convertExp e1) (Misc.power 2 k)
		    |	_ -> Printf.sprintf "(%s (%s,%s))" (Expression.binaryOpToVanillaString op) (convertExp e1) (convertExp e2)
		  end
	      |	Expression.RShift ->
		  begin
		    match (e2) with
		      Expression.Constant (Constant.Int k) -> (* divide by 2^k *)
			Printf.sprintf "(%s / %d)" (convertExp e1) (Misc.power 2 k) 
		    |	_ -> Printf.sprintf "(%s (%s,%s))" (Expression.binaryOpToVanillaString op) (convertExp e1)  (convertExp e2)
		  end	 
	      |	_ -> 
              if Expression.isRelOp op 
	      then 
		raise AtomInExpressionException
	      else 
		failwith ("convertExp: (CVC) This operator should not be handled here "^(Expression.toString e))
	  end

    | Expression.Constant (c) -> 
	begin
	  match c with
		Constant.Int i ->  
	          string_of_int i
	  | Constant.String s -> 
	      if (not (Hashtbl.mem varCvcHashtbl "STRING")) then 
		begin
		  add_cvc_ground ("STRING"^" : REAL ;");
		  Hashtbl.add varCvcHashtbl "STRING"  1
		end;
	      ("STRING") 
	  | _ -> failwith ("unknown constant type")
	end
    | Expression.Unary (uop, e) -> 
	begin
	  match uop with 
	      Expression.UnaryMinus -> Printf.sprintf ("( - %s)") (convertExp e)
	    | Expression.UnaryPlus -> Printf.sprintf ("( + %s)") (convertExp e)
	    | Expression.Address -> 
	      begin
	        let e_cvc = convertExp e in
	        let expr = Printf.sprintf "(addrof(%s))" e_cvc in
	          (* Also add an assertion relating deref and address for this expression. *)
                  add_cvc_ground (Printf.sprintf "ASSERT (select (%s, 0) = %s);" expr e_cvc) ;
	          expr
              end
	  |	_ -> failwith "unknown unop" 
	end
    | Expression.FunctionCall(Expression.Lval (Expression.Symbol f), args) ->
        begin
          Printf.sprintf "(%s ( %s ))" f (Misc.strList (List.map convertExp args))
        end
    | _ -> failwith ("convertExp (Cvc): " ^ (Expression.toString e) ^ "not handled")
  and convertLval l =
   match l with
   | Expression.This -> failwith "This not handled in convertLval"
   | Expression.Symbol(str) -> 
       begin
	 let cvc_name = (Str.global_replace (Str.regexp "@") "LOCAL" str) in
	   if (Hashtbl.mem varCvcHashtbl cvc_name) then cvc_name
	   else 
	     begin
	       Hashtbl.add varCvcHashtbl cvc_name  1; (* number 1 doesn't matter *)
	       add_cvc_ground (cvc_name ^ ": REAL ;") ;
	       cvc_name
	     end
       end
   | Expression.Dereference drfe -> "select( "^(convertExp drfe)^ ", 0)"
   | Expression.Access(Expression.Dot, e, s) -> 
       begin
	  if (not (Hashtbl.mem varCvcHashtbl ("FLD"^s))) then 
	    begin
	      add_cvc_ground ("FLD"^s^" : REAL;");
	      Hashtbl.add varCvcHashtbl ("FLD"^s)  1
	    end;
	 Printf.sprintf "(select(%s,%s))" (convertExp e) ("FLD"^s)
       end
    | Expression.Access(Expression.Arrow, e, s) ->
	begin
	  if (not (Hashtbl.mem varCvcHashtbl ("FLD"^s))) then 
	    begin
	      add_cvc_ground ("FLD"^s^" : REAL;");
	      Hashtbl.add varCvcHashtbl ("FLD"^s)  1
	    end;
	  (Printf.sprintf "(select(%s,%s))" (convertExp e) ("FLD"^s))
	end
    | _ -> failwith ("convertLval (Cvc): " ^ (Expression.lvalToString l) ^ "not handled")
	 
  and convertAtom e =
   flush stdout;
    match e with
      Expression.Binary (op, e1, e2) -> 
	begin
	  let opstring = 
	    match op with 
	      Expression.Eq  -> "="
	    | Expression.Ne  -> "/="
	    | Expression.Gt  -> ">"
	    | Expression.Lt  -> "<"
	    | Expression.Ge  -> ">="
	    | Expression.Le  -> "<="
	    | _   -> "" 
    in
        if opstring <> "" 
	  then
	    begin
              try
	        Printf.sprintf "(%s %s %s)"  (convertExp e1) opstring (convertExp e2)
              with AtomInExpressionException ->
		begin
		  if (opstring = "=" or opstring = "/=") then
		    begin
                      match e1 with
			Expression.Lval _ | Expression.Chlval _ | Expression.Constant (Constant.Int _) ->
			  begin
			    let s1 = convertExp e1 in
			    let s2 = convertAtom e2 in
                            if (opstring = "=") then
                                Printf.sprintf "((%s /= 0) <=> (%s))" s1 s2
                            else Printf.sprintf "((%s = 0) <=> (%s))" s1 s2
			   (* Printf.sprintf "(OR (AND %s  (%s %s  1)) (AND (NOT %s) (%s %s 0)))" s2 opstring s1 s2 opstring s1  *)
			    (* Printf.sprintf "(%s %s (%s 0))" opstring s1 (get_fresh_unknown ()) *)
			  end
                      | _ -> 
			  begin
			    match e2 with
			      Expression.Lval _ | Expression.Chlval _ | Expression.Constant (Constant.Int _) ->
				begin
				  let s1 = convertExp e2 in
				  let s2 = convertAtom e1 in
				    if (opstring = "=") then
                                      Printf.sprintf "((%s /= 0) <=> (%s))" s1 s2
				    else Printf.sprintf "((%s = 0) <=> (%s))" s1 s2
                            				end
			    | _ -> 
                                Printf.sprintf ""
			  end 
		    end
		  else
                    Printf.sprintf ""
		end
            end
	  else (Message.msg_string Message.Debug "failed opthing";Message.msg_string Message.Debug (Expression.toString e); 
               Printf.sprintf "")
	end
    | Expression.FunctionCall(Expression.Lval (Expression.Symbol f), args) ->
	begin
          Printf.sprintf "(%s ( %s ) = 0)" f (Misc.strList (List.map convertExp args))
        end
    | _ -> (Message.msg_string Message.Debug "failed opthing2";Message.msg_string Message.Debug (Expression.toString e);  Printf.sprintf "")
 	  
  and convertPred e =
   flush stdout;
    match e with
      Predicate.True -> "TRUE"
    | Predicate.False -> "FALSE"
    | Predicate.And plist -> 
	List.fold_left 
	  (fun a -> fun b -> Printf.sprintf " (%s AND %s) " (convertPred b) a) 
	  "TRUE" (plist)
    | Predicate.Or plist -> 
	List.fold_left 
	  (function a -> function b -> Printf.sprintf "(%s OR %s)" (convertPred b) a )
	    "FALSE" (plist)
    | Predicate.Implies (e1, e2) -> Printf.sprintf "(%s => %s)" (convertPred e1) (convertPred e2)
	  
    | Predicate.Not e1 -> Printf.sprintf "(NOT %s)" (convertPred e1)
    | Predicate.Atom e ->  (convertAtom e)
    | Predicate.All (s,e1) -> Printf.sprintf "(FORALL (%s : REAL): %s)" s (convertPred e1)
    | _ -> failwith ("CVC: Cannot handle this predicate "^(Predicate.toString e))
	  

  in
  (convertPred (Predicate.constantfold exp (Hashtbl.create 1)))
    
    
    
let cvcQueryNum = ref 1
let cvcQueryNumPf = ref 1
    
    
let rec cvc_output_list oc lst =
   match lst with 
   [] -> () 
   | a::rest -> 
   begin
   Message.msg_string Message.Debug ("cvc input: "^a^"\n");
   secure_output_string oc (Printf.sprintf "%s\n" a) ;
   output_string filed  (Printf.sprintf "%s\n" a) ;
   flush oc;
   flush filed;
   cvc_output_list oc rest 
   end 

let queryExpUsingCvc exp = 
  Message.msg_string Message.Debug ("cvcl: querying exp:");
  Message.msg_printer Message.Debug Predicate.print exp ;
  let cvcInput = convertToCvcSyntax exp
  in 
  Message.msg_string Message.Debug ("cvcl: "^cvcInput);
  flush stdout;
  flush stderr;
   let result = 
    try
      let ic,oc = getCvcServer () in
      cvc_output_list  oc !cvc_ground_term_list; 
      secure_output_string oc ("PUSH; QUERY ("^cvcInput^"); POP;\n") ;
      output_string filed ("PUSH; QUERY ("^cvcInput^"); POP;\n") ;
      Message.msg_string Message.Debug ("cvc input: "^"QUERY ("^cvcInput^");\n");
      reset_cvc_ground ();
      flush oc ; 
      flush filed;
      askCvcIfValid ic
    with e ->
      failwith ("Cvc raised exception "^(Printexc.to_string e)^" for "^(Predicate.toString exp)^". Check that cvcl is in your path.")
  in
  Message.msg_string Message.Debug (if result then "true" else "false");
  result

    
let queryExpUsingCvcGreedy exp =
  let _ = Message.msg_string Message.Debug ("querying exp in queryExpUsingCvcGreedy :"^(Predicate.toString exp)^"\n") in
   (* first we have to convert the formula and collect all the ground terms that appear in it *)

  let fdgreedy = open_out "blast-trace-greedy.cvc" in

  reset_cvc_ground ();
  let formulas = match exp with Predicate.And l -> l | _ -> failwith ("bad call to queryExpUsingCvcGreedy") in
  let all_asserts = List.map (convertToCvcSyntax) formulas
  in
	     (* at this point all_asserts is a list of all assertions, and !cvc_ground_term_list is
	        the list of all type declarations.
	     *)
  let rec output_asserts oc predlist =
    match predlist with 
      [] -> () 
    | a::rest -> 
	begin
          output_string oc (Printf.sprintf "ASSERT %s;\n" a) ; 
          output_asserts oc rest 
	end
  in 
  let rec output_list oc lst =
    match lst with 
      [] -> () 
    | a::rest -> 
        begin
          output_string oc (Printf.sprintf "%s\n" a) ; 
          output_list oc rest 
        end
  in
  let ic,oc = getCvcServer () in
  let rec getUsefulPredicates usefulPreds  =
    let rec _worker assertlist =
      match assertlist with 
	[] -> failwith "Exception : Formula is not valid!"
      | a :: rest ->
	  output_string oc (Printf.sprintf "ASSERT %s;\nECHO \"Query %s\";\nQUERY FALSE;\n" a (string_of_int !cvcQueryNum)) ;

	  output_string fdgreedy (Printf.sprintf "ASSERT %s;\nECHO \"Query %s\";\n\n" a (string_of_int !cvcQueryNum)) ;

          incr cvcQueryNum ;
          flush oc ;
          flush fdgreedy ;
	  let result = 
	     try
   	       askCvcIfValid ic
	     with e ->
	       failwith ("Cvc raised exception "^(Printexc.to_string e)^" for "^(Predicate.toString exp))
          in
          if result then a
          else
             _worker rest
    in
    output_string oc "PUSH 1;\n" ;
    output_string fdgreedy "PUSH 1;\n" ;
    output_list oc !cvc_ground_term_list ;
    output_asserts oc usefulPreds ;
    output_list fdgreedy !cvc_ground_term_list ;
    output_asserts fdgreedy usefulPreds ;
    output_string oc (Printf.sprintf "ECHO \"Query %s\" ; \n" (string_of_int !cvcQueryNum)) ;
    incr cvcQueryNum ;
    output_string oc "QUERY FALSE; \n" ;
    flush oc ;
    flush fdgreedy ;
    let result = 
      try
	askCvcIfValid ic
      with e ->
	failwith ("Cvc raised exception "^(Printexc.to_string e)^" for "^(Predicate.toString exp))
    in
    if result then usefulPreds 
    else 
      begin
        (* go on adding predicates from assertlist and check for validity.
	   When valid, add the last formula to usefulPreds and repeat
           *)
        let nextpred = _worker all_asserts in
        output_string oc "POPTO 1; \n" ;
        output_string fdgreedy "POPTO 1; \n" ;
        getUsefulPredicates (nextpred :: usefulPreds) 
      end
    in
  let listOfPreds = getUsefulPredicates [] 
  in
  output_string oc "POPTO 1;\n" ;
  output_string fdgreedy "POPTO 1;\n" ;
  flush oc ;
  output_string fdgreedy "ECHO\"Predicates are \" ;\n" ;
  output_list fdgreedy (List.map (fun s -> "ECHO \""^s^"\";\n") listOfPreds) ;
  flush fdgreedy ;
  close_out fdgreedy ;
   listOfPreds 

    

    
let queryExpUsingCvcProof exp =

   Message.msg_string Message.Debug ("CVC querying exp:"^(Predicate.toString (Predicate.convertDNF exp))^"\n");

  reset_cvc_ground ();
  let formulas = exp::[] in 
  let all_asserts = List.map convertToCvcSyntax formulas
  in
  let rec output_asserts oc predlist =
    match predlist with 
      [] -> () 
    | a::rest -> 
        begin
          output_string oc (Printf.sprintf "ASSERT %s;\n" a) ; 
          output_asserts oc rest
        end
  in 
  let rec output_list oc lst =
    match lst with 
      [] -> () 
    | a::rest -> 
        begin
          output_string oc (Printf.sprintf "%s\n" a) ; 
          output_list oc rest
        end
   in
(*  let ic,oc = getCvcServerWithProof () in - RuGang*)
   (* what are dump assertions? *)
   Message.msg_string Message.Debug "Asking CVC (Proof)";
   let ic,oc = getCvcServer () in
   output_string oc (Printf.sprintf "ECHO \"Query %s\";\nPUSH 1;\n" (string_of_int !cvcQueryNumPf)) ;
  output_string filed (Printf.sprintf "ECHO \"Query %s\";\nPUSH 1;\n" (string_of_int !cvcQueryNumPf)) ;
  output_list oc !cvc_ground_term_list ;
  output_list filed !cvc_ground_term_list ;
   Message.msg_string Message.Debug "after cvc ground crap";
   flush oc;
  let rec assert_loop asrtlist =
    match asrtlist with
      [] ->
        begin
          output_string oc "\nDUMP_ASSUMPTIONS; QUERY FALSE;\nPOPTO 1;\n" ;
          output_string filed "\nDUMP_ASSUMPTIONS; QUERY FALSE;\nPOPTO 1;\n" ;
          incr cvcQueryNumPf ;
          flush oc ; 
          flush filed ;
          let result = 
	     try
	       askCvcIfValid ic
	     with e ->
	      failwith ("Cvc raised exception "^(Printexc.to_string e)^" for "^(Predicate.toString exp))
          in
          reset_cvc_ground () ;
          flush oc;
          result       
        end
    | a::rest ->
       begin 
         output_string oc (Printf.sprintf "ASSERT %s;\n QUERY FALSE;\n" a);
         output_string filed (Printf.sprintf "ASSERT %s;\n QUERY FALSE;\n" a);
         flush oc ;
         let result = 
	     try
	       askCvcIfValid ic
	     with e ->
	       failwith ("Cvc raised exception "^(Printexc.to_string e)^" for "^(Predicate.toString exp))
         in
         if result then
           begin
             output_string oc "\nDUMP_ASSUMPTIONS; \nPOPTO 1;\n" ;
             output_string filed "\nDUMP_ASSUMPTIONS; \nPOPTO 1;\n" ;
             incr cvcQueryNumPf ;
             flush oc ; 
             flush filed ;
             reset_cvc_ground () ;
             flush oc;
             result      
           end
         else  
           assert_loop rest
       end
  in
  ignore (assert_loop all_asserts);
  output_asserts oc all_asserts ;
  output_list filed !cvc_ground_term_list ;
  output_asserts filed all_asserts ;
  output_string oc "\nDUMP_ASSUMPTIONS; QUERY FALSE;\nPOPTO 1;\n" ;
  output_string filed "\nDUMP_ASSUMPTIONS; QUERY FALSE;\nPOPTO 1;\n" ;
  incr cvcQueryNumPf ;
  flush oc ; 
  flush filed ;
  let result = 
    try
      askCvcIfValid ic
    with e ->
      failwith ("Cvc raised exception "^(Printexc.to_string e)^" for "^(Predicate.toString exp))
  in
  reset_cvc_ground () ;
  flush oc;
  result       



(* This implements the greedy algorithm (SLAM, Ruess, ...) to get back a list
   of useful predicates.
*)
exception CVCException 

let getProofsFromCvc exp =
  let _ = Message.msg_string Message.Debug ("querying exp in getProofsFromCvc :"^(Predicate.toString exp)^"\n") in
   (* first we have to convert the formula and collect all the ground terms that appear in it *)
  try  
  reset_cvc_ground ();
  let formulas = 
    match exp with 
      Predicate.And l -> l 
    | _ -> failwith ("bad call to getProofsFromCvc -- we only handle listd of conjunections") in
  let all_blast_asserts_array = Array.of_list formulas in
  let all_asserts = List.map (convertToCvcSyntax) formulas
  in
	     (* at this point all_asserts is a list of all assertions, and !cvc_ground_term_list is
	        the list of all type declarations.
	     *)
  let rec output_asserts oc predlist =
    match predlist with 
      [] -> () 
    | a::rest -> 
	begin
          output_string oc (Printf.sprintf "ASSERT %s;\n" a) ; 
          output_asserts oc rest 
	end
  in 
  let rec output_list oc lst =
    match lst with 
      [] -> () 
    | a::rest -> 
        begin
          output_string oc (Printf.sprintf "%s\n" a) ; 
          output_list oc rest 
        end
  in
  let ic,oc = getCvcServer () in
  let rec usefulPredicates usefulPreds  usefulBlastPreds =
    let rec _worker assertlist num =
      match assertlist with 
	[] -> 
	  begin
	    Message.msg_string Message.Error "Exception : Formula is not valid! Reverting to Vampyre" ;
	    raise CVCException 
	  end
      | a :: rest ->
	  output_string oc (Printf.sprintf "ASSERT %s;\nECHO \"Query %s\";\nQUERY FALSE;\n" a (string_of_int !cvcQueryNum)) ;
	  incr cvcQueryNum ;
          flush oc ;
	  let result = 
	     try
   	       askCvcIfValid ic
	     with e ->
	       failwith ("Cvc raised exception "^(Printexc.to_string e)^" for "^(Predicate.toString exp))
          in
          if result then (a, num)
          else
             _worker rest (num+1)
    in
    output_string oc "PUSH 1;\n" ;
    output_list oc !cvc_ground_term_list ;
    output_asserts oc usefulPreds ;
    output_string oc (Printf.sprintf "ECHO \"Query %s\" ; \n" (string_of_int !cvcQueryNum)) ;
    incr cvcQueryNum ;
    output_string oc "QUERY FALSE; \n" ;
    flush oc ;
    let result = 
      try
	askCvcIfValid ic
      with e ->
	failwith ("Cvc raised exception "^(Printexc.to_string e)^" for "^(Predicate.toString exp))
    in
    if result then usefulBlastPreds 
    else 
      begin
        (* go on adding predicates from assertlist and check for validity.
	   When valid, add the last formula to usefulPreds and repeat
           *)
        let (nextpred, i) = _worker all_asserts 0 in
        output_string oc "POPTO 1; \n" ;
        usefulPredicates (nextpred :: usefulPreds) ((all_blast_asserts_array.(i)) :: usefulBlastPreds)
      end
    in
  let listOfPreds = usefulPredicates [] []
  in
  output_string oc "POPTO 1;\n" ;
  flush oc ;
  List.map 
    (fun a -> match a with Predicate.Not b -> b | _ -> a)
    listOfPreds
with _ ->
	    begin
	      let _eToCheck = Predicate.implies exp (Predicate.False) in
              let ans = queryExpUsingVampyre _eToCheck
              in
              let _ = Message.msg_string Message.Debug "Got UsefulPred" in
              if ans = false
              then
		begin
		  failwith "Vampyre could not prove formula valid!" ; []
		end
              else
                begin
		  let _lst = Engine.getUsefulPredicates () in
            (*
	      let inferred_atoms = 
	      match (!proof) with 
	      Some(prf) -> 
	      Engine.getInferredAtoms (prf)
	      | None -> []
	      in
	      *)
		  List.map convertToPredicate _lst
		end
	    end

(*************************************************************************************)
(* Interface to Simplify.                                                            *)
(* Simplify is run as a server and queries are piped to it.                          *)
(* Thanks to Wes Weimer for the code.                                                *)
(*************************************************************************************)

(* Return a "\n" terminated string corresponding to simplify input *)
let clear_varHashtbl_flag = ref true (* the table will get cleared only if this flag is true *)
		    (* moreover -- when simplify_assume is called the flag gets set to false *)
		    (* when pop is called, the flag is again set to true *)

let varHashtbl = Hashtbl.create 53 
let next_var = ref 0 
let chlvalHashtbl = Hashtbl.create 53 

let unknown_var_counter = ref 0

let get_fresh_unknown () = 
  let rv = !unknown_var_counter in
    unknown_var_counter := !unknown_var_counter + 1;
    ("UNK"^(string_of_int rv))

(* Original *)

let convertToSimplifySyntax () exp =
(*Simplify has issues with variable names starting with "_", so we map all vars to v<num> *)

  Message.msg_string Message.Debug ("convertToSimplifySyntax exp = ");
  Message.msg_printer Message.Debug Predicate.print exp;  (*RuGang*)
  let get_next_var () = next_var := !next_var + 1; !next_var in
  let get_current_var () = !next_var in
  if (!clear_varHashtbl_flag) then 
    begin
      Hashtbl.clear varHashtbl;next_var :=0; 
      Hashtbl.clear chlvalHashtbl
    end ;
  let fieldId str =
    if Hashtbl.mem varHashtbl str then
      Hashtbl.find varHashtbl str
    else
      begin
	let new_id = "v"^(string_of_int (get_next_var ())) in
	Hashtbl.add varHashtbl str  new_id;
	new_id
      end
  in
  let id_of_var lv =
    match lv with
      Expression.Symbol s -> 
	begin
	  try
	    let vid = Hashtbl.find varHashtbl s in
	    int_of_string (Misc.chop_after_prefix "v" vid)
	  with 
	    Not_found -> failwith ("id_of_var: (Simplify) variable not found "^(s))
	end
	  
    | _ -> failwith ("id_of_var: (Simplify) lvalue not a symbol. "^(Expression.lvalToString lv))
  in
  let id_of_fld s =
    
    let vid = try Hashtbl.find varHashtbl s 
    with Not_found -> failwith ("id_of_fld: (Simplify) variable not found "^(s))	
    in
    int_of_string (Misc.chop_after_prefix "v" vid)
	
  in

  let rec  convertExp e =
    let chlval_convert_recursive (lv,s) = Printf.sprintf "(chlval %s %s)" (convertLval lv) (fieldId s) in  
    let chlval_convert_table e = 
      try (Hashtbl.find chlvalHashtbl e) 
      with 
	  Not_found ->
	    begin
	      let s_e = get_next_var () (* Expression.toString e*) in
              let new_id = "chlval"^(string_of_int s_e) in
		Hashtbl.add chlvalHashtbl e new_id;
		new_id
	    end
    in
      match e with
	| Expression.Lval lv -> convertLval lv
	| Expression.Chlval (lv,s)  -> 
	    if (Options.getValueOfBool "clock") then chlval_convert_recursive (lv,s) else chlval_convert_table e
	| Expression.Binary (op, e1, e2) -> 
	    begin
	      match op with 
		  Expression.Minus -> Printf.sprintf "(- %s %s)" (convertExp e1) (convertExp e2)
		|	Expression.Plus  -> Printf.sprintf "(+ %s %s)" (convertExp e1) (convertExp e2)
			  (* |     Expression.Ordset_Succ ->  Printf.sprintf "(SUCC %s)" (convertExp e1)*)
		|	Expression.Mul   -> Printf.sprintf "(* %s %s)" (convertExp e1) (convertExp e2)

		|	Expression.RShift ->
		    (* second term / shifted value must be a constant *)
		    begin
		    match (e2) with
		      Expression.Constant (Constant.Int _) 
		    | _ ->  Printf.sprintf "(%s %s %s)" 
			  (Expression.binaryOpToVanillaString op) (convertExp e1) (convertExp e2)
			  end
		|	Expression.BitAnd 
		|	Expression.BitOr 
		|	Expression.Div 
		|	Expression.Rem 
		|	Expression.Xor 
		|	Expression.Comma ->
		  (* since vampyre knows NOTHING about this things, we shall 
		      put in a brand new symbol here 
		  *)
              Printf.sprintf "(%s %s %s)" (Expression.binaryOpToVanillaString op) (convertExp e1) (convertExp e2) 
	  |	Expression.LShift ->
	      begin
		match e2 with 
		  Expression.Constant (Constant.Int k) -> 
		    Printf.sprintf "(* %s %d)" (convertExp e1) (Misc.power 2 k)
		| _ -> Printf.sprintf "(%s %s %s)" (Expression.binaryOpToVanillaString op) (convertExp e1) (convertExp e2)
	      end

	  |     Expression.FieldOffset ->
	      (* the second argument to field offset is a field id *)
	      let fld =
		match e2 with
		  Expression.Lval (Expression.Symbol s) -> s
		| _ -> failwith ("convertExp : (Simplify) Second operator of field offset not a field id "^
				 (Expression.toString e))
	      in
	      Printf.sprintf "(foffset %s %s)" (convertExp  e1) (fieldId fld)
	  |	_ -> 
	      if Expression.isRelOp op 
	      then 
		raise AtomInExpressionException
	      else 
		failwith ("convertExp: (Simplify) This operator "^(Expression.binaryOpToVanillaString op)^" should not be handled here "^(Expression.toString e))
	end
    | Expression.Constant (c) -> 
	begin
	  match c with
	    Constant.Int i ->  if (i < 0) then "(- 0 "^(string_of_int (-i))^")" else string_of_int i 
	  | Constant.String s -> 
              let c = get_string_constant s in
              ("(_STRINGCONSTANT "^c^")")
	  | Constant.Float f -> 
	      if (f -. (ceil f) = 0.0) then string_of_int (int_of_float f)
	      else ("FLOATINGPT"^(string_of_int (get_next_var ())))
	  | _ -> failwith ("Cannot handle constant "^(Constant.toString c))
	end

    | Expression.Unary (uop, e1) -> 
	begin
	  match uop with 
	    Expression.UnaryMinus -> "(- 0 "^(convertExp e1)^")"
	  | Expression.UnaryPlus -> convertExp e1
	  | Expression.Address -> 
	      begin
		match e1 with
		  Expression.Lval lv -> 
		    let l = convertLval lv
		    in
		    let id = (id_of_var lv) in
		    Printf.sprintf "(addrOf %s %d)" l id
		| Expression.Chlval (lv, f) -> 
		    (* let lstr = (convertLval lv) in
		    let fstr = (fieldId f) in
		    let id = (id_of_fld f) in
		       *)
		    let chlv = convertExp e1 in
		    let id = int_of_string (Misc.chop_after_prefix "chlval" chlv) in
		    Printf.sprintf "(addrOf %s %d)" chlv id
		      
		| _ -> failwith ("convertExp: (Simplify) Address: strange expression "^(Expression.toString e))
	      end
	  | Expression.BitNot -> Printf.sprintf "(BitNot %s)" (convertExp e1)
	  |	_ -> "("^Expression.unaryOpToVanillaString uop^" "^(convertExp e1)^")"
	end
    | Expression.FunctionCall (Expression.Lval (Expression.Symbol f), args) ->
        Printf.sprintf "(%s %s)" f (Misc.string_list_cat " " (List.map convertExp args))
    | _ -> failwith ("convertExp (Simplify): " ^ (Expression.toString e) ^ "not handled")
	 
  and convertLval l =
    match l with
    | Expression.This -> failwith "This not handled in convertLval"
    | Expression.Symbol(str) -> 
	if Hashtbl.mem varHashtbl str then
	  (Hashtbl.find varHashtbl str)
	else
	  begin
	    let new_id = "v"^(string_of_int (get_next_var ())) in
	    Hashtbl.add varHashtbl str new_id;
	    (*Message.msg_string Message.Debug ("NOT in hashtbl:"^str^":"^new_id);*)
	    new_id
	  end
    | Expression.Dereference drfe  -> "(select "^(convertExp drfe)^" 0)"

    | Expression.Access(Expression.Dot, e, s) ->
	     (Printf.sprintf "(select %s %s)" (convertExp e) (fieldId s))
    | Expression.Access(Expression.Arrow, e, s) ->
	let fieldId str =
	  if Hashtbl.mem varHashtbl str then
	    Hashtbl.find varHashtbl str
	  else
	    begin
	      let new_id = "v"^(string_of_int (get_next_var ())) in
	      Hashtbl.add varHashtbl str new_id;
	      new_id
	    end
	in
	(Printf.sprintf "(select %s %s)" (convertExp e) (fieldId s))
	  (* JHALA: MAJOR BUG!! this should not be (+ e1 e2) surely *)
    | Expression.Indexed (e1, e2) -> Printf.sprintf "(select %s %s)" (convertExp e1) (convertExp e2)

  and convertAtom e =   
    match e with
      Expression.Binary (op, e1, e2) -> 
	begin
	  let opstring = 
	    match op with 
		Expression.Eq  -> "EQ"
	      |	Expression.Ne  -> "NEQ"
	      |	Expression.Gt  -> ">"
	      |	Expression.Lt  -> "<"
	      |	Expression.Ge  -> ">="
	      |	Expression.Le  -> "<="
	      |	_   -> Message.msg_string Message.Error ("Strange binop: "^(Expression.binaryOpToVanillaString op)); "" (**** see below ** failwith ("convertAtom: Cannot handle this expression: "^(Expression.toString e)) *)
	  in
          if opstring <> "" 
	  then
	    begin
              try
	        Printf.sprintf "(%s %s %s)" opstring (convertExp e1) (convertExp e2)
              with AtomInExpressionException ->
		begin
		  if (opstring = "EQ" or opstring = "NEQ") then
		    begin
                      match e1 with
			Expression.Lval _ | Expression.Chlval _ | Expression.Constant (Constant.Int _) ->
			  begin
			    let s1 = convertExp e1 in
			    let s2 = convertAtom e2 in
			    Printf.sprintf "(OR (AND %s  (%s %s  1)) (AND (NOT %s) (%s %s 0)))" s2 opstring s1 s2 opstring s1 
			  end
                      | _ -> 
			  begin
			    match e2 with
			      Expression.Lval _ | Expression.Chlval _ | Expression.Constant (Constant.Int _) ->
				begin
				  let s1 = convertExp e2 in
				  let s2 = convertAtom e1 in
				  Printf.sprintf "(OR (AND %s  (%s %s  1)) (AND (NOT %s) (%s %s 0)))" s2 opstring s1 s2 opstring s1 
				end
			    | _ -> 
				Printf.sprintf "(EQ %s 0)" (get_fresh_unknown ()) 
			  end 
		    end
		  else
		    Printf.sprintf "(EQ %s 0)" (get_fresh_unknown () )
		end
            end
	  else (Message.msg_string Message.Debug "failed opthing";Message.msg_string Message.Debug (Expression.toString e); 
               (* convertExp e *)  
               Printf.sprintf "(EQ %s 0)" (get_fresh_unknown ()) )
	end
    | Expression.FunctionCall (Expression.Lval (Expression.Symbol f), args) ->
        Printf.sprintf "(EQ (%s %s) 0)" f (Misc.string_list_cat " " (List.map convertExp args))
    | _ -> (Message.msg_string Message.Debug "failed opthing2";Message.msg_string Message.Debug (Expression.toString e); 
            (* convertExp e *) Printf.sprintf "(EQ %s 0)" (get_fresh_unknown () ) )
	(*** instead i shall simply call convert expression on it *** failwith ("convertAtom: "^(Expression.toString e)^ " Not a binary relation operator") *)
  
  and convertPred e =
    match e with
      Predicate.True -> "TRUE"
    | Predicate.False -> "FALSE"
    | Predicate.And plist -> 
	List.fold_left (function a -> function b -> "(AND "^(convertPred b)^" "^a^")") 
	  "TRUE" (plist)
    | Predicate.Or plist -> 
	List.fold_left (function a -> function b -> "(OR "^(convertPred b)^" "^a^")") 
	  "FALSE" (plist)
    | Predicate.Implies (e1, e2) -> "(IMPLIES "^(convertPred e1)^" "^(convertPred e2)^")"
    | Predicate.Iff (e1,e2) -> Printf.sprintf "(IFF %s %s)" (convertPred e1) (convertPred e2)
    | Predicate.Not e1 -> "(NOT "^(convertPred e1)^")"
    | Predicate.Atom e ->  (convertAtom e) 
    | Predicate.All (s,e1) -> Printf.sprintf "(FORALL (%s) %s)" 
    (convertLval (Expression.Symbol s)) (convertPred e1)
    | _ -> failwith ("Simplify: Cannot handle this predicate "^(Predicate.toString e))
  in
  (convertPred (Predicate.constantfold exp (Hashtbl.create 1)))^"\n"

(* keep around info: do we have a simplify server (= other process) 
 * running? *)




let bits =  ref 4


let assertBitVector s =
  let rec assertBitV n s = 
    if (n = 0) then ""
    else 
      begin
	let f = Printf.sprintf "(OR (EQ %s_%d 0) (EQ %s_%d 1)) " s n s n in
	  (f ^ (assertBitV (n-1) s))
      end in
    begin
      ("(BG_PUSH (AND "^(assertBitV !bits s)^"))")
    end
      
let shift_R res a b_num =
  let fu = ref "" in
    for i = 0 to (b_num-1) do
      fu := !fu ^ (Printf.sprintf "(EQ %s_%d 0) " res (!bits - i))
    done;
    
    for i = b_num to (!bits-1) do
      fu := !fu ^ (Printf.sprintf "(EQ %s_%d %s_%d) " res (!bits - i) a (!bits - i + b_num ))
    done;
    !fu

let rShiftVar res a b = 
  let everything = ref "(AND " in
    for i = 0 to !bits do
      everything := !everything ^ (Printf.sprintf "(IFF (EQ %s %d) (AND %s)) " 
				     b i (shift_R res a i))
    done;
    !everything ^ (Printf.sprintf "(IFF (> %s %d) (AND %s)))" 
		     b !bits (shift_R res a !bits))
      
(* let () = print_string ("(BG_PUSH "^(rShiftVar "d" "a" "b")^")\n") *)

let rShiftCons res a b_cons =
  (Printf.sprintf "(AND %s) " (shift_R res a b_cons))

let condAnd res a b = 
  let foo = ref "(AND " in
    foo := !foo ^ (Printf.sprintf "(EQ %s 1) " res);
    foo := !foo ^ (Printf.sprintf "(IFF (EQ %s 0) (OR " res);
    foo := !foo ^ "(AND ";
    for i = 1 to !bits do
      foo := !foo ^ (Printf.sprintf "(EQ %s_%d 0) " a i)
    done;
    foo := !foo ^ ") ";
    foo := !foo ^ "(AND ";
    for i = 1 to !bits do
      foo := !foo ^ (Printf.sprintf "(EQ %s_%d 0) " b i)
    done;
    foo := !foo ^ ")))) ";
    !foo


let rec convToBin x =
  if (x = 0) then []
  else (x mod 2)::(convToBin (x/2))
    

(* let () = print_string ("(BG_PUSH "^(condAnd "d" "a" "b") ^ ")\n") *)

let rec bitEqBlist n a b  =
  match b with
      [] -> ""
    | numb::rest -> ((Printf.sprintf "(EQ %s_%d %d) " a n numb) ^ 
	(bitEqBlist (n-1) a rest))

let rec bitNEqBlist n a b  =
  match b with
      [] -> ""
    | numb::rest -> ((Printf.sprintf "(NEQ %s_%d %d) " a n numb) ^ 
	(bitNEqBlist (n-1) a rest))


let bitEqInt a b = 
  let blist = convToBin b in
  let foo = ref "(AND " in
  let i = List.length blist in
    for j = !bits downto (i+1) do
      foo := !foo ^ (Printf.sprintf "(EQ %s_%d 0) " a j)
    done;
    foo := !foo ^ (bitEqBlist i a (List.rev blist));
    !foo ^ ")"


let bitNEqInt a b = 
  let blist = convToBin b in
  let foo = ref "(OR " in
  let i = List.length blist in
    for j = !bits downto (i+1) do
      foo := !foo ^ (Printf.sprintf "(NEQ %s_%d 0) " a j)
    done;
    foo := !foo ^ (bitNEqBlist i a (List.rev blist));
    !foo ^ ")"

    (*let () = print_string ("(BG_PUSH "^(bitEqInt "a" 10) ^ ")\n") *)

let bitEqS a b =
  let foo = ref "(AND " in
  for i = 1 to !bits do
    foo := !foo ^ (Printf.sprintf "(EQ %s_%d %s_%d) " a i b i)
  done;
    !foo ^ ")"
      
let bitNEqS a b =
  let foo = ref "(OR " in
  for i = 1 to !bits do
    foo := !foo ^ (Printf.sprintf "(NEQ %s_%d %s_%d) " a i b i)
  done;
    !foo ^ ")"

let fixed_simplify_axioms = ref 
	    [
(*
             "(DEFPRED (BitAnd x y))\n" ;
	     "(DEFPRED (BitOr x y))\n" ;
	     "(DEFPRED (BitNot x))\n" ;
	     "(DEFPRED (Xor x y))\n" ;
	     "(DEFPRED (addrOf x y))\n" ;
	     (* "(DEFPRED (select y x))\n" ; *)
	     "(DEFPRED (foffset x f))\n" ;
	     "(DEFPRED (RShift x y))\n" ;
	     "(DEFPRED (LShift x y))\n" ;
	     "(DEFPRED (Div x y))\n" ;
	     "(DEFPRED (Rem x y))\n" ;
*)
  (*
  "(BG_PUSH (FORALL (x y z) (IMPLIES (AND (NEQ y 0) (EQ z (BitOr x y))) (NEQ (BitAnd z y) 0))))\n" ;
*)
  "(BG_PUSH (FORALL (x y) (EQ (select (addrOf x y) 0) x)))" ;
  "(BG_PUSH (FORALL (x y d1) (IMPLIES (EQ (foffset x d1) (foffset y d1)) (EQ  x y))))\n" ;
  "(BG_PUSH (FORALL (x y) (NEQ (addrOf x y) 0)))" ;
  "(BG_PUSH (FORALL (x y) (EQ (* (Div x y) y) x )) ) " ;
  "(BG_PUSH (FORALL (x) (NEQ (_STRINGCONSTANT x) 0 ))) " ;
(*
  "(BG_PUSH (FORALL (x y z) (IMPLIES (EQ z (addrOf x)) (EQ (select z 0) x))))\n" ;
  "(BG_PUSH (FORALL (x) (EQ (addrOf (select x 0)) x)))" ;
  "(BG_PUSH (FORALL (x y d1 d2) (IMPLIES (EQ (addrOf (select x d1)) (addrOf (select y d1))) (EQ  (addrOf x) (addrOf y)))))\n"
*)
	]



let simplify_stack_counter = ref 0
let current_simplify_stack = ref [] (* a string list *)
let simplifyServer = ref None
let simplifyServerCount = ref 0 
let simplifyCacheCount = ref 0 

let simplify_query_flag = ref false (* true *)

let set_simplify_query_flag () = 
  simplify_query_flag := false (* true *)


let rec addAxioms axiom_list =
  Message.msg_string Message.Minor "(Adding axioms)" ;
  let oc  = match (!simplifyServer) with
      None -> snd (getServer ()) 
    | Some(_,oc1,_) -> oc1
  in
    try
      List.iter 
	(function x -> (Message.msg_string Message.Debug (x^"\n"); 
			secure_output_string oc (x^"\n")))
	axiom_list ;
      if (!simplify_query_flag) then
	List.iter 
	  (function x -> ( 
	     Misc.append_to_file "simp_query" (x^"\n")))
	  axiom_list ;
      ()
    with ChannelException -> failwith "Channel Exception in addAxioms!"

  and  getServer () =
  bits := Options.getValueOfInt "bits";
    match !simplifyServer with
       None -> begin
	  Message.msg_string Message.Normal "Forking Simplify process..." ;
	  let ic,oc,ec = Unix.open_process_full "Simplify -nosc" 
			   (Unix.environment ()) in
	    simplifyServer := Some(ic,oc,ec) ;
    (* add axioms to simplify  -- assume axioms are in $BLASTHOME/bin/axioms.in *)
	    addAxioms (!fixed_simplify_axioms);
	    Message.msg_string Message.Normal "done!\n";
	    flush stdout ;
	    at_exit (fun () -> 
		       ignore ((*kill_simplify_server () *)Unix.close_process_full (ic,oc,ec)  ) ;  ) ;
	    (ic,oc)
	end
      | Some(a,b,c) -> (a,b)
  and  kill_simplify_server () = 
  match !simplifyServer with
      None -> ()
    | Some (ic,oc,ec) -> 
	begin
	  try
	    Message.msg_string Message.Error "Killing simplify server...";
	    (* ignore (Unix.close_process_full (ic,oc,ec) ); garbage process ... how to collect *)
	    Sys.command "killall Simplify";
	    simplifyServer := None;
	    Message.msg_string Message.Error "Done killing ..." ;
	    Message.msg_string Message.Normal "Done killing ..."
	  with _ -> failwith " failed to kill simplify !"
	end

let convertToSimplifyBitSyntax () exp =
  (*Simplify has issues with variable names starting with "_", so we map all vars to v<num> *)
  
  Message.msg_string Message.Debug ("convertToSimplifySyntax exp = ");
  Message.msg_printer Message.Debug Predicate.print exp;  (*RuGang*)
  
  let  isAtom e =   
    match e with
	Expression.Binary (op, e1, e2) -> 
	  begin
	    match op with 
		Expression.Eq  | Expression.Ne | Expression.Gt 
	      | Expression.Lt | Expression.Ge  | Expression.Le 
		  -> true
	      | _ -> false
	  end
      | _ -> false in

  let get_next_var () = next_var := !next_var + 1; !next_var in
  let get_current_var () = !next_var in
    if (!clear_varHashtbl_flag) then 
      begin
	Hashtbl.clear varHashtbl;next_var :=0; 
	Hashtbl.clear chlvalHashtbl
      end ;
    let fieldId str =
      if Hashtbl.mem varHashtbl str then
	Hashtbl.find varHashtbl str
      else
	begin
	  let new_id = "v"^(string_of_int (get_next_var ())) in
	    Hashtbl.add varHashtbl str  new_id;
	    new_id
	end
    in
    let id_of_var lv =
      match lv with
	  Expression.Symbol s -> 
	    begin
	      try
		let vid = Hashtbl.find varHashtbl s in
		  int_of_string (Misc.chop_after_prefix "v" vid)
	      with 
		  Not_found -> failwith ("id_of_var: (Simplify) variable not found "^(s))
	    end
	      
	| _ -> failwith ("id_of_var: (Simplify) lvalue not a symbol. "^(Expression.lvalToString lv))
    in
    let id_of_fld s =
      
      let vid = try Hashtbl.find varHashtbl s 
      with Not_found -> failwith ("id_of_fld: (Simplify) variable not found "^(s))	
      in
	int_of_string (Misc.chop_after_prefix "v" vid)
	  
    in
      
    let newVar str = 
      if Hashtbl.mem varHashtbl str then
	(Hashtbl.find varHashtbl str)
      else
	begin
	  let new_id = "v"^(string_of_int (get_next_var ())) in
	  let (_, oc) = getServer () in
	    Hashtbl.add varHashtbl str new_id;
	    (*Message.msg_string Message.Debug ("NOT in hashtbl:"^str^":"^new_id);*)
	    let s =  ((assertBitVector new_id)^"\n") in
	      Message.msg_string Message.Debug ("new bitvector: "^s);
	      let _ = current_simplify_stack := s :: !current_simplify_stack in
	      let _ = simplify_stack_counter := !simplify_stack_counter + 1 in
		output_string oc s;
		new_id
	end in

    let rec  convertExp e =
      let chlval_convert_recursive (lv,s) = Printf.sprintf "(chlval %s %s)" (convertLval lv) (fieldId s) in  
      let chlval_convert_table e = 
	try (Hashtbl.find chlvalHashtbl e) 
	with 
	    Not_found ->
	      begin
		let s_e = get_next_var () (* Expression.toString e*) in
		let new_id = "chlval"^(string_of_int s_e) in
		  Hashtbl.add chlvalHashtbl e new_id;
		  new_id
	      end
      in
	match e with
	  | Expression.Lval lv -> convertLval lv
	  | Expression.Chlval (lv,s)  -> 
	      chlval_convert_table e
		
	  | Expression.Constant (c) -> 
	      begin
		match c with
		    Constant.Int i ->  if (i < 0) then "(- 0 "^(string_of_int (-i))^")" else string_of_int i 
		  | _ -> failwith ("Cannot handle constant "^(Constant.toString c))
	      end
		
	  | Expression.Unary (uop, e1) -> 
	      begin
		match uop with 
		    Expression.UnaryMinus -> "(- 0 "^(convertExp e1)^")"
		  | Expression.UnaryPlus -> convertExp e1
		  | Expression.Address -> failwith ("unimplemented unary")
		  | Expression.BitNot -> failwith ("unimplemented unary")
		  |	_ -> failwith ("unimplemented unary")
	      end
          | Expression.FunctionCall (Expression.Lval (Expression.Symbol f), args) ->
              Printf.sprintf "(%s %s)" f (Misc.string_list_cat " " (List.map convertExp args))
	  | _ -> failwith ("convertExp (Simplify): " ^ (Expression.toString e) ^ "not handled")
	      
    and convertLval l =
      match l with
	| Expression.Symbol(str) -> 
	    newVar str
	| _ ->failwith ("unimplemented lval")  
	    
    and convertAtom e =   
    match e with
	Expression.Binary (Expression.Eq, e1, e2 ) -> 
	  begin
	    match e1 with
		Expression.Lval _ | Expression.Chlval _ ->
		  let s1 = convertExp e1 in
		    begin 
		      match e2 with
			  Expression.Binary _ ->
			    let newRes = newVar "result" in
			      Printf.sprintf "(AND %s %s)"  (bitEqS s1 newRes) (convertExpRes newRes e2)		  
			| Expression.Lval _
			| Expression.Chlval _ -> (bitEqS s1 (convertExp e2))
			| Expression.Constant (Constant.Int i) -> 
			    (bitEqInt s1 i)
			| _ -> failwith ("not implemented "^(Expression.toString e2))
		    end
	      | Expression.Constant (Constant.Int j) ->
		  begin 
		    match e2 with
			Expression.Binary _ ->
			  let newRes = newVar "result" in
			    Printf.sprintf "(AND %s %s)"  (bitEqInt newRes j) (convertExpRes newRes e2)		  
		      | Expression.Lval _
		      | Expression.Chlval _ -> Printf.sprintf "%s"  (bitEqInt (convertExp e2) j)
		      | Expression.Constant (Constant.Int i) -> 
			  if (i = j) then "TRUE" else "FALSE"
		      | _ -> failwith ("not implemented "^(Expression.toString e2))
		  end
	      | Expression.Binary _ -> 
		  begin
		    match e2 with
			Expression.Binary _ -> failwith ("not implemented "^(Expression.toString e2))
		      | _ -> convertAtom (Expression.Binary (Expression.Eq, e2, e1))
		  end
	      | _ ->  failwith ("not implemented")
	  end
      | Expression.Binary (Expression.Ne, e1,  e2) -> 
	  begin
	    match e1 with
		Expression.Lval _ | Expression.Chlval _ ->
		  let s1 = convertExp e1 in
		    begin 
		      match e2 with
			  Expression.Binary _ ->
			    let newRes = newVar "result" in
			      Printf.sprintf "(AND %s %s)"  
				(bitNEqS s1 newRes) (convertExpRes newRes e2) 
			| Expression.Lval _
			| Expression.Chlval _ -> (bitNEqS s1 (convertExp e2))
			| Expression.Constant (Constant.Int i) -> 
			    (bitNEqInt s1 i)
			| _ -> failwith ("not implemented "^(Expression.toString e2))
		    end
	      | Expression.Constant (Constant.Int j) ->
		    begin 
		      match e2 with
			  Expression.Binary _ ->
			    let newRes = newVar "result" in
			      Printf.sprintf "(AND %s %s)"  (bitNEqInt newRes j) (convertExpRes newRes e2)
			| Expression.Lval _
			| Expression.Chlval _ -> Printf.sprintf "%s"  (bitNEqInt (convertExp e2) j)
			| Expression.Constant (Constant.Int i) -> 
			    if (j = i) then "TRUE" else "FALSE" 
			| _ -> failwith ("not implemented "^(Expression.toString e2))
		    end
	      | Expression.Binary _ -> 
		  begin
		    match e2 with
			Expression.Binary _ -> failwith ("not implemented "^(Expression.toString e2))
		      | _ -> convertAtom (Expression.Binary (Expression.Ne, e2, e1))
		  end
	      | _ ->  failwith ("not implemented "^(Expression.toString e1))
	  end
      | _ -> failwith ("not implemented "^(Expression.toString e))
  
	  
    and convertExpRes res e = 
      match e with 
      | Expression.FunctionCall (Expression.Lval (Expression.Symbol f), args) ->
          Printf.sprintf "(EQ (%s %s) 0)" f (Misc.string_list_cat " " (List.map convertExp args))
      |	Expression.Binary (op, e1, e2) -> 
	    begin
	      match op with 
		| Expression.RShift -> 
		    begin
		      match e2 with
			Expression.Constant (Constant.Int i) ->
			  begin 
			    match e1 with
			      Expression.Binary _ ->
				let newRes = newVar "result" in
				let s = convertExpRes newRes e1 in
				(Printf.sprintf "(AND %s %s)" (rShiftCons res newRes i) s)
				| _ -> (* e1 is just a lval? *)
				    let s = convertExp e1 in
				    (Printf.sprintf "%s" (rShiftCons res s i))
			  end
		      | Expression.Lval _ | Expression.Chlval _  ->
			    let s2 = convertExp e2 in
			    begin 
			      match e1 with
				Expression.Binary _ ->
				  let newRes = newVar "result" in
				  let s = convertExpRes newRes e1 in
				  (Printf.sprintf "(AND %s %s)" (rShiftVar res newRes s2) s)
			      | _ -> (* e1 is just a lval? *)
				  let s = convertExp e1 in
				  (Printf.sprintf "%s" (rShiftVar res s s2))
			    end
			| _ -> failwith ("not implemented "^(Expression.toString e2))
		    end
		| Expression.Minus 
		| Expression.Plus   
		| Expression.Mul   
		| Expression.BitAnd 
		| Expression.BitOr 
		| Expression.Div 
		| Expression.Rem 
		| Expression.Xor 
		| Expression.FieldOffset 
		| Expression.Comma -> failwith ("not implemented")
		  (* since vampyre knows NOTHING about this things, we shall 
		      put in a brand new symbol here 
		  *)
		| _ -> 
		    if Expression.isRelOp op 
		    then 
		      raise AtomInExpressionException
		    else 
		      failwith ("convertExp: (Simplify) This operator "^(Expression.binaryOpToVanillaString op)^" should not be handled here "^(Expression.toString e))
	    end

    and convertPred e =
      match e with
	  Predicate.True -> "TRUE"
	| Predicate.False -> "FALSE"
	| Predicate.And plist -> 
	    List.fold_left (function a -> function b -> "(AND "^(convertPred b)^" "^a^")") 
	      "TRUE" (plist)
	| Predicate.Or plist -> 
	    List.fold_left (function a -> function b -> "(OR "^(convertPred b)^" "^a^")") 
	      "FALSE" (plist)
	| Predicate.Implies (e1, e2) -> "(IMPLIES "^(convertPred e1)^" "^(convertPred e2)^")"
        | Predicate.Iff (e1,e2) -> Printf.sprintf "(IFF %s %s)" (convertPred e1) (convertPred e2)
        | Predicate.Not e1 -> "(NOT "^(convertPred e1)^")"
	| Predicate.Atom e ->  (convertAtom e)
        | Predicate.All (s,e1) -> "(FORALL s "^(convertPred e1)^")" 
	| _ -> failwith ("SimplifyBit: Cannot handle this predicate "^(Predicate.toString e))
    in
      (convertPred (Predicate.constantfold exp (Hashtbl.create 1)))^"\n"


(*
let simplify_line_tap = ref [] (* string list *)
*)

let dump_line_tap () = ()
(*
  let oc = open_out "line_tap" in
    Marshal.to_channel oc ( List.rev !simplify_line_tap) []; 
    close_out oc;
    Message.msg_string Message.Error "Done writing line_tap";
    failwith "simplify choked!"
 *) 






let restart_simplify () = 
  Message.msg_string Message.Error "restarting simplify ...";
  Message.msg_string Message.Normal "restarting simplify ...";
  kill_simplify_server ();
  let (ic,oc) = getServer () in
    try
      List.iter (secure_output_string oc) (List.rev !current_simplify_stack)
    with ChannelException -> failwith "Simplify fails on restart!"


let cvc_push () =
  cvcPushes:=!cvcPushes+1;
  let (_,oc) = getCvcServer () in
    output_string oc "PUSH;\n";
    output_string filed "PUSH;\n";
    flush oc;
    flush filed
      
      
let cvc_assume pred = 
  Message.msg_string Message.Debug ("cvc_assume "^(Predicate.toString pred));
  clear_varCvcHashtbl_flag := false;
  let cvcInput = convertToCvcSyntax pred;
  in 
    Message.msg_string Message.Debug ("cvc Assuming: "^cvcInput);
    flush stdout;
    flush stderr;
    let s = ("PUSH; ASSERT ("^cvcInput^");\n") in (* add PUSH? *)
    cvcPushes:=!cvcPushes+1; 
      let (_,oc) = getCvcServer () in
	cvc_output_list  oc !cvc_ground_term_list; 
	Message.msg_string Message.Debug ("cvc input: "^s); 
	output_string oc s;
	output_string filed s;
	flush oc;
	flush filed;
	reset_cvc_ground ();
	Message.msg_string Message.Debug "Done Assume"; 
	flush stdout
   
let simplify_assume () pred = 
(* deadchange *)
  Message.msg_string Message.Debug ("simplify_assume");
  Message.msg_string Message.Debug ("sstk:"^(string_of_int !simplify_stack_counter));
   
  let simplifyInput = 
    if (!tp = "simplify-bits") then
      let simplifyBitInput = Bstats.time "convertToSimplifySyntax" (convertToSimplifyBitSyntax ()) pred in
      let simplifyOldInput = (convertToSimplifySyntax ()) pred in
	(Printf.sprintf "(AND %s %s)" simplifyBitInput simplifyOldInput) 
    else  Bstats.time "convertToSimplifySyntax" (convertToSimplifySyntax ()) pred
  in
    clear_varHashtbl_flag := false;
    Message.msg_string Message.Debug ("Simplify Assuming: "^simplifyInput);
    flush stdout;
    flush stderr;
    let s = ("(BG_PUSH "^simplifyInput^" )\n") in
    let _ = current_simplify_stack := s :: !current_simplify_stack in
    let _ = simplify_stack_counter := !simplify_stack_counter + 1 in
    let (_,oc) = getServer () in
    try
      secure_output_string oc s ;
      flush oc ;     
      Message.msg_string Message.Debug "Done Assume"; 
      flush stdout;
      if (!simplify_query_flag) then Misc.append_to_file "simp_query" ("\n"^s^"\n")
    with  ChannelException ->
      begin
	Message.msg_string Message.Error "ChannelException: in simp_assume";
	restart_simplify ()
      end
	
let cvc_pop () = 
   Message.msg_string Message.Debug ("cvc pop");
   Message.msg_string Message.Debug ("sstk:"^(string_of_int !cvcPushes));
   let (_,oc) = getCvcServer () in
   if (!cvcPushes > 0) then 
   begin
   cvcPushes:=!cvcPushes-1;
   output_string oc "POP;\n";
   output_string filed "POP;\n";
   flush filed;
   flush oc
   end else 
   Message.msg_string Message.Debug ("ERROR: cvc pop trying to pop nothing!")
   

let simplify_pop () = 
(* deadchange *)  Message.msg_string Message.Debug ("simplify_pop");

  Message.msg_string Message.Debug ("sstk:"^(string_of_int !simplify_stack_counter));
  let _ = 
    if (!simplify_stack_counter = 0) then failwith "bad simplify stack!"
    else
      simplify_stack_counter := !simplify_stack_counter - 1 
  in
  let top_of_stack = List.hd !current_simplify_stack in
  let _ = current_simplify_stack := List.tl !current_simplify_stack in
    Message.msg_string Message.Debug ("Simplify Popping");
    let (_,oc) = getServer () in
    let s = "(BG_POP)\n" in
    clear_varHashtbl_flag := true;
    try
      secure_output_string oc s;
      Message.msg_string Message.Debug "Done pop";
      if (!simplify_query_flag) then Misc.append_to_file "simp_query" ("\n"^s^"\n");
      flush oc
    with ChannelException ->
      begin
	Message.msg_string Message.Error "ChannelException: in simp_pop";
	restart_simplify ()
      end

(* doesnt clear the varHashtbl *)
let simplify_pop_noflush () =  
(* deadchange *)  Message.msg_string Message.Debug ("simplify_pnf");
  Message.msg_string Message.Debug ("sstk:"^(string_of_int !simplify_stack_counter));
   let _ = 
    if (!simplify_stack_counter = 0) then failwith "bad simplify stack_nf!"
    else
      simplify_stack_counter := !simplify_stack_counter - 1 
  in
  Message.msg_string Message.Debug ("Simplify Popping");
  let (_,oc) = getServer () in
  let s = "(BG_POP)\n" in
  let _ = current_simplify_stack := List.tl !current_simplify_stack in
    try
      secure_output_string oc s;
      Message.msg_string Message.Debug "done pop_pnf";
      if (!simplify_query_flag) then Misc.append_to_file "simp_query" s;
 flush oc
 with ChannelException ->
   begin
     Message.msg_string Message.Error "ChannelException: in simp_pop_pnf";
     restart_simplify ()
   end
 


let rec isValid ic = 
(* deadchange *)
  let  _ =  Message.msg_string Message.Debug ("Simplify : in isValid") in
  let  _ = flush stdout in
  (* let (_,oc) = getServer () in
  let _ = flush oc in*)
  let line = secure_input_line ic in (* any failure here will be caught further up *)
  let _ = Message.msg_string Message.Debug ("Simplify says: "^line) in
  let _ = 
    if Misc.is_substring line "Bad input" then 
      begin
	Message.msg_string Message.Error "Simplify poisoned!";
	exit 1
      end
  in
  let result = if String.contains line 'V' then 
    true
  else if String.contains line 'I' then
    false
  else isValid ic
  in
    result
      
let ques_flag = ref false (* to see if simplify is failing forever *)

let rec queryExpUsingSimplify pred = 
(* deadchange *)
  Message.msg_string Message.Debug ("querying exp:");
  Message.msg_printer Message.Debug Predicate.print pred ;
  let simplifyInput =
    if (!tp = "simplify-bits") then
      let simplifyBitInput = Bstats.time "convertToSimplifySyntax" (convertToSimplifyBitSyntax ()) pred in
      let simplifyOldInput = (convertToSimplifySyntax ()) pred in
	(Printf.sprintf "(AND %s %s)" simplifyBitInput simplifyOldInput) 
    else  Bstats.time "convertToSimplifySyntax" (convertToSimplifySyntax ()) pred
  in 
    Message.msg_string Message.Debug ("Simplify: "^simplifyInput);
    flush stdout;
    flush stderr;
    let result = 
    try
      let ic,oc = getServer () in
	secure_output_string oc (simplifyInput) ;
	if (!simplify_query_flag) 
	then Misc.append_to_file "simp_query" ("\n"^simplifyInput^"\n");
        flush oc ; 
        isValid ic
   with ChannelException ->
     begin
       if (!ques_flag) then failwith "Simplify fails again!";
       ques_flag := true;
       restart_simplify ();
       queryExpUsingSimplify pred
     end
   | e ->
       failwith ("Simplify raised exception "^(Printexc.to_string e)^" for "^(Predicate.toString pred)^". Check that Simplify is in your path.")
 in
  Message.msg_string Message.Debug (if result then "true" else "false");
  ques_flag := false;
  result



	       
let getProofsFromSimplify exp =
  let _ = Message.msg_string Message.Debug("querying exp in getProofsFromSimplify :"^(Predicate.toString exp)^"\n") in
   (* first we have to convert the formula and collect all the ground terms that appear in it *)
  let formulas = 
    match exp with 
      Predicate.And l -> l 
    | _ -> failwith ("bad call to getProofsFromSimplify -- we only handle listd of conjunections") in
  let ic,oc = getServer () in
  let counter = ref 0 in
  let output_asserts oc lst =
    List.iter (fun pred -> simplify_assume () pred; incr counter) lst
  in
  let rec usefulPredicates usefulPreds  =
    try
      let rec _worker assertlist =
	match assertlist with 
	  [] -> 
	    begin
	      Message.msg_string Message.Error "Exception : Formula is not valid! Simplify cannot prove validity." ;
	      Message.msg_string Message.Debug "Exception : Formula is not valid! Simplify cannot prove validity." ;
	      Message.msg_string Message.Debug "\nThe formula was :" ;
	      Message.msg_string Message.Debug (Predicate.toString exp) ;
	      
	      raise Not_found
	    end
	| a :: rest ->
	    simplify_assume () a ;
            incr counter ;
	    secure_output_string oc "FALSE\n" ;
            (* flush oc ;*)
	    if (!simplify_query_flag) then Misc.append_to_file "simp_query" 
("\n FALSE\n");
	    let result = 
	      try
   		isValid ic
	      with e ->
		failwith ("Simplify raised exception "^(Printexc.to_string e)^" for "^(Predicate.toString exp))
          in
            if result then a
            else
              _worker rest 
      in
      begin
        Message.msg_string Message.Debug "In useFulPredicates, List is " ;
	List.iter (fun p -> Message.msg_string Message.Debug (Predicate.toString p)) usefulPreds ;
        Message.msg_string Message.Debug "end List. " ;
	
        output_asserts oc usefulPreds ;
        secure_output_string oc "FALSE\n" ;
        (* flush oc ;*) 
	if (!simplify_query_flag) then Misc.append_to_file "simp_query" "\n FALSE \n";
	
        Message.msg_string Message.Debug "Asserted useful predicates. " ;
	
        let result = try isValid ic with e -> failwith "Simplify raised exception -- in getUseful" in
        if result then 
	  begin
            while !counter > 0 do simplify_pop () ; decr counter done ;
            usefulPreds 
          end
        else 
          begin 
           (* go on adding predicates from assertlist and check for validity.
	      When valid, add the last formula to usefulPreds and repeat
              *)
            let nextpred = _worker formulas  in
            while !counter > 0 do
              simplify_pop () ; decr counter 
            done ;
            usefulPredicates (nextpred ::  usefulPreds) 
          end
      end
    with
      Not_found -> begin (* failwith "simplify" ;*) while !counter >0 do simplify_pop (); decr counter done; formulas end
  in
  let listOfPreds = usefulPredicates [] 
  in
  (* iterate simplify pop to get rid of assumptions *)
  List.map 
    (fun a -> match a with Predicate.Not b -> b | _ -> a)
    listOfPreds

      
(* Simplify incremental interface *)




(* end simplify incremental *)



    
(*************************************************************************************)

(* let simplify_fun () e = queryExpUsingSimplify e -RuGang*)

let block_assert p = 
  Message.msg_string Message.Debug ("bs_assert: ");
  Message.msg_printer Message.Debug Predicate.print p;
  simp_ctr := !simp_ctr + 1;
   if (!tp = "simplify"
   || !tp = "simplify-bits") then simplify_assume () p
   else if (!tp = "cvc") then cvc_assume p

let block_pop () =
   Message.msg_string Message.Debug  "bs_pop"; 
   assert (!simp_ctr > 0);
   simp_ctr := !simp_ctr - 1;
   if (!tp = "simplify"
     || !tp = "simplify-bits") then 
   simplify_pop_noflush ()
   else if (!tp = "cvc") then cvc_pop ()

    
let block_reset () = 
  Message.msg_string Message.Debug "bs_reset";
  if (!tp = "simplify"
      || !tp = "simplify-bits"
     ) then
    begin
      Misc.repeat_fn (simplify_pop) (!simp_ctr);
      simp_ctr := 0
    end
  else if  (!tp = "cvc") then 
    begin
      Hashtbl.clear varCvcHashtbl;
      Hashtbl.clear chlvalCvcHashtbl;
      clear_varCvcHashtbl_flag := true;
      simp_ctr :=0;
      cvcPushes:=0;
      reset_cvc_ground ();
      cvcServer := None (* kill cvc; let garbage collect do it *)
      (*      Misc.repeat_fn cvc_pop  (!simp_ctr); simp_ctr := 0 *)
    end

let _is_contra () = 
  if (!tp = "simplify"
      || !tp = "simplify-bits"
     ) then 
    begin
      Message.msg_string Message.Debug "simplify check contra";
      flush stdout;
      try
	let rv = queryExpUsingSimplify (Predicate.False) in
	  Message.msg_string Message.Debug ("Contra ? "^(string_of_bool rv));
	  rv
      with e -> 
	failwith ("Simplify raised exception "^(Printexc.to_string e)^" in simplify_is_contra")
    end
  else if (!tp = "cvc") then
    begin 
      Message.msg_string Message.Debug "cvc check contra";
      let rv = queryExpUsingCvc (Predicate.False) in
	Message.msg_string Message.Debug ("Contra ? "^(string_of_bool rv));
	rv
    end
  else
    failwith ("pick a theorem prover")

let check_unsat p =
    let _ = block_assert p in
    let rv = _is_contra () in
    let _ = block_pop () in
      rv



let rec _assume pred =
   if (!tp = "simplify"
      || !tp = "simplify-bits"
      ) then simplify_assume () pred
   else if (!tp = "cvc") then cvc_assume pred
   else simplify_assume () pred

let _pop () = 
   if (!tp = "simpify"
       || !tp = "simplify-bits"
      ) then simplify_pop ()
   else if (!tp = "cvc") then cvc_pop ()
   else simplify_pop ()


let rec queryExp exp =

  Message.msg_string Message.Debug ("Query : ");
  Message.msg_printer Message.Debug Predicate.print exp; 
  if (!tp = "ics") then
    begin
      Message.msg_string Message.Debug ("Query is "^(Predicate.toString exp)^"\n");
      let icsans =  queryExpUsingIcs exp and
	  vamans =  queryExpUsingVampyre exp 
      in
      Message.msg_string Message.Debug ("Vampyre says "^(if vamans then "true" else "false")^"\n");
      Message.msg_string Message.Debug ("Ics says "^(if icsans then "true" else "false")^"\n");
      if (not(vamans = icsans)) then 
	failwith ("TheoremProvers disagree on "^(Predicate.toString exp)^"!! Vampyre="^(if vamans then "true" else "false")^", Ics ="^(if icsans then "true" else "false"))
      else 
	vamans
    end
  else 
    if (!tp = "vampyre") then
      queryExpUsingVampyre exp
    else if (!tp="cvc") then
      queryExpUsingCvc exp
    else
      if (!tp = "simplify"
	  || !tp = "simplify-bits"
	 ) then
	(Bstats.time "simplify querytime" (queryExpUsingSimplify) exp) 
      else
	begin
	  tp :=  Options.getValueOfString "s";
	  if (!tp = "ics" || !tp = "vampyre" || !tp = "simplify" || !tp = "cvc" || !tp = "simplify-bits") then
	    queryExp exp
	  else
	    failwith ("TheoremProver: queryExp: Unknown theorem prover "^(!tp)^". Options are ics, simplify, and vampyre.")
	end
	
let queryExpContext query =
  if (!tp = "") then tp := "simplify";
  if (!tp = "simplify" ||  !tp = "simplify-bits") then
    (Bstats.time "simplify querytime" (queryExpUsingSimplify) query) 
  else if (!tp = "cvc") then queryExpUsingCvc query
  else
    begin 
      let vam_query = Input.pushNeg (convertToVampyreSyntax query) in
	vampyre_prove_context vam_query
    end

(* get the hash table of theorems. For those theorems that are true,
   get proofs from Vampyre and put in another hash table. Finally Marshal
   the hashtable and write out to disk.
*)  
let generateProofsSimple htbl =
  let pfHt = Hashtbl.create 511 in
  let genPf thm ans =
    if (ans = true) then
      let pf = 
	try
	  let vampyreInput = convertToVampyreSyntax thm
	  in
	  try
	    vampyreProver "foo" (Input.pushNeg vampyreInput)
	  with Engine.Failure ->((Message.msg_string Message.Error "Bug in Vampyre.\n") ; failwith "Vampyre bug in proofgen")
	with
	  VampyreErrormsg.Error -> (failwith ("Bug in Vampyre ...Proof gen"))
      in
      Hashtbl.add pfHt thm pf
  in
  Hashtbl.iter genPf htbl ;
  let pfString = Marshal.to_string pfHt [] in
  Message.msg_string Message.Error ("Proof size is "^(string_of_int (String.length pfString)))


(***********************************************************************************)
(* Jhala: Adding the proof functions ... *)
(***********************************************************************************)


 let queryExpUsingVampyre_proof exp =
   Message.msg_string Message.Debug "in queryExpUsingVampyre_proof";
   let correct_ans = queryExpUsingSimplify exp in
   let vampyreInput = convertToVampyreSyntax exp in
   let vampyreInput_pushneg = (Input.pushNeg  vampyreInput) in  
     try
       let proof = vampyreProver "foo" (vampyreInput_pushneg)
       in
	 if (not correct_ans) then 
	   begin
	     Message.msg_string Message.Debug ("PROOF ERROR: TP MISMATCH + !: "^(Predicate.toString exp)); 
	     (*VampyrePretty.printf "Vampyre Saw: %a \n" Input.dp vampyreInput_pushneg;
	     VampyrePretty.printf "Vampyre Proof: %a \n" Proof.dp proof;*)
	     None
	end
      else Some(proof)
    with Engine.Failure -> 
      begin
	(Message.msg_string Message.Debug "Proof failed.\n") ; if (correct_ans) then 
	  begin
	    Message.msg_string Message.Debug ("PROOF ERROR: TP MISMATCH - !: "^(Predicate.toString exp)); 
            (* VampyrePretty.printf "Vampyre Saw: %a \n" Input.dp vampyreInput_pushneg; *)
	    Some (Proof.Unimplemented)
	  end
	else None
      end
      | VampyreErrormsg.Error -> 
	  begin
	    Message.msg_string Message.Debug ("PROOF ERROR: BUG IN VAMPYRE !: "^(Predicate.toString exp)); 
	    VampyrePretty.printf "Vampyre Saw: %a \n" Input.dp vampyreInput_pushneg;
	    let backup_ans = queryExpUsingSimplify exp in
	      if backup_ans then Some (Proof.Unimplemented) else None
	  end
 (* (failwith ("Bug in Vampyre ...")) *)

let queryExp_proof exp = 
  Message.msg_string Message.Debug ("Query _proof : "^(Predicate.toString exp)); 
  queryExpUsingVampyre_proof exp
  
let queryExpContext_proof query = 
  let vam_query = Input.pushNeg (convertToVampyreSyntax query) in
    vampyre_prove_context_proof vam_query
 
(****************************************************************************)
(************************  interface to UCLID *******************************)
(****************************************************************************)
let uclid_ground_term_list = ref [] 

let add_uclid_ground str = 
  if List.mem str (!uclid_ground_term_list) 
  then () 
  else uclid_ground_term_list := str::(!uclid_ground_term_list)
    
let reset_uclid_ground () = 
  uclid_ground_term_list := 
  ["zero : TERM;";
   "select : FUNC[2];"; 
   "select_dot : FUNC[2];";
   "guf : FUNC[2];";
   "gur : PRED[2];";
   "guf_plus : FUNC[2];";
   "guf_minus : FUNC[2];";
  ]
    
let convertToUclidSyntax exp =
(*Simplify has issues with variable names starting with "_", so we map all vars to v<num> *)
  let get_next_var () = next_var := !next_var + 1; !next_var in
  let _ = if (false) then (Hashtbl.clear varHashtbl;next_var :=0) in
  let rec convertExp e =
    match e with
	Expression.Binary (op, e1, e2) -> 
	  begin
	    match op with 
		Expression.Minus -> 
		  begin
		    match e2 with
	                Expression.Constant (Constant.Int 0) -> (convertExp e1)
		      |	Expression.Constant (Constant.Int i) -> 
			  if (i > 0) then Printf.sprintf "(pred^%s(%s))" (string_of_int i) (convertExp e1)
			  else Printf.sprintf "(succ^%s(%s))" (string_of_int (-i)) (convertExp e1)
		      | _ -> Printf.sprintf "guf_minus(%s,%s)" (convertExp e1) (convertExp e2)
		  end
	      |	Expression.Plus  -> 
		  begin
		    match e2 with
                        Expression.Constant (Constant.Int 0) -> (convertExp e1)
		      | Expression.Constant (Constant.Int i) -> 
			  if (i > 0) then Printf.sprintf "(succ^%s(%s))" (string_of_int i) (convertExp e1)
			  else Printf.sprintf "(pred^%s(%s))" (string_of_int (-i)) (convertExp e1)
		      | _ -> Printf.sprintf "guf_plus(%s,%s)" (convertExp e1) (convertExp e2)
		  end
              
	      |	Expression.Mul   
	      |	Expression.BitAnd 
	      |	Expression.BitOr 
	      |	Expression.LShift 
	      |	Expression.Div 
	      |	Expression.Rem 
	      |	Expression.RShift 
	      |	Expression.Xor 
	      |	Expression.Comma ->
	        Printf.sprintf "guf(%s,%s)" (convertExp e1) (convertExp e2)
	 
	      |	_ -> failwith ("convertExp: (Uclid) This operator should not be handled here "^(Expression.toString e))
	  end
      | Expression.Constant (c) -> 
	  begin
	    match c with
		Constant.Int i ->  
	          if (i = 0) then "zero" 
	          else
		    if (i < 0) then Printf.sprintf "(pred^%s(zero)" (string_of_int (-i)) 
		    else Printf.sprintf "succ^%s(zero)" (string_of_int i) 
	      | Constant.String s -> add_uclid_ground (s^" : TERM;");(s) 
	      | _ -> failwith ("unknown constant type")
	  end

      | Expression.Unary (uop, e) -> 
	  begin
	    match uop with 
		Expression.UnaryMinus -> failwith "cannot handle uminus"
	      | Expression.UnaryPlus -> failwith "cannot handle uplus"
	      | Expression.Address -> failwith "cannot handle address"
	      |	_ -> failwith "unknown unop" 
	  end
      | Expression.Lval l -> convertLval l
      | Expression.FunctionCall _ -> failwith "ConvertExp (Uclid): Does not handle function call" 
      | _ -> failwith ("convertExp: " ^ (Expression.toString e) ^ "not handled")

  and convertLval l =
    match l with
    | Expression.This -> failwith "This not handled in convertLval"
    | Expression.Symbol(str) -> 
	begin
	  (if Misc.is_prefix "b_p" str then add_uclid_ground (str^" : TRUTH;") else add_uclid_ground (str^" : TERM;"));
	  str 
	end
    | Expression.Access(Expression.Dot, e, s) -> 
	begin
	  add_uclid_ground (s ^" : TERM;");
	  (Printf.sprintf "(select_dot(%s,%s))" (convertExp e) s)
	end
    | Expression.Access(Expression.Arrow, e, s) ->
	begin
	  add_uclid_ground (s ^" : TERM;"); 
	  (Printf.sprintf "(select_dot(%s,%s))" (convertExp e) s)
	end
    | Expression.Dereference e -> 
	begin
	  match e with
	    Expression.Binary(Expression.Plus,e1,e2) -> Printf.sprintf "(select(%s,%s))" (convertExp e1) (convertExp e2)
	  | _ -> Printf.sprintf "(select(%s,zero))" (convertExp e)
	end
    | Expression.Indexed (e1, e2) ->
	begin
	  Printf.sprintf "(select (%s, %s))" (convertExp e1) (convertExp e2)
	end
  and convertAtom e =
    match e with
      Expression.Binary (op, e1, e2) -> 
	begin
	  let opstring = 
	    match op with 
	      Expression.Eq  -> "="
	    | Expression.Ne  -> "!="
	    | Expression.Gt  -> ">"
	    | Expression.Lt  -> "<"
	    | Expression.Ge  -> ">="
	    | Expression.Le  -> "<="
	    | _   -> "" 
	  in
          if not (opstring = "") then
	    Printf.sprintf "(%s %s %s)" (convertExp e1) opstring (convertExp e2)
	  else (convertExp e)
	end
    | _ -> convertExp e 
	  
and convertPred e =
    match e with
	Predicate.True -> "1"
      | Predicate.False -> "0"
      | Predicate.And plist -> 
	    List.fold_left 
	      (fun a -> fun b -> Printf.sprintf " (%s & %s) " (convertPred b) a) 
	      "1" (plist)
      | Predicate.Or plist -> 
	  List.fold_left 
	    (function a -> function b -> Printf.sprintf "(%s | %s)" (convertPred b) a )
	    "0" (plist)
      | Predicate.Implies (e1, e2) -> Printf.sprintf "(%s => %s)" (convertPred e1) (convertPred e2)
      | Predicate.Not e1 -> Printf.sprintf "(~ %s)" (convertPred e1)
      | Predicate.Atom e ->  (convertAtom e)
      
      | Predicate.Iff (e1,e2) ->
          let p1 = Predicate.Implies (e1,e2) in
          let p2 = Predicate.Implies (e2,e1) in
          Printf.sprintf " (%s & %s) " (convertPred p1) (convertPred p2)
      | _ -> failwith ("UCLID: Cannot handle this predicate "^(Predicate.toString e))
  
      in
	(convertPred exp)


          (* this function checks if the formula given below is satisfiable -- 
	     right now it just dumps a trace into a file called blast-trace.ucl *)

let uclidSat fmla = 
   (* first we have to convert the formula and collect all the ground terms that appear in it *)
   reset_uclid_ground ();
   
   let block_counter = ref 0 in     
   let block_temp_list = ref [] in
   let formula_const _fmla = 
       block_counter := !block_counter + 1;
       block_temp_list := ("blast_temp_"^(string_of_int !block_counter))::(!block_temp_list);
       Printf.sprintf "blast_temp_%s := %s; \n" (string_of_int !block_counter) (convertToUclidSyntax _fmla)
   in
   let formulas = match fmla with Predicate.And l -> l | _ -> failwith ("bad call to uclidSat") in
   
   let def_string = String.concat "\n" (List.map formula_const formulas) in	     
   let root_string = " \n root := "^(String.concat " & " (!block_temp_list))^";\n" in
   let const_def_string = List.fold_left (function a -> function b -> Printf.sprintf "%s \n  %s" b a )
	  "" !uclid_ground_term_list
   in
   let uclid_string = def_string^root_string in
   let send_to_file = 
       Printf.sprintf "MODEL blasttrace \n CONTROL \n EXTVAR \n STOREVAR \n VAR \n CONST \n %s \n DEFINE %s \n EXEC \n decide(~root); \n" const_def_string uclid_string 
   in 
   Misc.write_to_file "blast-trace.ucl" send_to_file;
   let uclidret = Sys.command "runuclid blast-trace.ucl" in
     (*if uclidret = 127 then failwith "uclid is NOT in the path or something!"
     else*)
     begin
       Message.msg_string Message.Debug "running smvsat...";
       let smvsat_ret =  Sys.command "smvsat blastrace.cnf foobaz" in
	 if (smvsat_ret <> 1) then failwith "issues with smvsat" 
	 else
	   begin
	     Message.msg_string Message.Debug "...smvsat finished";
	     
	     Sys.command "grep \"b\_p\" blasttrace.map > tmp-blast-bpvars";
	     Sys.command "cut -d ' ' -f 1 ztrace.variables > tmp-blast-pfvars";
	     let useful_vars = List.map (List.hd) (Misc.grep_file "tmp-blast-bpvars" "tmp-blast-pfvars") in
	       useful_vars
	   end 
       end

let uclid_output modelname  fmla comment_string = 
  (* first we have to convert the formula and collect all the ground terms that appear in it *)
   reset_uclid_ground ();
   let block_counter = ref 0 in     
   let block_temp_list = ref [] in
   let formula_const _fmla = 
       block_counter := !block_counter + 1;
       block_temp_list := ("blast_temp_"^(string_of_int !block_counter))::(!block_temp_list);
       Printf.sprintf "blast_temp_%s := %s; \n" (string_of_int !block_counter) (convertToUclidSyntax _fmla)
   in
   let formulas = [fmla] (* match fmla with Predicate.And l -> l | _ -> failwith ("bad call to uclidSat:"^(Predicate.toString fmla)) *) in
   
   let def_string = String.concat "\n" (List.map formula_const formulas) in	     
   let root_string = " \n root := "^(String.concat " & " (!block_temp_list))^";\n" in
   let const_def_string = List.fold_left (function a -> function b -> Printf.sprintf "%s \n  %s" b a )
	  "" !uclid_ground_term_list
   in
   let uclid_string = def_string^root_string in
   let send_to_file = 
       Printf.sprintf "MODEL %s \n CONTROL \n EXTVAR \n STOREVAR \n VAR \n CONST \n %s \n DEFINE %s \n EXEC \n decide(~root); \n (* %s *)" modelname const_def_string uclid_string comment_string 
   in 
   Misc.write_to_file (modelname^".ucl") send_to_file

(*****************************************************************)
(** Implement a simple "predicate flow analysis" on the useful   *)
(*  predicates. This does a congruence closure on the variables  *)
(*  in the set of useful predicates and returns extra predicates *)
(*  that may be useful.                                          *)
(*****************************************************************)

let scope_pred_table = Hashtbl.create 37

let predicate_closure predlist = 
  let cong_classes = ref [] in
  let new_predicates = ref [] in
  let set_of_elt e = (* puts e in a new set if its not anywhere and if it is then removes it*)
    let _temp_set = List.filter (List.mem e) (!cong_classes) in
    let _cong_classes = List.filter (fun x -> not (List.mem e x)) (!cong_classes) in
    cong_classes := _cong_classes;
    if _temp_set = [] then ([e]) else List.hd _temp_set
  in
  
  let process_pred pred = 
    match pred with
      Predicate.Atom (Expression.Binary (Expression.Eq,e1,e2)) -> 
	let s1 = set_of_elt e1 in
	if (List.mem e2 s1) then cong_classes := s1:: !cong_classes
	else 
	  let u = Misc.union s1 (set_of_elt e2) in
	  cong_classes := u::(!cong_classes)
    | _ -> ()
  in
  let _ = List.iter process_pred predlist 
  in
    let scope_of_exp exp =
      match exp with
	Expression.Lval Expression.Symbol(s) -> 
	  begin
	    let names = Misc.chop s "@" in
	    if (List.length names = 1) then "__BLAST_GLOBAL" else (List.nth names 1)
	  end
      | _ -> "__BLAST_GLOBAL"
    in
    
    let process_cong_class cc = 
      let _ = print_string "cc: \n";List.iter (fun x -> print_string ((Expression.toString x)^"  \n")) cc in  
  (* this should return a list of predicates *)
      let _ = Hashtbl.clear scope_pred_table in
      let process_exp e = 
	let scp = scope_of_exp e in
	let _ =  Message.msg_string Message.Debug (Printf.sprintf "exp: %s   scope: %s" (Expression.toString e) scp) in
	let present_set = try Hashtbl.find scope_pred_table scp with Not_found -> [] in
	Hashtbl.replace scope_pred_table scp (e::present_set)
      in
      let proc_scope_terms globals scp lst =
	if (not (scp = "__BLAST_GLOBAL")) then
	  begin
	    let pairs = Misc.cross_product lst globals (* Misc.make_all_pairs (lst@globals)*) in
	    new_predicates := (List.map (fun (x,y) -> Predicate.Atom (Expression.Binary(Expression.Eq,x,y))) pairs )@(!new_predicates)
	  end
	else 
	  ()
      in
      List.iter process_exp cc;
      let globals_cc = try Hashtbl.find scope_pred_table "__BLAST_GLOBAL" with Not_found -> [] in
      let _ = print_string "globals: \n";List.iter (fun x -> print_string ((Expression.toString x)^"  \n")) globals_cc    
      in      
      Hashtbl.iter (proc_scope_terms globals_cc) scope_pred_table 
    in  
    let _ = List.iter process_cong_class (!cong_classes) in
    !new_predicates
    (* very inefficient congruence closure algorithm *)
   
(**********************************************************)      
(* Get the list of useful predicates from the last proof. *)
(**********************************************************)
let getUsefulPredicates pred =
  let getPredsFromProof pred =
    if ("cvc" = Options.getValueOfString "s") 
    then
      getProofsFromCvc pred    
    else 
    if ("simplify" = Options.getValueOfString "s"
    || "simplify-bits" = Options.getValueOfString "s") then
      getProofsFromSimplify pred 
    else
      begin
        let _eToCheck = Predicate.implies pred (Predicate.False) in
        let ans = queryExpUsingVampyre _eToCheck
        in
        let _ = Message.msg_string Message.Debug "Got UsefulPred" in
        if ans = false
        then
          []
        else
          begin
            let _lst = Engine.getUsefulPredicates () in
	    List.map convertToPredicate _lst
          end
      end 
  in
  let plist = getPredsFromProof pred in
	  Message.msg_string Message.Debug "Leaf predicates.\n";
          List.iter (fun p -> Message.msg_string Message.Debug (Predicate.toString p)) (plist); 
      (* let inf_plist = List.map convertToPredicate inferred_atoms in *)
      if(Options.getValueOfBool "scope") then
	begin
	  let pred_closure_list = predicate_closure plist in
      
	  Message.msg_string Message.Debug "Closure predicates.\n";
   	  List.iter (fun p -> Message.msg_string Message.Debug (Predicate.toString p)) (pred_closure_list); 
	  plist @ pred_closure_list 
	end
      else
	plist

	       
(*************************************************************************************)
(*************************************************************************************)
(*                               Interface to FOCI                                   *)
(*************************************************************************************)
(*************************************************************************************)
let stats_nb_foci_queries = ref 0

exception FociSatException
exception FociSatExceptionAsgn of string list

(* FOCI VERSION DIFF -- BEGIN *)
(* FOR FOCI 1 
module FociPredicate = FociAst.Predicate
module FociExpression = FociAst.Expression
module FociConstant = FociAst.Constant

let foci_interpolate (a,b) = 
  stats_nb_foci_queries := 1 + !stats_nb_foci_queries;
  try Foci.interpolate (a,b) with Foci.SAT -> raise FociSatException 
let foci_interpolateN _ = failwith "foci_interpolateN not supported in foci 1 !"

let foci_interpolateN_symm _ = failwith "foci_sym_interpolateN not supported in foci 1/2"

let foci_interpolateN_dasdill _ _ = failwith "foci_interpolateN_dasdill not supported in foci1/2"
let foci_theory_clauses () = failwith "foci_theory_clauses not supported in foci 1/2"  


END FOCI 1 *)  

(* FOR FOCI 2 *) 
module FociPredicate = FociAst
module FociExpression = FociAst
module FociConstant = FociAst.Constant

let foci_interpolate (a,b) =  
  stats_nb_foci_queries := 1 + !stats_nb_foci_queries;
  let a' = FociDag.pred_dag_of_tree a in
  let b' = FociDag.pred_dag_of_tree b in
  let itp' = 
    try  Foci.interpolate (a',b') with Foci.SAT _ -> raise FociSatException 
  in
  let itp = FociDag.pred_tree_of_dag itp' in
  itp

let get_pred_symbol fp = 
  match fp with
  FociPredicate.Atom s -> Some s
  | _ -> None

let itpN itp_fun plist = 
  stats_nb_foci_queries := 1 + !stats_nb_foci_queries;
  let p'_list = List.map FociDag.pred_dag_of_tree plist in
   let get_itp_list () = 
   try Stats.time "foci itpn" itp_fun (*Foci.interpolateN*) p'_list 
    with Foci.SAT (plist',_) ->
        if plist' = [] 
        then (Message.msg_string Message.Normal "Foci SAT: Empty"; 
                raise FociSatException)
          else 
            let plist'_s = Misc.strList (List.map FociDag.Predicate.toString plist') in 
            let _ = Message.msg_string Message.Normal ("Foci SAT: "^plist'_s) in
            let plist = List.map (FociDag.pred_tree_of_dag) plist' in
            let pslist = Misc.map_partial get_pred_symbol plist in
            raise (FociSatExceptionAsgn pslist)
  in
  let itp'_list  = Stats.time "compute itpN" get_itp_list () in
  let itp_list = List.map FociDag.pred_tree_of_dag itp'_list in
  itp_list

let foci_interpolateN = itpN Foci.interpolateN 

(* Uncomment for Foci 1/2 
let foci_interpolateN_symm _ = failwith "foci_sym_interpolateN not supported in foci 1/2"
let foci_interpolateN_dasdill _ _ = failwith "foci_interpolateN_dasdill not supported in foci1/2"
let foci_theory_clauses () = failwith "foci_theory_clauses not supported in foci 1/2"  
let foci_instance_ref = None 
let foci_inc_pop () = failwith "foci_inc_pop: not implemented"
let foci_inc_push p = failwith "foci_inc_push: not implemented"
let foci_inc_check_sat () = failwith "foci_inc_check_sat: not implemented"
let foci_inc_reset () = failwith "foci_inc_reset: not implemented"
END FOCI 2 *)

(* the next is there only in FOCI 3 *)
let foci_interpolateN_symm = itpN Foci.interpolateN_symm 

let foci_theory_clauses () = 
  let fp_list_list = Foci.theoryClauses () in  
  let tc_size = List.length (List.flatten fp_list_list) in
  Message.msg_string Message.Normal (Printf.sprintf "Foci says: %d theory clauses" tc_size);
  List.map (List.map (FociDag.pred_tree_of_dag)) fp_list_list

let foci_interpolateN_dasdill t_list c_list = 
   stats_nb_foci_queries := 1 + !stats_nb_foci_queries;
  let t'_list = List.map FociDag.pred_dag_of_tree t_list in
  let c'_list = List.map FociDag.pred_dag_of_tree c_list in
  let get_dd_itp_list () =
   try 
     Message.msg_string Message.Error "Calling dd_itpn";
     Stats.time "foci dd_itpn" (FociDasDill.refine t'_list) c'_list  
    with Foci.SAT (plist',_) ->
        if plist' = [] 
        then (Message.msg_string Message.Normal "Foci SAT: Empty"; 
                raise FociSatException)
          else 
            let plist'_s = Misc.strList (List.map FociDag.Predicate.toString plist') in 
            let _ = Message.msg_string Message.Normal ("Foci SAT: "^plist'_s) in
            let plist = List.map (FociDag.pred_tree_of_dag) plist' in
            let pslist = Misc.map_partial get_pred_symbol plist in
            raise (FociSatExceptionAsgn pslist)
  in
  let itp'_list  = Stats.time "compute itpN" get_dd_itp_list () in
  let _ = Message.msg_string Message.Error "done dd_itpn" in
  let itp_list = List.map FociDag.pred_tree_of_dag itp'_list in
  itp_list





  
(* END FOCI 3 *)  
(* FOCI VERSION DIFF -- END *)

let foci_expression_toString = FociAst.Expression.toString
let foci_predicate_toString = FociAst.Predicate.toString


type foci_return = Sat | Unsat of (Predicate.predicate (* * Predicate.predicate *))

let fctr = ref 0 

(* For the benefit of foci_symm -- we need to rename the UFs used with this suffix so that they are disjoint *)


let foci_make_disjoint_flag = ref false
let set_disjoint_flag () = 
  Message.msg_string Message.Debug "Set Disjoint flag";
  foci_make_disjoint_flag := true

let unset_disjoint_flag () = 
  Message.msg_string Message.Debug "Unset Disjoint flag";
  foci_make_disjoint_flag := false

let foci_UF_idx = ref (0)

let foci_UF_idx_tick () = foci_UF_idx := !foci_UF_idx + 1

let foci_UF_idx_reset () =  foci_UF_idx := 0
 
let foci_UF_string s =
  let rv = if (!foci_make_disjoint_flag) then s^"_"^(string_of_int !foci_UF_idx) else s in
  Message.msg_string Message.Debug (Printf.sprintf "foci_UF_string: in = %s, out = %s" s rv);
  rv

let clean_foci_UF_string s = 
  if not (!foci_make_disjoint_flag) then s
  else
    let pieces = Misc.chop s "_" in
    if (List.length pieces > 1) then String.concat "_" (Misc.chop_last pieces) 
    else s
 
let new_foci_var () = 
  fctr := !fctr + 1;
  "foci"^(string_of_int !fctr)

let listOfAddrVarsFoci = ref []
let chlval_table = Hashtbl.create 109
let chlval_map_table  = Hashtbl.create 109
let int_table_to = Hashtbl.create 23
let int_table_from = Hashtbl.create 23
let int_table_seen = Hashtbl.create 23

let max_int = 999999 (* some RANDOM number I made up at 1am -- RJ *)
let num_offset = ref 100000
let make_new_num i =
  let s_i = string_of_int i in
  let s_i_c = Misc.k_suffix 2 s_i in
  let rv = int_of_string s_i_c in
    rv
  (*num_offset := !num_offset + 1; 
  !num_offset*)

let sign i = if i > 0 then 1 else if i = 0 then 0 else -1 
 
let int_map_to i =
  (* Message.msg_string Message.Debug ("int_map_to"^(string_of_int i));*)
  (* add i to the list of constants we've passed to foci *)
  let _ = Hashtbl.replace int_table_seen i true in 
  let abs_i = abs i in
  let ret_val = 
   (sign i) * 
   (if abs_i < max_int then abs_i
    else
      begin
	if (Hashtbl.mem int_table_to abs_i) 
	then 
	  (Hashtbl.find int_table_to abs_i)
	else
	  let key = make_new_num abs_i in
	  let _ = (Hashtbl.add int_table_to abs_i key) in
	  let _ = (Hashtbl.add int_table_from key abs_i) in
	    key
      end)
 in
   (*Message.msg_string Message.Debug ("returns: "^(string_of_int ret_val));*)
   
   let _ = if (ret_val <> i) then Message.msg_string Message.Normal "WARNING :int_maps_to overflow!"
   in
     ret_val

let int_map_from i = 
  Message.msg_string Message.Debug ("int_map_from"^(string_of_int i));
  let abs_i = abs i in
  let _ = 
    let focimax = Options.getValueOfInt "focimax"  in
    if (focimax <> -1) && (not (Hashtbl.mem int_table_seen i)) && (abs_i > focimax)
    then 
      (Message.msg_string Message.Debug ("Invalid Interpolant -- large constant:
        "^(string_of_int i));
      raise FociSatException)
    (* this will trickle up all the way and be caught by whoever *)
  in
  let ret_val =
    (try
      Hashtbl.find int_table_from abs_i
    with Not_found -> abs_i) * (sign i)
  in
    Message.msg_string Message.Debug ("returns: "^(string_of_int ret_val));
    ret_val
 
(* in theory one could even use simplify or something else to "simplify" *)
let isConstant e = 
  match e with
      Expression.Constant (Constant.Int i) -> Some i
    | Expression.Unary (Expression.UnaryMinus,Expression.Constant (Constant.Int i)) -> 
	Some (-i)
    | Expression.Unary (Expression.UnaryPlus,Expression.Constant (Constant.Int i)) -> 
	Some (i)
    | _ -> None
	
let rec convertExpFoci e =
  let convert_times e1 e2 = 
    match (isConstant e1, isConstant e2) with
	(Some i1,Some i2) -> FociExpression.Constant (FociConstant.Int (i1 * i2))
      | (Some i1, None) -> FociExpression.Coeff (FociConstant.Int i1, (convertExpFoci e2))
      | (None, Some i2) -> FociExpression.Coeff (FociConstant.Int i2, (convertExpFoci e1))
      | (None, None) -> 
	  begin
	    Message.msg_string Message.Debug (Printf.sprintf "WARNING! convertExpFoci: times of %s * %s  " (Expression.toString e1) (Expression.toString e2));
	    let ufs = foci_UF_string "times" in
            FociExpression.Application(ufs,List.map convertExpFoci [e1;e2])
	  end
  in
  let foci_minus e' =
    match isConstant e' with
	Some i -> FociExpression.Constant (FociConstant.Int (int_map_to (-i)))
      | None -> FociExpression.Coeff (FociConstant.Int (-1),(convertExpFoci e'))
  in
  let chlval_convert_table e = 
    try FociExpression.Variable (Hashtbl.find chlval_map_table e) 
    with 
	Not_found ->
	  let s_e = new_foci_var () (* Expression.toString e*) in
	    Hashtbl.add chlval_table s_e e;
	    Hashtbl.add chlval_map_table e s_e;
	    FociExpression.Variable s_e
  in
  let chlval_convert_recursive (l,s) = 
    let ufs = foci_UF_string "chlval" in
    FociExpression.Application (ufs, [ convertLvalFoci l; FociExpression.Variable ("_"^s) ])
  in
    match e with
	Expression.Lval lv -> convertLvalFoci lv
      | Expression.Chlval (l, s) -> 
	  if (Options.getValueOfBool "fociUF") then chlval_convert_recursive (l,s)
	  else chlval_convert_table e
	  
      (* So the chlval thing didnt work out so well. look at test/foci/ex3.foci 
	 so now we just convert it to string and for reconversion, stuff the result in a hashtbl!
	FociExpression.Application ("chlval", [ convertLvalFoci l; FociExpression.Variable ("_"^s) ])
      *)
  | Expression.Binary (op, e1, e2) -> 
      begin
	match op with 
	    Expression.Minus -> 
	      if (Options.getValueOfBool "uif") then
                let ufs = foci_UF_string "UIminus" in
		FociExpression.Application (ufs, [convertExpFoci e1; convertExpFoci e2])
	      else
		FociExpression.Sum [convertExpFoci e1; foci_minus e2]
	  | Expression.Plus  -> 
	      if (Options.getValueOfBool "uif") then
                let ufs = foci_UF_string "UIplus" in
		FociExpression.Application (ufs, [convertExpFoci e1; convertExpFoci e2])
	      else
		FociExpression.Sum[convertExpFoci e1; convertExpFoci e2]
	  | Expression.Mul   -> convert_times e1 e2
	  |	Expression.BitAnd 
	  |	Expression.BitOr 
	  |	Expression.Div 
	  |	Expression.Rem 
	  |	Expression.RShift 
	  |	Expression.Xor 
	  |	Expression.Comma ->
	        let ufs = foci_UF_string (Expression.binaryOpToVanillaString op ) in
                    FociExpression.Application(ufs, List.map convertExpFoci [e1;e2])
	  |	Expression.LShift ->
	      begin
		match e2 with
		  Expression.Constant (Constant.Int k) -> convert_times e1 (Expression.Constant (Constant.Int (Misc.power 2 k)))
		| _ -> 	  
                    let ufs = foci_UF_string (Expression.binaryOpToVanillaString op ) in
                    FociExpression.Application(ufs, List.map convertExpFoci [e1;e2])
	      end
	  |     Expression.FieldOffset ->
	     let v = convertExpFoci e1 in
	      (* the second argument to field offset is a field id *)
	     let s =
	       match e2 with
	         Expression.Lval (Expression.Symbol s) -> s
	       | _ -> failwith ("convertExp : (FOCI) Second operator of field offset not a field id "^
				 (Expression.toString e))
	     in
	     let ufs_fof = foci_UF_string "foffset" in
             let ufs_fld = foci_UF_string ("@field_"^s) in
             FociExpression.Application(ufs_fof, [v; FociExpression.Variable ufs_fld])
	     
	  |	_ -> 
	      if Expression.isRelOp op then raise AtomInExpressionException
	      else failwith ("convertExp (Foci): This operator should not be handled here "^(Expression.toString e))
      end
  | Expression.Constant (c) -> 
      begin
	match c with
	    Constant.Int i -> FociExpression.Constant (FociConstant.Int (int_map_to i))
	  | Constant.String s -> 
	      begin
		Message.msg_string Message.Debug ("convertExpFoci: String Constant! in"^(Expression.toString e));
                let c = get_string_constant s in
		FociExpression.Variable ("_STRINGCONSTANT"^c)
	      end
	  | Constant.Float f ->
	     	    if (f -. (ceil f) = 0.0) then FociExpression.Constant (FociConstant.Int (int_of_float f))
	    else FociExpression.Variable "_FLOATINGPTCONSTANT"
	  | _ -> failwith ("Cannot handle constant "^(Constant.toString c))
      end
  | Expression.Unary (uop, e) -> 
      begin
	match uop with 
	    Expression.UnaryMinus -> foci_minus e
	  | Expression.UnaryPlus -> convertExpFoci e
	  | Expression.Address -> 
	    let v = convertExpFoci e in
            let ufs = foci_UF_string "addrOf" in
	    if (List.mem v (!listOfAddrVarsFoci)) then () 
	    else
	      listOfAddrVarsFoci := v::(!listOfAddrVarsFoci);
	    FociExpression.Application(ufs, [v])
	  |	_ -> 
              let ufs = foci_UF_string (Expression.unaryOpToVanillaString uop)
              in
              FociExpression.Application(ufs, [convertExpFoci e])
      end
  | Expression.FunctionCall(Expression.Lval (Expression.Symbol f), args) ->
      FociExpression.Application(foci_UF_string f, List.map convertExpFoci args)
  | _ -> failwith ("convertExpFoci: " ^ (Expression.toString e) ^ "not handled")
	
  and convertLvalFoci l =
    match l with
    | Expression.This -> failwith "This not handled in convertLval"
    | Expression.Symbol(str) -> FociExpression.Variable (str) 
    | Expression.Access(Expression.Dot, e, s) -> (* NEEDS DEBUGGING *) 
	let v = convertExpFoci e in (* hack to get around foci bug! *)
        let ufs = foci_UF_string ("select@field_"^s) in
	FociExpression.Application( ufs, [v(*; FociExpression.Variable ("@field_"^s)*)])
    | Expression.Access(Expression.Arrow, e, s) ->
	failwith "foci doesnt like Arrow"
    | Expression.Dereference e ->
        let ufs = foci_UF_string "select" in
	FociExpression.Application(ufs, [convertExpFoci e; FociAst.zero])
    | Expression.Indexed (e1, e2) -> 
        let ufs = foci_UF_string "select" in
	FociExpression.Application(ufs, List.map convertExpFoci [e1;e2])
	(* WARNING!! need to check this -- RJ: Whats the difference between sel4 and select ? *)

  and 
    convertAtomFoci e = (* this returns a focipredicate *)
  let make_eq e1' e2' = 
    try 
      FociPredicate.Equality (convertExpFoci e1', convertExpFoci e2') 
    with AtomInExpressionException ->
      begin
        match e1' with
	  Expression.Lval _ | Expression.Chlval _ | Expression.Constant (Constant.Int _) ->
	    begin
	      let s1 = convertExpFoci e1' in
	      let s2 = convertAtomFoci e2' in
	      FociPredicate.Or [ (FociPredicate.And [s2 ; FociPredicate.Equality (s1, FociExpression.Constant (FociConstant.Int 1)) ] ) ; 
				 (FociPredicate.And [FociPredicate.Not s2 ; FociPredicate.Equality (s1, FociAst.zero) ] ) ]
	    end
        | _ -> 
	    begin
	      match e2' with
		Expression.Lval _ | Expression.Chlval _ | Expression.Constant (Constant.Int _) ->
		  begin
		    let s1 = convertExpFoci e2' in
		    let s2 = convertAtomFoci e1' in
		    FociPredicate.Or [ (FociPredicate.And [s2 ; FociPredicate.Equality (s1, FociExpression.Constant (FociConstant.Int 1)) ] ) ; 
				       (FociPredicate.And [FociPredicate.Not s2 ; FociPredicate.Equality (s1, FociAst.zero) ] ) ]
		  end
		    
	      | _ -> failwith ("convertAtomFoci: Cannot handle this expression: "^(Expression.toString e))
		    
	    end 
      end
  in
  let make_leq e1' e2' = 
    FociPredicate.Leq (convertExpFoci e1', convertExpFoci e2') 
  in
  match e with
    Expression.Binary (op, e1, e2) -> 
      begin
	match op with 
	  Expression.Eq  -> make_eq e1 e2 
	|	Expression.Ne  -> FociPredicate.Not (make_eq e1 e2)
	|	Expression.Le  -> make_leq e1 e2	      
	|	Expression.Lt  -> FociPredicate.Not (make_leq e2 e1)
		  (*FociPredicate.And 
		    [(make_leq e1 e2); FociPredicate.Not (make_eq e1 e2)]*)
	|	Expression.Gt  -> FociPredicate.Not (make_leq e1 e2)
		  (*FociPredicate.And 
		    [(make_leq e2 e1); FociPredicate.Not (make_eq e2 e1)]*)
	|	Expression.Ge  -> make_leq e2 e1 
	|	_   -> failwith ("convertAtomFoci: Cannot handle this expression: "^(Expression.toString e))
      end	
  | Expression.FunctionCall(Expression.Lval (Expression.Symbol f), args) ->
      make_eq e (Expression.Constant (Constant.Int 0))
  | Expression.Lval (Expression.Symbol s) -> FociPredicate.Atom s
  | _ -> failwith ("convertAtomFoci: "^(Expression.toString e)^ " Not a binary relation operator")
and convertPredFoci e =
  match e with
    Predicate.True -> FociPredicate.True
  | Predicate.False -> FociPredicate.False
  | Predicate.And plist -> 
      FociPredicate.And (List.map convertPredFoci plist)
  | Predicate.Or plist -> 
      FociPredicate.Or (List.map convertPredFoci plist)
  | Predicate.Implies (e1, e2) -> FociPredicate.Or [FociPredicate.Not (convertPredFoci e1); (convertPredFoci e2)]
  | Predicate.Not e1 -> FociPredicate.Not (convertPredFoci e1)
  
  | Predicate.Atom e ->  (convertAtomFoci e)
  | Predicate.Iff (e1,e2) ->  convertPredFoci 
        (Predicate.Or [(Predicate.And [e1;e2]);(Predicate.And [Predicate.Not e1 ; Predicate.Not e2])])
  | Predicate.All (s,e1) -> FociPredicate.Forall ([s],convertPredFoci e1)  
  | _ -> failwith ("Foci Cannot handle this predicate "^(Predicate.toString e))
	
and convertToFociSyntax (e: Predicate.predicate) =
  listOfAddrVarsFoci := [];
  let v_exp = convertPredFoci (Predicate.constantfold e (Hashtbl.create 1))
  in
  let ufs_sel = foci_UF_string "select" in
  let ufs_aof = foci_UF_string "addrOf" in
  let genAddrOfEq b =
        FociPredicate.Equality(b, FociExpression.Application(ufs_sel, 
							 [FociExpression.Application(ufs_aof, [b]); FociAst.zero]))
  in
  let genAddrPositive b = FociPredicate.Not  (FociPredicate.Leq 
				( FociExpression.Application(ufs_aof, [b]), (FociAst.zero))) in
    begin
    let vv = FociPredicate.And ((v_exp
                      ::(List.map (fun b -> FociPredicate.Not (FociPredicate.Equality ( 
                                                                       (FociExpression.Application(ufs_aof, [b])) ,
                                                                       FociAst.zero)))
                                  !listOfAddrVarsFoci))
                       @ (List.map genAddrOfEq !listOfAddrVarsFoci)
                       @ (List.map genAddrPositive !listOfAddrVarsFoci)
                       ) in
      vv
  end

(* oh no! now we have to write the convert from foci to BLAST. ugh. *)
let rec convertFociLval l = 
  (* this is weird ... since the arg: "l" is supposed to be a foci expression that is really an lval ! *)
  match convertFociExp l with
      Expression.Lval l' -> l'
    | _ -> failwith ("convertFociLval failed on argument"^(foci_expression_toString l))
    
and make_bin op e1 e2 = 
  Expression.Binary(op,convertFociExp e1, convertFociExp e2)
and make_eq e1 e2 = make_bin (Expression.Eq) e1 e2
and make_leq e1 e2 = make_bin (Expression.Le) e1 e2

and gatherFociSum el = 
  match el with
      [] -> Expression.Constant (Constant.Int 0)
    | h::tl -> 
		  List.fold_left 
	(fun a b -> Expression.Binary (Expression.Plus,a,convertFociExp b)) 
	(convertFociExp h) tl
	
and convertFociExp e = 
  match e with
	FociExpression.Constant (FociConstant.Int i) -> Expression.Constant (Constant.Int (int_map_from i))
    | FociExpression.Variable v -> 
	begin
	  match (Hashtbl.mem chlval_table v) with
	      true -> Hashtbl.find chlval_table v 
	    | false -> 
                if Misc.is_prefix "_STRINGCONSTANT" v then
                  let c = Misc.chop_after_prefix "_STRINGCONSTANT" v in 
                  let s = get_const_string c in
                  Expression.Constant (Constant.String s)
                else 
                    Expression.Lval (Expression.Symbol v)
	end
    | FociExpression.Application (f',el) ->
	begin
          let f = clean_foci_UF_string f' in
          match f with
	      "chlval" -> 
		begin
		  try
		    let [l;(FociExpression.Variable s)] = el in
		      Expression.Chlval (convertFociLval l, Misc.chop_after_prefix "_" s)
		  with _ -> (* which should be a match failure *)
		    failwith ("unknown chlval args in convertFociExp! "^(foci_expression_toString e))
		end
	      | "*" -> 
		  begin
		    try
		      make_bin Expression.Mul (List.hd el) (List.nth el 2)
		    with _ ->
		      failwith ("unknown * args in convertFociExp! "^(foci_expression_toString e))
		  end
		  (* Hopefully we wont have to deal with wacko spl cases here! *)
	      | "addrOf" ->
		  begin
		    try 
		      let [e1] = el in
			Expression.Unary(Expression.Address, convertFociExp e1) 
		    with _ ->
		      failwith ("unknown & args in convertFociExp! "^(foci_expression_toString e))
		  end
	      | "select" ->
		  begin
		    try 
		      let [e1;e2] = el in
			begin
			  match e2 with
			      FociExpression.Variable s ->
				begin
				  if (Misc.is_prefix "@field_" s)
				  then
				    Expression.Lval (Expression.Access(Expression.Dot, convertFociExp e1, (Misc.chop_after_prefix "@field_" s)))
				  else
				    Expression.Lval (Expression.Indexed(convertFociExp e1,convertFociExp e2))
				end

			    | _ ->
				if (e2 = FociAst.zero) then
				  Expression.Lval (
				  Expression.Dereference(convertFociExp e1))
				else
				  Expression.Lval (
				    Expression.Indexed(convertFociExp e1,convertFociExp e2))
			end
		    with _ -> 
		      failwith ("unknown select args in convertFociExp! "^(foci_expression_toString e))
				  
		  end
	      | "+" ->
		  begin
		    Message.msg_string Message.Normal "foci uses + in app";
		    gatherFociSum el
		  end
	      | "UIplus" ->
		  begin
		    Message.msg_string Message.Normal "foci uses UIplus in app";
		    gatherFociSum el
		  end
	      | "UIminus" ->
		  if (List.length el <> 2) then failwith "Bad use of UIminus";
		  let [e1;e2] = List.map convertFociExp el in
		    Expression.Binary(Expression.Minus,e1,e2)
              | "foffset" ->
                  begin
                    match el with
                      [e1;(FociExpression.Variable s)] -> 
                        let fld = if Misc.is_prefix "@field_" s then Misc.chop_after_prefix "@field_" s else failwith "bad foffset!" in
                        Expression.Binary (Expression.FieldOffset, (convertFociExp e1), (Expression.Lval (Expression.Symbol fld)))
                    | _ -> failwith "bad number of args to foffset!"
                  end
	      | _ -> 
		  begin
		    if (Misc.is_prefix "select@field_" f )
		    then
		      (* a field "access" thing *)
		      let field_name = Misc.chop_after_prefix "select@field_" f in
			try
			  let [e] = el in
			    Expression.Lval (Expression.Access(Expression.Dot, convertFociExp e, field_name))
			with _ -> failwith "multiple args to select@field_ ! ?"
                    else
                      try 
                        let op = Expression.binaryOpFromVanillaString f in
                        let [e1;e2] = el in
                        Expression.Binary (op, convertFociExp e1,convertFociExp e2)
                      with _ ->
			Expression.FunctionCall(Expression.Lval (Expression.Symbol f), List.map convertFociExp el)
		  end
	end

      | FociExpression.Sum el -> gatherFociSum el

      | FociExpression.Coeff ((FociConstant.Int i),e') ->
	  if (i < 0) then 
	    if (i = (-1)) then
	      Expression.Unary (Expression.UnaryMinus, (convertFociExp e'))
	    else 
	      Expression.Binary(Expression.Mul, Expression.Constant (Constant.Int (int_map_from i)),
						   convertFociExp e')
	  else
	    if (i = 1) then convertFociExp e'
	    else
	      (Expression.Binary(Expression.Mul, Expression.Constant (Constant.Int (int_map_from i)), convertFociExp e'))
	      

and convertFociPred p =
  match p with
      FociPredicate.True -> Predicate.True
    | FociPredicate.False -> Predicate.False
    | FociPredicate.And plist -> Predicate.And (List.map convertFociPred plist)
    | FociPredicate.Or plist -> Predicate.Or (List.map convertFociPred plist)
    | FociPredicate.Not p' -> Predicate.Not (convertFociPred p')
    | FociPredicate.Implies (p1,p2) -> Predicate.Implies (convertFociPred p1,convertFociPred p2)
    | FociPredicate.Equality (e1,e2) -> Predicate.Atom (make_eq e1 e2)
    | FociPredicate.Leq (e1,e2) -> Predicate.Atom (make_leq e1 e2)
    | FociPredicate.Atom s -> Predicate.Atom (Expression.Lval (Expression.Symbol s))
	(* jhala: actually I'm being lazy... whats wrong with atomic predicates! *)


(* now the somewhat non-trivial bit *)
(* cons_pre and cons_post are the two sets of constraints. This function returns the predicates
appearing in the fwd interpolant  *)


let check_interpolant 
  (itp : Predicate.predicate) 
  (phi_1: Predicate.predicate) (phi_2 : Predicate.predicate) =  
  (* for now, only check that every symbol occurring in itp is in phi_1, and in phi_2 *)
  let syms_phi_1 = Predicate.allVarExps phi_1 in
  let syms_phi_2 = Predicate.allVarExps phi_2 in
  let syms_itp = Predicate.allVarExps itp in
  let rv = 
    List.filter (fun x -> not ((List.mem x syms_phi_1) && (List.mem x syms_phi_2))) syms_itp
  in
    if rv <> [] then
      begin
	let s = Printf.sprintf "Interpolant issue: %s \n phi1: %s \n phi2: %s \n" 
		  (Predicate.toString itp) (Predicate.toString phi_1) (Predicate.toString phi_1)
	in
	  Message.msg_string Message.Normal s;
	  Message.msg_string Message.Error s;
	  List.iter (fun e -> Message.msg_string Message.Normal (Expression.toString e)) rv
      end
    
      
(* itp_process_function: Predicate.predicate -> 'a *)

let extract_and_process_interpolant itp_process_function (cons_pre,cons_post) = 
  let _ = Hashtbl.clear chlval_table in (* reset the string -> chlval map *)
  let _ = Hashtbl.clear chlval_map_table in
  let _ = Message.msg_string Message.Debug "In extract_and_process_interpolant"
  in
  let cons_l = List.map (fun x -> Predicate.And x) [cons_pre;cons_post] in
  
  let _ = List.iter (fun x -> Message.msg_printer Message.Debug Predicate.print x) cons_l 
  in
  let [foci_c_pre; foci_c_post] = List.map convertToFociSyntax cons_l in
    try
      let _ = 
	Message.msg_string Message.Debug ">>>>>>>>>>>> calling foci";
	Message.msg_string Message.Debug (foci_predicate_toString foci_c_pre);
	Message.msg_string Message.Debug ";";
	Message.msg_string Message.Debug (foci_predicate_toString foci_c_post);
	Message.msg_string Message.Debug ">>>>>>>>>>>> foci done" 
      in
      let foci_file_string = Printf.sprintf "\n \n \n %s \n ; \n  %s \n \n \n" (foci_predicate_toString foci_c_pre) (foci_predicate_toString foci_c_post)
          in
	if (false) then
          Misc.append_to_file "fociqs" foci_file_string ;
 let fwd_ip = 
      try 
        foci_interpolate (foci_c_pre,foci_c_post) 
      with e -> 
	(Message.msg_string Message.Error ("Foci FAILS!(eap_itp) raises exception!!"^ (Printexc.to_string e));
	 Message.msg_string Message.Normal ("Foci FAILS!(eap_itp) raises exception!!"^ (Printexc.to_string e));
	 Misc.append_to_file "foci_failed_q" foci_file_string;
	FociPredicate.True (* FociPredicate.And [foci_c_pre;foci_c_post ] *))
    in
      let _ = Message.msg_string Message.Debug ("Foci yields Interpolant: "^(foci_predicate_toString fwd_ip))
      in
      let fwd_ip_b = convertFociPred fwd_ip in
      let _ = Message.msg_string Message.Debug ("Foci yields b_Interpolant: "^(Predicate.toString fwd_ip_b))
      in
	(* check if the interpolant is clean *)
      let _ = 
	match cons_l with
	    [c1;c2] -> check_interpolant fwd_ip_b c1 c2 
	  | _ -> failwith "this cannot happen"
      in
	itp_process_function fwd_ip_b
    with
	FociSatException -> (* should be that the thing was SAT *)
	  failwith "Foci raised SAT!"
      | e -> raise e(*failwith "Foci raises some other exception..."*)



let extract_foci_preds = extract_and_process_interpolant Predicate.getAtoms

let extract_foci_preds_parity = 
  extract_and_process_interpolant (Predicate.getAtoms_parity true)

let extract_foci_interpolant = extract_and_process_interpolant (fun x -> x)



(* cut and paste from the above function -- but here the argumentes are not lists!*)
let interpolant_sat (cons_pre,cons_post) =
let _ = Hashtbl.clear chlval_table in (* reset the string -> chlval map *)
  let _ = Hashtbl.clear chlval_map_table in
  let cons_l = [cons_pre;cons_post] in
  let _ = Message.msg_string Message.Debug "In interpolant_sat:" in
  let _ = List.iter (fun x -> Message.msg_string Message.Debug (Predicate.toString x)) cons_l 
  in
  let [foci_c_pre; foci_c_post] = List.map convertToFociSyntax cons_l in
  let foci_file_string = Printf.sprintf "\n \n \n %s \n ; \n  %s \n \n \n" (foci_predicate_toString foci_c_pre) (foci_predicate_toString foci_c_post)
  in
    try  
      let fwd_ip = foci_interpolate (foci_c_pre,foci_c_post) in
      let _ = Message.msg_string Message.Debug ("Foci yields Interpolant: "^(foci_predicate_toString fwd_ip))
      in
      let fwd_ip_b = convertFociPred fwd_ip in
      let _ = Message.msg_string Message.Debug ("Foci yields b_Interpolant: "^(Predicate.toString fwd_ip_b))
      in
	Unsat(fwd_ip_b)
    with 
	FociSatException -> (Message.msg_string Message.Debug "Foci says SAT!"; Sat)
      | FociSatExceptionAsgn _ -> (Message.msg_string Message.Debug "Foci says SATa!"; Sat)
      | e ->
	  begin
	    Message.msg_string Message.Error ("Foci FAILS!(isat) raises exception!!"^ (Printexc.to_string e));
	    Message.msg_string Message.Error ("Foci FAILS!(isat) raises exception!!"^ (Printexc.to_string e));
	    Misc.append_to_file "foci_failed_q" foci_file_string;
	  raise e
 end
	      	  
let interpolant_implies a b = 
  match interpolant_sat (a, (Predicate.negate b)) with
      Sat -> false
    | Unsat _ -> true
  
    
let foci_plist_toString pl = String.concat "\n ; \n" (List.map foci_predicate_toString pl) 

(* may raise "FociSatException" *)
let big_fq_count_ref = ref 0
let fq_dump num_cons cons_l_string =
  let foci_file_string = Printf.sprintf "\n \n \n %s \n \n \n" (cons_l_string) in
  let _file = 
    if num_cons > 2 then 
      (big_fq_count_ref:= !big_fq_count_ref + 1;
      "fociq."^(string_of_int !big_fq_count_ref))
    else "fociqs"
  in
  let file = ((Options.getValueOfString "mainsourcename")^"."^_file) in
  let _ = Message.msg_string Message.Debug ("writing to: "^file) in
  Misc.write_to_file file foci_file_string
    

let extract_and_process_interpolantN itp_fun foci_syn_conv itp_process_function cons_l : Predicate.predicate list = 
  let _ = Hashtbl.clear chlval_table in (* reset the string -> chlval map *)
  let _ = Hashtbl.clear chlval_map_table in
  let _ = Message.msg_string Message.Debug "In extract_and_process_interpolantN:" in
  let _ = List.iter (fun x -> Message.msg_printer Message.Debug Predicate.print x) cons_l 
  in
  let n_cons = List.length cons_l in
  let foci_cons_l = List.map foci_syn_conv cons_l in
  let cons_l_string = foci_plist_toString foci_cons_l in
  let _ =
    Message.msg_string Message.Debug ">>>>>>>>>>>> calling foci";
    Message.msg_string Message.Debug cons_l_string ;
    Message.msg_string Message.Debug ">>>>>>>>>>>> foci done" 
  in
  let _ = if (Options.getValueOfBool "fqdump") then fq_dump n_cons cons_l_string in
  let itp_l = 
    try 
      let rv = (* foci_interpolateN *) itp_fun foci_cons_l in
      Message.msg_string Message.Debug "FOCI returns";
      rv
    with
    FociSatException -> (Message.msg_string Message.Normal "Foci says SAT!"; raise FociSatException)             
  | FociSatExceptionAsgn a -> (Message.msg_string Message.Normal "Foci says SATa!"; raise (FociSatExceptionAsgn a))
         | e -> 
	(Message.msg_string Message.Normal ("Foci FAILS!(eap_itpN) raises exception!!"^ (Printexc.to_string e));
	 Message.msg_string Message.Error ("Foci FAILS!(eap_itpN) raises exception!!"^ (Printexc.to_string e));
	 let foci_file_string = Printf.sprintf "\n \n \n %s \n \n \n" (cons_l_string) in
         Misc.append_to_file "foci_failed_q" foci_file_string;
         raise FociSatException)
    in
    let itp_l_string = foci_plist_toString itp_l in 
    let _ = Message.msg_string Message.Debug ("Foci yields Interpolants: \n"^itp_l_string) in
    let itp_b_l = List.map convertFociPred itp_l in
    let itp_b_l_string = String.concat "\n ; \n" (List.map Predicate.toString itp_b_l) in
    let _ = Message.msg_string Message.Debug ("Foci yields b_Interpolant: "^itp_b_l_string) in
	List.map itp_process_function itp_b_l


let conv_fun p = 
  if (Options.getValueOfBool "fmc") then 
    (foci_UF_idx_tick (); convertToFociSyntax p) 
  else (convertToFociSyntax p)
  
let extract_foci_interpolantN = 
  extract_and_process_interpolantN foci_interpolateN conv_fun  (fun x -> x)

let extract_foci_interpolantN_noconv =
  extract_and_process_interpolantN foci_interpolateN convertToFociSyntax (fun x -> x)
  
let extract_foci_interpolantN_symm arg =
 set_disjoint_flag ();
 try 
   let rv =  extract_and_process_interpolantN foci_interpolateN_symm conv_fun (fun x -> x) arg in
   unset_disjoint_flag ();
   (* let rv_tc =
   let tc = foci_theory_clauses () in
   List.map (List.map convertFociPred) tc
 in *)
 (rv,[])
  with e -> unset_disjoint_flag (); raise e
  

let extract_foci_interpolantN_dasdill trel_cons_l cube_cons_l =
  assert (List.length trel_cons_l + 1 = List.length cube_cons_l);
  let _ = Hashtbl.clear chlval_table in (* reset the string -> chlval map *)
  let _ = Hashtbl.clear chlval_map_table in
  let _ = Message.msg_string Message.Debug "In extract_and_process_interpolantN_dasdill:" in
  let _ = List.iter (fun x -> Message.msg_printer Message.Debug Predicate.print x) trel_cons_l in
  let _ = List.iter (fun x -> Message.msg_printer Message.Debug Predicate.print x) cube_cons_l in
  let n_trel_cons = List.length trel_cons_l in
  let foci_trel_cons_l = List.map conv_fun trel_cons_l in
  let foci_cube_cons_l = List.map conv_fun cube_cons_l in
  let trel_cons_l_string = foci_plist_toString foci_trel_cons_l in
  let cube_cons_l_string = foci_plist_toString foci_cube_cons_l in
  let _ =
    Message.msg_string Message.Debug ">>>>>>>>>>>> calling foci:trel";
    Message.msg_string Message.Debug trel_cons_l_string ;
    Message.msg_string Message.Debug ">>>>>>>>>>>>cube>";
    Message.msg_string Message.Debug cube_cons_l_string ;
    Message.msg_string Message.Debug ">>>>>>>>>>>> done"
  in
  let itp_l = 
    try 
      let rv = foci_interpolateN_dasdill foci_trel_cons_l foci_cube_cons_l in
      Message.msg_string Message.Debug "FOCI DasDill returns";
      rv
    with
    FociSatException -> (Message.msg_string Message.Normal "Foci says SAT!"; raise FociSatException)             
  | FociSatExceptionAsgn a -> (Message.msg_string Message.Normal "Foci says SATa!"; raise (FociSatExceptionAsgn a))
  | e -> 
	(Message.msg_string Message.Normal ("Foci FAILS!(eap_itpN_dd) raises exception!!"^ (Printexc.to_string e));
	 Message.msg_string Message.Error ("Foci FAILS!(eap_itpN_dd) raises exception!!"^ (Printexc.to_string e));
         raise FociSatException)
    in
    let itp_l_string = foci_plist_toString itp_l in 
    let _ = Message.msg_string Message.Debug ("FociDD yields Interpolants: (Negate if not TRUE) \n"^itp_l_string) in
    let itp_b_l = List.map convertFociPred itp_l in
    let itp_b_l_string = String.concat "\n ; \n" (List.map Predicate.toString itp_b_l) in
    let _ = Message.msg_string Message.Debug ("Foci yields b_Interpolant: "^itp_b_l_string) in
       itp_b_l


(* incremental foci interface -- using foci just as a DP 
0. reset: unit -> unit
1. push: pred-> unit 
2. pop: unit -> unit
3. check_sat: unit -> bool
*)


(* foci 3 stuff *)
let foci_instance_ref = 
  let inst = Foci2.new_instance 1 false true in
  ref (inst,0)
      

let foci_inc_pop () = 
  let (inst,ht) = !foci_instance_ref in 
  if ht > 0 then 
    begin
      Message.msg_string Message.Debug "fociinc:pop";
    Foci2.pop inst;
    foci_instance_ref := (inst, ht-1)
    end
  
let foci_inc_push p =
  let (foci_inst,ht) = !foci_instance_ref in
  let p' = FociDag.pred_dag_of_tree (convertToFociSyntax p) in
  Foci2.push foci_inst;
  Foci2.add foci_inst 0 p';
  Message.msg_string Message.Debug "fociinc:push";
  foci_instance_ref := (foci_inst,ht+1)

let foci_inc_check_sat () =
  let (inst,ht) = !foci_instance_ref in 
  try ignore(Foci2.solve inst); 
  Message.msg_string Message.Debug ("fociinc: ht="^(string_of_int ht)^".");
  false 
  with Foci2.SAT (_,_) -> 
    true

  
let foci_inc_reset () =
  let inst = Foci2.new_instance 1 false true in
  foci_instance_ref := (inst,0)
(*  let (_,ht) = !foci_instance_ref in
  Misc.repeat_fn foci_inc_pop ht *)  


 
(******************************  End Interface to FOCI  ******************************)
  


(************************************************************************************)

let generateTest p =
	failwith "Test generation not supported"
(*
  (* Data structures used *)
(* let ac_generateTest p = *)
  let id_to_sym_ht = Hashtbl.create 31 in
  let sym_to_id_ht = Hashtbl.create 31 in
  let number_of_vars = ref 0 in
  let get_next_id () = incr number_of_vars ; !number_of_vars
  in 
  (* Helper functions *)
  let find_variables ap =
    match ap with 
      Predicate.Atom aex | Predicate.Not (Predicate.Atom aex) ->
	let rec _trawl_exp aex =
	  match aex with
	    Expression.Lval _ (* add to hash table *)
	  | Expression.Chlval _ -> (* add to hash table *)
	      if Hashtbl.mem sym_to_id_ht aex then
		()
	      else 
		begin
		  let n = get_next_id () in
		  Hashtbl.add sym_to_id_ht aex n ;
		  Hashtbl.add id_to_sym_ht n aex ;
		  ()
		end
	  | Expression.Binary (Expression.FieldOffset, e1, e2) -> failwith "Do not handle field offset"
	  | Expression.Binary (_, e1, e2) ->
	      _trawl_exp e1 ; _trawl_exp e2
	  | Expression.Constant _ -> ()
	  | Expression.Unary (_, e) -> _trawl_exp e
	in
	_trawl_exp aex
    | _ -> (* cannot happen *) failwith "In find_variables: atomic predicate expected"
  in
  let get_lp_constraints aex =
    match aex with
      Expression.Binary (relop, e1, e2) ->
	begin
	  let (coeffs : int array) = Array.make (!number_of_vars + 1) 0 in (* coefficients for variables *)
	  let const = ref 0 in
	  let rec get_coeffs e pos_or_neg =
	    match e with 
	      Expression.Constant c ->
		begin
		  match c with Constant.Int i -> const := !const + (if pos_or_neg then i else -i) 
		  | _ -> failwith "add_lp_constraints: Only handle integers"
		end
	    | Expression.Lval _
	    | Expression.Chlval _ -> 
		let id = Hashtbl.find sym_to_id_ht e in
		coeffs.(id) <- coeffs.(id) + (if pos_or_neg then 1 else -1)
	    | Expression.Binary (Expression.Plus, e1, e2) ->
		get_coeffs e1 pos_or_neg ; get_coeffs e2 pos_or_neg
	    | Expression.Binary (Expression.Minus, e1, e2) ->
		get_coeffs e1 pos_or_neg ; get_coeffs e2 (not pos_or_neg)
	    | Expression.Binary (Expression.Minus, e1, e2) ->
		get_coeffs e1 pos_or_neg ; get_coeffs e2 (not pos_or_neg)
	    | Expression.Binary (Expression.Mul, Expression.Constant (Constant.Int i), Expression.Lval l)
	    | Expression.Binary (Expression.Mul, Expression.Lval l, Expression.Constant (Constant.Int i)) ->
		let id = Hashtbl.find sym_to_id_ht (Expression.Lval l) in
		coeffs.(id) <- coeffs.(id) + (if pos_or_neg then i else -i)
	    | Expression.Binary (Expression.Mul, Expression.Constant (Constant.Int i), Expression.Chlval (l,s))
	    | Expression.Binary (Expression.Mul, Expression.Chlval(l,s), Expression.Constant (Constant.Int i)) ->
		let id = Hashtbl.find sym_to_id_ht (Expression.Chlval (l,s)) in
		coeffs.(id) <- coeffs.(id) + (if pos_or_neg then i else -i)
	    | Expression.Binary (Expression.Mul, Expression.Constant (Constant.Int i), Expression.Constant (Constant.Int j)) ->
		let ij = i * j in const := !const + (if pos_or_neg then ij else -ij) 
	    | Expression.Unary (Expression.UnaryPlus, e1) -> get_coeffs e1 pos_or_neg
	    | Expression.Unary (Expression.UnaryMinus, e1) -> get_coeffs e1 pos_or_neg
	    | _ -> failwith ("get_coeffs: failing on "^(Expression.toString e))
	  in
	  get_coeffs e1 true ;
	  get_coeffs e2 false ;
	  (coeffs, !const, relop) 
	end
    | _ -> failwith ("get_lp_constraints: Strange expression "^(Expression.toString aex))
  in

  (* main function *)
  match p with
    Predicate.And l ->
      begin
	if (List.exists 
	     (fun ap -> match ap with 
		Predicate.Atom _ | Predicate.Not (Predicate.Atom _) -> false | _ -> true) l) then
	  failwith "generateTest: Assumes conjunction of atomic predicates"
	else 
	  begin (* Now start the work *)
	    (* first do a trawl and find out the number of variables. In the process, also
	       populate the two hash tables name -> id, and id -> name
	       *)
	    List.iter find_variables l ;
	    (* then for each atomic formula, generate the corresponding LP constraint *)
            let get_aexp ap =
              match ap with Predicate.Atom aex -> aex 
              | Predicate.Not (Predicate.Atom aex) -> Expression.negateRel aex
              | _ -> failwith "Non atomic!"
            in
	    let constraints = List.map (fun ap -> get_lp_constraints (get_aexp ap)) l in 
	    (* first cut : assume no Neq *)
	    let linprog = Lp.lp_make_lp 0 !number_of_vars in
	    let add_constraints (coeffs, const, relop) =
	      match relop with
	      |	Expression.Le -> Lp.lp_add_constraint linprog (Array.map float_of_int coeffs) 0 (float_of_int (-const)) 
	      |	Expression.Ge -> Lp.lp_add_constraint linprog (Array.map float_of_int coeffs) 2 (float_of_int (-const)) 
	      |	Expression.Eq -> Lp.lp_add_constraint linprog (Array.map float_of_int coeffs) 1 (float_of_int (-const)) 
	      |	Expression.Gt -> Lp.lp_add_constraint linprog (Array.map float_of_int coeffs) 2 (float_of_int (-const+ 1)) 
	      |	Expression.Lt -> Lp.lp_add_constraint linprog (Array.map float_of_int coeffs) 0 (float_of_int (-const- 1)) 
	      |	Expression.Ne -> if (Random.float 1.0< 0.5) then Lp.lp_add_constraint linprog (Array.map float_of_int coeffs) 2 (float_of_int (-const+ 1)) 
				else Lp.lp_add_constraint linprog (Array.map float_of_int coeffs) 0 (float_of_int (-const- 1)) 
	      |	_ -> failwith "add_con: not handled"
	    in
	    List.iter (fun c -> ignore (add_constraints c)) constraints ;
	    (* for each variable, set lower bound to -infinity *)
	    Lp.lp_set_obj_fn linprog (Array.make (!number_of_vars + 1) 0.0) ;
	    (* if (Options.getValueOfBool "quiet") then () else Lp.lp_print_lp linprog ; *)
            let stillnotdone = ref true in

            let set_minus_inf = ref false in
            while !stillnotdone do
	    begin 
              stillnotdone := false ;
              Message.msg_string Message.Normal "Next round of solving Linprog" ;
	      (* now solve linprog *)
	      let retcode = (Lp.lp_solve linprog) in
	      if retcode = 2 then (* infeasible! *) begin
                (if !set_minus_inf then failwith "LP is infeasible!"
                else set_minus_inf := true );
	        let (minus_infinity : float) =  -. (Lp.lp_get_infinite (linprog) ) in
	        for i = 1 to !number_of_vars do ignore (Lp.lp_set_lowbo linprog i minus_infinity) done ; 
	        stillnotdone := true
                end
              else 
              begin (* check that the solution is integral *) 
(*
	      if (Options.getValueOfBool "quiet") then () else Lp.lp_print_lp linprog ;
	      if (Options.getValueOfBool "quiet") then () else Lp.lp_print_solution linprog ;
*)
              let sol_array = Lp.lp_get_solution linprog in
              for i = 1 to !number_of_vars do
                if (floor sol_array.(i)  <> sol_array.(i) ) then
                begin
	          Message.msg_string Message.Normal "Solution not integral!";
                  stillnotdone := true;
                  ignore (Lp.lp_set_int linprog i 1 ) ; 
                end ; 
              done
              end
            end done ;
	    (* ok now we have the map from lvalues to values *)
	    let sol = Lp.lp_get_solution linprog in
	    let make_list ht sol =
              let l = ref [] in
              for i = 1 to !number_of_vars do
                l := (Hashtbl.find ht i, sol.(i)) :: !l ;
              done ;
              !l
            in
	    make_list id_to_sym_ht sol
	  end
      end
  | _ -> failwith "generateTest: Assumes conjunction of atomic predicates (not a conjunction)"



*)
