(*
 * Description:
 *
 * Copyright (c) 2000 by
 *  David Blei			blei@cs.berkeley.edu
 *  Chris Harrelson		chrishtr@cs.berkeley.edu
 *  Ranjit Jhala		jhala@cs.berkeley.edu
 *  Rupak Majumdar		rupak@cs.berkeley.edu
 *  George C. Necula	necula@cs.berkeley.edu
 *  Shree P. Rahul		sprahul@cs.berkeley.edu
 * 	Wes Weimer			weimer@cs.berkeley.edu
 *  Dror Weitz			dror@cs.berkeley.edu
 *   
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software for research purposes only is hereby granted, 
 * provided that the following conditions are met: 
 * 1.  Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 3. The name of the authors may not be used to endorse or promote products derived from 
  *  this software without specific prior written permission. 
 *
 * DISCLAIMER:
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)
module P = Proof
module I = Input
module E = VampyreErrormsg
    
open VampyrePretty
open UnionFind

module UF = UnionFind

exception Contradiction of P.proof
    
type    node = 
    {         id   : int;               (* Node identifier. Unique. *)
              op   : operator;          (* The operator *)
              args : node list;         (* The arguments (children) *)
      mutable root : node;               (* The root of the class *)
              
      mutable eqClass : int list;
      mutable use : int list;            (* The list of its "parents" *)
      mutable ac_use : int list;         (* The "actual" list of its parents *)
      mutable isAsserted : bool;      (* Whether or not its been asserted *) 
      mutable mesg : proofInfo;     (* if the above is true then this has the info about who made the assertion ... *)
    } 
   
    (*[Jhala]: set of all the people I am equated with node list is too extravagant in either case ... just an int list should suffice. corresp the same with the root datatype ... *)
      (* extend here with what you need to support your decision procedures. 
       * Do not add too many things, though. Prefer to keep additional data 
       * structures in your decision procedures. It is Ok to add the 
       * stuff for the congruence closure here because that will always be 
       * part of the prover. *)
    

and operator = OpName of string 
      (* [Jhala] one could also store the arity of the operator over here, but that is implicitly stored as the length of the list of vars *)

  | Int of int 
  | Var of Input.var                     (* An integer constant *)
  | Plus
  | Minus
  | Times

  | Div of int
  | Mod of int
  | Floor of int
  | Ceil of int

  | Fun of string 

                                        (* Some boolean nodes *)
  | BFun of string
  | NotBFun of string 
  | Eq
  | Neq
  | Geq
  | Gt
  | Leq
  | Lt

type nodeExpType = DAtom of I.atom | DExp of I.exp
exception Jhala

let isExp n =
    match n.op with
      (BFun(_)
    | NotBFun(_) 
    | Eq
    | Neq
    | Geq
    | Gt
    | Leq
    | Lt ) -> false
    | _ -> true
	  

let rec dummyNode = {id = -10;op = OpName "";args =[];root = dummyNode; eqClass =[]; use = []; ac_use = []; isAsserted = false; mesg = dummyMesg}

let dummyMesg = dummyMesg    
     
let sillyProof =  P.G("true",[])         

let debug = ref (ref false) (* set to point to Equality.debug *)


     
(* 
 * [weimer]
 * newId gives a new (ever-increasing) integer each time it is called.
 * Used to make unique node ids. newNode returns a fresh node in a new
 * equivalence class with with some default values.
 *)

(* [Jhala] My snapshot should put onto a stack a tuple of (idCounter,dagTable) -- undo merely pops this stack ... and uses whatever was popped to reset the globals idCounter and dagTable *)	  
let snapMark = (-1,-1)
    
(* make this size larger ...*)    
let tableInitSize = 10009
let tableInitSize1 = 1009
let tableInitSize2 = 113

let edgeSetSize = 11
    
let idCounter = ref 0 (* reinit *)
let discoveryList = ref [] (* reinit *)
    
let dagTable = Hashtbl.create tableInitSize 
let dagTableIndex = Hashtbl.create tableInitSize
let ufTable = Hashtbl.create tableInitSize
let disEqTable = Hashtbl.create tableInitSize2
let _ = Hashtbl.add disEqTable (-1,-1) dummyMesg
  let setTable = Hashtbl.create tableInitSize

    
let undoStack = ref [snapMark]
let disEqStack = ref [snapMark]
let assertStack = ref [snapMark]
    
let newId () =
	idCounter := !idCounter + 1 ;
	!idCounter
    
let dagSize () = !idCounter
    
    
let dagReset () =
  idCounter := 0;
  discoveryList := [];
  Hashtbl.clear dagTable;
  Hashtbl.clear dagTableIndex;
  Hashtbl.clear ufTable;
  Hashtbl.clear disEqTable; 
  Hashtbl.clear setTable;
  undoStack := [];
  disEqStack := [];
  assertStack := []

let idf x = List.map (function p -> p.id) x
(*
 *  [dror]
 *  function to return an existing node by its id
 *)
let findDag i =
  (* if (!(!debug)) then (
    print_string "CC: in findDag with ";print_int i; print_string " \n");*)
  try
    !(Hashtbl.find dagTable i)
  with
    Not_found -> E.s (E.bug "CC: invalid query to findDag!! ");print_int i;print_string "...\n"; dummyNode
let rec print_list l =
  match l with
    [] -> ()
  | h::tl -> print_int h;print_string "; ";print_list tl

let temp = ref [] 	(* reinit *)
let showEqClass () =
  Hashtbl.iter (function x -> function n -> (UF.print_path n;print_string "\n")) ufTable ;
  Hashtbl.iter (function x -> function n -> print_int x;print_string "...";print_list ((findDag x).eqClass);print_string "\n") ufTable;
  temp := [];
  Hashtbl.iter (function x -> function n -> (temp := (n.UF.idt,n.UF.proof,n.UF.sP)::(!temp))) ufTable;(!temp)

	      
let nodeEqClass n = List.map findDag (!(Hashtbl.find dagTable (UF.find (Hashtbl.find ufTable n.id))).eqClass)
   
(* [JHALA] theres more to do ... you need also to update the use of the canonical node of the eq class corresp to snode ... - probably need to update all along the path ...*)    
let updateUse parent snode =
  (snode.use <- parent::snode.use);
  (snode.ac_use <- parent::snode.ac_use);
  let sn_root = Hashtbl.find dagTable (UF.find (Hashtbl.find ufTable snode.id))in
  (!sn_root).use <- parent::(!sn_root).use 
  

let rec list_delete x l = match l with
  [] -> []
| h::t -> if h=x then (list_delete x t) else (h::(list_delete x t))
    
let intersect2 l1 l2 =
  List.flatten (List.map (function x -> if (List.mem x l2) then [x] else []) l1)
let rec intersectl l =
  match l with
    [e] -> e
  | [e1;e2] -> intersect2 e1 e2
  | e1::e2::e3::tl -> intersect2 (intersect2 e1 e2) (intersectl (e3::tl))  
  | [] -> []


let knownEq x1 y1 =
  let x1n = UF.find (Hashtbl.find ufTable x1.id) in
  let y1n = UF.find (Hashtbl.find ufTable y1.id) in
  x1n=y1n

let congruent x y =
  let nx = Hashtbl.find dagTable x in
  let ny = Hashtbl.find dagTable y in
  try 
    if ((!nx).op = (!ny).op) & (List.for_all2 knownEq (!nx).args (!ny).args) then true else false
  with _ -> false
	
let findPossibleLink n =
  let useList = list_delete n.id (intersectl (List.map (function x -> !(Hashtbl.find dagTable (UF.find (Hashtbl.find ufTable x))).use)  (idf n.args))) in

  let _ = if (!(!debug)) & (n.id = 37) then (print_string " candidates:";print_list useList) in 
  List.find  (congruent n.id) useList
    
let cross l1 l2 =
  let makePair n1 n2 =
    let _ =
      if (n1,n2) = (32,49) then
	print_string "pairing ... 32,49"
      else ()
    in
    (n1,n2)
  in
  List.flatten (List.map (function x -> (List.map (makePair x) l2)) l1)  
    
let rec merge (t1,t2) info =
  let useMerge slist x =
    let useMerge1 x1 y1 =
      if ((not (UF.find (Hashtbl.find ufTable x1) = UF.find (Hashtbl.find ufTable y1))) & (congruent x1 y1)) then (merge (x1,y1) dummyMesg)
    in
    List.iter (useMerge1 x) slist
  in
  let ufn1 = (Hashtbl.find ufTable t1) in
  let ufn2 = (Hashtbl.find ufTable t2) in 
  let (ufun,other) = UF.tr_union ufn1 ufn2 info in
  (*let _ = print_string "linking...";print_int ufn1.idt;print_string "
and ";print_int ufn2.idt;print_string " \n" in *)
  (* fraudy engine bug ... [Jhala] *)
  let _ = (undoStack := (ufun.idt,other.idt)::(!undoStack)) in
  let nu = !(Hashtbl.find dagTable (ufun.idt)) in
  let nother = !(Hashtbl.find dagTable (other.idt)) in
  let _ =
    try 
      if (ufun.idt != other.idt) then (List.iter (useMerge nother.use)) nu.use
    with Stack_overflow -> E.s (E.bug "in here")
  in
  let _ = nu.use <- (nu.use @ nother.use) in
  let _ = (discoveryList := (cross nu.eqClass nother.eqClass)@(!discoveryList)) in
  let _ = nu.eqClass <-nu.eqClass @ nother.eqClass in
  (*print_string "Exit Merge ...";*)
  ()
   

    
let newNode op args =
  
  let argIdList = idf args in
  if (Hashtbl.mem dagTableIndex (op, argIdList)) then  !(Hashtbl.find dagTable (Hashtbl.find dagTableIndex (op,argIdList))) else (
  let newNum = newId () in
  let rec node = { id = newNum ; op = op ; args = args ; root = node ; use = []; ac_use = []; eqClass = [newNum];isAsserted = false;mesg = dummyMesg} in
  let foo = List.map (updateUse newNum) args (* the use structure gets added to everytime a new parent shows up *) 
  in
  let _ = Hashtbl.add dagTable newNum (ref node);
    Hashtbl.add ufTable newNum (UF.newNode newNum);
    Hashtbl.add dagTableIndex (op, argIdList) newNum
  in

  let _ =
    try
      
      let linkTo = findPossibleLink node in
      merge (newNum,linkTo) dummyMesg
    with
      _ -> (*if (newNum = 37) then (
	print_string " 37: args"; print_list (idf args);
	print_string " 25: root ";print_int (UF.find (Hashtbl.find ufTable 25));print_string " 26 : use "; print_list !(Hashtbl.find dagTable 26).use;
print_string " 25 : use "; print_list !(Hashtbl.find dagTable 25).use;
	print_string "screwed again at 37!!");*) ()
  in
  
  let _ = List.map (function x-> if not(List.mem newNum x.use) then E.s (E.bug "parent bug")) args in
  if (newNum = 37)& false  then (
    if (List.mem 36 (!(Hashtbl.find dagTable 26)).use) then print_string "26-36 now !!! " else (E.s (E.bug "parent bug here"))
      );
  node )
    
    
    (* it should be clear that there was no previous binding in the table... *)

                                     (* Internalize an atomic predicate *)
(* 
 * [weimer] 
 * I put in some skeletal support for internAtom because I need it in order
 * to test general tacticals on abstract functions. The cogruence closure
 * person should probably change this later. 
 *)

   
let rec internAtom (p : I.atom) : node = 
  match p with
		I.Eq(a,b) -> newNode Eq [ internExp a ; internExp b ]
        |	I.Neq(a,b) -> newNode Neq [ internExp a ; internExp b ]
	|	I.Geq(a,b) -> newNode Geq [ internExp a ; internExp b ]
	|	I.Gt(a,b) -> newNode Gt [ internExp a ; internExp b ]
	|	I.Leq(a,b) -> newNode Leq [ internExp a ; internExp b ]
	|	I.Lt(a,b) -> newNode Lt [ internExp a ; internExp b ]
	|	I.BFun(str,expl) -> newNode (BFun(str)) (List.map internExp expl)
	|	I.NotBFun(str,expl) -> newNode (NotBFun(str)) (List.map internExp expl)      
  (* | _ -> E.s (E.unimp "dag.internAtom") *)

                                   (* Internalize an atomic expression *)
(* 
 * [weimer] 
 * I put in some skeletal support for internExp.  The cogruence closure
 * person should probably change this later. 
 *)
    
and internExp (p : I.exp) : node = 
	match p with
    I.Var(v) 			-> newNode (Var(v)) []
  | I.Int(i) 			-> newNode (Int(i)) [] 
  | I.Plus(a,b) 	-> newNode Plus [ internExp(a) ; internExp(b) ]
  | I.Minus(a,b) 	-> newNode Minus [ internExp(a) ; internExp(b) ]
  | I.Times(a,b) 	-> newNode Times [ internExp(a) ; internExp(b) ]
  | I.Div(i,e) 		-> newNode (Div(i)) [ internExp(e) ]
  | I.Mod(i,e) 		-> newNode (Mod(i))[ internExp(e) ]
  | I.Floor(i,e) 	-> newNode (Floor(i)) [ internExp(e) ]
  | I.Ceil(i,e) 	-> newNode (Ceil(i)) [ internExp(e) ]
  | I.Fun(s,expl) -> newNode (Fun(s)) (List.map internExp expl)
	(* | _ -> E.s (E.unimp "dag.internExp") *)

(* [Jhala]: for Dror - takes a node and returns the expression corresponding to it ... *)	
let nodeVal node = 
  let rec nodeExp node =
    match node.op with
      OpName(s) -> I.Var(s)
    | Var(v) -> I.Var(v) 
    | Int(i) -> I.Int(i) 
    | Plus   -> I.Plus(nodeExp (List.nth node.args 0),nodeExp (List.nth node.args 1)) 
    | Minus  -> I.Minus(nodeExp (List.nth node.args 0),nodeExp (List.nth node.args 1)) 
    | Times  -> I.Times(nodeExp (List.nth node.args 0),nodeExp (List.nth node.args 1)) 
    | Div(i) -> I.Div(i,nodeExp (List.nth node.args 0))
    | Mod(i) -> I.Mod(i,nodeExp (List.nth node.args 0))
    | Floor(i) -> I.Floor(i,nodeExp (List.nth node.args 0))
    | Ceil(i)  -> I.Ceil(i,nodeExp (List.nth node.args 0))
    | Fun(s) -> I.Fun(s,(List.map nodeExp node.args))
    | _ -> raise (Failure "bug in jhala's code #2")
  in
  let nodeAtom node =
    match node.op with
  | BFun(s) -> I.BFun(s,List.map nodeExp node.args)
  | NotBFun (s) -> I.NotBFun(s,List.map nodeExp node.args)
  | Eq -> I.Eq(nodeExp (List.nth node.args 0),nodeExp (List.nth node.args 1)) 
  | Neq -> I.Neq(nodeExp (List.nth node.args 0),nodeExp (List.nth node.args 1)) 
  | Geq -> I.Geq(nodeExp (List.nth node.args 0),nodeExp (List.nth node.args 1)) 
  | Gt -> I.Gt(nodeExp (List.nth node.args 0),nodeExp (List.nth node.args 1)) 
  | Leq ->I.Leq(nodeExp (List.nth node.args 0),nodeExp (List.nth node.args 1))
  | Lt -> I.Lt(nodeExp (List.nth node.args 0),nodeExp (List.nth node.args 1))
  | _ -> raise (Failure "bug in jhala's code")
  in
  match node.op with
    (OpName(_)
  |Int(_) 
  | Var(_)                     (* An integer constant *)
  | Plus
  | Minus
  | Times

  | Div(_)
  | Mod(_)
  | Floor(_)
  | Ceil(_)

  | Fun(_))-> DExp (nodeExp node)

                                        (* Some boolean nodes *)
  | (BFun(_)
  | NotBFun(_)     
  | Eq
  | Neq
  | Geq
  | Gt
  | Leq
  | Lt) -> DAtom(nodeAtom node)
	
  
(* [weimer] pretty print a Dag.operator *)
    
let dop = function
    OpName(f) -> dprintf "OpName(%s)" f
  | Int(i) -> dprintf "Int(%d)" i
  | Var(iv) -> dprintf "Var(%s)" iv
  | Plus -> text "Plus"
  | Minus -> text "Minus"
  | Times -> text "Times"
  | Div(i) -> dprintf "Div(%d)" i
  | Mod(i) -> dprintf "Mod(%d)" i
  | Floor(i) -> dprintf "Floor(%d)" i
  | Ceil(i) -> dprintf "Ceil(%d)" i
  | Fun(s) -> text s
  | BFun(s) -> text s
  | NotBFun(s) -> text s
  | Eq -> text "Eq"
  | Neq -> text "Neq" 
  | Geq -> text "Geq" 
  | Gt -> text "Gt" 
  | Leq -> text "Leq" 
  | Lt -> text "Lt" 
                                        (* VampyrePretty print a node *)
(* 
 * [weimer] 
 * We may want a recursive printout (for nodes that have edges to other
 * nodes) here, it's hard to say. 
 *)
(*		
let dn () (n : node) = 
  dprintf "[node %d %a%s]" n.id dop n.op
  (if n.id != n.root.id then " root=" ^ string_of_int n.root.id else "")
*)
let dn () n =
  let foo = nodeVal n in
  match foo with
    DAtom a -> I.da () a
  | DExp e -> I.de () e

(*------------------------------------------------------------------------*)

let snapshot () =
  (if (!(!debug)) then 
    ignore (printf "CC: nodes in dag: %d ...\n" (!idCounter-1)));
  undoStack := snapMark :: (!undoStack);
  disEqStack := snapMark :: (!disEqStack);
  assertStack := snapMark :: (!assertStack)
			      
let rec undo () =
  let cutLink (t1,t2) = UF.cut (Hashtbl.find ufTable t1) (Hashtbl.find ufTable t2) in
  let rec popAsserts () =
    match (!assertStack) with
      [] ->
	let _ = E.s (E.bug "WARNING: CC Assert stack popped while empty ...") in ()
    | (-1,-1)::t -> (assertStack := t)
    | (n,0)::t ->
	let d = findDag n in
	d.isAsserted <- false;(assertStack := t);popAsserts ()
    | (n,1)::t ->
	let d = findDag n in
	d.mesg <- dummyMesg;(assertStack := t);popAsserts ()
    | _::t -> let _ =  E.s (E.bug "WARNING: CC bad elt in assert stack ...") in ()
  in
	     
  let rec popDisEq () =
    match (!disEqStack) with
      [] -> let _ = E.s (E.bug "WARNING: CC DisEq stack popped while empty ..." ) in ()
    | h::t -> if h = snapMark then (disEqStack := t) else (Hashtbl.remove disEqTable h;disEqStack := t;popDisEq ())
  in
	
  let recomputeSets () =
    let recomputeNode rooti setList =
      let getUse j = (!(Hashtbl.find dagTable j)).ac_use in
      let nodei = !(Hashtbl.find dagTable rooti) in
      let _ = (nodei.eqClass <- setList) in
      let _ = (nodei.use <- List.flatten (List.map getUse setList)) in
      
      ()
    in

    let _ = Hashtbl.clear setTable in
    for i = 1 to (!idCounter) do
    let n_i = Hashtbl.find ufTable i in
    let n_i_node = !(Hashtbl.find dagTable i) in
    
    let n_i_root = UF.find n_i in
    let n_i_root_node = findDag n_i_root in
    let _ = n_i_node.root <- n_i_root_node in
    (
    if (Hashtbl.mem setTable n_i_root) then 
      Hashtbl.add setTable n_i_root (i::(Hashtbl.find setTable n_i_root))
    else
      Hashtbl.add setTable n_i_root [i])
    done;
    Hashtbl.iter recomputeNode setTable 
  in
  let printNode n p_n =
    print_string "node ";
    print_int n;print_string "parents ";
    List.iter (function x -> print_int x;print_string ",") (!p_n).use;
    print_string "children";
    List.iter (function x -> print_int x;print_string ",") (idf (!p_n).args);
    print_string "\n"
  in
      
  let showDagTable () = ();(*Hashtbl.iter printNode dagTable;*)
    (* let n26 = Hashtbl.find dagTable 26 in *)
    let _ = () (*if not (List.mem 36 (!n26).use) then E.s (E.bug "here parent trap") else print_string "dammit it is there ..." *)in
    ()
  in
  
  let finishUp () =
    try
      (try
	recomputeSets ()
      with Not_found -> E.s (E.bug "CC Error in Recompute Sets"));
      (try
	popDisEq ()
      with
	Not_found -> E.s (E.bug "CC Error in popDisEq"));
      (try
	popAsserts ()
      with Not_found -> E.s (E.bug "CC Error in popAsserts"))
	;
      (try
	showDagTable()
      with Not_found -> E.s (E.bug "CC Error in showDagTable")) 
    with Not_found -> E.s (E.bug "CC error in finishUp")
  in
  if (!undoStack = []) then (let _ = E.s (E.bug  "WARNING:CC undoing past an empty stack !!") in (finishUp ())) (* E.s will raise Error *)
  else (
    if (List.hd (!undoStack) = snapMark)
    then (undoStack := List.tl (!undoStack);finishUp ())
    else (
      (cutLink (List.hd (!undoStack)); (undoStack := List.tl (!undoStack)); undo ())))
        
(* All the above is in George's dag.ml *)
(* [Jhala] the DAG itself is an set of nodes that must be accessible (fast) using the id of the node - the first node is id 0 all the way through (idCounter-1) *)
(* --------------------------- to equality ----------------------------------------------*)
